/*****************************************************************************
 *
 * FILE:	aio.c
 * DESCRIPTION:	WTL: Advanced I/O functions
 * DATE:	Thu, Mar 27 2003
 * UPDATED:	Sun, Aug 19 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2003-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2003-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: aio.c,v 1.18 2007/08/19 05:57:22 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <sys/uio.h>
#include <errno.h>
#if	ENABLE_SSL
#include "wt-ssl.h"
#else
#include "wt.h"
#endif	/* ENABLE_SSL */
#include "wt-sub.h"

/******************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wt_netfd_readn
 * Description:		read "n" bytes exactly
 * Arguments:		netfd, buffer, length, timeout
 * Return Value:	successful: read size, otherwise: -1
 */
ssize_t
wt_netfd_readn(nfd, buf, nbyte, timeout)
	wt_netfd_t *	nfd;
	void *		buf;
	size_t		nbyte;
	wt_utime_t	timeout;
{
  char *	ptr;
  size_t	nleft;

  ptr	= (char *)buf;
  nleft	= nbyte;
  while (nleft > 0) {
    ssize_t	n;
    
    n = read(nfd->osfd, ptr, nleft);
    if (n < 0) {
      if (errno == EINTR) {
	continue;	/* call read() again */
      }
      if (!WT_IO_NOT_READY) {
       	return -1;
      }
      /* wait until the socket becomes readable */
      if (wt_netfd_poll(nfd, WT_POLL_IN, timeout) == -1) {
	return -1;
      }
    }
    else if (n == 0) {
      break;	/* EOF */
    }

    nleft -= n;
    ptr	  += n;
  }

  return (ssize_t)(nbyte - nleft);
}

/*
 * Function Name:	wt_netfd_writen
 * Description:		write "n" bytes exactly
 * Arguments:		netfd, buffer, length, timeout
 * Return Value:	successful: printed size, otherwise: -1
 */
ssize_t
wt_netfd_writen(nfd, buf, nbyte, timeout)
	wt_netfd_t *	nfd;
	const void *	buf;
	size_t		nbyte;
	wt_utime_t	timeout;
{
  const char *	ptr;
  size_t	nleft;

  ptr	= buf;
  nleft	= nbyte;
  while (nleft > 0) {
    ssize_t	n;
    
    n = write(nfd->osfd, buf, nbyte);
    if (n <= 0) {
      if (errno == EINTR) {
	continue;	/* call write() again */
      }
      if (!WT_IO_NOT_READY) {
       	return -1;
      }
      /* wait until the socket becomes writable */
      if (wt_netfd_poll(nfd, WT_POLL_OUT, timeout) == -1) {
	return -1;
      }
    }

    nleft -= n;
    ptr	  += n;
  }

  return (ssize_t)nbyte;
}

/*
 * Function Name:	wt_netfd_recv_fd
 * Description:		receive file descriptor
 * Arguments:		netfd pointer, returned descriptor pointer, timeout
 * Return Value:	successful: 0, otherwise: -1
 */
ssize_t
wt_netfd_recv_fd(nfd, recvfd, timeout)
	wt_netfd_t *	nfd;
	int *		recvfd;
	wt_utime_t	timeout;
{
  struct msghdr		msg;
  struct iovec		iov[1];
  char			buf[1];
  ssize_t		n;
  union {
    struct cmsghdr	cm;
    char		control[CMSG_SPACE(sizeof(int))];
  } control_un;
  struct cmsghdr *	cmptr;

  msg.msg_control	= control_un.control;
  msg.msg_controllen	= sizeof(control_un.control);
  msg.msg_name		= NULL;
  msg.msg_namelen	= 0;

  iov[0].iov_base	= buf;
  iov[0].iov_len	= sizeof(buf);
  msg.msg_iov		= iov;
  msg.msg_iovlen	= 1;

  while ((n = recvmsg(nfd->osfd, &msg, 0)) <= 0) {
    if (errno == EINTR) {
      continue;
    }
    if (!WT_IO_NOT_READY) {
      return -1;
    }
    /* wait until the socket becomes readable */
    if (wt_netfd_poll(nfd, WT_POLL_IN, timeout) == -1) {
      return -1;
    }
  }

  if ((cmptr = CMSG_FIRSTHDR(&msg)) != NULL &&
      cmptr->cmsg_len == CMSG_LEN(sizeof(int))) {
    if (cmptr->cmsg_level != SOL_SOCKET) {
#if	DEBUG
      fprintf(stderr, "DEBUG: control level != SOL_SOCKET\n");
#endif	/* DEBUG */
      exit(1);
    }
    if (cmptr->cmsg_type != SCM_RIGHTS) {
#if	DEBUG
      fprintf(stderr, "DEBUG: control type != SCM_RIGHTS\n");
#endif	/* DEBUG */
      exit(1);
    }
    *recvfd = *((int *)CMSG_DATA(cmptr));
  }
  else {
    *recvfd = -1;	/* descriptor was not passed */
  }

  return 0;
}

/*
 * Function Name:	wt_netfd_send_fd
 * Description:		send file descriptor
 * Arguments:		netfd pointer, sent descriptor, timeout
 * Return Value:	successful: 0, otherwise: -1
 */
ssize_t
wt_netfd_send_fd(nfd, sendfd, timeout)
	wt_netfd_t *	nfd;
	int		sendfd;
	wt_utime_t	timeout;
{
  struct msghdr		msg;
  struct iovec		iov[1];
  char			buf[1];
  ssize_t		n;
  union {
    struct cmsghdr	cm;
    char		control[CMSG_SPACE(sizeof(int))];
  } control_un;
  struct cmsghdr *	cmptr;

  msg.msg_control	= control_un.control;
  msg.msg_controllen	= sizeof(control_un.control);

  cmptr			= CMSG_FIRSTHDR(&msg);
  cmptr->cmsg_len	= CMSG_LEN(sizeof(int));
  cmptr->cmsg_level	= SOL_SOCKET;
  cmptr->cmsg_type	= SCM_RIGHTS;
  *((int *)CMSG_DATA(cmptr)) = sendfd;

  msg.msg_name		= NULL;
  msg.msg_namelen	= 0;

  iov[0].iov_base	= buf;
  iov[0].iov_len	= sizeof(buf);
  msg.msg_iov		= iov;
  msg.msg_iovlen	= 1;

  while ((n = sendmsg(nfd->osfd, &msg, 0)) < 0) {
    if (errno == EINTR) {
      continue;
    }
    if (!WT_IO_NOT_READY) {
      return -1;
    }
    /* wait until the socket becomes writable */
    if (wt_netfd_poll(nfd, WT_POLL_OUT, timeout) == -1) {
      return -1;
    }
  }

  return 0;
}

/*
 * Function Name:	wt_netfd_readv
 * Description:		read data
 * Arguments:		netfd, iovec buffers, iovec size, timeout
 * Return Value:	successful: read size, otherwise: -1
 */
ssize_t
wt_netfd_readv(nfd, iov, iovcnt, timeout)
	wt_netfd_t *		nfd;
	const struct iovec *	iov;
	int			iovcnt;
	wt_utime_t		timeout;
{
  ssize_t	n;

  while ((n = readv(nfd->osfd, iov, iovcnt)) < 0) {
    if (errno == EINTR) {
      continue;
    }
    if (!WT_IO_NOT_READY) {
      return -1;
    }
    /* wait until the socket becomes readable */
    if (wt_netfd_poll(nfd, WT_POLL_IN, timeout) == -1) {
      return -1;
    }
  }

  return n;
}

/*
 * Function Name:	wt_netfd_writev
 * Description:		write data
 * Arguments:		netfd, iovec buffers, iovec size, timeout
 * Return Value:	successful: printed size, otherwise: -1
 */
ssize_t
wt_netfd_writev(nfd, iov, iovcnt, timeout)
	wt_netfd_t *		nfd;
	const struct iovec *	iov;
	int			iovcnt;
	wt_utime_t		timeout;
{
  int		i;
  ssize_t	nbytes;
  size_t	resid;
  ssize_t	n;

  for (nbytes = 0, i = 0; i < iovcnt; i++) {
    nbytes += iov[i].iov_len;
  }

  resid = nbytes;
  while (resid > 0) {
    n = writev(nfd->osfd, iov, iovcnt);
    if (n == -1) {
      if (errno == EINTR) {
	continue;
      }
      if (!WT_IO_NOT_READY) {
       	return -1;
      }
      /* wait until the socket becomes writable */
      if (wt_netfd_poll(nfd, WT_POLL_OUT, timeout) == -1) {
	return -1;
      }
    }
    else {
      resid -= n;
      break;
    }
  }

  if (resid > 0) {
    for (i = 0; i < iovcnt; i++) {
      if (n >= iov[i].iov_len) {	/* written */
	n -= iov[i].iov_len;
      }
      else {
	ssize_t	l;

#if	DEBUG
	fprintf(stderr, "DEBUG[wt_netfd_writev] incompleted: i=%d len=%zu\n",
		i, iov[i].iov_len);
#endif	/* DEBUG */
	l = wt_netfd_write(nfd, iov[i].iov_base, iov[i].iov_len, timeout);
	if (l == -1) {
	  break;
	}
	n     -= l;
	resid -= l;
      }
    }
  }

  return (ssize_t)(nbytes - resid);
}
