/*****************************************************************************
 *
 * FILE:	cache.c
 * DESCRIPTION:	Wyvern: cache library
 * DATE:	Sun, Jun 29 2003
 * UPDATED:	Sun, Aug 19 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2003-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2003-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: cache.c,v 1.13 2007/08/19 05:57:22 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#include "wc.h"

/******************************************************************************
 *
 *	Local functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wc_new_cache
 * Description:		create a new cache
 * Arguments:		cache size
 * Return Value:	successful: cache pointer, otherwise: NULL
 */
Cache *
wc_new_cache(size)
	size_t	size;
{
  Cache *	new;

			   /* +1 for a entry */
  new = (Cache *)calloc(size + 1, sizeof(Cache));
  if (new != NULL) {
    unsigned int	i;

    new->next = new->prev = new;
    for (i = 0; i < size + 1; i++) {
      wc_insert_cache(new, &new[i]);
    }
  }

  return new;
}

/*
 * Function Name:	wc_insert_cache
 * Description:		insert 'new' beore 'pos'
 * Arguments:		position, new slot
 * Return Value:	none
 */
void
wc_insert_cache(pos, new)
	register Cache *	pos;
	register Cache *	new;
{
  pos->prev->next = new;
  new->prev	  = pos->prev;
  new->next	  = pos;
  pos->prev	  = new;
}

/*
 * Function Name:	wc_remove_cache
 * Description:		remove the slot from cache
 * Arguments:		removed slot
 * Return Value:	removed slot
 */
Cache *
wc_remove_cache(x)
	register Cache *	x;
{
  x->prev->next	= x->next;
  x->next->prev	= x->prev;
  return x;
}

/*
 * Function Name:	wc_search_cache
 * Description:		search the data from cache
 * Arguments:		cache top, data, compare function
 * Return Value:	successul: data pointer, otherwise: NULL
 */
void *
wc_search_cache(top, data, compar)
	register Cache *	top;
	register void *		data;
	int			(*compar)(const void *, const void *);
{
  register Cache *	cp;

  for (cp = top->prev; cp != top; cp = cp->prev) {
    if (cp->data != NULL) {
      if ((*compar)(cp->data, data) == 0) {
	wc_insert_cache(top, wc_remove_cache(cp));
	return cp->data;
      }
    }
  }

  return NULL;
}

/*
 * Function Name:	wc_flush_cache
 * Description:		flush the data from cache
 * Arguments:		cache top, trashcan function
 * Return Value:	none
 */
void
wc_flush_cache(top, trashcan)
	Cache *	top;
	void	(*trashcan)(const void *);
{
  register Cache *	cp;

  for (cp = top->prev; cp != top; cp = cp->prev) {
    if (cp->data != NULL) {
      (*trashcan)(cp->data);
    }
  }
}

/*****************************************************************************/

#if	DEBUG_CACHE
#include <stdio.h>
#include <string.h>

#define	INFO_TABLE_SIZE	(3)
typedef struct _info {
  const char *	name;
  const char *	email;
} Info;

static void
trashcan(x)
	const void *	x;
{
  Info *	ip = (Info *)x;
  fprintf(stderr, "[trashcan]\t\t%s => %s\n", ip->name, ip->email);
  free((char *)ip->name);
  free((char *)ip->email);
  free((Info *)ip);
  ip = NULL;
}

static int
compare(x, y)
	register const void *	x;
	register const void *	y;
{
  if (x && y) {
    if (strcmp(((Info *)x)->name, ((Info *)y)->name) == 0) {
      return 0;
    }
  }
  return -1;
}

static void
dump(top)
	register Cache *	top;
{
  register Cache *	cp;

  for (cp = top->prev; cp != top; cp = cp->prev) {
    if (cp->data) {
      Info *	info = (Info *)cp->data;
      fprintf(stderr, "[dump]\t\t%s => %s\n", info->name, info->email );
    }
  }
}

static void
insert(top, name, email)
	register Cache *	top;
	const char *		name;
	const char *		email;
{
  Cache *	new = wc_remove_cache(top->next);
  if (new) {
    Info *	info = (Info *)malloc(sizeof(Info));

    info->name  = strdup(name);
    info->email = strdup(email);
    if (new->data) {
      fprintf(stderr, "[free]\t\t%s => %s\n",
	      ((Info *)new->data)->name, ((Info *)new->data)->email);
      free((char *)((Info *)new->data)->name);
      free((char *)((Info *)new->data)->email);
      free((Info *)new->data);
    }
    new->data	= info;
    wc_insert_cache(top, new);
    fprintf(stderr, "[insert]\t%s => %s\n", name, email);
  }
}

static void
search(top, name)
	register Cache *	top;
	const char *		name;
{
  Info		info;
  Info *	p;

  info.name = name;
  p = (Info *)wc_search_cache(top, &info, compare);
  if (p) {
    fprintf(stderr, "[search]\t%s => %s\n", p->name, p->email);
  }
  else {
    fprintf(stderr, "[search]\t%s not found\n", name);
  }
}

int
main(argc, argv)
	int	argc;
	char *	argv[];
{
  Cache *	infoList;

  infoList = wc_new_cache(INFO_TABLE_SIZE);
  if (infoList) {
    insert(infoList, "kouichi", "kouichi@mysticwall.com"); dump(infoList);
    search(infoList, "kouichi");
    insert(infoList, "john", "john@example.com"); dump(infoList);
    search(infoList, "kouichi");
    insert(infoList, "tom", "tom@example.net"); dump(infoList);
    search(infoList, "kouichi");
    insert(infoList, "ken", "ken@example.org"); dump(infoList);
    search(infoList, "kouichi");
    insert(infoList, "jack", "jack@example.co.jp"); dump(infoList);
  }
  wc_flush_cache(infoList, trashcan);
  free((Cache *)infoList);

  return 0;
}
#endif	/* DEBUG_CACHE */
