/*****************************************************************************
 *
 * FILE:	cond.c
 * DESCRIPTION:	WTL: condition variable functions
 * DATE:	Thu, Mar 27 2003
 * UPDATED:	Fri, Jun 10 2005
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2003-2005 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2003-2005 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: cond.c,v 1.24 2005/06/14 23:51:18 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#include <errno.h>
#if	ENABLE_SSL
#include "wt-ssl.h"
#else
#include "wt.h"
#endif	/* ENABLE_SSL */
#include "wt-sub.h"

/******************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static int	wt_cond_signal(wt_cond_t, bool);

/******************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wt_thread_cond_init
 * Description:		create a new condition variable list
 * Arguments:		new condvar returned pointer
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_thread_cond_init(cvar)
	wt_cond_t *	cvar;
{
  wt_cond_t	new;

  *cvar = new = (wt_cond_t)calloc(1, sizeof(wt_clist_t));
  if (new != NULL) {
    wt_clist_init(new, NULL);
    return 0;
  }
  errno = ENOMEM;

  return -1;
}

/*
 * Function Name:	wt_thread_cond_destroy
 * Description:		destroy a condition variable list
 * Arguments:		condition variable list
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_thread_cond_destroy(cvar)
	wt_cond_t 	cvar;
{
  if (cvar == NULL) {
    errno = EINVAL;
    return -1;
  }

  if (cvar->next != cvar) {
    errno = EBUSY;
    return -1;
  }
  free((wt_cond_t)cvar);

  return 0;
}

/*
 * Function Name:	wt_cond_signal
 * Description:		send a signal to waiting threads
 * Arguments:		condition variable list, broadcast flag: true/false
 * Return Value:	successful: 0, otherwise: -1
 */
static int
wt_cond_signal(cvar, broadcast)
	wt_cond_t	cvar;
	bool		broadcast;
{
  register wt_cond_t	cp;

  if (cvar == NULL) {
    errno = EINVAL;
    return -1;
  }

  for (cp = cvar->next; cp->body != NULL; cp = cp->next) {
    wt_thread_t	thd = wt_thread_ptr(cp);
    if (thd->state == WT_ST_WAIT_COND) {
      if (thd->flags & WT_FL_ON_SLEEPQ) {
	wt_remove_sleep_q(thd);
      }
      wt_append_run_q(thd);
      if (broadcast == false) {
	break;
      }
    }
  }

  return 0;
}

/*
 * Function Name:	wt_thread_cond_broadcast
 * Description:		send a signal to all waiting threads
 * Arguments:		condition variable list
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_thread_cond_broadcast(cvar)
	wt_cond_t 	cvar;
{
  return wt_cond_signal(cvar, true);
}

/*
 * Function Name:	wt_thread_cond_signal
 * Description:		send a signal to a waiting thread
 * Arguments:		condition variable list
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_thread_cond_signal(cvar)
	wt_cond_t 	cvar;
{
  return wt_cond_signal(cvar, false);
}

/*
 * Function Name:	wt_thread_cond_timedwait
 * Description:		wait a thread until time is out
 * Arguments:		condition variable list, timeout
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_thread_cond_timedwait(cvar, timeout)
	wt_cond_t 	cvar;
	wt_utime_t	timeout;
{
  wt_thread_t	me = wt_current_thread;

  if (cvar == NULL) {
    errno = EINVAL;
    return -1;
  }

  wt_clist_append_link(&me->wait_q, cvar);

  if (timeout != WT_UTIME_NO_TIMEOUT) {
    wt_remove_run_q(me);
    wt_append_sleep_q(me, timeout);
  }
  me->state = WT_ST_WAIT_COND;

  wt_context_switch(me);

  wt_clist_remove_link(&me->wait_q);

  if (me->flags & WT_FL_TIMEDOUT) {
    me->flags &= ~WT_FL_TIMEDOUT;
    errno = ETIMEDOUT;
    return -1;
  }

  return 0;
}

