/*****************************************************************************
 *
 * FILE:	httpd2.c
 * DESCRIPTION:	httpd server with WTL
 * DATE:	Wed, Jun 11 2003
 * UPDATED:	Sun, Jun 11 2006
 * AUTHOR:	Kouichi ABE (WALL) / $B0$It9/0l(B
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2003-2006 $B0$It9/0l!?(BKouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2003-2006 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: httpd2.c,v 1.10 2006/06/11 13:22:30 kouichi Exp $
 *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <signal.h>
#include "wt.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	CRLF	"\015\012"
#define	LF	'\012'
#define	CR	'\015'
#define	EOL	'\0'

#define	MAX_IP	(2)

typedef struct {
  wt_thread_t	id;	/* thread ID */
  unsigned long	count;	/* #connections handled */
} Thread;

typedef struct {
  wt_netfd_t *	nsd;
  socklen_t	addrlen;
  Thread *	thread;
  wt_mutex_t	mlock;
} ListenAddr;

typedef struct {
  /**/
#define	DEF_PORT	"80"
  char *	port;
  char *	docdir;
  char *	logfile;
  uid_t		uid;
  /**/
#define	DEF_THREADS	(5)
  int		nthreads;
  /**/
  ListenAddr	listen[MAX_IP];
} Httpd;

enum {
  EST400,	/* bad request */
  EST403,	/* forbidden */
  EST404,	/* not found */
  EST405,	/* method not allowed */
  EST500	/* internal server error */
};

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static Httpd	httpd = {
  DEF_PORT,	/* Port */
  NULL,		/* Document Root */
  NULL,		/* Access/Error Log File */
  65535,	/* User */
  DEF_THREADS,	/* Max Threads */
  {
    { NULL, 0, NULL, WT_MUTEX_INITIALIZER },
    { NULL, 0, NULL, WT_MUTEX_INITIALIZER }
  }
};

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	make_threads(int, int);
static void	thread_main(int *);
static void	web_child(wt_netfd_t *);
static char *	get_uri(wt_netfd_t *);
static void	http_response(wt_netfd_t *, const char *);
static void	http_mime(wt_netfd_t *, const char *);
static void	http_error(wt_netfd_t *, int);
static void	http_date(wt_netfd_t *);

static void	set_trap(void);
static void	trap(int);
static void	print_thread_count(int);
static void	print_cpu_time(void);

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

static void
make_threads(n, i)
	int	n;
	int	i;
{
  wt_thread_attr_t	attr = wt_thread_attr_defaults;
  int *			v = malloc(sizeof(int));

  *v = n * 10 + i;
#if	DEBUG
  fprintf(stderr, "DEBUG[make thread] v=%d n=%d i=%d\n", *v, n, i);
#endif	/* DEBUG */
  asprintf((char **)&attr.name, "thread%d.%d", n, i + 1);
  if (wt_thread_create(&httpd.listen[n].thread[i].id, &attr,
		       (void *)&thread_main, (void *)v) == -1) {
    perror("wt_thread_create");
    exit(1);
  }
}

static void
thread_main(arg)
	int *	arg;
{
  struct sockaddr *	cliaddr;
  int			n = *(int *)arg / 10;
  int			i = *(int *)arg % 10;

#if	DEBUG
  fprintf(stderr, "DEBUG[thread_main] v=%d n=%d i=%d\n", *(int *)arg, n, i);
#endif	/* DEBUG */
  cliaddr = malloc(httpd.listen[n].addrlen);
  if (cliaddr == NULL) {
    perror("malloc");
    exit(1);
  }
  free((int *)arg);

  for (;;) {
    wt_netfd_t *	nfd;
    socklen_t		clilen = httpd.listen[n].addrlen;

    wt_thread_mutex_lock(httpd.listen[n].mlock);
    nfd = wt_netfd_accept(httpd.listen[n].nsd, cliaddr, &clilen,
			  WT_UTIME_NO_TIMEOUT);
    wt_thread_mutex_unlock(httpd.listen[n].mlock);

    httpd.listen[n].thread[i].count++;

    web_child(nfd);		/* process the request */
    wt_netfd_close(nfd);
  }

  free(cliaddr);
}

static void
web_child(nfd)
	wt_netfd_t *	nfd;
{
  char *	uri;

#if	DEBUG
  fprintf(stderr, "DEBUG: Read Data From Client (descriptor #%d)\n", nfd->osfd);
#endif /* DEBUG */

  uri = get_uri(nfd);
  if (uri == NULL) {
    return;
  }

  for (;;) {
    char	req[BUFSIZ];
    ssize_t	n;

    memset(req, 0, sizeof(req));
    n = wt_netfd_readline(nfd, req, sizeof(req), WT_UTIME_NO_TIMEOUT);
    if (n < 0) {	/* error */
      http_error(nfd, EST500);
    }
    if (memcmp(req, CRLF, 2) == 0) {
      break;	/* request end */
    }
  }
  http_response(nfd, uri);
  free(uri);
}

static char *
get_uri(nfd)
	register wt_netfd_t *	nfd;
{
  char		req[BUFSIZ];
  ssize_t	n;
  int		begin;
  int		end;
  size_t	len;
  char *	uri;

  memset(req, 0, sizeof(req));
  n = wt_netfd_readline(nfd, req, sizeof(req), WT_UTIME_NO_TIMEOUT);
  if (n <= 0) {
    http_error(nfd, EST400);
    return NULL;
  }

  /* HTTP method */
  if (strncmp(req, "GET ", 4) == 0) {
    begin = 4;
  }
  else if (strncmp(req, "HEAD ", 5) == 0) {
    begin = 5;
  }
  else {
    http_error(nfd, EST405);
    return NULL;
  }

  /* End the request of line */
  end = strstr(req, CRLF) - req;

  /* HTTP version: HTTP/x.y */
  end -= 9;
  switch (req[end + 8]) {
    case '9':
      break;
    case '0':
      break;
    case '1':
      break;
    default:
      http_error(nfd, EST400);
      return NULL;
  }

  len = end - begin;
  uri = (char *)malloc(len + 1);
  memcpy(uri, req + begin, len);
  uri[len] = EOL;

#if	DEBUG
  fprintf(stderr, "uri: %s\n", uri);
#endif	/* DEBUG */

  return uri;
}

static void
http_response(nfd, uri)
	register wt_netfd_t *	nfd;
	const char *		uri;
{
  wt_netfd_t *	fd;

  while (uri[0] == '/') {
    uri++;
  }
  if (uri == NULL) {
    http_error(nfd, EST404);
    return;
  }

  fd = wt_netfd_open(uri, O_RDONLY, 0400);
  if (fd != NULL) {
    const char	head[] = "HTTP/1.0 200 OK\r\nServer: httpd\r\n";
    struct stat	sbuf;
    char	buf[BUFSIZ];
    int		n;

    if (fstat(wt_netfd_fileno(fd), &sbuf) == -1) {
      wt_netfd_close(fd);
      return;
    }
    wt_netfd_write(nfd, head, strlen(head), WT_UTIME_NO_TIMEOUT);
    memset(buf, 0, sizeof(buf));
    n = snprintf(buf, sizeof(buf), "Content-Length: %ld\r\n",
		 (long)sbuf.st_size);
    wt_netfd_write(nfd, buf, (size_t)n, WT_UTIME_NO_TIMEOUT);
    http_date(nfd);
    memset(buf, 0, sizeof(buf));
    strftime(buf, sizeof(buf), "Last-Modified: %a, %d %h %Y %H:%M:%S %Z\r\n",
	     gmtime(&sbuf.st_mtime));
    wt_netfd_write(nfd, buf, strlen(buf), WT_UTIME_NO_TIMEOUT);
    http_mime(nfd, uri);

    wt_netfd_write(nfd, CRLF, 2, WT_UTIME_NO_TIMEOUT);
    memset(buf, 0, sizeof(buf));
    while ((n = wt_netfd_read(fd, buf, sizeof(buf), WT_UTIME_NO_TIMEOUT)) > 0) {
      wt_netfd_write(nfd, buf, (size_t)n, WT_UTIME_NO_TIMEOUT);
      memset(buf, 0, sizeof(buf));
    }

    wt_netfd_close(fd);
  }
  else {
    http_error(nfd, EST404);
  }
}

static void
http_mime(nfd, uri)
	register wt_netfd_t *	nfd;
	const char *		uri;
{
  const char	text_plain[] = "text/plain";
  const char	text_html[]  = "text/html";
  const char	text_css[]   = "text/css";
  const char	image_png[]  = "image/png";
  const char	image_gif[]  = "image/gif";
  const char	image_jpeg[] = "image/jpeg";
  char		buf[BUFSIZ];
  char *	p;

  p = strrchr(uri, '.');
  if (p == NULL) {
    return;
  }
  p++;

  memset(buf, 0, sizeof(buf));
  if (strcmp(p, "html") == 0 || strcmp(p, "htm") == 0) {
    snprintf(buf, sizeof(buf), "Content-Type: %s\r\n", text_html);
  }
  else if (strcmp(p, "css") == 0) {
    snprintf(buf, sizeof(buf), "Content-Type: %s\r\n", text_css);
  }
  else if (strcmp(p, "png") == 0) {
    snprintf(buf, sizeof(buf), "Content-Type: %s\r\n", image_png);
  }
  else if (strcmp(p, "gif") == 0) {
    snprintf(buf, sizeof(buf), "Content-Type: %s\r\n", image_gif);
  }
  else if (strcmp(p, "jpeg") == 0 || strcmp(p, "jpg") == 0) {
    snprintf(buf, sizeof(buf), "Content-Type: %s\r\n", image_jpeg);
  }
  else {
    snprintf(buf, sizeof(buf), "Content-Type: %s\r\n", text_plain);
  }
  wt_netfd_write(nfd, buf, strlen(buf), WT_UTIME_NO_TIMEOUT);
#if	DEBUG
  fprintf(stderr, "URI: %s\t%s", uri, buf);
#endif	/* DEBUG */
}

static void
http_error(nfd, eno)
	register wt_netfd_t *	nfd;
	int			eno;
{
  struct {
    const char *	line;
    const char *	text;
  } emsg[] = {
    { "HTTP/1.0 400 Bad Request\r\nServer: httpd\r\n",
      "<html><head><title>Error</title></head><body><h1>Error 400</h1><p>Your request has bad syntax or is inherently impossible to satisy.</p></body></html>\r\n" },
    { "HTTP/1.0 403 Forbidden\r\nServer: httpd\r\n",
      "<html><head><title>Error</title></head><body><h1>Error 403</h1><p>You do not have permission to get the request URL.</p></body></html>\r\n" },
    { "HTTP/1.0 404 Not Found\r\nServer: httpd\r\n",
      "<html><head><title>Error</title></head><body><h1>Error 404</h1><p>Not found - file does not exist or is read protected</p></body></html>\r\n" },
    { "HTTP/1.0 405 Method Not Allowed\r\nServer: httpd\r\nAllow: HEAD,GET\r\n",
      "<html><head><title>Error</title></head><body><h1>Error 405</h1><p>This server answer only to GET and Head requests</p></body></html>\r\n" },
    { "HTTP/1.0 500 Internal Server Error\r\nServer: httpd\r\n",
      "<html><head><title>Error</title></head><body><h1>Error 500</h1><p>There was an unusual problem serving the requested URL.</p></body></html>\r\n" }
  };
  char	text_html[] = "Content-Type: text/html\r\n\r\n";

  wt_netfd_write(nfd, emsg[eno].line, strlen(emsg[eno].line),
		 WT_UTIME_NO_TIMEOUT);
  http_date(nfd);
  wt_netfd_write(nfd, text_html, strlen(text_html), WT_UTIME_NO_TIMEOUT);
  wt_netfd_write(nfd, CRLF, 2, WT_UTIME_NO_TIMEOUT);
  wt_netfd_write(nfd, emsg[eno].text, strlen(emsg[eno].text),
		 WT_UTIME_NO_TIMEOUT);
}

static void
http_date(nfd)
	register wt_netfd_t *	nfd;
{
  time_t	now = time(NULL);
  char		date[80];

  strftime(date, sizeof(date), "Date: %a, %d %h %Y %H:%M:%S %Z\r\n",
	   gmtime(&now));
  wt_netfd_write(nfd, date, strlen(date), WT_UTIME_NO_TIMEOUT);
}

static void
set_trap(void)
{
  struct sigaction	act;

  /* quit */
  sigemptyset(&act.sa_mask);
  act.sa_flags	 = 0;
  act.sa_handler = (void *)trap;
  sigaction(SIGQUIT, &act, NULL);
  sigaction(SIGTERM, &act, NULL);
  /**/
  sigaction(SIGBUS,  &act, NULL);
  sigaction(SIGSEGV, &act, NULL);

  /* print statistics */
  act.sa_handler = (void *)print_thread_count;
  sigaction(SIGINT,  &act, NULL);
}

static void
print_thread_count(signo)
	int	signo;
{
  register int	n;

  print_cpu_time();
  for (n = 0; n < MAX_IP; n++) {
    register int	i;

    for (i = 0; i < httpd.nthreads; i++) {
#if	DEBUG
      fprintf(stderr, "thread %d, %ld connections\n",
	      i, httpd.listen[n].thread[i].count);
#endif	/* DEBUG */
    }
  }

  exit(0);
}

static void
print_cpu_time(void)
{
  double	user;
  double	sys;
  struct rusage	myusage;
  struct rusage	childusage;

  if (getrusage(RUSAGE_SELF, &myusage) < 0) {
    perror("getrusage");
    exit(1);
  }
  if (getrusage(RUSAGE_CHILDREN, &childusage) < 0) {
    perror("getrusage");
    exit(1);
  }

  user = (double)myusage.ru_utime.tv_sec + myusage.ru_utime.tv_usec / 1000000.0;
  user += (double)childusage.ru_utime.tv_sec
				+ childusage.ru_utime.tv_usec / 1000000.0;
  sys = (double)myusage.ru_stime.tv_sec + myusage.ru_stime.tv_usec / 1000000.0;
  sys += (double)childusage.ru_stime.tv_sec
				+ childusage.ru_stime.tv_usec / 1000000.0;
#if	DEBUG
  fprintf(stderr, "\nuser time = %g, sys time = %g\n", user, sys);
#endif	/* DEBUG */
}

static void
trap(signo)
	int	signo;
{
#if	DEBUG
  fprintf(stderr, "killed by sig#%d\n", signo);
#endif	/* DEBUG */
  exit(0);
}

static void
usage(void)
{
  fprintf(stderr, "usage: httpd [-p port][-d docdir][-l logfile][-u uid]\n");
  exit(64);
}

int
main(argc, argv)
	int	argc;
	char *	argv[];
{
  register int	n;
  int		ch;

  httpd.uid = getuid();
  while ((ch = getopt(argc, argv, "p:d:l:u:t:")) != -1) {
    switch (ch) {
      case 'p':
	httpd.port	= strdup(optarg);
	break;
      case 'd':
	httpd.docdir	= strdup(optarg);
	break;
      case 'l':
	httpd.logfile	= strdup(optarg);
	break;
      case 'u':
	httpd.uid	= (uid_t)atoi(optarg);
	break;
      case 't':
	httpd.nthreads	= atoi(optarg);
	break;
      case '?':
      default:
	usage();
    }
  }
  argc -= optind;
  argv += optind;

  if (httpd.docdir == NULL) {
    httpd.docdir = getenv("HOME");
  }
  if (chdir(httpd.docdir) == -1) {
    perror("chdir");
    exit(1);
  }

  set_trap();

  if (wt_init(WT_VP_SS_MEDIUM) == -1) {
    perror("wt_init");
    exit(1);
  }

  httpd.listen[0].nsd = wt_netfd_tcp_server("127.0.0.1", httpd.port, 16);
  httpd.listen[1].nsd = wt_netfd_tcp_server("192.168.1.2", httpd.port, 16);
  for (n = 0; n < MAX_IP; n++) {
    register int	i;

    if (httpd.listen[n].nsd == NULL) {
      perror("wt_init_port");
      exit(1);
    }
    httpd.listen[n].addrlen = wt_netfd_socklen(httpd.listen[n].nsd);
    httpd.listen[n].thread  = (Thread *)calloc((size_t)httpd.nthreads,
					       sizeof(Thread));
    if (httpd.listen[n].thread == NULL) {
      perror("calloc");
      exit(1);
    }
    if (wt_thread_mutex_init(&httpd.listen[n].mlock) == -1) {
      perror("wt_thread_mutex_init");
      exit(1);
    }

    for (i = 0; i < httpd.nthreads; i++) {
      make_threads(n, i);
    }
  }

  for (;;) {
    wt_thread_run();
//    pause();
  }

  return 0;	/* not reached */
}

