/*****************************************************************************
 *
 * FILE:	httpd3.c
 * DESCRIPTION:	httpd server with WTL
 * DATE:	Wed, Jun 11 2003
 * UPDATED:	Sun, Jun 11 2006
 * AUTHOR:	Kouichi ABE (WALL) / $B0$It9/0l(B
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2003-2006 $B0$It9/0l!?(BKouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2003-2006 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: httpd3.c,v 1.12 2006/06/11 13:22:30 kouichi Exp $
 *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <signal.h>
#include "wt.h"
#include "wl.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	CRLF	"\015\012"
#define	LF	'\012'
#define	CR	'\015'
#define	EOL	'\0'

#define	DEF_LOGFILE	"/dev/null"

typedef struct _ListenAddr {
  const char *		ipaddr;
  const char *		port;
  wt_netfd_t *		nsd;
  socklen_t		addrlen;
  wt_thread_t *		thread;
  wt_mutex_t		mlock;
  struct _ListenAddr *	next;
} ListenAddr;

typedef struct {
#define	DEF_ADDR	"192.168.1.23"
#define	DEF_PORT	"80"
  char *	addr;
  char *	port;
  char *	docdir;
  char *	logfile;
  uid_t		uid;
  /**/
#define	DEF_THREADS	(5)
  int		nthreads;
  /**/
  ListenAddr *	listen;
  /**/
  FILE *	logger;
} Httpd;

enum {
  EST400,	/* bad request */
  EST403,	/* forbidden */
  EST404,	/* not found */
  EST405,	/* method not allowed */
  EST500	/* internal server error */
};

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static Httpd	httpd = {
  DEF_ADDR,	/* Address */
  DEF_PORT,	/* Port */
  NULL,		/* Document Root */
  DEF_LOGFILE,	/* Access/Error Log File */
  65535,	/* User */
  DEF_THREADS,	/* Max Threads */
  NULL,
  NULL
};

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static ListenAddr *	new_listen(const char *, const char *);
static void		make_threads(ListenAddr *, int);
static void		thread_main(void *);
static void		web_child(wt_netfd_t *);
static char *		get_uri(wt_netfd_t *);
static void		http_response(wt_netfd_t *, const char *);
static void		http_mime(wt_netfd_t *, const char *);
static void		http_error(wt_netfd_t *, int);
static void		http_date(wt_netfd_t *);

static void		set_trap(void);
static void		trap(int);
static void		print_thread_count(int);
static void		print_cpu_time(void);

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

static ListenAddr *
new_listen(ipaddr, port)
	const char *	ipaddr;
	const char *	port;
{
  ListenAddr *	new;

  new = (ListenAddr *)malloc(sizeof(ListenAddr));
  if (new == NULL) {
    perror("malloc");
    exit(1);
  }
  new->ipaddr	= strdup(ipaddr);
  new->port	= strdup(port);
  new->nsd	= NULL;
  new->addrlen	= 0;
  new->thread	= NULL;
  new->mlock	= WT_MUTEX_INITIALIZER;
  new->next	= httpd.listen;
  httpd.listen	= new;

  return new;
}

static void
make_threads(lap, i)
	ListenAddr *	lap;
	int		i;
{
  if (wt_thread_create(&lap->thread[i], NULL, (void *)&thread_main,
			(void *)lap) == -1) {
    perror("wt_thread_create");
    exit(1);
  }
}

static void
thread_main(arg)
	void *	arg;
{
  struct sockaddr *	cliaddr;
  ListenAddr *		lap = (ListenAddr *)arg;

  cliaddr = malloc(lap->addrlen);
  if (cliaddr == NULL) {
    perror("malloc");
    exit(1);
  }

  for (;;) {
    wt_netfd_t *	nfd;
    socklen_t		clilen = lap->addrlen;
    char *		peer;

    wt_thread_mutex_lock(lap->mlock);
    nfd = wt_netfd_accept(lap->nsd, cliaddr, &clilen, WT_UTIME_NO_TIMEOUT);
    peer = wt_sock_ntop(cliaddr, true);
#if	DEBUG
    fprintf(httpd.logger, "DEBUG: Peer: %s\n", peer);
#endif	/* DEBUG */
    wt_thread_mutex_unlock(lap->mlock);

    web_child(nfd);		/* process the request */
    wt_netfd_close(nfd);
  }

  free(cliaddr);
}

static void
web_child(nfd)
	wt_netfd_t *	nfd;
{
  char *	uri;

  uri = get_uri(nfd);
  if (uri == NULL) {
    return;
  }

  for (;;) {
    char	req[BUFSIZ];
    ssize_t	n;

    memset(req, 0, sizeof(req));
    n = wt_netfd_readline(nfd, req, sizeof(req), WT_UTIME_NO_TIMEOUT);
    if (n < 0) {	/* error */
      http_error(nfd, EST500);
    }
    if (memcmp(req, CRLF, 2) == 0) {
      break;	/* request end */
    }
  }
  http_response(nfd, uri);
  free(uri);
}

static char *
get_uri(nfd)
	register wt_netfd_t *	nfd;
{
  char		req[BUFSIZ];
  ssize_t	n;
  int		begin;
  int		end;
  size_t	len;
  char *	uri;

  memset(req, 0, sizeof(req));
  n = wt_netfd_readline(nfd, req, sizeof(req), WT_UTIME_NO_TIMEOUT);
  if (n <= 0) {
    http_error(nfd, EST400);
    return NULL;
  }

  /* HTTP method */
  if (strncmp(req, "GET ", 4) == 0) {
    begin = 4;
  }
  else if (strncmp(req, "HEAD ", 5) == 0) {
    begin = 5;
  }
  else {
    http_error(nfd, EST405);
    return NULL;
  }

  /* End the request of line */
  end = strstr(req, CRLF) - req;

  /* HTTP version: HTTP/x.y */
  end -= 9;
  switch (req[end + 8]) {
    case '9':
      break;
    case '0':
      break;
    case '1':
      break;
    default:
      http_error(nfd, EST400);
      return NULL;
  }

  len = end - begin;
  uri = (char *)malloc(len + 1);
  memcpy(uri, req + begin, len);
  uri[len] = EOL;

  return uri;
}

static void
http_response(nfd, uri)
	register wt_netfd_t *	nfd;
	const char *		uri;
{
  wt_netfd_t *	fd;

  while (uri[0] == '/') {
    uri++;
  }
  if (uri == NULL) {
    http_error(nfd, EST404);
    return;
  }

  fd = wt_netfd_open(uri, O_RDONLY, 0400);
  if (fd != NULL) {
    const char	head[] = "HTTP/1.0 200 OK\r\nServer: httpd\r\n";
    struct stat	sbuf;
    char	buf[BUFSIZ];
    int		n;

    if (fstat(wt_netfd_fileno(fd), &sbuf) == -1) {
      wt_netfd_close(fd);
      return;
    }
    wt_netfd_write(nfd, head, strlen(head), WT_UTIME_NO_TIMEOUT);
    memset(buf, 0, sizeof(buf));
    n = snprintf(buf, sizeof(buf), "Content-Length: %ld\r\n",
		 (long)sbuf.st_size);
    wt_netfd_write(nfd, buf, (size_t)n, WT_UTIME_NO_TIMEOUT);
    http_date(nfd);
    memset(buf, 0, sizeof(buf));
    strftime(buf, sizeof(buf), "Last-Modified: %a, %d %h %Y %H:%M:%S %Z\r\n",
	     gmtime(&sbuf.st_mtime));
    wt_netfd_write(nfd, buf, strlen(buf), WT_UTIME_NO_TIMEOUT);
    http_mime(nfd, uri);
    wt_netfd_write(nfd, CRLF, 2, WT_UTIME_NO_TIMEOUT);

#if	NO_SENDFILE
    memset(buf, 0, sizeof(buf));
    while ((n = wtNetfdRead(fd, buf, sizeof(buf))) > 0) {
      wtNetfdWrite(nfd, buf, (size_t)n);
      memset(buf, 0, sizeof(buf));
    }
#else
    {
      off_t	sbytes;
      wt_netfd_sendfile(fd, nfd, 0, 0, NULL, &sbytes, 0, WT_UTIME_NO_TIMEOUT);
#if	DEBUG
      fprintf(httpd.logger, "DEBUG: sendfile = %d\n", (int)sbytes);
#endif	/* DEBUG */
    }
#endif	/* NO_SENDFILE */

    wt_netfd_close(fd);
  }
  else {
    http_error(nfd, EST404);
  }
}

static void
http_mime(nfd, uri)
	register wt_netfd_t *	nfd;
	const char *		uri;
{
  const char	text_plain[] = "text/plain";
  const char	text_html[]  = "text/html";
  const char	text_css[]   = "text/css";
  const char	image_png[]  = "image/png";
  const char	image_gif[]  = "image/gif";
  const char	image_jpeg[] = "image/jpeg";
  char		buf[BUFSIZ];
  char *	p;

  p = strrchr(uri, '.');
  if (p == NULL) {
    return;
  }
  p++;

  memset(buf, 0, sizeof(buf));
  if (strcmp(p, "html") == 0 || strcmp(p, "htm") == 0) {
    snprintf(buf, sizeof(buf), "Content-Type: %s\r\n", text_html);
  }
  else if (strcmp(p, "css") == 0) {
    snprintf(buf, sizeof(buf), "Content-Type: %s\r\n", text_css);
  }
  else if (strcmp(p, "png") == 0) {
    snprintf(buf, sizeof(buf), "Content-Type: %s\r\n", image_png);
  }
  else if (strcmp(p, "gif") == 0) {
    snprintf(buf, sizeof(buf), "Content-Type: %s\r\n", image_gif);
  }
  else if (strcmp(p, "jpeg") == 0 || strcmp(p, "jpg") == 0) {
    snprintf(buf, sizeof(buf), "Content-Type: %s\r\n", image_jpeg);
  }
  else {
    snprintf(buf, sizeof(buf), "Content-Type: %s\r\n", text_plain);
  }
  wt_netfd_write(nfd, buf, strlen(buf), WT_UTIME_NO_TIMEOUT);
#if	DEBUG
  fprintf(httpd.logger, "DEBUG: URI: %s\t%s", uri, buf);
#endif	/* DEBUG */
}

static void
http_error(nfd, eno)
	register wt_netfd_t *	nfd;
	int			eno;
{
  struct {
    const char *	line;
    const char *	text;
  } emsg[] = {
    { "HTTP/1.0 400 Bad Request\r\nServer: httpd\r\n",
      "<html><head><title>Error</title></head><body><h1>Error 400</h1><p>Your request has bad syntax or is inherently impossible to satisy.</p></body></html>\r\n" },
    { "HTTP/1.0 403 Forbidden\r\nServer: httpd\r\n",
      "<html><head><title>Error</title></head><body><h1>Error 403</h1><p>You do not have permission to get the request URL.</p></body></html>\r\n" },
    { "HTTP/1.0 404 Not Found\r\nServer: httpd\r\n",
      "<html><head><title>Error</title></head><body><h1>Error 404</h1><p>Not found - file does not exist or is read protected</p></body></html>\r\n" },
    { "HTTP/1.0 405 Method Not Allowed\r\nServer: httpd\r\nAllow: HEAD,GET\r\n",
      "<html><head><title>Error</title></head><body><h1>Error 405</h1><p>This server answer only to GET and Head requests</p></body></html>\r\n" },
    { "HTTP/1.0 500 Internal Server Error\r\nServer: httpd\r\n",
      "<html><head><title>Error</title></head><body><h1>Error 500</h1><p>There was an unusual problem serving the requested URL.</p></body></html>\r\n" }
  };
  char	text_html[] = "Content-Type: text/html\r\n\r\n";

  wt_netfd_write(nfd, emsg[eno].line, strlen(emsg[eno].line),
		 WT_UTIME_NO_TIMEOUT);
  http_date(nfd);
  wt_netfd_write(nfd, text_html, strlen(text_html), WT_UTIME_NO_TIMEOUT);
  wt_netfd_write(nfd, CRLF, 2, WT_UTIME_NO_TIMEOUT);
  wt_netfd_write(nfd, emsg[eno].text, strlen(emsg[eno].text),
		 WT_UTIME_NO_TIMEOUT);
}

static void
http_date(nfd)
	register wt_netfd_t *	nfd;
{
  time_t	now = time(NULL);
  char		date[80];

  strftime(date, sizeof(date), "Date: %a, %d %h %Y %H:%M:%S %Z\r\n",
	   gmtime(&now));
  wt_netfd_write(nfd, date, strlen(date), WT_UTIME_NO_TIMEOUT);
}



static void
set_trap(void)
{
  struct sigaction	act;

  /* quit */
  sigemptyset(&act.sa_mask);
  act.sa_flags	 = 0;
  act.sa_handler = (void *)trap;
  sigaction(SIGQUIT, &act, NULL);
  sigaction(SIGTERM, &act, NULL);
  /**/
  sigaction(SIGBUS,  &act, NULL);
  sigaction(SIGSEGV, &act, NULL);

  /* print statistics */
  act.sa_handler = (void *)print_thread_count;
  sigaction(SIGINT,  &act, NULL);
}

static void
print_thread_count(signo)
	int	signo;
{
  print_cpu_time();
  wl_closelog(httpd.logger);

  exit(0);
}

static void
print_cpu_time(void)
{
  double	user;
  double	sys;
  struct rusage	myusage;
  struct rusage	childusage;

  if (getrusage(RUSAGE_SELF, &myusage) < 0) {
    perror("getrusage");
    exit(1);
  }
  if (getrusage(RUSAGE_CHILDREN, &childusage) < 0) {
    perror("getrusage");
    exit(1);
  }

  user = (double)myusage.ru_utime.tv_sec + myusage.ru_utime.tv_usec / 1000000.0;
  user += (double)childusage.ru_utime.tv_sec
				+ childusage.ru_utime.tv_usec / 1000000.0;
  sys = (double)myusage.ru_stime.tv_sec + myusage.ru_stime.tv_usec / 1000000.0;
  sys += (double)childusage.ru_stime.tv_sec
				+ childusage.ru_stime.tv_usec / 1000000.0;
  fprintf(httpd.logger, "\nuser time = %g, sys time = %g\n", user, sys);
}

static void
trap(signo)
	int	signo;
{
  fprintf(httpd.logger, "killed by sig#%d\n", signo);
  wl_closelog(httpd.logger);
  exit(0);
}

static void
usage(void)
{
  fprintf(stderr,
	  "usage: httpd [-a addr][-p port][-d docdir][-l logfile][-u uid]\n");
  exit(64);
}

int
main(argc, argv)
	int	argc;
	char *	argv[];
{
  int			ch;
  register ListenAddr *	lap;

  httpd.uid = getuid();
  while ((ch = getopt(argc, argv, "a:p:d:l:u:t:")) != -1) {
    switch (ch) {
      case 'a':
	httpd.addr	= strdup(optarg);
	break;
      case 'p':
	httpd.port	= strdup(optarg);
	break;
      case 'd':
	httpd.docdir	= strdup(optarg);
	break;
      case 'l':
	httpd.logfile	= strdup(optarg);
	break;
      case 'u':
	httpd.uid	= (uid_t)atoi(optarg);
	break;
      case 't':
	httpd.nthreads	= atoi(optarg);
	break;
      case '?':
      default:
	usage();
    }
  }
  argc -= optind;
  argv += optind;

  if (httpd.docdir == NULL) {
    httpd.docdir = getenv("HOME");
  }
  if (chdir(httpd.docdir) == -1) {
    perror("chdir");
    exit(1);
  }

  set_trap();

  if (wt_init(WT_VP_SS_MEDIUM) == -1) {
    perror("wt_init");
    exit(1);
  }

  httpd.logger = wl_openlog(httpd.logfile);
  if (httpd.logger == NULL) {
    perror("wl_openlog");
    exit(1);
  }

  new_listen("127.0.0.1", httpd.port);
  new_listen(httpd.addr, httpd.port);

  for (lap = httpd.listen; lap != NULL; lap = lap->next) {
    register int	i;

    lap->nsd = wt_netfd_tcp_server(lap->ipaddr, lap->port, 48);
    if (lap->nsd == NULL) {
      perror("wt_init_tcp_server");
      exit(1);
    }
    lap->addrlen = wt_netfd_socklen(lap->nsd);
    lap->thread  = (wt_thread_t *)calloc((size_t)httpd.nthreads,
					 sizeof(wt_thread_t));
    if (lap->thread == NULL) {
      perror("calloc");
      exit(1);
    }
    if (wt_thread_mutex_init(&lap->mlock) == -1) {
      perror("wt_thread_mutex_init");
      exit(1);
    }

    for (i = 0; i < httpd.nthreads; i++) {
      make_threads(lap, i);
    }
  }

  for (;;) {
    for (lap = httpd.listen; lap != NULL; lap = lap->next) {
      if (wt_netfd_poll(lap->nsd, WT_POLL_IN, WT_UTIME_NO_TIMEOUT) == 0) {
       	;	/* nothing to do */
      }
    }
  }

  return 0;	/* not reached */
}

