/*****************************************************************************
 *
 * FILE:	ping4.c
 * DESCRIPTION:	ping with WTL
 * DATE:	Wed, Jun 11 2003
 * UPDATED:	Sun, Jun 27 2004
 * AUTHOR:	Kouichi ABE (WALL) / $B0$It9/0l(B
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2003-2004 $B0$It9/0l!?(BKouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2003-2004 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: ping4.c,v 1.4 2004/11/13 16:38:22 kouichi Exp $
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/ip_icmp.h>
#include <errno.h>
#include "wt.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	PHDR_LEN	((int)sizeof(struct timeval))
#define	DEFDATALEN	(64 - PHDR_LEN)	/* default data length */
#define	MAXIPLEN	(sizeof(struct ip) + MAX_IPOPTLEN)
#define	MINICMPLEN	ICMP_MINLEN
#define	MAXPAYLOAD	(IP_MAXPACKET - MAXIPLEN - MINICMPLEN)
#define	PING_COUNT	(5)
#define	PING_TIMEOUT	(3)

typedef struct {
  void			(*fproc)(char *, ssize_t, struct timeval *);
  void			(*fsend)(void);
  struct sockaddr *	sasend;	/* sockaddr{} for send, from getaddrinfo */
  struct sockaddr *	sarecv;	/* sockaddr{} fro receiving */
  socklen_t		salen;	/* length of sockaddr{}s */
  int			icmpproto;	/* IPPROTO_XXX value for ICMP */
} ping_t;

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void		readloop(int);
static void		proc_v4(char *, ssize_t, struct timeval *);
static void		tv_sub(struct timeval *, struct timeval *);
static void		send_v4(void);
static unsigned short	in_cksum(unsigned short *, int);
static void		sig_alarm(int);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
/* data that goes with ICMP echo request */
static const int	datalen = DEFDATALEN;
static pid_t		pid;	/* our PID */
static wt_netfd_t *	nfd = NULL;

static ping_t	ping_v4 = {proc_v4, send_v4, NULL, NULL, 0, IPPROTO_ICMP};
static ping_t *	ping;

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/
int
main(argc, argv)
	int	argc;
	char *	argv[];
{
  const char *		host;
  struct addrinfo *	ai;

  if (argc != 2) {
    fprintf(stderr, "usage: %s <hostname>\n", argv[0]);
    exit(1);
  }
  host = argv[argc - 1];

  /* initialize thread engine */
  if (wt_init(WT_VP_SS_LARGE) == -1) {
    perror("wt_init");
    exit(1);
  }

  ai = wt_host_serv(host, NULL, 0, 0);
  if (ai) {
    fprintf(stderr, "PING %s (%s) %d data bytes\n",
	    ai->ai_canonname ? ai->ai_canonname : host,
	    wt_sock_ntop(ai->ai_addr, false), datalen);

    /* initialize according to protocol */
    if (ai->ai_family == AF_INET) {
      ping = &ping_v4;
    }
    else {
      fprintf(stderr, "Error: unknown address family %d\n", ai->ai_family);
      exit(1);
    }

    ping->sasend = ai->ai_addr;
    ping->sarecv = calloc(1, ai->ai_addrlen);
    ping->salen  = ai->ai_addrlen;

    readloop(PING_COUNT);
  }

  return 0;
}

static void
readloop(max)
	int	max;	/* max count */
{
  int	soc;
  int	size;

  soc = socket(ping->sasend->sa_family, SOCK_RAW, ping->icmpproto);
  if (soc == -1) {
    perror("socket");
    exit(1);
  }
  setuid(getuid());	/* don't need special permissions any more */

  size = IP_MAXPACKET + 128;	/* OK if setsockopt fails */
  (void)setsockopt(soc, SOL_SOCKET, SO_RCVBUF, &size, sizeof(size));
  nfd = wt_netfd_fdopen(soc);

  pid = getpid() & 0xffff;
  signal(SIGALRM, sig_alarm);
  do {
    struct timeval	interval = { 1L, 0L };
    struct timeval	value	 = { 1L, 0L };
    struct itimerval	itv	 = { interval, value };

    setitimer(ITIMER_REAL, &itv, NULL);
  } while (0);

  for (;max > 0; max--) {
    socklen_t		len;
    unsigned char	recvbuf[MINICMPLEN + MAXPAYLOAD];
    ssize_t		n;
    struct timeval	tval;
    
    len = ping->salen;
    n = wt_netfd_recvfrom(nfd, recvbuf, sizeof(recvbuf), 0, ping->sarecv, &len,
			  WT_UTIME_SEC(PING_TIMEOUT));
    if (n < 0) {
      if (errno == EINTR) {
	continue;
      }
      else {
	perror("wt_netfd_revfrom");
	exit(1);
      }
    }
    gettimeofday(&tval, NULL);
    (*ping->fproc)(recvbuf, n, &tval);
  }
  wt_netfd_close(nfd);
}

static void
proc_v4(ptr, len, tvrecv)
	char *			ptr;
	ssize_t			len;
	struct timeval *	tvrecv;
{
  struct ip *		ip;
  int			hlen1;
  struct icmp *		icmp;
  int			icmplen;

  ip	= (struct ip *)ptr;	/* start of IP header */
  hlen1	= ip->ip_hl << 2;	/* length of IP header */

  icmp	= (struct icmp *)(ptr + hlen1);	/* start of ICMP header */
  icmplen = len - hlen1;
  if (icmplen < 8) {
    fprintf(stderr, "icmplen (%d) < 8\n", icmplen);
    exit(1);
  }

  if (icmp->icmp_type == ICMP_ECHOREPLY) {
    struct timeval *	tvsend;
    double		rtt;

    if (icmp->icmp_id != pid) {
      return;	/* not a response to our ECHO_REQUEST */
    }
    if (icmplen < 16) {
      fprintf(stderr, "icmplen (%d) < 16\n", icmplen);
      exit(1);
    }

    tvsend = (struct timeval *)icmp->icmp_data;
    tv_sub(tvrecv, tvsend);
    rtt	= tvrecv->tv_sec * 1000.0 + tvrecv->tv_usec / 1000.0;

    fprintf(stdout, "%d bytes from %s: seq=%u, ttl=%d, rtt=%.3f ms\n",
	    icmplen, wt_sock_ntop(ping->sarecv, false),
	    icmp->icmp_seq, ip->ip_ttl, rtt);
  }
}

static void
tv_sub(out, in)
	register struct timeval *	out;
	register struct timeval *	in;
{
  out->tv_usec -= in->tv_usec;
  if (out->tv_usec < 0) {
    --out->tv_sec;
    out->tv_usec += 1000000;
  }
  out->tv_sec -= in->tv_sec;
}

static void
send_v4(void)
{
  if (nfd) {
    static int		nsent = 0;	/* add 1 for each sendto() */
    unsigned char	sendbuf[MINICMPLEN + MAXPAYLOAD];
    int			len;
    struct icmp *	icmp;

    icmp = (struct icmp *)sendbuf;
    icmp->icmp_type	= ICMP_ECHO;
    icmp->icmp_code	= 0;
    icmp->icmp_cksum	= 0;
    icmp->icmp_id	= pid;
    icmp->icmp_seq	= nsent++;
    gettimeofday((struct timeval *)icmp->icmp_data, NULL);

    len = MINICMPLEN + datalen;	/* checksum ICMP header and data */
    icmp->icmp_cksum = in_cksum((unsigned short *)icmp, len);

    (void)wt_netfd_sendto(nfd, sendbuf, (size_t)len, 0, ping->sasend,
			  ping->salen, WT_UTIME_SEC(PING_TIMEOUT));
  }
}

static unsigned short
in_cksum(addr, len)
	unsigned short *	addr;
	int			len;
{
  int			nleft	= len;
  int			sum	= 0;
  unsigned short *	w	= addr;
  unsigned short	answer	= 0;

  /*
   * Our algorithm is simple, using a 32 bit accumulator (sum),
   * we add sequential 16 bit words to it, and at the end,
   * fold back all the carry bits from the top 16 bits into the lower 16 bits.
   */
  while (nleft > 1) {
    sum	  += *w++;
    nleft -= 2;
  }

  /* 4mop up an odd byte, if neccessary */
  if (nleft == 1) {
#if	0
    *(unsigned char *)(&answer) = *(unsigned char *)w;
    sum += answer;
#else
    union {
      unsigned short	us;
      unsigned char	uc[2];
    } last;
    last.uc[0] = *(unsigned char *)w;
    last.uc[1] = 0;
    sum += last.us;
#endif
  }

  sum	 = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
  sum	+= (sum >> 16);			/* add carry */
  answer = ~sum;			/* truncate to 16 bits */

  return answer;
}

static void
sig_alarm(signo)
	int	signo;
{
  (*ping->fsend)();
}

