/*****************************************************************************
 *
 * FILE:	io.c
 * DESCRIPTION:	WTL: I/O functions
 * DATE:	Thu, Mar 27 2003
 * UPDATED:	Wed, Sep  5 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2003-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2003-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: io.c,v 1.59 2007/09/04 17:09:03 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#if	HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif	/* HAVE_SYS_TYPES_H */
#if	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif	/* HAVE_SYS_TIME_H */
#if	HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif	/* HAVE_SYS_RESOURCE_H */
#include <sys/mman.h>
#include <signal.h>
#include <errno.h>
#if	HAVE_SYS_SENDFILE
#include <sys/sendfile.h>
#endif	/* HAVE_SYS_SENDFILE */
#if	ENABLE_SSL
#include "wt-ssl.h"
#else
#include "wt.h"
#endif	/* ENABLE_SSL */
#include "wt-sub.h"

/******************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
/* network file descriptor object list */
static wt_netfd_t *	wt_netfd_freelist = NULL;
/* maximum number of file descriptors that the process can open */
static int		wt_osfd_limit = -1;

/******************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static wt_netfd_t *	wt_netfd_new(int osfd, bool nonblock);
static void		wt_netfd_free(wt_netfd_t * nfd);
static ssize_t		one_read(wt_netfd_t *, char *, wt_utime_t);
static int		wt_netfd_readbuf(wt_netfd_t *, wt_utime_t);

/******************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wt_init_io
 * Description:		initialize I/O
 * Arguments:		none
 * Return Value:	successful: max files, otherwise: -1
 */
int
wt_init_io(void)
{
  struct sigaction	sa;
  struct rlimit		rl;

  /* ignore SIGPIPE */
  sigemptyset(&sa.sa_mask);
  sa.sa_flags	= SA_RESTART;
  sa.sa_handler = SIG_IGN;
  if (sigaction(SIGPIPE, &sa, NULL) == -1) {
    return -1;
  }

  /* set maximum number of open file descriptors */
  if (getrlimit(RLIMIT_NOFILE, &rl) == -1) {
    return -1;
  }
  rl.rlim_cur = rl.rlim_max;
  if (setrlimit(RLIMIT_NOFILE, &rl) == -1) {
    return -1;
  }
  wt_osfd_limit = (int)rl.rlim_max;

  return wt_osfd_limit;
}

/*****************************************************************************/

/*
 * Function Name:	wt_netfd_open
 * Description:		open file
 * Arguments:		path, flag, mode
 * Return Value:	successful: netfd, otherwise: NULL
 */
wt_netfd_t *
wt_netfd_open(path, flags, mode)
	const char *	path;
	int		flags;
	mode_t		mode;
{
  int		fd;
  wt_netfd_t *	newfd;

  while ((fd = open(path, flags|O_NONBLOCK, mode)) < 0) {
    if (errno != EINTR) {
      return NULL;
    }
  }

  newfd = wt_netfd_new(fd, false);
  if (newfd == NULL) {
    int	saved_errno = errno;
    close(fd);
    errno = saved_errno;
  }

  return newfd;
}

/*
 * Function Name:	wt_netfd_fdopen
 * Description:		associate a netfd with the existing descriptor
 * Arguments:		descriptor
 * Return Value:	successful: new netfd, otherwise: NULL
 */
wt_netfd_t *
wt_netfd_fdopen(osfd)
	int	osfd;
{
  return wt_netfd_new(osfd, true);
}

/*
 * Function Name:	wt_netfd_close
 * Description:		close netfd
 * Arguments:		netfd
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_netfd_close(nfd)
	register wt_netfd_t *	nfd;
{
  int	status;

  status = close(nfd->osfd);
  wt_netfd_free(nfd);

  return status;
}

/*
 * Function Name:	wt_netfd_seek
 * Description:		reposition read/write file offset
 * Arguments:		netfd, offset, whence
 * Return Value:	successful: offset, otherwise: -1
 */
off_t
wt_netfd_seek(nfd, offset, whence)
	register wt_netfd_t *	nfd;
	off_t			offset;
	int			whence;
{
  if (nfd->saddr != NULL) {
    errno = EINVAL;
    return -1;
  }
  return lseek(nfd->osfd, offset, whence);
}

/*
 * Function Name:	wt_netfd_read0
 * Description:		read data without inner netfd buffer
 * Arguments:		netfd, buffer, buffer size, timeout
 * Return Value:	successful: read size, otherwise: -1
 */
ssize_t
wt_netfd_read0(nfd, buf, nbytes, timeout)
	wt_netfd_t *	nfd;
	void *		buf;
	size_t		nbytes;
	wt_utime_t	timeout;
{
  ssize_t	n;

  while ((n = read(nfd->osfd, buf, nbytes)) < 0) {
    if (errno == EINTR) {
      continue;
    }
    if (!WT_IO_NOT_READY) {
      return -1;
    }
    if (wt_netfd_poll(nfd, WT_POLL_IN, timeout) == -1) {
      return -1;
    }
  }

  return n;
}

/*
 * Function Name:	wt_netfd_read
 * Description:		read data
 * Arguments:		netfd, buffer, buffer size, timeout
 * Return Value:	successful: read size, otherwise: -1
 */
ssize_t
wt_netfd_read(nfd, buf, nbytes, timeout)
	wt_netfd_t *	nfd;
	void *		buf;
	size_t		nbytes;
	wt_utime_t	timeout;
{
  register ssize_t	n   = 0;
  register char *	ptr = (char *)buf;

  do {
    if (wt_netfd_readbuf(nfd, timeout) < 0) {
      return -1;
    }
    if (nfd->rb.total == 0) {
      break;
    }
    *ptr++ = *nfd->rb.bufp++;
    nfd->rb.total--;
  } while ((size_t)n++ < nbytes && nfd->rb.total > 0);

  return n;
}

/*
 * Function Name:	wt_netfd_write
 * Description:		write data
 * Arguments:		netfd, buffer, buffer size, timeout
 * Return Value:	successful: printed size, otherwise: -1
 */
ssize_t
wt_netfd_write(nfd, buf, nbytes, timeout)
	wt_netfd_t *	nfd;
	const void *	buf;
	size_t		nbytes;
	wt_utime_t	timeout;
{
  size_t	resid = nbytes;
  ssize_t	n;

  while (resid > 0) {
    n = write(nfd->osfd, buf, resid);
    if (n == -1) {
      if (errno == EINTR) {
	continue;
      }
      if (!WT_IO_NOT_READY) {
       	return -1;
      }
      /* wait until the socket becomes writable */
      if (wt_netfd_poll(nfd, WT_POLL_OUT, timeout) == -1) {
	return -1;
      }
    }
    else {
      resid -= n;
      if (resid == 0) {
	break;
      }
      buf = (const void *)((const char *)buf + n);
    }
  }

  return (ssize_t)(nbytes - resid);
}

/*****************************************************************************/

/*
 * Function Name:	wt_netfd_accept
 * Description:		accept a connection on a socket
 * Arguments:		netfd, addr, addrlen, timeout
 * Return Value:	successful: new netfd, otherwise: NULL
 */
wt_netfd_t *
wt_netfd_accept(nfd, addr, addrlen, timeout)
	wt_netfd_t *		nfd;
	struct sockaddr *	addr;
	socklen_t *		addrlen;
	wt_utime_t		timeout;
{
  int		sd;
  wt_netfd_t *	newfd;

  while ((sd = accept(nfd->osfd, addr, (socklen_t *)addrlen)) < 0) {
    if (errno == EINTR) {
      continue;
    }
    if (!WT_IO_NOT_READY) {
      return NULL;
    }
    /* wait until the socket becomes readable */
    if (wt_netfd_poll(nfd, WT_POLL_IN, timeout) == -1) {
      return NULL;
    }
  }

  newfd = wt_netfd_new(sd, true);
  if (newfd == NULL) {
    int	saved_errno = errno;
    close(sd);
    errno = saved_errno;
  }

  return newfd;
}

/*
 * Function Name:	wt_netfd_connect
 * Description:		initiate a connection on a socket
 * Arguments:		netfd, addr, addrlen, timeout
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_netfd_connect(nfd, addr, addrlen, timeout)
	wt_netfd_t *		nfd;
	const struct sockaddr * addr;
	socklen_t		addrlen;
	wt_utime_t		timeout;
{
  int	err = 0;
  while (connect(nfd->osfd, addr, addrlen) < 0) {
    if (errno != EINTR) {
      int	n;

      if (errno != EINPROGRESS && (errno != EADDRINUSE || err == 0)) {
	return -1;
      }
      /* wait until the socket becomes writable */
      if (wt_netfd_poll(nfd, WT_POLL_OUT, timeout) == -1) {
	return -1;
      }
      /* try to find out whether the connection setup succeeded or failed */
      n = sizeof(int);
      if (getsockopt(nfd->osfd, SOL_SOCKET, SO_ERROR, (char *)&err,
		     (socklen_t *)&n) == -1) {
	return -1;
      }
      if (err) {
	errno = err;
	return -1;
      }
      break;
    }
    err = 1;	/* interrupted */
  }

  return 0;
}

/*
 * Function Name:	wt_netfd_recvfrom
 * Description:		receive message
 * Arguments:		netfd, message, message length, flags (unused),
 *			from, fromlen, timeout
 * Return Value:	successful: received size, otherwise: -1
 */
ssize_t
wt_netfd_recvfrom(nfd, msg, len, flags, from, fromlen, timeout)
	wt_netfd_t *		nfd;
	const void *		msg;
	size_t			len;
	int			flags;
	struct sockaddr *	from;
	socklen_t *		fromlen;
	wt_utime_t		timeout;
{
  ssize_t	n;

  while ((n = recvfrom(nfd->osfd, (char *)msg, len, flags,
		       from, fromlen)) < 0) {
    if (errno == EINTR) {
      continue;
    }
    if (!WT_IO_NOT_READY) {
      return -1;
    }
    /* wait until the socket becomes readable */
    if (wt_netfd_poll(nfd, WT_POLL_IN, timeout) == -1) {
      return -1;
    }
  }

  return n;
}

/*
 * Function Name:	wt_netfd_sendto
 * Description:		send message
 * Arguments:		netfd, message, message length, flags (unused),
 *			to, tolen, timeout
 * Return Value:	successful: sent size, otherwise: -1
 */
ssize_t
wt_netfd_sendto(nfd, msg, len, flags, to, tolen, timeout)
	wt_netfd_t *		nfd;
	const void *		msg;
	size_t			len;
	int			flags;
	const struct sockaddr * to;
	socklen_t		tolen;
	wt_utime_t		timeout;
{
  ssize_t	n;

  while ((n = sendto(nfd->osfd, msg, len, flags, to, tolen)) < 0) {
    if (errno == EINTR) {
      continue;
    }
    if (!WT_IO_NOT_READY) {
      return -1;
    }
    /* wait until the socket becomes writable */
    if (wt_netfd_poll(nfd, WT_POLL_OUT, timeout) == -1) {
      return -1;
    }
  }

  return n;
}

/*
 * Function Name:	wt_netfd_sendfile
 * Description:		senf a file to a socket
 * Arguments:		nfd, nsd, offset, nbytes, hdtr, sbytes,
 *			flags (always: 0), timeout
 * Return Value:	successful: 0, otherwise: -1
 */
#if	HAVE_FREEBSD_SENDFILE
int
wt_netfd_sendfile(nfd, nsd, init_offset, nbytes, hdtr, ret_sbytes, flags,
		  timeout)
	wt_netfd_t *		nfd;
	wt_netfd_t *		nsd;
	off_t			init_offset;
	size_t			nbytes;
	struct sf_hdtr *	hdtr;
	off_t *			ret_sbytes;
	int			flags;
	wt_utime_t		timeout;
{
  off_t	offset;
  off_t	sbytes;

  offset = init_offset;
  while (sendfile(nfd->osfd, nsd->osfd, offset, nbytes, hdtr, &sbytes,
		  flags) < 0) {
    offset	+= sbytes;
    nbytes	-= sbytes;

    if (errno == EINTR) {
      continue;
    }
    if (!WT_IO_NOT_READY) {
      return -1;
    }
    /* wait until the socket becomes writable */
    if (wt_netfd_poll(nfd, WT_POLL_OUT, timeout) == -1) {
      return -1;
    }
  }
  if (ret_sbytes != NULL) {
    *ret_sbytes = offset + sbytes;
  }

  return 0;
}
#else
#if	HAVE_LINUX_SENDFILE
/*
 * I have no linux environments. Is this routine ok?
 */
int
wt_netfd_sendfile(nfd, nsd, init_offset, nbytes, hdtr, ret_sbytes, flags,
		  timeout)
	wt_netfd_t *	nfd;
	wt_netfd_t *	nsd;
	off_t		init_offset;
	size_t		nbytes;
	void *		hdtr;
	off_t *		ret_sbytes;
	int		flags;
	wt_utime_t	timeout;
{
  off_t	offset;
  off_t	sbytes;

  offset = init_offset;
  while (sendfile(nfd->osfd, nsd->osfd, &sbytes, nbytes) < 0) {
    offset	+= sbytes;
    nbytes	-= sbytes;

    if (errno == EINTR) {
      continue;
    }
    if (!WT_IO_NOT_READY) {
      return -1;
    }
    /* wait until the socket becomes writable */
    if (wt_netfd_poll(nfd, WT_POLL_OUT, timeout) == -1) {
      return -1;
    }
  }
  if (ret_sbytes != NULL) {
    *ret_sbytes = offset + sbytes;
  }

  return 0;
}
#else	/* No sendfile(2) */
int
wt_netfd_sendfile(nfd, nsd, offset, nbytes, hdtr, sbytes, flags, timeout)
	wt_netfd_t *	nfd;
	wt_netfd_t *	nsd;
	off_t		offset;
	size_t		nbytes;
	void *		hdtr;
	off_t *		sbytes;
	int		flags;
	wt_utime_t	timeout;
{
  void *	buf;
  size_t	len;
  size_t	n;
  int		status = 0;

  if (nbytes == 0) {
    /* get whole file size */
    len = (size_t)wt_netfd_seek(nfd, 0, WT_SEEK_END);
    if (len == -1) {
      return -1;
    }
  }
  else {
    len = nbytes;
  }
  buf = mmap(NULL, len, PROT_READ, MAP_PRIVATE, nfd->osfd, offset);
  if (buf == MAP_FAILED) {
    return -1;
  }
  n = wt_netfd_write(nsd, buf, len, timeout);
  if (n == -1) {
    status = -1;
  }
  if (sbytes != NULL) {
    *sbytes = n;
  }
  munmap(buf, len);

  return status;
}
#endif	/* HAVE_LINUX_SENDFILE */
#endif	/* HAVE_FREEBSD_SENDFILE */

/*
 * Function Name:	wt_netfd_readline
 * Description:		read a line
 * Arguments:		netfd, buffer, max buffer length, timeout
 * Return Value:	successful: read size, otherwise: -1
 */
ssize_t
wt_netfd_readline(nfd, buf, maxlen, timeout)
	wt_netfd_t *	nfd;
	void *		buf;
	size_t		maxlen;
	wt_utime_t	timeout;
{
  register ssize_t	n;
  register char *	ptr = (char *)buf;

  for (n = 1; (size_t)n < maxlen; n++) {
    char *	c;
    int		status;

    c = (char *)alloca(sizeof(char));
    if (c == NULL) { return -1; }

    status = one_read(nfd, c, timeout);
    if (status == 1) {
      *ptr++ = *c;
      if (*c == LF)  { break; }
      if (*c == EOL) { return 0; }	/* EOF?, null data read */
    }
    else if (status == 0) {
      if (n == 1) { return 0; }		/* EOF, no data read */
      else	  { break; }		/* EOF, some data was read */
    }
    else {
      return -1;	/* error */
    }
  }
  *ptr = EOL;

  return n;
}

static ssize_t
one_read(nfd, ptr, timeout)
	wt_netfd_t *	nfd;
	char *		ptr;
	wt_utime_t	timeout;
{
  if (wt_netfd_readbuf(nfd, timeout) < 0) {
    return -1;
  }
  if (nfd->rb.total == 0) {
    return 0;
  }
  nfd->rb.total--;
  *ptr = *nfd->rb.bufp++;

  return 1;
}

static int
wt_netfd_readbuf(nfd, timeout)
	wt_netfd_t *	nfd;
	wt_utime_t	timeout;
{
  if (nfd->rb.total <= 0) {
    memset(nfd->rb.buf, 0, sizeof(nfd->rb.buf));
    while ((nfd->rb.total = read(nfd->osfd,
				 nfd->rb.buf, sizeof(nfd->rb.buf))) < 0) {
      if (errno == EINTR) {
	continue;
      }
      if (!WT_IO_NOT_READY) {
	return -1;
      }
      /* wait until the socket becomes readable */
      if (wt_netfd_poll(nfd, WT_POLL_IN, timeout) == -1) {
	return -1;
      }
    }
    nfd->rb.bufp = nfd->rb.buf;
  }

  return nfd->rb.total;
}

/*****************************************************************************/

static wt_netfd_t *
wt_netfd_new(osfd, nonblock)
	int	osfd;
	bool	nonblock;
{
  wt_netfd_t *	nfd;

  if (wt_netfd_freelist != NULL) {
    nfd		      = wt_netfd_freelist;
    wt_netfd_freelist = wt_netfd_freelist->next;
  }
  else {
    nfd = (wt_netfd_t *)calloc(1, sizeof(wt_netfd_t));
  }

  if (nfd != NULL) {
    nfd->osfd	  = osfd;
    nfd->inuse	  = true;
    nfd->saddr	  = NULL;
    nfd->addrlen  = 0;
    nfd->udpath	  = NULL;
    nfd->rb.total = 0;
    nfd->rb.bufp  = nfd->rb.buf;
    memset(nfd->rb.buf, 0, sizeof(nfd->rb.buf));
    nfd->next	  = NULL;

    if (nonblock) {
      int	flags;

      if ((flags = fcntl(osfd, F_GETFL, 0)) == -1 ||
	  fcntl(osfd, F_SETFL, flags|O_NONBLOCK) == -1) {
	wt_netfd_free(nfd);
	return NULL;
      }
    }
  }

  return nfd;
}

static void
wt_netfd_free(nfd)
	register wt_netfd_t *	nfd;
{
  if (nfd->inuse) {
    nfd->osfd  = -1;
    nfd->inuse = false;
    wt_free(nfd->saddr);
    nfd->addrlen = 0;
    if (nfd->udpath != NULL) {
      unlink(nfd->udpath);
      free((char *)nfd->udpath);
      nfd->udpath = NULL;
    }
    nfd->rb.total = 0;
    nfd->rb.bufp  = nfd->rb.buf;
    memset(nfd->rb.buf, 0, sizeof(nfd->rb.buf));
    nfd->next = wt_netfd_freelist;
    wt_netfd_freelist = nfd;
  }
}

/*****************************************************************************/

/*
 * Function Name:	wt_netfd_poll
 * Description:		wait for I/O on a single descriptor
 * Arguments:		netfd, how, timeout
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_netfd_poll(nfd, how, timeout)
	wt_netfd_t *	nfd;
	unsigned short	how;
	wt_utime_t	timeout;
{
  wt_pollfd_t	pfd;
  int		status = 0;

  /* set pollfd */
  pfd.fd	= nfd->osfd;
  pfd.events	= how;
  pfd.revents	= WT_POLL_NONE;

  switch (wt_poll(&pfd, timeout)) {
    case 0:	/* time out */
      errno = ETIMEDOUT;
    case -1:	/* error */
      status = -1;
    default:
      if (pfd.revents & WT_POLL_INVAL) {
	errno  = EBADF;
	status = -1;
      }
      break;
  }

  return status;
}
