/*****************************************************************************
 *
 * FILE:	key.c
 * DESCRIPTION:	WTL: thread private key functions
 * DATE:	Thu, Mar 27 2003
 * UPDATED:	Sun, Aug 19 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2003-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2003-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: key.c,v 1.13 2007/08/19 05:57:22 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <errno.h>
#if	ENABLE_SSL
#include "wt-ssl.h"
#else
#include "wt.h"
#endif	/* ENABLE_SSL */
#include "wt-sub.h"

/******************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
			/* destructor table for per-thread private data */
static wt_destructor_t	destructors[WT_KEYS_MAX];
static wt_key_t		key_max = 0;

/******************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wt_thread_key_create
 * Description:		create a key to be used for thread specific data
 * Arguments:		pointer of private key, destructor
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_thread_key_create(key, destructor)
	wt_key_t *	key;
	wt_destructor_t	destructor;
{
  if (key_max >= WT_KEYS_MAX) {
    errno = EAGAIN;
    return -1;
  }

  *key = key_max++;
  destructors[*key] = destructor;

  return 0;
}

/*
 * Function Name:	wt_thread_key_delete
 * Description:		delete thread specific data pointed by key
 * Arguments:		private key
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_thread_key_delete(key)
	wt_key_t	key;
{
  wt_thread_t	me = wt_current_thread;

  if (key >= key_max) {
    errno = EINVAL;
    return -1;
  }

  if (me->data[key] != NULL && destructors[key] != NULL) {
    (*destructors[key])(me->data[key]);
    me->data[key] = NULL;
  }

  return 0;
}

/*
 * Function Name:	wt_thread_getspecific
 * Description:		get thread specific data pointed by key
 * Arguments:		private key
 * Return Value:	successful: private data, otherwise: NULL
 */
void *	
wt_thread_getspecific(key)
	wt_key_t	key;
{
  wt_thread_t	me = wt_current_thread;

  if (key >= key_max) {
    return NULL;
  }

  return me->data[key];
}

/*
 * Function Name:	wt_thread_setspecific
 * Description:		set thread specific data with key
 * Arguments:		private key, private value
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_thread_setspecific(key, value)
	wt_key_t	key;
	void *		value;
{
  wt_thread_t	me = wt_current_thread;

  if (key >= key_max) {
    errno = EINVAL;
    return -1;
  }

  if (me->data[key] != value) {
    /* free previously set non-NULL data value */
    if (me->data[key] != NULL && destructors[key] != NULL) {
      (*destructors[key])(me->data[key]);
    }
    me->data[key] = value;
  }

  return 0;
}

/*
 * Function Name:	wt_thread_cleanup
 * Description:		free up all per-thread private data
 * Arguments:		thread
 * Return Value:	none
 */
void
wt_thread_cleanup(thd)
	register wt_thread_t	thd;
{
  register int	key;

  for (key = 0; key < WT_KEYS_MAX; key++) {
    if (thd->data[key] != NULL && destructors[key] != NULL) {
      (*destructors[key])(thd->data[key]);
      thd->data[key] = NULL;
    }
  }
}
