/*****************************************************************************
 *
 * FILE:	log.c
 * DESCRIPTION:	Wyvern: log library
 * DATE:	Tue, Jun  3 2003
 * UPDATED:	Wed, Mar 30 2005
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2003-2005 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2003-2005 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: log.c,v 1.11 2005/03/31 14:49:23 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#if	HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif	/* HAVE_NETINET_IN_H */
#if	HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif	/* HAVE_ARPA_INET_H */
#if	HAVE_NETDB_H
#include <netdb.h>
#endif	/* HAVE_NETDB_H */
#include <errno.h>
#include "wl.h"

/******************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	PATH_BIN_SH	"/bin/sh"
#define	SECURE_PATH	"/bin:/usr/bin"

#ifndef	HAVE_SETLINEBUF
#define	setlinebuf(x)	setvbuf((x), (char *)NULL, _IOLBF, 0)
#endif	/* HAVE_SETLINEBUF */

/******************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static int	open_file_log(const char *, int);
static int	open_pipe_log(const char *);
static int	open_net_log(const char *);

/******************************************************************************
 *
 *	Local functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wl_openlog
 * Description:		open log stream
 * Arguments:		path for log stream
 * Return Value:	successful: file pointer, otherwise: NULL
 */
FILE *
wl_openlog(path)
	const char *	path;
{
  register int		fd;
  register FILE *	fp;

  if (path[0] == '|') {
    fd = open_pipe_log(path + 1);
  }
  else if (path[0] == ':') {
    fd = open_net_log(path + 1);
  }
  else {
    int	flags;

#if	O_NOFOLLOW
#if	O_EXLOCK
    flags = O_WRONLY|O_CREAT|O_APPEND|O_EXLOCK|O_NOFOLLOW;
#else
    flags = O_WRONLY|O_CREAT|O_APPEND|O_NOFOLLOW;
#endif	/* O_EXLOCK */
#else
#if	O_EXLOCK
    flags = O_WRONLY|O_CREAT|O_APPEND|O_EXLOCK;
#else
    flags = O_WRONLY|O_CREAT|O_APPEND;
#endif	/* O_EXLOCK */
#endif	/* O_NOFOLLOW */
    fd = open_file_log(path, flags);
    if (fd == -1 && errno == EOPNOTSUPP) {	/* for /dev/null */
#if	O_EXLOCK
      flags &= ~O_EXLOCK;
#endif	/* O_EXLOCK */
      fd = open_file_log(path, flags);
    }
  }

  if (fd == -1) {
    return NULL;
  }

  fp = fdopen(fd, "w");
  if (fp == NULL) {
    return NULL;
  }
  setlinebuf(fp);

  return fp;
}

static int
open_file_log(path, flags)
	const char *	path;
	int		flags;
{
  int	fd;

#if	O_NOFOLLOW
  fd = open(path, flags, 0644);
#else
  struct stat	sbuf;

  if (lstat(path, &sbuf) == 0 && S_ISLNK(sbuf.st_mode)) {
    fd = -1;
  }
  else {
    fd = open(path, flags, 0644);
  }
#endif	/* O_NOFOLLOW */

  return fd;
}

#define	PIPE_READER	(0)
#define	PIPE_WRITER	(1)
static int
open_pipe_log(cmd)
	const char *	cmd;
{
  int	pfds[2];

  if (pipe(pfds) == -1) {
    return -1;
  }

  switch (fork()) {
    case -1:
      close(pfds[PIPE_READER]);
      close(pfds[PIPE_WRITER]);
      return -1;
    case 0: {	/* child */
	char *	args[4];
	char *	env[] = { "PATH=" SECURE_PATH, NULL };

	args[0] = "sh";
	args[1] = "-c";
	args[2] = (char *)cmd;
	args[3] = NULL;

	close(pfds[PIPE_WRITER]);
	if (pfds[PIPE_READER] != STDIN_FILENO) {
	  dup2(pfds[PIPE_READER], STDIN_FILENO);
	  close(pfds[PIPE_READER]);
	}
	execve(PATH_BIN_SH, args, env);
	_exit(127);
      }
    default:	/* parent */
      close(pfds[PIPE_READER]);
      return pfds[PIPE_WRITER];
  }
}

static int
open_net_log(node)
	const char *	node;
{
  char *		host;
  char *		port;
  struct addrinfo	hints;
  struct addrinfo *	res;
  struct addrinfo *	res_save;
  int			sd;

  host = strdup(node);
  if (host == NULL) {
    return -1;
  }
  port = strchr(host, ':');
  if (port == NULL) {
    goto error;
  }
  *port++ = '\0';

  memset((char *)&hints, 0, sizeof(hints));
  hints.ai_flags	= AI_CANONNAME;
  hints.ai_family	= AF_UNSPEC;
  hints.ai_socktype	= SOCK_STREAM;
  hints.ai_protocol	= IPPROTO_TCP;

  if (getaddrinfo(host, port, &hints, &res) != 0) {
    goto error;
  }

  res_save = res;

  do {
    sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if (sd == -1) {
      continue;
    }
    if (connect(sd, res->ai_addr, res->ai_addrlen) == 0) {
      break;
    }
    close(sd);
  } while ((res = res->ai_next) != NULL);

  if (res == NULL) {
    sd = -1;
  }
  freeaddrinfo(res_save);

  return sd;

error:
  free(host);
  return -1;
}

/*
 * Function Name:	wl_closelog
 * Description:		close log stream
 * Arguments:		log stream pointer
 * Return Value:	successful: 0, otherwise: -1
 */
int
wl_closelog(lp)
	register FILE *	lp;
{
  if (lp == NULL) {
    errno = EINVAL;
    return -1;
  }

  return fclose(lp);
}

