/*****************************************************************************
 *
 * FILE:	module.c
 * DESCRIPTION:	Wyvern: plugable module manager
 * DATE:	Wed, Sep 13 2000
 * UPDATED:	Sat, Feb 11 2006
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2000-2006 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2000-2006 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: module.c,v 1.3 2006/02/11 09:57:26 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <dlfcn.h>
#include <errno.h>
#include "wm.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
typedef struct _ModuleList {
  struct _ModuleList *	prev;
  struct _ModuleList *	next;
  const wmModule *	module;
} ModuleList;

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void *		load(const char *);
static void *		fetch(void *);
static int		verify(const wmModule *);
static ModuleList *	lookup(const char *);
static ModuleList *	make_new_module_list(const wmModule *);
static void		install(ModuleList *);
static int		unload(void *);
static void		uninstall(ModuleList *);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static ModuleList	moduleList = { &moduleList, &moduleList, NULL };

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

int
wm_load_module(path)
	const char *	path;
{
  void *	handle;
  wmModule *	module;

  if (path == NULL) { return -1; }

  /* load the module */
  handle = load(path);
  if (handle == NULL) {
    goto error1;
  }

  /* fetch the symbol function */
  module = (wmModule *)fetch(handle);
  if (module == NULL) {
    goto error;
  }

  /* verify the module */
  if (verify(module) == -1) {
    goto error;
  }

  /* install the module */
  module->handle = handle;
  install(make_new_module_list(module));

  return 0;

error:
  unload(handle);
error1:
  return -1;
}

static void *
load(path)
	const char *	path;
{
  void *	handle;

  if (path[0] != '/') {
#if	0
    fprintf(stderr, "error: No absolute path: %s\n", path);
#endif
    errno = EACCES;
    return NULL;
  }

  handle = dlopen(path, RTLD_NOW|RTLD_LOCAL);
  if (handle == NULL) {
#if	0
    fprintf(stderr, "error: %s: %s\n", path, (const char *)dlerror());
#endif
    errno = ENOENT;
    return NULL;
  }

  return handle;
}

static void *
fetch(handle)
	void *	handle;
{
  void *	(*func)(void);
#if defined(USCORE) && !defined(DLSYM_ADDS_USCORE)
  char *	usymb;

  if (asprintf(&usymb, "_%s", WML_MODULE_DLSYM_NAME) == -1) {
    return NULL;
  }
  func = dlsym(handle, usymb);
  free((char *)usymb);
#else
  func = dlsym(handle, WML_MODULE_DLSYM_NAME);
#endif /* USCORE */
  if (func == NULL) {
#if	0
    fprintf(stderr, "error: Not get a symbol: %s\n", WML_MODULE_DLSYM_NAME);
#endif
    errno = ENOENT;
    return NULL;
  }

  return func();
}

static int
verify(module)
	const wmModule *	module;
{
  if (module == NULL) { return -1; }

  /* check the API version */
  if (module->version != WML_MODULE_MAGIC_NUMBER_MAJOR) {
    return -1;
  }

  /* check the module name */
  if (module->name == NULL) {
    return -1;
  }

  /* check the duplicate loading */
  if (lookup(module->name) != NULL) {
    return -1;
  }

  return 0;
}

static ModuleList *
lookup(name)
	const char *	name;
{
  register ModuleList *	mlp;

  for (mlp = moduleList.next; mlp != &moduleList; mlp = mlp->next) {
    if (strcmp(mlp->module->name, name) == 0) {
      return mlp;
    }
  }

  return NULL;
}

static ModuleList *
make_new_module_list(module)
	const wmModule *	module;
{
  ModuleList *	new;

  new = (ModuleList *)malloc(sizeof(ModuleList));
  if (new == NULL) {
    perror("malloc");
    exit(1);
  }
  new->next   = new;
  new->prev   = new;
  new->module = module;

  return new;
}

static void
install(mlist)
	register ModuleList *	mlist;
{
  register ModuleList *	tail = moduleList.prev;

  mlist->next	    = tail->next;
  mlist->prev	    = tail;
  mlist->prev->next = mlist;
  mlist->next->prev = mlist;
}

/*****************************************************************************/

int
wm_unload_module(name)
	const char *	name;
{
  ModuleList *	mlp;
  int		ret;

  if (name == NULL) {
    errno = EINVAL;
    return -1;
  }

  mlp = lookup(name);
  if (mlp == NULL) {
    errno = ENOENT;
    return -1;
  }

  ret = unload(mlp->module->handle);
  uninstall(mlp);
  free((ModuleList *)mlp);

  return ret;
}

static int
unload(handle)
	void *	handle;
{
  if (handle != NULL) {
    return dlclose(handle);
  }

  return 0;
}

static void
uninstall(mlist)
	register ModuleList *	mlist;
{
  mlist->prev->next = mlist->next;
  mlist->next->prev = mlist->prev;
}

/*****************************************************************************/

void
wm_initialize_modules(opt)
	void *	opt;
{
  register ModuleList *	mlp;

  for (mlp = moduleList.next; mlp != &moduleList; mlp = mlp->next) {
    if (mlp->module->init != NULL) {
      mlp->module->init(opt);
    }
    mlp->module->registration();
  }
}

void
wm_remove_modules(void)
{
  register ModuleList *	mlp;
  register ModuleList *	mlp_next;

  for (mlp = moduleList.next; mlp != &moduleList; mlp = mlp_next) {
    mlp_next = mlp->next;
    if (mlp->module->finish != NULL) {
      mlp->module->finish();
    }
    uninstall(mlp);
    unload(mlp->module->handle);
    free((ModuleList *)mlp);
  }
}

