/*****************************************************************************
 *
 * FILE:	mutex.c
 * DESCRIPTION:	WTL: mutual exclusion functions
 * DATE:	Thu, Mar 27 2003
 * UPDATED:	Mon, Jul 11 2005
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2003-2005 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2003-2005 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: mutex.c,v 1.22 2005/07/18 12:53:22 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#include <errno.h>
#if	ENABLE_SSL
#include "wt-ssl.h"
#else
#include "wt.h"
#endif	/* ENABLE_SSL */
#include "wt-sub.h"

/******************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	MUTEX_OWNER(mtx)	((mtx)->body)

#define	wt_append_wait_q(thd, mtx) \
	do { \
		(thd)->state = WT_ST_WAIT_MUTEX; \
		wt_clist_append_link(&(thd)->wait_q, (mtx)); \
	} while (0)

#define	wt_remove_wait_q(thd) \
	do { \
		wt_clist_remove_link(&(thd)->wait_q); \
	} while (0)

/******************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wt_thread_mutex_init
 * Description:		create a mutex
 * Arguments:		new mutex returned pointer
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_thread_mutex_init(mtx)
	wt_mutex_t *	mtx;
{
  wt_mutex_t	new;

  *mtx = new = (wt_mutex_t)calloc(1, sizeof(wt_clist_t));
  if (new != NULL) {
    wt_clist_init(new, NULL);
    return 0;
  }
  errno = ENOMEM;

  return -1;
}

/*
 * Function Name:	wt_thread_mutex_destroy
 * Description:		free resources allocated for a mutex
 * Arguments:		mutex
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_thread_mutex_destroy(mtx)
	wt_mutex_t 	mtx;
{
  if (mtx == NULL) {
    errno = EINVAL;
    return -1;
  }

  if (MUTEX_OWNER(mtx) != NULL || mtx->next != mtx) {
    errno = EBUSY;
    return -1;
  }
  free((wt_mutex_t)mtx);

  return 0;
}

/*
 * Function Name:	wt_thread_mutex_lock
 * Description:		lock a mutex
 * Arguments:		mutex
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_thread_mutex_lock(mtx)
	wt_mutex_t 	mtx;
{
  wt_thread_t	me = wt_current_thread;

  if (mtx == NULL) {
    errno = EINVAL;
    return -1;
  }

  if (MUTEX_OWNER(mtx) == NULL) {
    MUTEX_OWNER(mtx) = me;	/* got the mutex */
    return 0;
  }

  if (MUTEX_OWNER(mtx) == me) {
    errno = EDEADLK;
    return -1;
  }

  wt_append_wait_q(me, mtx);
  wt_remove_run_q(me);

  wt_context_switch(me);

  wt_remove_wait_q(me);

  return 0;
}

/*
 * Function Name:	wt_thread_mutex_trylock
 * Description:		attempt to lock a mutex without blocking
 * Arguments:		mutex
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_thread_mutex_trylock(mtx)
	wt_mutex_t 	mtx;
{
  if (mtx == NULL) {
    errno = EINVAL;
    return -1;
  }

  if (MUTEX_OWNER(mtx) != NULL) {
    errno = EBUSY;
    return -1;
  }

  /* got the mutex */
  MUTEX_OWNER(mtx) = wt_current_thread;

  return 0;
}

/*
 * Function Name:	wt_thread_mutex_unlock
 * Description:		unlock a mutex
 * Arguments:		mutex
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_thread_mutex_unlock(mtx)
	wt_mutex_t 	mtx;
{
  wt_thread_t		me = wt_current_thread;
  register wt_mutex_t	mp;

  if (mtx == NULL) {
    errno = EINVAL;
    return -1;
  }

  if (MUTEX_OWNER(mtx) != me) {
    errno = EPERM;
    return -1;
  }

  for (mp = mtx->next; mp != mtx; mp = mp->next) {
    wt_thread_t	thd = (wt_thread_t)MUTEX_OWNER(mp);

    if (thd->state == WT_ST_WAIT_MUTEX) {
      MUTEX_OWNER(mtx) = thd;
      wt_append_run_q(thd);
      return 0;
    }
  }

  /* no thread waiting on this mutex */
  MUTEX_OWNER(mtx) = NULL;

  return 0;
}

