/*****************************************************************************
 *
 * FILE:	net.c
 * DESCRIPTION:	WTL: Network I/O functions
 * DATE:	Fri, Apr 18 2003
 * UPDATED:	Wed, Sep  5 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2003-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2003-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: net.c,v 1.36 2007/09/04 17:09:03 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#ifndef	HAVE_SNPRINTF
#include <stdarg.h>
#endif	/* HAVE_SNPRINTF */
#if	HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif	/* HAVE_NETINET_IN_H */
#if	HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif	/* HAVE_ARPA_INET_H */
#include <sys/un.h>
#if	HAVE_NET_IF_DL_H
#include <net/if_dl.h>
#endif	/* HAVE_NET_IF_DL_H */
#if	HAVE_NETDB_H
#include <netdb.h>
#endif	/* HAVE_NETDB_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#include <errno.h>
#if	ENABLE_SSL
#include "wt-ssl.h"
#else
#include "wt.h"
#endif	/* ENABLE_SSL */
#include "wt-sub.h"

/******************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	WT_LISTENQ	(1024)

/******************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static int	wt_tcp_connect(const char *, const char *, int);
static int	wt_tcp_client(const char *, const char *);
static int	wt_tcp_server(const char *, const char *, socklen_t *, int);
static int	wt_udp_connect(const char *, const char *);
static int	wt_udp_client(const char *, const char *, void **, socklen_t *);
static int	wt_udp_server(const char *, const char *, socklen_t *);
static int	wt_ust_client(const char *);
static int	wt_ust_server(const char *, int);
static int	wt_udg_connect(const char *, char *);
static int	wt_udg_client(const char *, char *, void **, socklen_t *);
static int	wt_udg_server(const char *);
static int	wt_tmpnam(char *);

/******************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wt_netfd_initport
 * Description:		initialize network socket
 * Arguments:		node name, service/port, role, <option value>
 *			<option value>: tcp/ust-server: listen queue
 *					-1: No use, 0: default, >0: listen queue
 *					tcp-connect: timeout seconds
 * Return Value:	successful: netfd, otherwise: NULL
 */
wt_netfd_t *
wt_netfd_initport(nodename, service, role, optval)
	const char *	nodename;
	const char *	service;
	int		role;
	int		optval;
{
  int			sd;
  wt_netfd_t *		nfd;
  struct sockaddr *	saddr;
  socklen_t 		addrlen;
  char			tmpath[] = "/tmp/.s_wtXXXXXX";

  switch (role) {
    case WT_TCP_CONNECT:
      sd = wt_tcp_connect(nodename, service, optval);
      break;
    case WT_TCP_CLIENT:
      sd = wt_tcp_client(nodename, service);
      break;
    case WT_TCP_SERVER:
      sd = wt_tcp_server(nodename, service, &addrlen, optval);
      break;
    case WT_UDP_CONNECT:
      sd = wt_udp_connect(nodename, service);
      break;
    case WT_UDP_CLIENT:
      sd = wt_udp_client(nodename, service, (void **)&saddr, &addrlen);
      break;
    case WT_UDP_SERVER:
      sd = wt_udp_server(nodename, service, &addrlen);
      break;
    case WT_UST_CLIENT:
      sd = wt_ust_client(nodename);
      break;
    case WT_UST_SERVER:
      sd = wt_ust_server(nodename, optval);
      break;
    case WT_UDG_CONNECT:
      sd = wt_udg_connect(nodename, tmpath);
      break;
    case WT_UDG_CLIENT:
      sd = wt_udg_client(nodename, tmpath, (void **)&saddr, &addrlen);
      break;
    case WT_UDG_SERVER:
      sd = wt_udg_server(nodename);
      break;
    default:
      return NULL;
  }
  if (sd == -1) {
    return NULL;
  }

  nfd = wt_netfd_fdopen(sd);
  if (nfd != NULL) {
    switch (role) {
      case WT_UDP_CONNECT:
      case WT_UDG_CONNECT:
	nfd->udpath = strdup(tmpath);
	break;
      case WT_UDP_CLIENT:
      case WT_UDG_CLIENT:
	nfd->saddr = saddr;
      case WT_TCP_SERVER:
      case WT_UDP_SERVER:
	nfd->addrlen = addrlen;
      default:
	break;
    }
  }

  return nfd;
}

/*
 * Function Name:	wt_tcp_connect
 * Description:		create a new socket descriptor
 * Arguments:		hostname, service, timeout
 * Return Value:	successful: descriptor, otherwise: -1
 * Reference:		UNIX Networking Programming, 2nd Edition, Vol.1, p398
 */
static int
wt_tcp_connect(hostname, service, nsec)
	const char *	hostname;
	const char *	service;
	int		nsec;
{
  struct addrinfo	hints;
  struct addrinfo *	res;
  struct addrinfo *	res_save;
  int			sd;

  memset((char *)&hints, 0, sizeof(hints));
  hints.ai_flags	= AI_CANONNAME;	/* always return canonical name */
  hints.ai_family	= AF_UNSPEC;
  hints.ai_socktype	= SOCK_STREAM;
  hints.ai_protocol	= IPPROTO_TCP;

  if (getaddrinfo(hostname, service, &hints, &res) != 0) {
    return -1;
  }

  res_save = res;
  do {
    int	flags;

    /* create a socket */
    sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if (sd == -1) {
      continue;	/*ignore this one */
    }

    flags = fcntl(sd, F_GETFL, 0);
    (void)fcntl(sd, F_SETFL, flags | O_NONBLOCK);
    /* initiate a connection on a socket */
    if (connect(sd, res->ai_addr, res->ai_addrlen) != 0) {
      if (errno == EINPROGRESS) {
	wt_pollfd_t	pfd;
	int		n;

	/* set pollfd */
	pfd.fd	    = sd;
	pfd.events  = WT_POLL_IN | WT_POLL_OUT;
	pfd.revents = WT_POLL_NONE;

again:
	n = wt_poll(&pfd, WT_UTIME_SEC(nsec));
	if (n == 0) {	/* timeout */
	  errno = ETIMEDOUT;
	}
	else if (n > 0) {
	  if ((pfd.revents & WT_POLL_IN) || (pfd.revents & WT_POLL_OUT)) {
	    int		optval;
	    socklen_t	optlen = sizeof(optval);

	    if (getsockopt(sd, SOL_SOCKET, SO_ERROR, &optval, &optlen) == 0) {
	      errno = optval;
	      if (optval == 0) {
		(void)fcntl(sd, F_SETFL, flags);
		break;	/* established */
	      }
	    }
	  }
	  else if (errno == EINTR) {
	    goto again;
	  }
	}
      }
    }
    else {	/* established */
      (void)fcntl(sd, F_SETFL, flags);
      break;
    }
    close(sd);	/* ignore this one */
    sd = -1;
  } while ((res = res->ai_next) != NULL);

  if (res == NULL) {	/* errno set from final connect() */
    sd = -1;
  }

  freeaddrinfo(res_save);

  return sd;
}

/*
 * Function Name:	wt_tcp_client
 * Description:		create a new socket descriptor
 * Arguments:		hostname, service
 * Return Value:	successful: descriptor, otherwise: -1
 */
static int
wt_tcp_client(hostname, service)
	const char *	hostname;
	const char *	service;
{
  struct addrinfo	hints;
  struct addrinfo *	res;
  struct addrinfo *	res_save;
  int			sd;

  memset((char *)&hints, 0, sizeof(hints));
  hints.ai_flags	= AI_CANONNAME;	/* always return canonical name */
  hints.ai_family	= AF_UNSPEC;
  hints.ai_socktype	= SOCK_STREAM;
  hints.ai_protocol	= IPPROTO_TCP;

  if (getaddrinfo(hostname, service, &hints, &res) != 0) {
    return -1;
  }

  res_save = res;
  do {
    /* create a socket */
    sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if (sd == -1) {
      continue;	/*ignore this one */
    }
    /* initiate a connection on a socket */
    if (connect(sd, res->ai_addr, res->ai_addrlen) == 0) {
      break;
    }
    close(sd);	/* ignore this one */
  } while ((res = res->ai_next) != NULL);

  if (res == NULL) {	/* errno set from final connect() */
    sd = -1;
  }

  freeaddrinfo(res_save);

  return sd;
}

/*
 * Function Name:	wt_tcp_server
 * Description:		create a new socket descriptor
 * Arguments:		hostname, service, socklen_t pointer, listen queue
 * Return Value:	successful: descriptor, otherwise: -1
 */
static int
wt_tcp_server(hostname, service, addrlenp, listenq)
	const char *	hostname;
	const char *	service;
	socklen_t *	addrlenp;
	int		listenq;
{
  struct addrinfo	hints;
  struct addrinfo *	res;
  struct addrinfo *	res_save;
  int			sd;

  if (listenq <= 0) {
    listenq = WT_LISTENQ;
  }

  memset((char *)&hints, 0, sizeof(hints));
  hints.ai_flags	= hostname == NULL ? AI_PASSIVE : AI_NUMERICHOST;
  hints.ai_family	= AF_UNSPEC;
  hints.ai_socktype	= SOCK_STREAM;
  hints.ai_protocol	= IPPROTO_TCP;

  if (getaddrinfo(hostname, service, &hints, &res) != 0) {
    return -1;
  }

  res_save = res;
  do {
    const int	on = 1;

    /* create a socket */
    sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if (sd < 0) {
      continue;	/*error, try next one */
    }
    /* allow reuse of local address */
    if (setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) == -1) {
      close(sd);
      continue;
    }
    /* bind a socket */
    if (bind(sd, res->ai_addr, res->ai_addrlen) == 0) {
      if (addrlenp) {
	*addrlenp = res->ai_addrlen;	/* return size of protocol address */
      }
      break;	/* success */
    }
    close(sd);	/* bind error, close and try next one */
  } while ((res = res->ai_next) != NULL);

  if (res == NULL) {	/* errno set from final connect() */
    sd = -1;
  }
  /* listen for connection on the socket */
  else if (listen(sd, listenq) == -1) {
    close(sd);
    sd = -1;
  }

  freeaddrinfo(res_save);

  return sd;
}

/*
 * Function Name:	wt_udp_client
 * Description:		create a new socket descriptor
 * Arguments:		hostname, service, sockaddr pointer, socklen_t pointer
 * Return Value:	successful: descriptor, otherwise: -1
 */
static int
wt_udp_client(hostname, service, saptr, lenp)
	const char *	hostname;
	const char *	service;
	void **		saptr;
	socklen_t *	lenp;
{
  struct addrinfo	hints;
  struct addrinfo *	res;
  struct addrinfo *	res_save;
  int			sd;

  memset((char *)&hints, 0, sizeof(hints));
  hints.ai_flags	= AI_CANONNAME;
  hints.ai_family	= AF_UNSPEC;
  hints.ai_socktype	= SOCK_DGRAM;
  hints.ai_protocol	= IPPROTO_UDP;

  if (getaddrinfo(hostname, service, &hints, &res) != 0) {
    return -1;
  }

  res_save = res;
  do {
    sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if (sd >= 0) {
      break;	/* success */
    }
  } while ((res = res->ai_next) != NULL);

  if (res == NULL) {	/* errno set from final socket() */
    sd = -1;
    goto done;
  }

  *saptr = calloc(1, res->ai_addrlen);
  if (*saptr == NULL) {
    close(sd);
    sd = -1;
    goto done;
  }
  memcpy(*saptr, res->ai_addr, res->ai_addrlen);
  *lenp = res->ai_addrlen;

done:
  freeaddrinfo(res_save);

  return sd;
}

/*
 * Function Name:	wt_udp_connect
 * Description:		create a new socket descriptor
 * Arguments:		hostname, service
 * Return Value:	successful: descriptor, otherwise: -1
 */
static int
wt_udp_connect(hostname, service)
	const char *	hostname;
	const char *	service;
{
  struct addrinfo	hints;
  struct addrinfo *	res;
  struct addrinfo *	res_save;
  int			sd;

  memset((char *)&hints, 0, sizeof(hints));
  hints.ai_flags	= AI_CANONNAME;
  hints.ai_family	= AF_UNSPEC;
  hints.ai_socktype	= SOCK_DGRAM;

  if (getaddrinfo(hostname, service, &hints, &res) != 0) {
    return -1;
  }

  res_save = res;
  do {
    sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if (sd == -1) {
      continue;	/* ignore this one */
    }
    if (connect(sd, res->ai_addr, res->ai_addrlen) == 0) {
      break;	/* success */
    }
    close(sd);	/* ignore this one */
  } while ((res = res->ai_next) != NULL);

  if (res == NULL) {	/* errno set from final socket() */
    sd = -1;
  }

  freeaddrinfo(res_save);

  return sd;
}

/*
 * Function Name:	wt_udp_server
 * Description:		create a new socket descriptor
 * Arguments:		hostname, service, socklen_t pointer
 * Return Value:	successful: descriptor, otherwise: -1
 */
static int
wt_udp_server(hostname, service, addrlenp)
	const char *	hostname;
	const char *	service;
	socklen_t *	addrlenp;
{
  struct addrinfo	hints;
  struct addrinfo *	res;
  struct addrinfo *	res_save;
  int			sd;

  memset((char *)&hints, 0, sizeof(hints));
  hints.ai_flags	= hostname == NULL ? AI_PASSIVE : AI_NUMERICHOST;
  hints.ai_family	= AF_UNSPEC;
  hints.ai_socktype	= SOCK_DGRAM;
  hints.ai_protocol	= IPPROTO_UDP;

  if (getaddrinfo(hostname, service, &hints, &res) != 0) {
    return -1;
  }

  res_save = res;
  do {
    const int	on = 1;

    sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if (sd == -1) {
      continue;	/* error, try next one */
    }
    /* allow reuse of local address */
    if (setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) == -1) {
      close(sd);
      continue;
    }
#ifdef	SO_REUSEPORT
    if (setsockopt(sd, SOL_SOCKET, SO_REUSEPORT, &on, sizeof(on)) == -1) {
      close(sd);
      continue;
    }
#endif	/* SO_REUSEPORT */
    if (bind(sd, res->ai_addr, res->ai_addrlen) == 0) {
      if (addrlenp) {
	*addrlenp = res->ai_addrlen;	/* return size of protocol address */
      }
      break;	/* success */
    }
    close(sd);	/* bind error, close and try next one */
  } while ((res = res->ai_next) != NULL);

  if (res == NULL) {	/* errno set from final socket() */
    sd = -1;
  }

  freeaddrinfo(res_save);

  return sd;
}

/*
 * Function Name:	wt_ust_client
 * Description:		create a new socket descriptor
 * Arguments:		unix path
 * Return Value:	successful: descriptor, otherwise: -1
 */
static int
wt_ust_client(path)
	const char *	path;
{
  int			sd;
  struct sockaddr_un	caddr;

  sd = socket(AF_LOCAL, SOCK_STREAM, 0);
  if (sd == -1) {
    return -1;
  }
  memset((char *)&caddr, 0, sizeof(caddr));
  caddr.sun_family = AF_LOCAL;
#if	HAVE_STRLCPY
  strlcpy(caddr.sun_path, path, sizeof(caddr.sun_path));
#else
  strncpy(caddr.sun_path, path, strlen(path));	/* XXX: vulnerability */
#endif	/* HAVE_STRLCPY */
  if (connect(sd, (struct sockaddr *)&caddr, sizeof(caddr)) == -1) {
    close(sd);
    return -1;
  }

  return sd;
}

/*
 * Function Name:	wt_ust_server
 * Description:		create a new socket descriptor
 * Arguments:		unix path, listen queue
 * Return Value:	successful: descriptor, otherwise: -1
 */
static int
wt_ust_server(path, listenq)
	const char *	path;
	int		listenq;
{
  int			sd;
  struct sockaddr_un	saddr;

  if (listenq <= 0) {
    listenq = WT_LISTENQ;
  }

  sd = socket(AF_LOCAL, SOCK_STREAM, 0);
  if (sd == -1) {
    return -1;
  }

  unlink(path);
  memset((char *)&saddr, 0, sizeof(saddr));
  saddr.sun_family = AF_LOCAL;
#if	HAVE_STRLCPY
  strlcpy(saddr.sun_path, path, sizeof(saddr.sun_path));
#else
  strncpy(saddr.sun_path, path, strlen(path));	/* XXX: vulnerability */
#endif	/* HAVE_STRLCPY */
  if (bind(sd, (struct sockaddr *)&saddr, sizeof(saddr)) == -1 ||
      listen(sd, listenq) < 0) {
    close(sd);
    return -1;
  }

  return sd;
}

/*
 * Function Name:	wt_udg_connect
 * Description:		create a new socket descriptor
 * Arguments:		unix path, temporary filename pointer
 * Return Value:	successful: descriptor, otherwise: -1
 */
static int
wt_udg_connect(path, tmpath)
	const char *	path;
	char *		tmpath;
{
  int			sd;
  struct sockaddr_un	caddr;
  struct sockaddr_un	saddr;

  /* make a temporary filename */
  if (wt_tmpnam(tmpath) == -1) {
    return -1;
  }

  /* create a socket */
  sd = socket(AF_LOCAL, SOCK_DGRAM, 0);
  if (sd == -1) {
    return -1;
  }

  /* bind an address for us */
  memset((char *)&caddr, 0, sizeof(caddr));
  caddr.sun_family = AF_LOCAL;
#if	HAVE_STRLCPY
  strlcpy(caddr.sun_path, tmpath, sizeof(caddr.sun_path));
#else
  strncpy(caddr.sun_path, tmpath, strlen(tmpath));	/* XXX: vulnerability */
#endif	/* HAVE_STRLCPY */
  if (bind(sd, (struct sockaddr *)&caddr, sizeof(caddr)) == -1) {
    close(sd);
    return -1;
  }

  /* fill in server's address */
  memset((char *)&saddr, 0, sizeof(saddr));
  saddr.sun_family = AF_LOCAL;
#if	HAVE_STRLCPY
  strlcpy(saddr.sun_path, path, sizeof(saddr.sun_path));
#else
  strncpy(saddr.sun_path, path, strlen(path));	/* XXX: vulnerability */
#endif	/* HAVE_STRLCPY */

  if (connect(sd, (struct sockaddr *)&saddr, sizeof(saddr)) == -1) {
    close(sd);
    return -1;
  }

  return sd;
}

/*
 * Function Name:	wt_udg_client
 * Description:		create a new socket descriptor
 * Arguments:		unix path, temporary filename pointer,
 *			server address pointer,	socket length pointer
 * Return Value:	successful: descriptor, otherwise: -1
 */
static int
wt_udg_client(path, tmpath, saptr, lenp)
	const char *	path;
	char *		tmpath;
	void **		saptr;
	socklen_t *	lenp;
{
  int			sd;
  struct sockaddr_un	caddr;
  struct sockaddr_un	saddr;

  /* make a temporary filename */
  if (wt_tmpnam(tmpath) == -1) {
    return -1;
  }

  /* create a socket */
  sd = socket(AF_LOCAL, SOCK_DGRAM, 0);
  if (sd == -1) {
    return -1;
  }

  /* bind an address for us */
  memset((char *)&caddr, 0, sizeof(caddr));
  caddr.sun_family = AF_LOCAL;
#if	HAVE_STRLCPY
  strlcpy(caddr.sun_path, tmpath, sizeof(caddr.sun_path));
#else
  strncpy(caddr.sun_path, tmpath, strlen(tmpath));	/* XXX: vulnerability */
#endif	/* HAVE_STRLCPY */
  if (bind(sd, (struct sockaddr *)&caddr, sizeof(caddr)) == -1) {
    close(sd);
    return -1;
  }

  /* fill in server's address */
  memset((char *)&saddr, 0, sizeof(saddr));
  saddr.sun_family = AF_LOCAL;
#if	HAVE_STRLCPY
  strlcpy(saddr.sun_path, path, sizeof(saddr.sun_path));
#else
  strncpy(saddr.sun_path, path, strlen(path));	/* XXX: vulnerability */
#endif	/* HAVE_STRLCPY */

  *saptr = calloc(1, sizeof(saddr));
  if (*saptr == NULL) {
    close(sd);
    return -1;
  }
  memcpy(*saptr, (struct sockaddr *)&saddr, sizeof(saddr));
  *lenp = sizeof(saddr);

  return sd;
}

/*
 * Function Name:	wt_udg_server
 * Description:		create a new socket descriptor
 * Arguments:		unix path
 * Return Value:	successful: descriptor, otherwise: -1
 */
static int
wt_udg_server(path)
	const char *	path;
{
  int			sd;
  struct sockaddr_un	saddr;

  sd = socket(AF_LOCAL, SOCK_DGRAM, 0);
  if (sd == -1) {
    return -1;
  }

  unlink(path);
  memset((char *)&saddr, 0, sizeof(saddr));
  saddr.sun_family = AF_LOCAL;
#if	HAVE_STRLCPY
  strlcpy(saddr.sun_path, path, sizeof(saddr.sun_path));
#else
  strncpy(saddr.sun_path, path, strlen(path));	/* XXX: vulnerability */
#endif	/* HAVE_STRLCPY */
  if (bind(sd, (struct sockaddr *)&saddr, sizeof(saddr)) == -1) {
    close(sd);
    return -1;
  }

  return sd;
}

/*****************************************************************************/

/*
 * Function Name:	wt_netfd_shutdown
 * Description:		shutdown netfd
 * Arguments:		netfd, how(WT_SHUT_RD,WT_SHUT_WR,WT_SHUT_RDWR)
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_netfd_shutdown(nfd, how)
	register wt_netfd_t *	nfd;
	register int		how;
{
  return shutdown(nfd->osfd, how);
}

/*
 * Function Name:	wt_host_serv
 * Description:		get address information simply
 * Arguments:		hostname, service, protocol family, socket type
 * Return Value:	successful: addrinfo pointer, otherwise: NULL
 */
struct addrinfo *
wt_host_serv(hostname, service, family, socktype)
	const char *	hostname;
	const char *	service;
	int		family;		/* AF_UNSPEC, AF_INET, AF_INET6 */
	int		socktype;	/* 0, SOCK_STREAM, SOCK_DGRAM */
{
  struct addrinfo	hints;
  struct addrinfo *	res;

  memset((char *)&hints, 0, sizeof(hints));
  hints.ai_flags	= AI_CANONNAME;	/* always return canonical name */
  hints.ai_family	= family;
  hints.ai_socktype	= socktype;

  return (getaddrinfo(hostname, service, &hints, &res) != 0 ? NULL : res);
}

/*
 * Function Name:	wt_sock_ntop
 * Description:		convert an address from a socket address
 * Arguments:		socket address, port
 * Return Value:	successful: nodename, otherwise: NULL
 */
char *
wt_sock_ntop(sa, has_port)
	const struct sockaddr *	sa;
	bool			has_port;
{
  char	port[7];
  char	host[128];	/* Unix domain is largest */

  memset(port, 0, sizeof(port));
  memset(host, 0, sizeof(host));

  switch (sa->sa_family) {
    case 255: {
	unsigned long	mask;
	unsigned int	index = 1 << 31;
	unsigned short	new_mask = 0;

	mask = ntohl(((struct sockaddr_in *)sa)->sin_addr.s_addr);
	while (mask & index) {
	  new_mask++;
	  index >>= 1;
	}
#if	HAVE_SNPRINTF
	if (snprintf(host, sizeof(host), "/%hu", new_mask) == -1) {
	  return NULL;
	}
#else
	if (sprintf(host, "/%hu", new_mask) == -1) {
	  return NULL;
	}
#endif	/* HAVE_SNPRINTF */
      }
      break;

    case AF_UNSPEC:
    case AF_INET: {
	struct sockaddr_in *	sin = (struct sockaddr_in *)sa;

	if (inet_ntop(AF_INET, &sin->sin_addr, host, sizeof(host)) == NULL) {
	  return NULL;
	}
	if (strcmp(host, "0.0.0.0") == 0 && has_port == false) {
#if	HAVE_STRLCPY
	  strlcpy(host, "default", sizeof(host));
#else
	  strncpy(host, "default", 7);	/* XXX: vulnerability */
#endif	/* HAVE_STRLCPY */
	}
	if (ntohs(sin->sin_port) != 0 && has_port == true) {
#if	HAVE_SNPRINTF
	  if (snprintf(port, sizeof(port), ".%d", ntohs(sin->sin_port)) == -1) {
	    return NULL;
	  }
#else
	  int	sport = ntohs(sin->sin_port);
	  if (sport > 99999) {	/* over the number with five figures */
	    return NULL;
	  }
	  if (sprintf(port, ".%d", sport) == -1) {
	    return NULL;
	  }
#endif	/* HAVE_SNPRINTF */
#if	HAVE_STRLCAT
	  strlcat(host, port, sizeof(host));
#else
	  strncat(host, port, strlen(port));	/* XXX: vulnerability */
#endif	/* HAVE_STRLCAT */
	}
      }
      break;

    case AF_INET6: {
	struct sockaddr_in6 *	sin6 = (struct sockaddr_in6 *)sa;

	if (inet_ntop(AF_INET6, &sin6->sin6_addr, host, sizeof(host)) == NULL) {
	  return NULL;
	}
	if (ntohs(sin6->sin6_port) != 0 && has_port == true) {
#if	HAVE_SNPRINTF
	  if (snprintf(port, sizeof(port), ".%d", ntohs(sin6->sin6_port)) == -1) {
	    return NULL;
	  }
#else
	  int	sport = ntohs(sin6->sin6_port);
	  if (sport > 99999) {	/* over the number with five figures */
	    return NULL;
	  }
	  if (sprintf(port, ".%d", ntohs(sin6->sin6_port)) == -1) {
	    return NULL;
	  }
#endif	/* HAVE_SNPRINTF */
#if	HAVE_STRLCAT
	  strlcat(host, port, sizeof(host));
#else
	  strncat(host, port, strlen(port));	/* XXX: vulnerability */
#endif	/* HAVE_STRLCAT */
	}
      }
      break;

    case AF_UNIX: {
	struct sockaddr_un *	unp = (struct sockaddr_un *)sa;

	if (unp->sun_path[0] == EOL) {
#if	HAVE_STRLCPY
	  strlcpy(host, "(no pathname found)", sizeof(host));
#else
	  strncpy(host, "(no pathname found)", 19);	/* XXX: vulnerability */
#endif	/* HAVE_STRLCPY */
	}
	else {
#if	HAVE_SNPRINTF
	  if (snprintf(host, sizeof(host), "%s", unp->sun_path) == -1) {
	    return NULL;
	  }
#else
	  if (sprintf(host, "%.*s", (int)(sizeof(host) - 1), unp->sun_path) == -1) {
	    return NULL;
	  }
#endif	/* HAVE_SNPRINTF */
	}
      }
      break;

#if	AF_LINK
    case AF_LINK: {
	struct sockaddr_dl *	sdl = (struct sockaddr_dl *)sa;

	if (sdl->sdl_nlen > 0) {
	  memset(host, 0, sizeof(host));
	  memcpy(host, sdl->sdl_data, sdl->sdl_nlen);
	}
	else {
#if	HAVE_SNPRINTF
	  snprintf(host, sizeof(host), "link#%d", sdl->sdl_index);
#else
	  sprintf(host, "link#%d", sdl->sdl_index);
#endif	/* HAVE_SNPRINTF */
	}
      }
      break;
#endif	/* AF_LINK */

    default:
      return NULL;
  }

  return strdup(host);
}

/*****************************************************************************/

/*
 * Function Name:	wt_netfd_tcp_servers
 * Description:		create new socket descriptors
 * Arguments:		hostname, service, listen queue
 * Return Value:	successful: netfds, otherwise: NULL
 */
wt_netfd_t *
wt_netfd_tcp_servers(hostname, service, listenq)
	const char *	hostname;
	const char *	service;
	int		listenq;
{
  struct addrinfo	hints;
  struct addrinfo *	res;
  struct addrinfo *	ai;
  wt_netfd_t *		nsd;
  wt_netfd_t *		retval = NULL;

  if (listenq <= 0) {
    listenq = WT_LISTENQ;
  }

  memset((char *)&hints, 0, sizeof(hints));
  hints.ai_flags	= hostname == NULL ? AI_PASSIVE : AI_NUMERICHOST;
  hints.ai_family	= AF_UNSPEC;
  hints.ai_socktype	= SOCK_STREAM;
  hints.ai_protocol	= IPPROTO_TCP;

  if (getaddrinfo(hostname, service, &hints, &res) != 0) {
    return NULL;
  }

  for (ai = res; ai != NULL; ai = ai->ai_next) {
    int		sd;
    const int	on = 1;

    /* create a socket */
    sd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
    if (sd == -1) {
      continue;	/* error, try next one */
    }
    /* allow reuse of local addresses */
    if (setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) == -1) {
      close(sd);
      continue;
    }
    /* bind a socket */
    if (bind(sd, ai->ai_addr, ai->ai_addrlen) == -1) {
      close(sd);
      continue;
    }
    /* listen for connection on the socket */
    if (listen(sd, listenq) == -1) {
      close(sd);
      continue;
    }
    /* associate a netfd */
    nsd = wt_netfd_fdopen(sd);
    wt_netfd_socklen(nsd) = ai->ai_addrlen;

    nsd->next = retval;
    retval    = nsd;
  }
  freeaddrinfo(res);

  return retval;
}

/*****************************************************************************/

static int
wt_tmpnam(template)
	char *	template;
{
  int	fd;
  int	status;

  fd = mkstemp(template);
  if (fd == -1) {
    return -1;
  }
  status = unlink(template);
  if (close(fd) == -1) {
    return -1;
  }

  return status;
}

