/*****************************************************************************
 *
 * FILE:	proc.c
 * DESCRIPTION:	WTL: process manipulation functions
 * DATE:	Tue, Jun 24 2003
 * UPDATED:	Wed, Sep  5 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2003-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2003-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: proc.c,v 1.27 2007/09/04 17:09:04 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif	/* HAVE_SYS_WAIT_H */
#if	HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif	/* HAVE_SYS_TYPES_H */
#if	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif	/* HAVE_SYS_TIME_H */
#if	HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif	/* HAVE_SYS_RESOURCE_H */
#include <signal.h>
#include <errno.h>
#if	ENABLE_SSL
#include "wt-ssl.h"
#else
#include "wt.h"
#endif	/* ENABLE_SSL */
#include "wt-sub.h"

/******************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
typedef struct _wt_procq {
  wt_clist_t	q;
  wt_thread_t	thread;
  int		status;	/* _exit(2) status */
  bool		on_q;	/* flag of state on queue */
  pid_t		pid;	/* PID of child process */
} wt_procq_t;
#define	wt_procq_ptr(x)		((wt_procq_t *)(x)->body)

/******************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	sigxcpu(int);
static void	sigxfsz(int);
static void	wt_append_proc_q(wt_procq_t *, wt_utime_t);
static void	wt_remove_proc_q(wt_procq_t *);

/******************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static wt_clist_t	vp_proc_q;		/* process queue */

/******************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/
/*
 * Function Name:	wt_init_proc
 * Description:		initialize process manager
 * Arguments:		signal handler
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_init_proc(handler)
	void	(*handler)(int);
{
  struct sigaction	sa;

  sigemptyset(&sa.sa_mask);
  sa.sa_flags	= SA_RESTART|SA_NOCLDSTOP;
  sa.sa_handler = (void *)handler;
  if (sigaction(SIGCHLD, &sa, NULL) == -1) {
    return -1;
  }
  wt_clist_init(&vp_proc_q, NULL);

  return 0;
}

int
vp_check_proc(void)
{
  int	n = 0;
  pid_t	pid;

  do {
    int		status;

    errno = 0;
    while ((pid = waitpid((pid_t)0, &status, WNOHANG|WUNTRACED)) > 0) {
      register wt_clist_t *	cp;
      register wt_clist_t *	cp_next;

      for (cp = vp_proc_q.next; cp != &vp_proc_q; cp = cp_next) {
	wt_procq_t *	pq = wt_procq_ptr(cp);

	cp_next = cp->next;
	if (pid == pq->pid) {
	  n++;
	  pq->status = WEXITSTATUS(status);
	  wt_remove_proc_q(pq);
	  wt_append_run_q(pq->thread);
#if	DEBUG
	  fprintf(stderr, "DEBUG[vp_check_proc] pid=%d status=%d\n",
		  pq->pid, pq->status);
#endif	/* DEBUG */
	}
      }
    }
  } while (pid == -1 && errno == EINTR);

  return n;
}

/*****************************************************************************/

/*
 * Function Name:	wt_spawn
 * Description:		create a new process and execute a command
 * Arguments:		working path, command, arguments, environments,
 *			error trap, stdin, stdout, stderr,
 *			maxcpu, maxfsiz, timeout
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_spawn(path, cmd, argv, envp, errtrap, fdin, fdout, fderr, maxcpu, maxfsiz,
	 timeout)
	const char *	path;
	const char *	cmd;
	char * const	argv[];
	char * const	envp[];
	void		(*errtrap)(int);
	int		fdin;
	int		fdout;
	int		fderr;
	int		maxcpu;		/* max cpu time */
	int		maxfsiz;	/* max file size */
	wt_utime_t	timeout;
{
  wt_thread_t	me = wt_current_thread;
  wt_procq_t *	pq;
  volatile int	status = 0;

  pq = (wt_procq_t *)calloc(1, sizeof(wt_procq_t));
  if (pq == NULL) {
    errno = ENOMEM;
    return -1;
  }
  wt_clist_init(&pq->q, pq);
  pq->thread	= me;
  pq->status	= 0;
  pq->on_q	= true;
  pq->pid	= fork();

  wt_remove_run_q(me);
  wt_append_proc_q(pq, timeout);
  switch (pq->pid) {
    case -1:	/* fork error */
      status = -1;
      goto done;
    case 0: {	/* child process */
	struct sigaction	sa;
	struct rlimit		rl;

	/* set signal trap */
	sigemptyset(&sa.sa_mask);
	sa.sa_flags   = 0;
	sa.sa_handler = (void *)sigxcpu;
	(void)sigaction(SIGXCPU, &sa, NULL);
	sa.sa_handler = (void *)sigxfsz;
	(void)sigaction(SIGXFSZ, &sa, NULL);

	/* set resource limits */
	if (getrlimit(RLIMIT_CORE, &rl) == 0) {
	  rl.rlim_cur = rl.rlim_max = 0;
	  (void)setrlimit(RLIMIT_CORE, &rl);
	}
	if (maxcpu > 0 && getrlimit(RLIMIT_CPU, &rl) == 0) {
	  rl.rlim_cur = maxcpu;	/* sec */
	  (void)setrlimit(RLIMIT_CPU, &rl);
	}
	if (maxfsiz > 0 && getrlimit(RLIMIT_FSIZE, &rl) == 0) {
	  rl.rlim_cur = maxfsiz;	/* byte */
	  (void)setrlimit(RLIMIT_FSIZE, &rl);
	}

	if (fdin  >= 0) { dup2(fdin,  STDIN_FILENO); }	/* standard input */
	if (fdout >= 0) { dup2(fdout, STDOUT_FILENO); }	/* standard output */
	if (fderr >= 0) { dup2(fderr, STDERR_FILENO); }	/* standard error */

	if (chdir(path) == -1 || execve(cmd, argv, envp) == -1) {
	  if (errtrap != NULL) {
	    (*errtrap)(errno);
	  }
	}
	_exit(errno);
	/* NOTREACHED */
      }
    default:
      break;
  }

  wt_context_switch(me);

  if (pq->on_q) {
    wt_remove_proc_q(pq);
  }
  /* check return value from execve() */
  if (pq->status != 0) {
    errno  = pq->status;
    status = -1;
  }
  /* Is time over? */
  if (me->flags & WT_FL_TIMEDOUT) {
#if	DEBUG
    fprintf(stderr, "DEBUG[wt_spawn] timedout: %s pid=%d\n", me->name, pq->pid);
#endif	/* DEBUG */
    if (kill(pq->pid, SIGTERM) == 0) {
      pid_t	pid;

      do {
	pid = waitpid(pq->pid, NULL, WNOHANG);
      } while (pid == -1 && errno == EINTR);
    }
    me->flags &= ~WT_FL_TIMEDOUT;
    errno  = ETIMEDOUT;
    status = -1;
  }

done:
  free((wt_procq_t *)pq);

  return status;
}

static void
sigxcpu(signo)
	int	signo;
{
#if	DEBUG
  fprintf(stderr, "DEBUG[WTL] cpu time limit exceeded\n");
#endif	/* DEBUG */
  exit(signo);	/* CPU uptime over */
}

static void
sigxfsz(signo)
	int	signo;
{
#if	DEBUG
  fprintf(stderr, "DEBUG[WTL] file size limit exceeded\n");
#endif	/* DEBUG */
  exit(signo);	/* file size over */
}

/*****************************************************************************/

static void
wt_append_proc_q(pq, timeout)
	wt_procq_t *	pq;
	wt_utime_t	timeout;
{
  wt_clist_append_link(&pq->q, &vp_proc_q);
  pq->thread->flags |= WT_FL_WAIT_PID;
#if	DEBUG
  fprintf(stderr, "DEBUG[+procq] %s\n", pq->thread->name);
#endif	/* DEBUG */
  if (timeout != WT_UTIME_NO_TIMEOUT) {
    wt_append_sleep_q(pq->thread, timeout);
  }
}

static void
wt_remove_proc_q(pq)
	wt_procq_t *	pq;
{
  wt_clist_remove_link(&pq->q);
  pq->thread->flags &= ~WT_FL_WAIT_PID;
#if	DEBUG
  fprintf(stderr, "DEBUG[-procq] %s\n", pq->thread->name);
#endif	/* DEBUG */
  if (pq->thread->flags & WT_FL_ON_SLEEPQ) {
    wt_remove_sleep_q(pq->thread);
  }
  pq->on_q = false;
}

