/*****************************************************************************
 *
 * FILE:	schd.c
 * DESCRIPTION:	WTL: scheduler
 * DATE:	Wed, Mar 26 2003
 * UPDATED:	Wed, Sep  5 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * NOTES:	I referred to the following url
 *		http://www.engelschall.com/pw/usenix/2000/pmt-html/
 *		(Portable Multithreading The Signal Stack Trick For User-Space
 *		Thread Creation) to implement multithreading mechanism.
 *		I also referred to "State Threads Library for Internet
 *		Applications" (http://state-threads.sourceforge.net/).
 * COPYRIGHT:	(c) 2003-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2003-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: schd.c,v 1.50 2007/09/04 17:09:04 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#if	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif	/* HAVE_SYS_TIME_H */
#if	HAVE_SYS_EVENT_H
#include <sys/event.h>
#endif	/* HAVE_SYS_EVENT_H */
#include <signal.h>
#include <errno.h>
#if	ENABLE_SSL
#include "wt-ssl.h"
#else
#include "wt.h"
#endif	/* ENABLE_SSL */
#include "wt-sub.h"

/******************************************************************************
 *
 *	Glocal variables declaration
 *
 *****************************************************************************/
const wt_thread_attr_t wt_thread_attr_defaults = {
  "user-thread",	/* thread name (useful for debug) */
  false,		/* does thread permit join? */
  NULL,			/* new signal stack address */
  WT_STACK_SIZE		/* signal stack size for thread */
};

/******************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#if	HAVE_KQUEUE
struct wt_kq {
  int			fd;	/* for kqueue() */
  struct kevent *	events;
  int			nevents;
  struct kevent *	revents;
  int			nrevents;
};
#endif	/* HAVE_KQUEUE */

/* virtual processor */
typedef struct {
  wt_thread_t	main_thread;	/* main thread */
  wt_thread_t	idle_thread;	/* idle thread */
  wt_utime_t	last_clock;	/* last time */
  wt_utime_t	min_sleep;	/* minimum sleep time */
  wt_clist_t	run_q;		/* run queue */
  wt_clist_t	sleep_q;	/* sleep queue */
  wt_clist_t	zombie_q;	/* zombie queue */
  wt_clist_t	io_q;		/* I/O queue */
#if	HAVE_KQUEUE
  struct wt_kq	kq;
#else
  int		maxfd;
  fd_set	rfds;
  fd_set	wfds;
#endif	/* HAVE_KQUEUE */
  volatile int	wait_child;
} wt_vp_t;

#define	vp_main_thread	(wt_this_vp.main_thread)
#define	vp_idle_thread	(wt_this_vp.idle_thread)
#define	vp_last_clock	(wt_this_vp.last_clock)
#define	vp_min_sleep	(wt_this_vp.min_sleep)
#define	vp_run_q	(wt_this_vp.run_q)
#define	vp_sleep_q	(wt_this_vp.sleep_q)
#define	vp_zombie_q	(wt_this_vp.zombie_q)
#define	vp_io_q		(wt_this_vp.io_q)
#if	HAVE_KQUEUE
#define	vp_kq_fd	(wt_this_vp.kq.fd)
#define	vp_kq_events	(wt_this_vp.kq.events)
#define	vp_kq_nevents	(wt_this_vp.kq.nevents)
#define	vp_kq_revents	(wt_this_vp.kq.revents)
#define	vp_kq_nrevents	(wt_this_vp.kq.nrevents)
#else
#define	vp_io_maxfd	(wt_this_vp.maxfd)
#define	vp_io_rfds	(wt_this_vp.rfds)
#define	vp_io_wfds	(wt_this_vp.wfds)
#endif	/* HAVE_KQUEUE */
#define	vp_wait_child	(wt_this_vp.wait_child)

/******************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static wt_vp_t			wt_this_vp;		/* this VP */
/* active thread count */
#define	NO_ACTIVE_THREAD	(-1)
static volatile int		vp_active_threads = NO_ACTIVE_THREAD;
static wt_thread_t volatile	wt_this_thread = NULL;	/* current thread */

/******************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	sigchld(int);
static void	vp_run_idle(void *);
static void	vp_check_io(void);
static void	vp_find_bad_fd(void);
static void	vp_check_clock(void);

static int	mctx_create(mctx_t *, void * (*)(void *), void *, void *,
			    size_t);
static void	mctx_create_trampoline(int);
static void	mctx_create_boot(void);
#if	HAVE_KQUEUE
static int	kq_init(int);
static void	kq_add(int, short);
static void	kq_del(int, short);
static bool	kq_check(int, short);
#endif	/* HAVE_KQUEUE */
#if	DEBUG
static void	debug_dump_q(void);
#endif	/* DEBUG */

/******************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wt_init
 * Description:		initialize wyvern thread library
 * Arguments:		magnification for signal stack 
 * Return Value:	successful: 0, otherwise: -1, already init: 1
 */
int
wt_init(ss_mag)
	int	ss_mag;
{
#if	HAVE_KQUEUE
  int			nfiles;
#endif	/* HAVE_KQUEUE */
  wt_thread_attr_t	attr = wt_thread_attr_defaults;
  int			status;

  if (vp_active_threads > NO_ACTIVE_THREAD) {
    return 1;	/* already initialized */
  }
#if	HAVE_KQUEUE
  nfiles = wt_init_io();
  if (nfiles == -1) {
    return -1;
  }
  if (kq_init(nfiles) == -1) {
    return -1;
  }
#else
  if (wt_init_io() == -1) {
    return -1;
  }
  vp_io_maxfd = -1;
  FD_ZERO(&vp_io_rfds);
  FD_ZERO(&vp_io_wfds);
#endif
  if (wt_init_proc(sigchld) == -1) {
    return -1;
  }
  vp_active_threads = 0;
  /* initialize virtual processor */
  vp_last_clock = wt_utime();
  vp_min_sleep	= WT_UTIME_NO_TIMEOUT;
  wt_clist_init(&vp_run_q, NULL);
  wt_clist_init(&vp_sleep_q, NULL);
  wt_clist_init(&vp_zombie_q, NULL);
  wt_clist_init(&vp_io_q, NULL);
  vp_wait_child = 0;
  /* create idle thread for scheduling */
  attr.name = "idler";
  attr.stacksize = ss_mag < 0 ? WT_STACK_SIZE : WT_VP_SS_SIZE(ss_mag);
  status = wt_thread_create(&vp_idle_thread, &attr, (void *)&vp_run_idle, NULL);
  if (status == -1) {
    return -1;
  }
  vp_idle_thread->flags = WT_FL_IDLE_THREAD;
  wt_remove_run_q(vp_idle_thread);
  vp_active_threads--;
  /* create main thread */
  attr.name = "main";
  status = wt_thread_create(&vp_main_thread, &attr, NULL, NULL);
  vp_main_thread->state = WT_ST_RUNNING;
  vp_main_thread->flags = WT_FL_PRIMORDIAL;
  wt_this_thread	= vp_main_thread;

  return status;
}

static void
sigchld(signo)
	int	signo __unused;
{
  vp_wait_child++;
#if	DEBUG
  fprintf(stderr, "DEBUG[sigchld] waiting child process: %u\n", vp_wait_child);
#endif	/* DEBUG */
}

/*
 * Function Name:	wt_thread_create
 * Description:		create a new thread
 * Arguments:		entry point, arguments, joinable flag
 * Return Value:	successful: new thread, otherwise: NULL
 */
int
wt_thread_create(ret_thd, attr, ep, arg)
	wt_thread_t *			ret_thd;
	const wt_thread_attr_t *	attr;
	void *				(*ep)(void *);
	void *				arg;
{
  wt_thread_t	new;

  if (ret_thd == NULL) {
    errno = EINVAL;
    return -1;
  }

  *ret_thd = new = (wt_thread_t)calloc(1, sizeof(mctx_t));
  if (new != NULL) {
    new->ss_size = attr != NULL ? attr->stacksize : WT_STACK_SIZE;
    new->ss_addr = calloc(new->ss_size, sizeof(char));
    if (new->ss_addr == NULL) {
      goto error0;
    }

    if (attr != NULL) {
      if (attr->name != NULL) {
	new->name = strdup(attr->name);
      }
      if (attr->joinable) {
	if (wt_thread_cond_init(&new->term) == -1) {
	  goto error1;
	}
      }
    }
    if (mctx_create(new, ep, arg, new->ss_addr, new->ss_size) == -1) {
      goto error1;
    }

    wt_clist_init(&new->q, new);
    wt_clist_init(&new->wait_q, new);
    wt_append_run_q(new);
    vp_active_threads++;

    return 0;

error1:
    if (new->term != NULL) {
      wt_thread_cond_destroy(new->term);
    }
    free(new->ss_addr);
error0:
    free(new);
  }
  *ret_thd = NULL;

  return -1;
}

/*
 * Function Name:	wt_thread_exit
 * Description:		terminate the calling thread
 * Arguments:		return value of thread
 * Return Value:	none
 */
void
wt_thread_exit(retval)
	void *	retval;
{
  wt_thread_t	me = wt_this_thread;

#if	DEBUG
  fprintf(stderr, "DEBUG[exit] %s\n", me->name);
#endif	/* DEBUG */
  me->retval = retval;
  vp_active_threads--;
  wt_remove_run_q(me);
  if (me->term != NULL) {
    /* put thread on zombie queue */
    wt_append_zombie_q(me);
    /* notify on your termination condition variable */
    wt_thread_cond_signal(me->term);
    /* switch context and come back later */
    wt_context_switch(me);
    /* continue the cleanup */
    wt_thread_cond_destroy(me->term);	me->term = NULL;
  }
  me->state = WT_ST_DEAD;

#if	DEBUG
  fprintf(stderr, "DEBUG[exit] %s finished\n", me->name);
#endif	/* DEBUG */
  wt_context_switch(me);
}

/*
 * Function Name:	wt_thread_detach
 * Description:		detach thread
 * Arguments:		target thread
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_thread_detach(thd)
	wt_thread_t	thd;
{
  if (thd->state == WT_ST_DEAD) {
    if (thd->name != wt_thread_attr_defaults.name) {
      free((char *)thd->name);
    }
    free(thd->ss_addr);
    free(thd);
    thd = NULL;
    return 0;
  }

  return -1;
}

/*
 * Function Name:	wt_thread_join
 * Description:		wait for thread termination
 * Arguments:		target thread, pointer of return value
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_thread_join(thd, retvalp)
	wt_thread_t	thd;
	void **		retvalp;
{
  wt_cond_t	term;

  if (thd == NULL) {
    errno = EINVAL;
    return -1;
  }

  term = thd->term;
  if (term == NULL) {	/* cannot join a non-joinable thread */
    errno = EINVAL;
    return -1;
  }
  if (thd == wt_this_thread) {
    errno = EDEADLK;
    return -1;
  }

  /* multiple threads cannot wait on the same joinable thread */
  if (term->next != term) {
    errno = EINVAL;
    return -1;
  }

#if	DEBUG
  fprintf(stderr, "DEBUG[join] %s wait for %s\n", wt_this_thread->name, thd->name);
#endif	/* DEBUG */
  while (thd->state != WT_ST_ZOMBIE) {
    if (wt_thread_cond_timedwait(term, WT_UTIME_NO_TIMEOUT) == -1) {
      return -1;
    }
  }

  if (retvalp != NULL) {
    *retvalp = thd->retval;
  }

  wt_remove_zombie_q(thd);
  wt_append_run_q(thd);

  return 0;
}

/*
 * Function Name:	wt_thread_self
 * Description:		get the calling thread
 * Arguments:		none
 * Return Value:	thread
 */
wt_thread_t
wt_thread_self(void)
{
  return wt_this_thread;
}

/*
 * Function Name:	wt_thread_yield
 * Description:		give a CPU time to other thread
 * Arguments:		none
 * Return Value:	none
 */
void
wt_thread_yield(void)
{
#if	DEBUG
  fprintf(stderr, "DEBUG[yield] %s\n", wt_this_thread->name);
#endif	/* DEBUG */
  if (wt_this_thread != vp_idle_thread) {
    wt_append_run_q(wt_this_thread);
  }
  wt_context_switch(wt_this_thread);
}

/*
 * Function Name:	wt_thread_cancel
 * Description:		cancel execution of a thread
 * Arguments:		target thread
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_thread_cancel(thd)
	wt_thread_t	thd;
{
  return 0;	/* Not yet implemented */
}

/*
 * Function Name:	wt_thread_once
 * Description:		dynamic package initialization
 * Arguments:		once control, init routine
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_thread_once(once_control, init_routine)
	wt_once_t *	once_control;
	void		(*init_routine)(void);
{
  if (once_control->state == WT_NEED_INIT) {
    init_routine();
    once_control->state = WT_DONE_INIT;
    return 0;
  }
  return -1;
}

/*
 * Function Name:	wt_thread_run
 * Description:		start thread engine
 * Arguments:		none
 * Return Value:	none
 */
void
wt_thread_run(void)
{
  sigset_t	sigs;

  wt_remove_run_q(vp_main_thread);
  vp_active_threads--;
#if	DEBUG
  fprintf(stderr, "DEBUG[run] %s\n", vp_main_thread->name);
#endif	/* DEBUG */
  wt_context_switch(vp_main_thread);

  sigemptyset(&sigs);
  sigsuspend(&sigs);
}

/*****************************************************************************/

/*
 * Function Name:	wt_utime
 * Description:		get time of day
 * Arguments:		none
 * Return Value:	value of current time
 */
wt_utime_t
wt_utime(void)
{
  struct timeval	tv;
  wt_utime_t		now;

  (void)gettimeofday(&tv, NULL);
  now = (wt_utime_t)WT_UTIME_SEC(tv.tv_sec) + (wt_utime_t)tv.tv_usec;

  return now;
}

/*
 * Function Name:	wt_usleep
 * Description:		suspend thread execution for an interval measured
 *			in microseconds
 * Arguments:		microseconds
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_usleep(usecs)
	wt_utime_t	usecs;
{
  wt_thread_t	me = wt_this_thread;

#if	DEBUG
  fprintf(stderr, "DEBUG[usleep] %s\n", me->name);
#endif	/* DEBUG */
  wt_remove_run_q(me);
  wt_append_sleep_q(me, usecs);

  wt_context_switch(me);

  return 0;
}

/*
 * Function Name:	wt_sleep
 * Description:		suspend thread execution for an interval measured
 *			in seconds
 * Arguments:		seconds
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_sleep(secs)
	int	secs;
{
  return wt_usleep(secs < 0 ? WT_UTIME_NO_TIMEOUT : WT_UTIME_SEC(secs));
}

/*****************************************************************************/

/*
 * Function Name:	wt_poll
 * Description:		synchronous I/O multiplexing
 * Arguments:		wt_pollfd_t, timeout
 * Return Value:	successful: ready fd, timeout: 0, otherwise: -1
 */
int
wt_poll(pfd, timeout)
	wt_pollfd_t *	pfd;
	wt_utime_t	timeout;
{
  wt_thread_t	me = wt_this_thread;
  wt_pollq_t *	pq;
  volatile int	fd = -1;

#if	DEBUG
  fprintf(stderr, "DEBUG[poll0] %s fd=%d\n", me->name, pfd->fd);
#endif	/* DEBUG */
#if	HAVE_KQUEUE
  if (pfd->events & WT_POLL_IN)	 { kq_add(pfd->fd, EVFILT_READ); }
  if (pfd->events & WT_POLL_OUT) { kq_add(pfd->fd, EVFILT_WRITE); }
#else
  if (pfd->events & WT_POLL_IN)	 { FD_SET(pfd->fd, &vp_io_rfds); }
  if (pfd->events & WT_POLL_OUT) { FD_SET(pfd->fd, &vp_io_wfds); }
  if (vp_io_maxfd < pfd->fd)	 { vp_io_maxfd = pfd->fd; }
#endif	/* HAVE_KQUEUE */

  pq = (wt_pollq_t *)calloc(1, sizeof(wt_pollq_t));
  if (pq == NULL) {
    errno = ENOMEM;
    return -1;
  }
  wt_clist_init(&pq->q, pq);
  pq->thread	= me;
  pq->poll	= *pfd;
  pq->on_q	= true;

  wt_remove_run_q(me);
  wt_append_io_q(pq, timeout);

  wt_context_switch(me);

#if	DEBUG
  fprintf(stderr, "DEBUG[poll1] %s fd=%d revents=%u\n",
	  pq->thread->name, pq->poll.fd, pq->poll.revents);
#endif	/* DEBUG */
  if (pq->on_q) {
    wt_remove_io_q(pq);
  }
  if (pq->poll.revents != WT_POLL_NONE) {
    fd		 = pq->poll.fd;	/* set ready descriptor */
    pfd->revents = pq->poll.revents;
  }
  free((wt_pollq_t *)pq);

  /* Is time over? */
  if (me->flags & WT_FL_TIMEDOUT) {
    me->flags &= ~WT_FL_TIMEDOUT;
    errno = ETIMEDOUT;
    fd	  = 0;
  }

  return fd;
}

/*****************************************************************************/

/*
 * Function Name:	wt_vp_scheduler
 * Description:		make a schedule of threads
 * Arguments:		none
 * Return Value:	none
 */
void
wt_vp_scheduler(void)
{
#if	DEBUG
  debug_dump_q();
  fprintf(stderr, "DEBUG[scheduler] %s -> ", wt_this_thread->name);
#endif	/* DEBUG */
  if (vp_run_q.next != &vp_run_q) {	/* change thread to next one */
    wt_this_thread = wt_thread_ptr(vp_run_q.next);
#if	DEBUG
    fprintf(stderr, "%s\n", wt_this_thread->name);	// DEBUG
#endif	/* DEBUG */
    wt_remove_run_q(wt_this_thread);
    wt_append_run_q(wt_this_thread);	// added on Sep 16, 2004
  }
  else {				/* no running threads */
    wt_this_thread = vp_idle_thread;
#if	DEBUG
    fprintf(stderr, "%s.\n", wt_this_thread->name);	// DEBUG
#endif	/* DEBUG */
  }
  wt_this_thread->state = WT_ST_RUNNING;
  wt_context_restore(wt_this_thread);
}

static void
vp_run_idle(dummy)
	void *	dummy __unused;
{
#if	DEBUG
  fprintf(stderr, "DEBUG[vp_run_idle] %s\n", wt_this_thread->name);
#endif	/* DEBUG */
  while (vp_active_threads > 0) {
    errno = 0;
    vp_check_io();
    if (vp_wait_child > 0) {
      int	n;

      n = vp_check_proc();
      if (n > 0) {
	vp_wait_child -= n;
#if	DEBUG
	fprintf(stderr, "DEBUG[vp_run_idle] vp_wait_child=%d\n", vp_wait_child);
#endif	/* DEBUG */
      }
      else if (errno == ECHILD) {
	vp_wait_child = 0;
#if	DEBUG
	fprintf(stderr, "DEBUG[vp_run_idle] no more unwaited child process\n");
#endif	/* DEBUG */
      }
    }
    vp_check_clock();
    wt_context_switch(wt_this_thread);
  }
#if	DEBUG
  fprintf(stderr, "all thread are finished: %d\n", vp_active_threads);
#endif	/* DEBUG */
  exit(0);
}

#if	HAVE_KQUEUE
static void
vp_check_io(void)
{
  struct timespec	timeout;
  struct timespec *	tvp;

  if (vp_wait_child > 0) {			/* polling */
    timeout.tv_sec	= 0L;
    timeout.tv_nsec	= 0L;
    tvp			= &timeout;
  }
  else if (wt_clist_is_empty(&vp_sleep_q)) {	/* blocking */
    tvp = NULL;
  }
  else {					/* timer */
    timeout.tv_sec	= (long)WT_UTIME_TO_SEC(vp_min_sleep);
    timeout.tv_nsec	= (long)WT_UTIME_TO_NSEC(vp_min_sleep % WT_UTIME_UNIT);
    tvp			= &timeout;
#if	DEBUG
    fprintf(stderr, "DEBUG[vp_check_io] timeout=(%ld.%ld)\n", timeout.tv_sec, timeout.tv_nsec);
#endif	/* DEBUG */
  }

  /* check I/O operation */
#if	DEBUG
  fprintf(stderr, "DEBUG[kevent0] %s blocking\n", wt_this_thread->name);
#endif	/* DEBUG */
  switch (kevent(vp_kq_fd, vp_kq_events, vp_kq_nevents,
		 vp_kq_revents, vp_kq_nrevents, tvp)) {
    case 0:
      errno = ETIMEDOUT;
      break;
    case -1:
      if (errno == EBADF) {
	vp_find_bad_fd();
      }
#if	DEBUG
      else if (errno == EINTR) {	/* set by wt_spawn() */
	fprintf(stderr, "DEBUG[kevent] %s interrupted\n", wt_this_thread->name);
      }
#endif	/* DEBUG */
      break;
    default: {
	register wt_clist_t *	cp;
	register wt_clist_t *	cp_next;

	for (cp = vp_io_q.next; cp != &vp_io_q; cp = cp_next) {
	  wt_pollq_t *		pq	= wt_pollq_ptr(cp);
	  unsigned short	revents	= WT_POLL_NONE;

	  cp_next = cp->next;
	  /* check events */
	  if ((pq->poll.events & WT_POLL_IN) &&
	      kq_check(pq->poll.fd, EVFILT_READ)) {
	    revents |= WT_POLL_IN;
#if	DEBUG
	    fprintf(stderr, "DEBUG[selected] fd=%d in\n", pq->poll.fd);
#endif	/* DEBUG */
	  }
	  if ((pq->poll.events & WT_POLL_OUT) &&
	      kq_check(pq->poll.fd, EVFILT_WRITE)) {
	    revents |= WT_POLL_OUT;
#if	DEBUG
	    fprintf(stderr, "DEBUG[selected] fd=%d out\n", pq->poll.fd);
#endif	/* DEBUG */
	  }
	  /* set revents */
	  pq->poll.revents = revents;
	  if (revents != WT_POLL_NONE) {
	    wt_remove_io_q(pq);
	    wt_append_run_q(pq->thread);
	  }
	}
      }
      break;
  }
  vp_kq_nevents = 0;	/* clear events */
#if	DEBUG
  fprintf(stderr, "DEBUG[kqueue1] %s errno=%d\n", wt_this_thread->name, errno);
#endif	/* DEBUG */
}
#else
static void
vp_check_io(void)
{
  struct timeval	timeout;
  struct timeval *	tvp;
  fd_set		rset = vp_io_rfds;
  fd_set		wset = vp_io_wfds;
  fd_set *		rp   = &rset;
  fd_set *		wp   = &wset;

  if (vp_io_maxfd == -1) {
    return;
  }

  if (vp_wait_child > 0) {			/* polling */
    timeout.tv_sec	= 0L;
    timeout.tv_usec	= 0L;
    tvp			= &timeout;
  }
  else if (wt_clist_is_empty(&vp_sleep_q)) {	/* blocking */
    tvp = NULL;
  }
  else {					/* timer */
    timeout.tv_sec	= (long)WT_UTIME_TO_SEC(vp_min_sleep);
    timeout.tv_usec	= (long)(vp_min_sleep % WT_UTIME_UNIT);
    tvp			= &timeout;
  }

  /* check I/O operation */
#if	DEBUG
  fprintf(stderr, "DEBUG[select0] %s blocking: maxfd=%d\n", wt_this_thread->name, vp_io_maxfd);
#endif	/* DEBUG */
  switch (select(vp_io_maxfd + 1, rp, wp, NULL, tvp)) {
    case 0:
      errno = ETIMEDOUT;
      break;
    case -1:
      if (errno == EBADF) {
	vp_find_bad_fd();
      }
      break;
    default: {
	register wt_clist_t *	cp;
	register wt_clist_t *	cp_next;

	vp_io_maxfd = -1;
	for (cp = vp_io_q.next; cp != &vp_io_q; cp = cp_next) {
	  wt_pollq_t *		pq	= wt_pollq_ptr(cp);
	  unsigned short	revents	= WT_POLL_NONE;

	  cp_next = cp->next;
	  /* check events */
	  if ((pq->poll.events & WT_POLL_IN) && FD_ISSET(pq->poll.fd, rp)) {
	    revents |= WT_POLL_IN;
#if	DEBUG
	    fprintf(stderr, "DEBUG[selected] fd=%d in\n", pq->poll.fd);
#endif	/* DEBUG */
	  }
	  if ((pq->poll.events & WT_POLL_OUT) && FD_ISSET(pq->poll.fd, wp)) {
	    revents |= WT_POLL_OUT;
#if	DEBUG
	    fprintf(stderr, "DEBUG[selected] fd=%d out\n", pq->poll.fd);
#endif	/* DEBUG */
	  }
	  /* set revents */
	  pq->poll.revents = revents;
	  if (revents != WT_POLL_NONE) {
	    wt_remove_io_q(pq);
	    wt_append_run_q(pq->thread);
	  }
	  /* update maxfd */
	  if (pq->poll.fd > vp_io_maxfd) {
	    vp_io_maxfd = pq->poll.fd;
	  }
	}
      }
      break;
  }
#if	DEBUG
  fprintf(stderr, "DEBUG[select1] %s errno=%d\n", wt_this_thread->name, errno);
#endif	/* DEBUG */
}
#endif	/* HAVE_KQUEUE */

static void
vp_find_bad_fd(void)
{
  register wt_clist_t *	cp;
  register wt_clist_t *	cp_next;

#ifndef	HAVE_KQUEUE
  vp_io_maxfd = -1;
#endif	/* HAVE_KQUEUE */
  for (cp = vp_io_q.next; cp != &vp_io_q; cp = cp_next) {
    wt_pollq_t *	pq = wt_pollq_ptr(cp);

    cp_next = cp->next;
    pq->poll.revents = WT_POLL_NONE;
    if (pq->poll.events != WT_POLL_NONE) {
      if (fcntl(pq->poll.fd, F_GETFL, 0) == -1) {
	pq->poll.revents = WT_POLL_INVAL;
	wt_remove_io_q(pq);
	wt_append_run_q(pq->thread);
      }
#ifndef	HAVE_KQUEUE
      else {
	/* update maxfd */
	if (pq->poll.fd > vp_io_maxfd) {
	  vp_io_maxfd = pq->poll.fd;
	}
      }
#endif	/* HAVE_KQUEUE */
    }
  }
}

static void
vp_check_clock(void)
{
  wt_utime_t		now;
  wt_utime_t		eltime;	/* elapsed time */
  register wt_clist_t *	sp;
  register wt_clist_t *	sp_next;

  now		= wt_utime();
  eltime	= now - vp_last_clock;
  vp_last_clock	= now;

  for (sp = vp_sleep_q.next; sp != &vp_sleep_q; sp = sp_next) {
    wt_thread_t	thd = wt_thread_ptr(sp);

    sp_next = sp->next;
    if (thd->state == WT_ST_SUSPENDED) {
      continue;
    }
    if (thd->flags & WT_FL_ON_SLEEPQ) {
      if (eltime < thd->sleep) {
	thd->sleep -= eltime;
	if (vp_min_sleep > thd->sleep) {
  	  vp_min_sleep = thd->sleep;
   	}
      }
      else {	/* expired */
	wt_remove_sleep_q(thd);
	if (thd->state == WT_ST_WAIT_COND ||
	    (thd->flags & WT_FL_WAIT_PID) || (thd->flags & WT_FL_WAIT_IO)) {
	  thd->flags |= WT_FL_TIMEDOUT;
	}
	wt_append_run_q(thd);
      }
    }
  }
}

/*****************************************************************************/

void
wt_append_run_q(thd)
	wt_thread_t	thd;
{
  thd->state = WT_ST_RUNNABLE;
  wt_clist_append_link(&thd->q, &vp_run_q);
#if	DEBUG
  fprintf(stderr, "DEBUG[+runq] %s\n", thd->name);
#endif	/* DEBUG */
}

void
wt_remove_run_q(thd)
	wt_thread_t	thd;
{
  wt_clist_remove_link(&thd->q);
#if	DEBUG
  fprintf(stderr, "DEBUG[-runq] %s\n", thd->name);
#endif	/* DEBUG */
}

void
wt_append_sleep_q(thd, timeout)
	wt_thread_t	thd;
	wt_utime_t	timeout;
{
  if (timeout < vp_min_sleep) {
    vp_min_sleep = timeout;
  }

  if (timeout == WT_UTIME_NO_TIMEOUT) {
    thd->state = WT_ST_SUSPENDED;
  }
  else {
    thd->state	= WT_ST_SLEEPING;
  }
  thd->flags |= WT_FL_ON_SLEEPQ;
  thd->sleep  = timeout;
  wt_clist_append_link(&thd->q, &vp_sleep_q);
#if	DEBUG
  fprintf(stderr, "DEBUG[+sleepq] %s (%ld[usec])\n", thd->name, thd->sleep);
#endif	/* DEBUG */
}

void
wt_remove_sleep_q(thd)
	wt_thread_t	thd;
{
  wt_clist_remove_link(&thd->q);
#if	DEBUG
  fprintf(stderr, "DEBUG[-sleepq] %s\n", thd->name);
#endif	/* DEBUG */

  thd->flags &= ~WT_FL_ON_SLEEPQ;
  thd->sleep = 0;
  if (!wt_clist_is_empty(&vp_sleep_q)) {
    register wt_clist_t *	sp;

    for (sp = vp_sleep_q.next; sp != &vp_sleep_q; sp = sp->next) {
      wt_thread_t	xthd = wt_thread_ptr(sp);

      if (vp_min_sleep > xthd->sleep) {
	vp_min_sleep = xthd->sleep;
      }
    }
  }
  else {
    vp_min_sleep = WT_UTIME_NO_TIMEOUT;
  }
}

void
wt_append_zombie_q(thd)
	wt_thread_t	thd;
{
  thd->state = WT_ST_ZOMBIE;
  thd->flags = 0;
  wt_clist_append_link(&thd->q, &vp_zombie_q);
#if	DEBUG
  fprintf(stderr, "DEBUG[+zombieq] %s\n", thd->name);
#endif	/* DEBUG */
}

void
wt_remove_zombie_q(thd)
	wt_thread_t	thd;
{
  wt_clist_remove_link(&thd->q);
#if	DEBUG
  fprintf(stderr, "DEBUG[-zombieq] %s\n", thd->name);
#endif	/* DEBUG */
}

void
wt_append_io_q(pq, timeout)
	wt_pollq_t *	pq;
	wt_utime_t	timeout;
{
  wt_clist_append_link(&pq->q, &vp_io_q);
  pq->thread->flags |= WT_FL_WAIT_IO;
#if	DEBUG
  fprintf(stderr, "DEBUG[+ioq] %s\n", pq->thread->name);
#endif	/* DEBUG */
  if (timeout != WT_UTIME_NO_TIMEOUT) {
    wt_append_sleep_q(pq->thread, timeout);
  }
}

void
wt_remove_io_q(pq)
	wt_pollq_t *	pq;
{
  wt_clist_remove_link(&pq->q);
  pq->thread->flags &= ~WT_FL_WAIT_IO;
#if	DEBUG
  fprintf(stderr, "DEBUG[-ioq] %s\n", pq->thread->name);
#endif	/* DEBUG */
  if (pq->thread->flags & WT_FL_ON_SLEEPQ) {
    wt_remove_sleep_q(pq->thread);
  }

  pq->on_q = false;
#if	HAVE_KQUEUE
  if (pq->poll.events & WT_POLL_IN)  { kq_del(pq->poll.fd, EVFILT_READ); }
  if (pq->poll.events & WT_POLL_OUT) { kq_del(pq->poll.fd, EVFILT_WRITE); }
#else
  if (pq->poll.events & WT_POLL_IN)  { FD_CLR(pq->poll.fd, &vp_io_rfds); }
  if (pq->poll.events & WT_POLL_OUT) { FD_CLR(pq->poll.fd, &vp_io_wfds); }
#endif	/* HAVE_KQUEUE */
}

/*****************************************************************************/

/* save machine context */
#define	mctx_save(mctx)	\
	( (mctx)->ecode = errno, \
	  sigprocmask(SIG_SETMASK, &(((mctx_t *)mctx)->sigs), NULL), \
	  sigsetjmp(((mctx_t *)mctx)->jb, 1) )

/* restore machine context */
#define	mctx_restore(mctx) \
	( errno = (mctx)->ecode, siglongjmp(((mctx_t *)mctx)->jb, 1))

/* switch machine context */
#define	mctx_switch(mctx_old,mctx_new) \
	if (mctx_save(mctx_old) == 0) \
		mctx_restore(mctx_new)

static volatile mctx_t		mctx_caller;
static volatile sig_atomic_t	mctx_called;
static volatile jmp_buf		mctx_trampoline;
static mctx_t * volatile	mctx_creating = NULL;
static void *			(* volatile mctx_creating_func)(void *) = NULL;
static void * volatile		mctx_creating_arg = NULL;
static volatile sigset_t	mctx_creating_sigs;

/* create machine context */
static int
mctx_create(mctx, ep_addr, ep_arg, sk_addr, sk_size)
	mctx_t *	mctx;
	void *		(*ep_addr)(void *);	/* entry point address */
	void *		ep_arg;			/* entry point arguments */
	void *		sk_addr;		/* stack address */
	size_t		sk_size;		/* stack size */
{
  struct sigaction	sa;
  struct sigaction	osa;
  struct sigaltstack	ss;
  struct sigaltstack	oss;
  sigset_t		sigs;
  sigset_t		osigs;

  /*
   * Step 1:
   * Preserve the current signal mask and block an arbitrary worker signal
   * (we use SIGUSR1, but any signal can be used for this - even an already
   * used one). This worker signal is later temporarily required for the
   * trampoline step.
   */
  sigemptyset(&sigs);
  sigaddset(&sigs, SIGUSR1);
  sigprocmask(SIG_BLOCK, &sigs, &osigs);

  /*
   * Step 2:
   * Preserve a possibly existing signal action for the worker signal and
   * configure a trampoline function as the new temporary signal action. The
   * signal delivery is configured to occur on an alternate signal stack
   * (see next step).
   */
  memset((void *)&sa, 0, sizeof(struct sigaction));
  sa.sa_handler	= mctx_create_trampoline;
  sa.sa_flags	= SA_ONSTACK;
  sigemptyset(&sa.sa_mask);
  if (sigaction(SIGUSR1, &sa, &osa) == -1) {
    return -1;
  }

  /*
   * Step 3:
   * Preserve a possibly active alternate signal stack and configure the
   * memory chunk starting at sk_addr as the new temporary alternate signal
   * stack of length sk_size.
   */
  ss.ss_sp	= sk_addr;
  ss.ss_size	= sk_size;
  ss.ss_flags	= 0;
  if (sigaltstack(&ss, &oss) == -1) {
    return -1;
  }

  /*
   * Step 4:
   * Save parameters for the trampoline step (mctx, ep_addr, ep_arg, etc.)
   * in global variables, send the current process the worker signal,
   * temporarily unblock it and this way allow it to be delivered on the
   * signal stack in order to transfer execution control to the trampoline
   * function.
   */
  mctx_creating		= mctx;
  mctx_creating_func	= ep_addr;
  mctx_creating_arg	= ep_arg;
  mctx_creating_sigs	= osigs;
  mctx_called		= false;
  kill(getpid(), SIGUSR1);
  sigfillset(&sigs);
  sigdelset(&sigs, SIGUSR1);
  while (!mctx_called) {
    sigsuspend(&sigs);
  }

  /*
   * Step 6:
   * Restore the preserved alternate signal stack, preserved signal action
   * and preserved signal mask for worker signal. This way an existing
   * application configuration for the worker signal is restored.
   */
  sigaltstack(NULL, &ss);
  ss.ss_flags = SS_DISABLE;
  if (sigaltstack(&ss, NULL) == -1) {
    return -1;
  }
  sigaltstack(NULL, &ss);
  if (!(ss.ss_flags & SS_DISABLE)) {
    return -1;
  }
  if (!(oss.ss_flags & SS_DISABLE)) {
    sigaltstack(&oss, NULL);
  }

  /* Restore the old sigusr1 signal handler and mask */
  sigaction(SIGUSR1, &osa, NULL);
  sigprocmask(SIG_SETMASK, &osigs, NULL);

  /* Initialize additional ingredients of the machine context structure */
  sigemptyset(&mctx->sigs);
  mctx->ecode = 0;

  /*
   * Step 7
   * Save the current machine context of mctx_create. This allows us to
   * return to this point after the next trampoline step.
   *
   * Step 8:
   * Restore the previously saved machine context of the trampoline function
   * (mctx) to again transfer execution control onto the alternate stack,
   * but this time without(!) signal handler scope.
   */
  if (mctx_save(&mctx_caller) == 0) {
    longjmp(*((jmp_buf *)&mctx_trampoline), 1);
  }

  /*
   * Step 14:
   * Return to the calling application.
   */
  return 0;
}

static void
mctx_create_trampoline(sig)
	int	sig __unused;
{
  sigset_t	sigs;

  /*
   * Unblock the SIGUSR1 signal that got us here because we have to avoid
   * that it's stored in blocked state in the machine context.
   */
  sigemptyset(&sigs);
  sigaddset(&sigs, SIGUSR1);
  sigprocmask(SIG_UNBLOCK, &sigs, NULL);

  /*
   * Step 5:
   * After the trampoline function asynchronously entered, save its machine
   * context in the mctx structure and immediately return from it to
   * terminate the signal handler scope.
   */
  if (setjmp(*((jmp_buf *)&mctx_trampoline)) == 0) {
    mctx_called = true;
    return;
  }

  /*
   * Step 9:
   * After reaching the trampoline function (mctx) again, immediately
   * bootstrap into a clean stack frame by just calling a second function.
   */
  mctx_create_boot();
}

static void
mctx_create_boot(void)
{
  mctx_t * volatile	mctx_starting;
  void *		(* volatile mctx_starting_func)(void *);
  void *		mctx_starting_arg;
  void *		retval;

  /*
   * Step 10:
   * Set the new signal mask to be the same as the original signal mask
   * which was active when mctx_create was called. This is required because
   * in the first trampoline step we usually had at least the worker signal
   * blocked.
   */
  sigprocmask(SIG_SETMASK, (sigset_t *)&mctx_creating_sigs, NULL);

  /*
   * Step 11:
   * Load the passed startup information (ep_addr, ep_arg) from mctx_create
   * into local (stack-based) variables. This is important because their
   * values have to be preserved in machine context dependent memory until
   * the created machine context is the first time restored by the application.
   */
  mctx_starting		= mctx_creating;
  mctx_starting_func	= mctx_creating_func;
  mctx_starting_arg	= mctx_creating_arg;

  /*
   * Step 12
   * Save the current machine context for later restoring by the calling
   * application.
   *
   * Step 13:
   * Restore the previously saved machine context of mctx_creat to transfer
   * execution control back to it.
   */
#if	DEBUG
  fprintf(stderr, "DEBUG[boot0] %s\n", mctx_starting->name);
#endif	/* DEBUG */
  mctx_switch(mctx_starting, &mctx_caller);
#if	DEBUG
  fprintf(stderr, "DEBUG[boot1] %s\n", mctx_starting->name);
#endif	/* DEBUG */

  /* The thread "magically" starts... */
  retval = mctx_starting_func(mctx_starting_arg);

  wt_thread_exit(retval);	/* really need? */

  /* NOTREACHED */
  abort();
}

/*****************************************************************************/

#if	HAVE_KQUEUE
static int
kq_init(nfiles)
	int	nfiles;
{
  vp_kq_fd = kqueue();
  if (vp_kq_fd == -1) {
    return -1;
  }

  vp_kq_events =
	(struct kevent *)calloc((size_t)nfiles, sizeof(struct kevent));
  if (vp_kq_events == NULL) {
    return -1;
  }
  vp_kq_nevents = 0;

  vp_kq_revents =
	(struct kevent *)calloc((size_t)nfiles, sizeof(struct kevent));
  if (vp_kq_revents == NULL) {
    free((struct kevent *)vp_kq_events);
    return -1;
  }
  vp_kq_nrevents = nfiles;

  return 0;
}

static void
kq_add(fd, filter)
	int	fd;
	short	filter;
{
  vp_kq_events[vp_kq_nevents].ident	= fd;
  vp_kq_events[vp_kq_nevents].filter	= filter;
  vp_kq_events[vp_kq_nevents].flags	= EV_ADD;
  vp_kq_events[vp_kq_nevents].fflags	= 0;
  vp_kq_events[vp_kq_nevents].data	= 0;
  vp_kq_events[vp_kq_nevents].udata	= NULL;
  vp_kq_nevents++;
}

static void
kq_del(fd, filter)
	int	fd;
	short	filter;
{
  vp_kq_events[vp_kq_nevents].ident	= fd;
  vp_kq_events[vp_kq_nevents].filter	= filter;
  vp_kq_events[vp_kq_nevents].flags	= EV_DELETE;
  vp_kq_events[vp_kq_nevents].fflags	= 0;
  vp_kq_events[vp_kq_nevents].data	= 0;
  vp_kq_events[vp_kq_nevents].udata	= NULL;
  vp_kq_nevents++;
}

static bool
kq_check(fd, filter)
	int	fd;
	short	filter;
{
  register int	i;

  for (i = 0; i < vp_kq_nrevents; i++) {
    if (vp_kq_revents[i].ident == fd && vp_kq_revents[i].filter == filter) {
      return true;
    }
  }

  return false;
}
#endif	/* HAVE_KQUEUE */

/*****************************************************************************/

#if	DEBUG
static void
debug_dump_q(void)
{
  register wt_clist_t *	cp;

  fprintf(stderr, "DEBUG[dump] ");
  if (!wt_clist_is_empty(&wt_this_thread->wait_q)) {
    fprintf(stderr, "wait_q: ");
  }
  fprintf(stderr, "%s\n", wt_this_thread->name);
  fprintf(stderr, "DEBUG[dump] runq:");
  for (cp = vp_run_q.next; cp != &vp_run_q; cp = cp->next) {
    fprintf(stderr, " %s", (wt_thread_ptr(cp))->name);
  }
  fprintf(stderr, "\n");

  fprintf(stderr, "DEBUG[dump] sleepq:");
  for (cp = vp_sleep_q.next; cp != &vp_sleep_q; cp = cp->next) {
    fprintf(stderr, " %s", (wt_thread_ptr(cp))->name);
  }
  fprintf(stderr, "\n");

  fprintf(stderr, "DEBUG[dump] zombieq:");
  for (cp = vp_zombie_q.next; cp != &vp_zombie_q; cp = cp->next) {
    fprintf(stderr, " %s", (wt_thread_ptr(cp))->name);
  }
  fprintf(stderr, "\n");

  fprintf(stderr, "DEBUG[dump] ioq:");
  for (cp = vp_io_q.next; cp != &vp_io_q; cp = cp->next) {
    wt_pollq_t *	pq = wt_pollq_ptr(cp);
    fprintf(stderr, " %s", pq->thread->name);
  }
  fprintf(stderr, "\n");
}
#endif	/* DEBUG */
