/*****************************************************************************
 *
 * FILE:	secure.c
 * DESCRIPTION:	Wyvern: secure library
 * DATE:	Wed, May 21 2003
 * UPDATED:	Tue, Sep 11 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2003-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2003-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: secure.c,v 1.26 2007/09/12 02:17:14 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#if	__STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif	/* __STDC__ */
#include <ctype.h>
#include <errno.h>
#include "ws.h"

/******************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
/* users/groups */
#define UID_ROOT	(0)	/* super user id in /etc/passwd */
#define GID_WHEEL	(0)	/* super user's group id in /etc/passwd */
#define UID_NOBODY	(65534)	/* nobody user id in /etc/passwd */
#define GID_NOGROUP	(65533)	/* nogroup id in /etc/passwd */

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	out_of_memory_error(void);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static uid_t	saved_uid;	/* effective user ID */
static gid_t	saved_gid;	/* effective group ID */
static void	(*nomem_callback)(void) = NULL;

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	ws_drop_privilege
 * Description:		drop the privilege of the process
 * Arguments:		uid, gid
 * Return Value:	successful: 0, error: -1, otherwise: 1
 */
int
ws_drop_privilege(uid, gid)
	uid_t	uid;
	gid_t	gid;
{
  /* drop privilege */
  if (uid == UID_ROOT) {
    uid = UID_NOBODY;
  }
  if (gid == GID_WHEEL) {
    gid = GID_NOGROUP;
  }

  /* change owner/group */
  if (getuid() == UID_ROOT) {
    struct passwd *	pwd;

    pwd = getpwuid(uid);
    if (pwd == NULL) {
      return -1;
    }

    if (initgroups(pwd->pw_name, (int)pwd->pw_gid) == -1) {
      return -1;
    }

    /* drop privilege */
    if (setgid(gid) == -1 || setuid(uid) == -1) {
      return -1;
    }

    return 0;
  }

  return 1;
}

/*
 * Function Name:	ws_revoke_privilege
 * Description:		demote the privilege of the process
 * Arguments:		uid, gid
 * Return Value:	successful: 0, error: -1, otherwise: 1
 */
int
ws_revoke_privilege(uid, gid)
	uid_t	uid;
	gid_t	gid;
{
  if (getuid() == UID_ROOT) {
    uid_t	exec_uid;
    gid_t	exec_gid;

    /* save effective uid/gid */
    saved_uid = geteuid();
    saved_gid = getegid();

    exec_uid = uid == UID_ROOT  ? UID_NOBODY  : uid;
    exec_gid = gid == GID_WHEEL ? GID_NOGROUP : gid;

    /* drop privilege */
    if (setegid(exec_gid) == -1 || seteuid(exec_uid) == -1) {
      return -1;
    }

    return 0;
  }

  return 1;
}

/*
 * Function Name:	ws_grant_privilege
 * Description:		raise the privilege of the process
 * Arguments:		none
 * Return Value:	successful: 0, error: -1, otherwise: 1
 */
int
ws_grant_privilege(void)
{
  if (getuid() == UID_ROOT) {
    if (setegid(saved_gid) == -1 || seteuid(saved_uid) == -1) {
      return -1;
    }

    return 0;
  }

  return 1;
}

/*****************************************************************************/

/*
 * Function Name:	ws_sanitize_uri
 * Description:		sanitize URI
 * Arguments:		uri string
 * Return Value:	successful: sanitized uri string, otherwise: NULL
 * Note:		The free() can releases the allocated uri string.
 */
char *
ws_sanitize_uri(uri)
	const char *	uri;
{
  char *		retp;
  size_t		len;
  register char *	s;
  char			c;

  /* empty uri? */
  if (uri == NULL || uri[0] == '\0') {
    errno = EINVAL;
    return NULL;
  }

  len	= (size_t)((double)strlen(uri) * 4.86);
  retp	= (char *)ws_calloc(len + 1, sizeof(char));
  s	= retp;
  while ((c = *uri) != '\0') {
    /* escape meta characters for cross-site scripting */
    if (c == '<') {
      *s++ = '&'; *s++ = 'l'; *s++ = 't'; *s++ = ';';
    }
    else if (c == '>') {
      *s++ = '&'; *s++ = 'g'; *s++ = 't'; *s++ = ';';
    }
    else if (c == '"') {
      *s++ = '&'; *s++ = 'q'; *s++ = 'u'; *s++ = 'o'; *s++ = 't'; *s++ = ';';
    }
    else if (c == '\'') {
      *s++ = '&'; *s++ = '#'; *s++ = '3'; *s++ = '9'; *s++ = ';';
    }
    else if (c == '(') {
      *s++ = '&'; *s++ = '#'; *s++ = '4'; *s++ = '0'; *s++ = ';';
    }
    else if (c == ')') {
      *s++ = '&'; *s++ = '#'; *s++ = '4'; *s++ = '1'; *s++ = ';';
    }
    else if (c == '&') {
      *s++ = '&'; *s++ = 'a'; *s++ = 'm'; *s++ = 'p'; *s++ = ';';
    }
    /* normal characters */
    else {
      *s++ = c;
    }
    uri++;
  }
  *s = '\0';

  return retp;
}

/*****************************************************************************/

/*
 * Function Name:	ws_safe_open
 * Description:		open file safety
 * Arguments:		filename, open flags, status buffer, symlink check
 * Return Value:	successful: file descriptor, otherwise: -1
 * Notes:		http://www.securityfocus.com/popups/forums/secprog/secure-programming.shtml
 */
int
ws_safe_open(filename, open_flags, ret_sbuf, allow_symlink)
	const char *	filename;
	int		open_flags;
	struct stat *	ret_sbuf;
	int		allow_symlink;	/* true: 1, false: 0 */
{
  struct stat	lsbuf;
  struct stat	fsbuf;
  int		fd;

  /* check filename */
  if (filename == NULL || filename[0] == '\0') {
    errno = EINVAL;
    return -1;
  }

  /* obtain the file's status information */
  if (lstat(filename, &lsbuf) == -1) {
    if (errno == ENOENT && !(open_flags & O_WRONLY)) {
      return -1;
    }
    memset((char *)&lsbuf, 0, sizeof(lsbuf));
  }
  else {
    /* make sure the file is regular file */
    if (!S_ISREG(lsbuf.st_mode)) {
      if (S_ISDIR(lsbuf.st_mode)) {
	errno = EISDIR;
	return -1;
      }
      if (S_ISLNK(lsbuf.st_mode) && !allow_symlink) {
	errno = EMLINK;
	return -1;
      }
      if (!S_ISLNK(lsbuf.st_mode) && allow_symlink) {
	errno = EACCES;
	return -1;
      }
    }

    /* make sure that the file is owned by user (not uid 0) */
    if (lsbuf.st_uid == UID_ROOT) {
      errno = EPERM;
    }
  }

  /* open the file */
again:
  fd = open(filename, open_flags, 0600);
  if (fd == -1) {
    if (errno == EOPNOTSUPP) {	/* no support locking mechanism */
#if	O_SHLOCK
      open_flags &= ~O_SHLOCK;	/* in the underlying filesystem */
#endif	/* O_SHLOCK */
#if	O_EXLOCK
      open_flags &= ~O_EXLOCK;	/* unset O_SHLOCK/O_EXLOCK and try again */
#endif	/* O_EXLOCK */
      goto again;
    }
    return -1;
  }

  /* now we fstat() the file, to make sure it's the same file still! */
  if (fstat(fd, &fsbuf) == -1) {
    close(fd);
    return -1;
  }

  /*
   * here we make sure the i-node and device numbers are the same in the file
   * we actually opened, compared to the file we performed the initial
   * lstat() call on.
   */
  if (!allow_symlink &&	/* not allow symbolic link */
      (lsbuf.st_ino != 0 && lsbuf.st_dev != 0) &&	/* not create newly */
      (lsbuf.st_ino != fsbuf.st_ino || lsbuf.st_dev != fsbuf.st_dev)) {
    close(fd);
    errno = EACCES;
    return -1;
  }
  *ret_sbuf = fsbuf;

  return fd;
}

/*
 * Function Name:	ws_safe_fopen
 * Description:		open file safety
 * Arguments:		filename, open mode
 * Return Value:	successful: file pointer, otherwise: NULL
 */
FILE *
ws_safe_fopen(filename, mode)
	const char *	filename;
	const char *	mode;
{
  int		fd;
  register int	m;
  register int	o;
  struct stat	sbuf;

  switch (*mode) {
    case 'r':
#if	O_SHLOCK
      m = O_RDONLY | O_SHLOCK;
#else
      m = O_RDONLY;
#endif	/* O_SHLOCK */
      o = 0;
      break;
    case 'w':
#if	O_EXLOCK
      m = O_WRONLY | O_EXLOCK;
#else
      m = O_WRONLY;
#endif	/* O_EXLOCK */
      o = O_CREAT  | O_TRUNC;
      break;
    case 'a':
#if	O_EXLOCK
      m = O_WRONLY | O_EXLOCK;
#else
      m = O_WRONLY;
#endif	/* O_EXLOCK */
      o = O_CREAT  | O_APPEND;
      break;
    default:
      errno = EINVAL;
      return NULL;
  }
  if (mode[1] == '+') {
    m = O_RDWR;
  }

  fd = ws_safe_open(filename, m | o, &sbuf, 0);
  if (fd == -1) {
    return NULL;
  }

  return fdopen(fd, mode);
}

/*****************************************************************************/

/*
 * Function Name:	ws_set_error_func
 * Description:		set user specified error handling function
 * Arguments:		none
 * Return Value:	none
 */
void
ws_set_error_func(f)
	void	(*f)(void);
{
  nomem_callback = f;
}

/*
 * Function Name:	out_of_memory_error
 * Description:		called when any of the memory functions fail
 * Arguments:		none
 * Return Value:	none
 */
static void
out_of_memory_error(void)
{
  if (nomem_callback != NULL) {
    nomem_callback();
  }
  else {
    fprintf(stderr, "Out of memory\n");
    exit(71);	/* system error */
  }
}

/*
 * Function Name:	ws_malloc
 * Description:		do as malloc(3)
 * Arguments:		buffer size
 * Return Value:	pointer to buffer allocated.
 *			never returns in case of error.
 */
void *
ws_malloc(size)
	size_t	size;
{
  void *	ret;

  ret = malloc(size);
  if (ret == NULL) {
    out_of_memory_error();
  }

  return ret;
}

/*
 * Function Name:	ws_calloc
 * Description:		do as calloc(3)
 * Arguments:		# of elements, # of bytes for each element
 * Return Value:	pointer to buffer allocated.
 *			never returns in case of error.
 */
void *
ws_calloc(nmemb, size)
	size_t	nmemb;
	size_t	size;
{
  void *	ret;

  ret = calloc(nmemb, size);
  if (ret == NULL) {
    out_of_memory_error();
  }

  return ret;
}

/*
 * Function Name:	ws_realloc
 * Description:		do as realloc(3)
 * Arguments:		pointer to buffer, new buffer size
 * Return Value:	new pointer to buffer allocated.
 *			never returns in case of error.
 */
void *
ws_realloc(ptr, size)
	void *	ptr;
	size_t	size;
{
  void *	ret;

  ret = realloc(ptr, size);
  if (ret == NULL) {
    out_of_memory_error();
  }

  return ret;
}

/*
 * Function Name:	ws_strdup
 * Description:		do as strdup(3)
 * Arguments:		string
 * Return Value:	pointer to buffer allocated.
 *			never returns in case of error.
 */
char *
ws_strdup(s)
	const char *	s;
{
  char *	ret;

  if (s == NULL) {
    errno = EINVAL;
    return NULL;
  }

  ret = strdup(s);
  if (ret == NULL) {
    out_of_memory_error();
  }

  return ret;
}

/*
 * Function Name:	ws_strndup
 * Description:		duplicate a string like as strdup()
 * Arguments:		string, length
 * Return Value:	pointer to buffer allocated.
 *			never returns in case of error.
 */
char *
ws_strndup(s, len)
	const char *	s;
	size_t		len;
{
  char *	dst;
  char *	ret;
  unsigned int	i;

  if (s == NULL) {
    errno = EINVAL;
    return NULL;
  }

  ret = dst = ws_malloc(len + 1);
  for (i = 0; i < len; i++) {
    *dst++ = *s++;
  }
  *dst = '\0';

  return ret;
}

/*
 * Function Name:	ws_asprintf
 * Description:		formatted output conversion
 * Arguments:		returned string, format string
 * Return Value:	successful: output number
 * Note:		modified '/usr/src/lib/libc/stdio/asprintf.c'
 */
#if	1
int
#if __STDC__
ws_asprintf(char **result, char const *fmt, ...)
#else
ws_asprintf(result, fmt, va_alist)
	char **		result;
	const char *	fmt;
	va_dcl
#endif
{
  const char *	p;
  size_t	width;
  va_list	ap;
  int		status = -1;

  /* empty fmt? */
  if (fmt == NULL || fmt[0] == '\0') {
    errno = EINVAL;
    return -1;
  }
  p	= fmt;
  width	= strlen(fmt) + 1;

#if	__STDC__
  va_start(ap, fmt);
#else
  va_start(ap);
#endif	/* __STDC__ */
  /* the following code comes from binutils' vasprintf.c */
  while (*p != '\0') {
    if (*p++ == '%') {
      while (strchr("-+ #0", *p)) {
	p++;
      }
      if (*p == '*') {
	p++;
	width += (size_t)abs(va_arg(ap, int));
      }
      else {
	width += (size_t)strtoul(p, (char **)&p, 10);
      }
      if (*p == '.') {
	p++;
	if (*p == '*') {
	  p++;
	  width += (size_t)abs(va_arg(ap, int));
	}
	else {
	  width += (size_t)strtoul(p, (char **)&p, 10);
	}
      }
      while (strchr("hlL", *p)) {
	p++;
      }
      /* Should be big enough for any format specifier except %s and floats. */
      width += 30;
      switch (*p) {
	case 'd':
	case 'i':
	case 'o':
	case 'u':
	case 'x':
	case 'X':
	case 'c':
	  va_arg(ap, int);
	  break;
	case 'f':
	case 'e':
	case 'E':
	case 'g':
	case 'G':
	  va_arg(ap, double);
	  /* Since an IEEE double can have exponent of 307, we'll make
	     the buffer wide enough to cover the gross case. */
	  width += 307;
	  break;
	case 's': {	/* XXX: modified on Tue, Sep 11 2007 */
	    char *	xp = va_arg(ap, char *);
	    size_t	xl = sizeof("(null)");

	    if (xp != NULL) {
	      xl = strlen(xp);
	    }
	    width += xl;
	  }
	  break;
	case 'p':
	case 'n':
	  va_arg(ap, char *);
	  break;
      }
      p++;
    }
  }
  va_end(ap);

  *result = (char *)ws_malloc(width);
  if (*result != NULL) {
    va_start(ap, fmt);
    status = vsprintf(*result, fmt, ap);
    va_end(ap);
  }

  return status;
}
#else
#if	__FreeBSD__
int
#if	__STDC__
ws_asprintf(char **result, const char *fmt, ...)
#else
ws_asprintf(result, fmt, va_alist)
	char **		result;
	const char *	fmt;
	va_dcl
#endif	/* __STDC__ */
{
  int		size;
  FILE 		f;
  va_list	ap;

  /* empty fmt? */
  if (fmt == NULL || fmt[0] == '\0') {
    errno = EINVAL;
    return -1;
  }

#if	__STDC__
  va_start(ap, fmt);
#else
  va_start(ap);
#endif	/* __STDC__ */
  f._file     = -1;
  f._flags    = __SWR | __SSTR | __SALC;
  f._bf._base = f._p = (unsigned char *)ws_malloc(128);
  f._bf._size = f._w = 127;
  size	= vfprintf(&f, fmt, ap);
  *f._p = '\0';
  va_end(ap);
  f._bf._base = ws_realloc(f._bf._base, (size_t)(f._bf._size + 1));
  *result = (char *)f._bf._base;
  return size;
}
#else
int
#if	__STDC__
ws_asprintf(char **result, const char *fmt, ...)
#else
ws_asprintf(result, fmt, va_alist)
	char **		result;
	const char *	fmt;
	va_dcl
#endif	/* __STDC__ */
{
  char *	buf;
  size_t	blen;	/* buffer length */
  int		size;
  va_list	ap;

  /* empty fmt? */
  if (fmt == NULL || fmt[0] == '\0') {
    errno = EINVAL;
    return -1;
  }

  /* Make a guess */
  blen	= strlen(fmt) + 128;
  buf	= (unsigned char *)ws_malloc(blen);
#if	__STDC__
  va_start(ap, fmt);
#else
  va_start(ap);
#endif	/* __STDC__ */
  size	= vsprintf(buf, fmt, ap);
  va_end(ap);
  if (size <= blen) {
    *result = buf;
    return size;
  }

  /* Too large */
  buf = ws_realloc(buf, (size_t)(size + 1));
#if	__STDC__
  va_start(ap, fmt);
#else
  va_start(ap);
#endif	/* __STDC__ */
  size	= vsprintf(buf, fmt, ap);
  va_end(ap);
  *result = buf;

  return size;
}
#endif	/* __FreeBSD__ */
#endif

/*****************************************************************************/

/*
 * Function Name:	ws_strcmp
 * Description:		compare two strings
 * Arguments:		string, string
 * Return Value:	successful: 0, otherwise: -1
 */
int
ws_strcmp(s1, s2)
	register const char *	s1;
	register const char *	s2;
{
  register size_t	n1;
  register size_t	n2;

  if (s1 == NULL || s2 == NULL) {
    errno = EINVAL;
    return -1;
  }

  n1 = strlen(s1);
  n2 = strlen(s2);
  if (n1 != n2) {
    return -1;
  }

  return ws_strncmp(s1, s2, n1);
}

/*
 * Function Name:	ws_strncmp
 * Description:		compare two strings
 * Arguments:		string, string, length
 * Return Value:	successful: 0, otherwise: -1
 */
int
ws_strncmp(s1, s2, n)
	register const char *	s1;
	register const char *	s2;
	register size_t		n;
{
  if (s1 == NULL || s2 == NULL) {
    errno = EINVAL;
    return -1;
  }

  if (n != 0) {
    do {
      if (*s1 != *s2++) {
	return -1;
      }
      if (*s1++ == '\0') {
	break;
      }
    } while (--n != 0);
  }

  return 0;
}

/*
 * Function Name:	ws_strcasecmp
 * Description:		compare two strings
 * Arguments:		string, string
 * Return Value:	successful: 0, otherwise: -1
 */
int
ws_strcasecmp(s1, s2)
	register const char *	s1;
	register const char *	s2;
{
  register size_t	n1;
  register size_t	n2;

  if (s1 == NULL || s2 == NULL) {
    errno = EINVAL;
    return -1;
  }

  n1 = strlen(s1);
  n2 = strlen(s2);
  if (n1 != n2) {
    return -1;
  }

  return ws_strncasecmp(s1, s2, n1);
}

/*
 * Function Name:	ws_strncasecmp
 * Description:		compare two strings
 * Arguments:		string, string, length
 * Return Value:	successful: 0, otherwise: -1
 */
int
ws_strncasecmp(s1, s2, n)
	register const char *	s1;
	register const char *	s2;
	register size_t		n;
{
  if (s1 == NULL || s2 == NULL) {
    errno = EINVAL;
    return -1;
  }

  if (n != 0) {
    register const unsigned char *	us1 = (const unsigned char *)s1;
    register const unsigned char *	us2 = (const unsigned char *)s2;

    do {
      if (tolower(*us1) != tolower(*us2++)) {
	return -1;
      }
      if (*us1++ == '\0') {
	break;
      }
    } while (--n != 0);
  }

  return 0;
}

#ifndef	HAVE_STRLCAT
/*
 * Function Name:	ws_strncat
 * Description:		boundary checked strncat(3)
 * Arguments:		dst, src, size
 * Return Value:	same as strncat(3)
 */
char *
ws_strncat(dst, src, size)
	char *		dst;
	const char *	src;
	size_t		size;	/* size of dst buffer */
{
  register char *	dp = dst;
  register const char *	sp = src;
  register size_t	n  = size;
  size_t		dlen;

  while (n-- != 0 && *dp != '\0') {
    dp++;
  }
  dlen	= dp - dst;
  n	= size - dlen;

  if (n == 0) {
    return dst;
  }
  while (*sp != '\0') {
    if (n != 1) {
      *dp++ = *sp;
      n--;
    }
    sp++;
  }
  *dp = '\0';

  return dst;
}
#endif	/* HAVE_STRLCAT */

#ifndef	HAVE_STRLCPY
/*
 * Function Name:	ws_strncpy
 * Description:		boundary checked strncpy(3)
 * Arguments:		dst, src, size
 * Return Value:	same as strncpy(3)
 */
char *
ws_strncpy(dst, src, size)
	char *		dst;
	const char *	src;
	size_t		size;	/* size of dst buffer */
{
  register char *	dp = dst;
  register const char *	sp = src;
  register size_t	n  = size;

  if (n != 0 && --n != 0) {
    do {
      if ((*dp++ = *sp++) == '\0') {
	break;
      }
    } while (--n != 0);
  }

  if (n == 0) {
    if (size != 0) {
      *dp = '\0';
    }
  }

  return dst;
}
#endif	/* HAVE_STRLCPY */

/*
 * Function Name:	ws_chmod
 * Description:		chmod(2) to check symbolic link
 * Arguments:		path, mode
 * Return Value:	same as chmod(2)
 */
#if	HAVE_LCHMOD
int
ws_chmod(path, mode)
	const char *	path;
	mode_t		mode;
{
  return lchmod(path, mode);
}
#else
int
ws_chmod(path, mode)
	const char *	path;
	mode_t		mode;
{
  struct stat	sbuf;

  if (lstat(path, &sbuf) == 0 && S_ISLNK(sbuf.st_mode)) {
    errno = ENOSYS;
    return -1;
  }

  return chmod(path, mode);
}
#endif	/* HAVE_LCHMOD */

/*
 * Function Name:	ws_chown
 * Description:		chown(2) to check symbolic link
 * Arguments:		file, uid, gid
 * Return Value:	same as chown(2)
 */
#if	HAVE_LCHOWN
int
ws_chown(file, uid, gid)
	const char *	file;
	uid_t		uid;
	gid_t		gid;
{
  return lchown(file, uid, gid);
}
#else
int
ws_chown(file, uid, gid)
	const char *	file;
	uid_t		uid;
	gid_t		gid;
{
  struct stat	sbuf;

  if (lstat(file, &sbuf) == 0 && S_ISLNK(sbuf.st_mode)) {
    errno = ENOSYS;
    return -1;
  }

  return chown(file, uid, gid);
}
#endif	/* HAVE_LCHOWN */
