/*****************************************************************************
 *
 * FILE:	ssl.c
 * DESCRIPTION:	WTL: secure socket layer functions
 * DATE:	Mon, Jun 30 2003
 * UPDATED:	Wed, Sep  5 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2003-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2003-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: ssl.c,v 1.32 2007/09/04 17:09:04 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif	/* HAVE_SYS_TYPES_H */
#include <sys/mman.h>
#include <errno.h>
#include "wt-ssl.h"
#include "wt-sub.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
/*
 * see ciphers(1)
 * Au:	RSA|DSS
 * Enc:	AES|RC4|RC2|3DES|DES|...
 * Mac:	SHA1|MD5
 */
#if	1
#define	SSL_CIPHER_SUITE \
	"ALL:!ADH:!EXPORT56:!SSLv2:RC4+RSA:+HIGH:+MEDIUM:+LOW:+EXP:+eNULL"
#else
#define	SSL_CIPHER_SUITE \
	"ALL:!ADH:!EXPORT56:!SSLv2:AES+RSA+SHA:@STRENGTH"
#endif
#define	SSL_SERVER_CRT		"server.crt"
#define	SSL_SERVER_KEY		"server.key"
#define	DH_PARAM_512_PEM	"dh_param_512.pem"
#define	DH_PARAM_1024_PEM	"dh_param_1024.pem"

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static SSL_CTX *	ssl_client_init(const char *, const char *,
					const char *);
static SSL_CTX *	ssl_server_init(const char *, const char *,
					const char *);
static int		init_rsa_keys(void);
static RSA *		tmp_rsa_callback(SSL *, int, int);
#if	USE_DH
static int		init_dh_keys(void);
static DH *		tmp_dh_callback(SSL *, int, int);
#endif	/* USE_DH */
static ssize_t		one_read(wt_netfd_t *, char *, wt_utime_t);
static int		wt_netfd_readbuf(wt_netfd_t *, wt_utime_t);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static RSA *	rsa_512	 = NULL;	/* set up ephemeral RSA stuff */
static RSA *	rsa_1024 = NULL;	/* set up ephemeral RSA stuff */
#if	USE_DH
static DH *	dh_512	 = NULL;	/* set up ephemeral DH stuff */
static DH *	dh_1024	 = NULL;	/* set up ephemeral DH stuff */
#endif	/* USE_DH */

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wt_netfd_ssl_tcp_client
 * Description:		initialize network socket for a SSL client
 * Arguments:		hostname, service
 * Return Value:	successful: netfd, otherwise: NULL
 */
wt_netfd_t *
wt_netfd_ssl_tcp_client(nodename, service)
	const char *	nodename;
	const char *	service;
{
  SSL_CTX *	ssl_ctx;
  wt_netfd_t *	nfd;
  register int	status;

  ssl_ctx = ssl_client_init(NULL, NULL, NULL);
  if (ssl_ctx == NULL) {
    return NULL;
  }

  nfd = wt_netfd_tcp_client(nodename, service);
  if (nfd == NULL) {
    SSL_CTX_free(ssl_ctx);
    return NULL;
  }

  nfd->ssl_ctx	= ssl_ctx;
  nfd->ssl	= SSL_new(ssl_ctx);
  if (nfd->ssl == NULL) {
    wt_netfd_ssl_close(nfd);
    return NULL;
  }
  SSL_set_fd(nfd->ssl, nfd->osfd);

  while ((status = SSL_connect(nfd->ssl)) <= 0) {
    switch (SSL_get_error(nfd->ssl, status)) {
      case SSL_ERROR_NONE:
	break;
      case SSL_ERROR_WANT_READ:
      case SSL_ERROR_WANT_WRITE:
	/* wait until the socket becoms writable */
	if (wt_netfd_poll(nfd, WT_POLL_OUT, WT_UTIME_NO_TIMEOUT) == -1) {
	  wt_netfd_ssl_close(nfd);
	  return NULL;
	}
	break;
      default: 
#if	DEBUG
	fprintf(stderr, "DEBUG[wt_netfd_ssl_tcp_client] fatal error=%d ret=%d\n",
		SSL_get_error(nfd->ssl, status), status);
#endif	/* DEBUG */
	wt_netfd_ssl_close(nfd);
	return NULL;
    }
  }

  return nfd;
}

/*
 * FUNCTION NAME:	ssl_client_init
 * DESCRIPTION:		initialize SSL connection for client
 * ARGUMENTS:		certificate file, key file, cipher list
 * RETURN VALUE:	successful: SSL context, otherwise: NULL
 */
static SSL_CTX *
ssl_client_init(certfile, keyfile, cipher)
	const char *	certfile;
	const char *	keyfile;
	const char *	cipher;
{
  SSL_CTX *	ssl_ctx;
  SSL_METHOD *	method = NULL;

  /* get meaningful error messages */
  SSL_load_error_strings();

  /* initialize SSL */
  OpenSSL_add_ssl_algorithms();	/* same as SSLeay_add_ssl_algorithms(); */

  /* create new SSL context */
#if	!defined(OPENSSL_NO_SSL2) && !defined(OPENSSL_NO_SSL3)
  method = SSLv23_client_method();
#elif	!defined(OPENSSL_NO_SSL3)
  method = SSLv3_client_method();
#elif	!defined(OPENSSL_NO_SSL2)
  method = SSLv2_client_method();
#endif
  ssl_ctx = (SSL_CTX *)SSL_CTX_new(method);
  if (ssl_ctx == NULL) {
    return NULL;
  }

  /* RSA keys for ephemeral key exchange */
  if (SSL_CTX_need_tmp_RSA(ssl_ctx) && init_rsa_keys() == 0) {
    SSL_CTX_set_tmp_rsa_callback(ssl_ctx, tmp_rsa_callback);
  }
#if	USE_DH
  /* DH keys for ephemeral key exchange */
  if (init_dh_keys() == 0) {
    SSL_CTX_set_tmp_dh_callback(ssl_ctx, tmp_dh_callback);
  }
#endif	/* USE_DH */

  /* set SSL Context options */
  SSL_CTX_set_options(ssl_ctx, SSL_OP_ALL|SSL_OP_NO_SSLv2);

  /* ceritification */
  if (certfile != NULL) {
    if (!SSL_CTX_use_certificate_file(ssl_ctx, certfile, SSL_FILETYPE_PEM)) {
      SSL_CTX_free(ssl_ctx);
      return NULL;
    }
    else {
      if (keyfile == NULL) {
	keyfile = certfile;
      }
      if (!SSL_CTX_use_PrivateKey_file(ssl_ctx, keyfile, SSL_FILETYPE_PEM) ||
	  !SSL_CTX_check_private_key(ssl_ctx)) {
	SSL_CTX_free(ssl_ctx);
	return NULL;
      }
    }
  }

  /* set cipher list */
  if (cipher == NULL) {
    cipher = SSL_CIPHER_SUITE;
  }
  if (SSL_CTX_set_cipher_list(ssl_ctx, cipher) == 0) {
    SSL_CTX_free(ssl_ctx);
    return NULL;
  }

#if	SSLEAY_VERSION_NUMBER >= 0x00906000L
  SSL_CTX_set_mode(ssl_ctx, SSL_MODE_ENABLE_PARTIAL_WRITE |
		   SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER |
		   SSL_MODE_AUTO_RETRY);
  SSL_CTX_set_verify(ssl_ctx, SSL_VERIFY_NONE, NULL);
#endif

  return ssl_ctx;
}

/*
 * Function Name:	wt_netfd_ssl_tcp_server
 * Description:		initialize network socket for a SSL server
 * Arguments:		hostname, service, socklen_t pointer, listen queue
 *			certificate file, key file, cipher list
 * Return Value:	successful: netfd, otherwise: NULL
 */
wt_netfd_t *
wt_netfd_ssl_tcp_server(nodename, service, listenq, certfile, keyfile, cipher)
	const char *	nodename;
	const char *	service;
	int		listenq;
	const char *	certfile;
	const char *	keyfile;
	const char *	cipher;
{
  SSL_CTX *	ssl_ctx;
  wt_netfd_t *	nsd;

  ssl_ctx = ssl_server_init(certfile, keyfile, cipher);
  if (ssl_ctx == NULL) {
    return NULL;
  }

  nsd = wt_netfd_tcp_server(nodename, service, listenq);
  if (nsd == NULL) {
    SSL_CTX_free(ssl_ctx);
    return NULL;
  }

  nsd->ssl_ctx	= ssl_ctx;
  nsd->ssl	= SSL_new(ssl_ctx);
  if (nsd->ssl == NULL) {
    wt_netfd_ssl_close(nsd);
    return NULL;
  }
  SSL_set_fd(nsd->ssl, nsd->osfd);

  return nsd;
}

/*
 * Function Name:	wt_netfd_ssl_tcp_servers
 * Description:		initialize network socket for a SSL server
 * Arguments:		hostname, service, socklen_t pointer, listen queue
 *			certificate file, key file, cipher list
 * Return Value:	successful: netfds, otherwise: NULL
 */
wt_netfd_t *
wt_netfd_ssl_tcp_servers(nodename, service, listenq, certfile, keyfile, cipher)
	const char *	nodename;
	const char *	service;
	int		listenq;
	const char *	certfile;
	const char *	keyfile;
	const char *	cipher;
{
  SSL_CTX *	ssl_ctx;
  wt_netfd_t *	nsd;
  wt_netfd_t *	newfd = NULL;

  ssl_ctx = ssl_server_init(certfile, keyfile, cipher);
  if (ssl_ctx == NULL) {
    return NULL;
  }

  nsd = wt_netfd_tcp_servers(nodename, service, listenq);
  if (nsd == NULL) {
    SSL_CTX_free(ssl_ctx);
    return NULL;
  }

  for (newfd = nsd; nsd != NULL; nsd = nsd->next) {
    nsd->ssl_ctx = ssl_ctx;
    nsd->ssl	 = SSL_new(ssl_ctx);
    if (nsd->ssl == NULL) {
      goto failed;
    }
    SSL_set_fd(nsd->ssl, nsd->osfd);
  }

  return newfd;

failed:
  do {
    wt_netfd_t *	nsd_next;

    for (nsd = newfd; nsd != NULL; nsd = nsd_next) {
      nsd_next = nsd->next;
      wt_netfd_ssl_close(nsd);
    }
  } while (0);

  return NULL;
}

/*
 * FUNCTION NAME:	ssl_server_init
 * DESCRIPTION:		initialize SSL connection for server
 * ARGUMENTS:		certificate file, key file, cipher list
 * RETURN VALUE:	successful: SSL context, otherwise: NULL
 */
static SSL_CTX *
ssl_server_init(certfile, keyfile, cipher)
	const char *	certfile;
	const char *	keyfile;
	const char *	cipher;
{
  SSL_CTX *	ssl_ctx;
  SSL_METHOD *	method = NULL;
  const char	sid_ctx[] = "wtl-ssl";

  /* get meaningful error messages */
  SSL_load_error_strings();

  /* initialize SSL */
  OpenSSL_add_ssl_algorithms();	/* same as SSLeay_add_ssl_algorithms(); */

  /* create new SSL context */
#if	!defined(OPENSSL_NO_SSL2) && !defined(OPENSSL_NO_SSL3)
  method = SSLv23_server_method();
#elif	!defined(OPENSSL_NO_SSL3)
  method = SSLv3_server_method();
#elif	!defined(OPENSSL_NO_SSL2)
  method = SSLv2_server_method();
#endif
  ssl_ctx = (SSL_CTX *)SSL_CTX_new(method);
  if (ssl_ctx == NULL) {
    return NULL;
  }

  /* set the context session id */
  SSL_CTX_set_session_id_context(ssl_ctx, sid_ctx, strlen(sid_ctx));

  /* RSA keys for ephemeral key exchange */
  if (SSL_CTX_need_tmp_RSA(ssl_ctx) && init_rsa_keys() == 0) {
    SSL_CTX_set_tmp_rsa_callback(ssl_ctx, tmp_rsa_callback);
  }
#if	USE_DH
  /* DH keys for ephemeral key exchange */
  if (init_dh_keys() == 0) {
    SSL_CTX_set_tmp_dh_callback(ssl_ctx, tmp_dh_callback);
  }
#endif	/* USE_DH */

  /* set SSL Context options */
  SSL_CTX_set_options(ssl_ctx, SSL_OP_ALL|SSL_OP_NO_SSLv2);

  /* ceritification */
  if (certfile != NULL) {
    if (!SSL_CTX_use_certificate_file(ssl_ctx, certfile, SSL_FILETYPE_PEM)) {
      SSL_CTX_free(ssl_ctx);
      return NULL;
    }
    if (keyfile == NULL) {
      keyfile = certfile;
    }
    if (!SSL_CTX_use_PrivateKey_file(ssl_ctx, keyfile, SSL_FILETYPE_PEM) ||
	!SSL_CTX_check_private_key(ssl_ctx)) {
      SSL_CTX_free(ssl_ctx);
      return NULL;
    }
  }

  /* set cipher list */
  if (cipher == NULL) {
    cipher = SSL_CIPHER_SUITE;
  }
  if (SSL_CTX_set_cipher_list(ssl_ctx, cipher) == 0) {
    SSL_CTX_free(ssl_ctx);
    return NULL;
  }

#if	SSLEAY_VERSION_NUMBER >= 0x00906000L
  SSL_CTX_set_mode(ssl_ctx, SSL_MODE_ENABLE_PARTIAL_WRITE |
		   SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER |
		   SSL_MODE_AUTO_RETRY);
  SSL_CTX_set_verify(ssl_ctx, SSL_VERIFY_NONE, NULL);
  SSL_CTX_set_session_cache_mode(ssl_ctx, SSL_SESS_CACHE_SERVER);
#endif
#if	DEBUG
  fprintf(stderr, "DEBUG[ssl_server_init] X509 dir: %s\n",
	      X509_get_default_cert_dir());
#endif	/* DEBUG */

  return ssl_ctx;
}

/*
 * FUNCTION NAME:	init_rsa_keys
 * DESCRIPTION:		initialize RSA keys
 * ARGUMENTS:		none
 * RETURN VALUE:	successful: 0, otherwise: -1
 */
static int
init_rsa_keys(void)
{
  /* initialize temporary rsa keys */
#if	SSLEAY_VERSION_NUMBER >= 0x0900
  rsa_512 = RSA_generate_key(512, RSA_F4, NULL, NULL);
#else
  rsa_512 = RSA_generate_key(512, RSA_F4, NULL);
#endif
  if (rsa_512 == NULL) {
    return -1;
  }

#if	SSLEAY_VERSION_NUMBER >= 0x0900
  rsa_1024 = RSA_generate_key(1024, RSA_F4, NULL, NULL);
#else
  rsa_1024 = RSA_generate_key(1024, RSA_F4, NULL);
#endif
  if (rsa_1024 == NULL) {
    return -1;
  }

  return 0;
}

/*
 * FUNCTION NAME:	tmp_rsa_callback
 * DESCRIPTION:		make a temporary private key
 * ARGUMENTS:		SSL connect, export flag, key length
 * RETURN VALUE:	key 512 / key 1024
 */
static RSA *
tmp_rsa_callback(s, is_export, keylength)
	SSL *	s __unused;
	int	is_export __unused;
	int	keylength;
{
  RSA *	rsa_tmp = NULL;

  switch (keylength) {
    case 512:
      rsa_tmp = rsa_512;
      break;
    case 1024:
      rsa_tmp = rsa_1024;
      break;
    default:
      /* generateing a key on the fly is very costly, so use what is there */
      if (rsa_1024) {
	rsa_tmp = rsa_1024;
      }
      else {
	rsa_tmp = rsa_512;	/* use at least a shorter key */
      }
      break;
  }

  return rsa_tmp;
}

#if	USE_DH
/*
 * FUNCTION NAME:	init_dh_keys
 * DESCRIPTION:		initialize DH keys
 * ARGUMENTS:		none
 * RETURN VALUE:	successful: 0, otherwise: -1
 */
static int
init_dh_keys(void)
{
  register FILE *	fin;

  /* initialize temporary rsa keys */
  fin = wsSafeFopen(DH_PARAM_512_PEM, "r");
  if (fin == NULL) {
    return -1;
  }
  dh_512 = PEM_read_DHparams(fin, NULL, NULL, NULL);
  fclose(fin);

  fin = wsSafeFopen(DH_PARAM_1024_PEM, "r");
  if (fin == NULL) {
    return -1;
  }
  dh_1024 = PEM_read_DHparams(fin, NULL, NULL, NULL);
  fclose(fin);

  return 0;
}

/*
 * FUNCTION NAME:	tmp_dh_callback
 * DESCRIPTION:		make a temporary private key
 * ARGUMENTS:		SSL connect, export flag, key length
 * RETURN VALUE:	key 512 / key 1024
 */
static DH *
tmp_dh_callback(s, is_export, keylength)
	SSL *	s;
	int	is_export;
	int	keylength;
{
  DH *	dh_tmp = NULL;

  switch (keylength) {
    case 512:
      dh_tmp = dh_512;
      break;
    case 1024:
      dh_tmp = dh_1024;
      break;
    default:
      /* generateing a key on the fly is very costly, so use what is there */
      if (dh_1024) {
	dh_tmp = dh_1024;
      }
      else {
	dh_tmp = dh_512;	/* use at least a shorter key */
      }
      break;
  }

  return dh_tmp;
}
#endif	/* USE_DH */


/*
 * FUNCTION NAME:	wt_netfd_ssl_close
 * DESCRIPTION:		close SSL connection
 * ARGUMENTS:		netfd
 * RETURN VALUE:	successful: 0 netfd, otherwise: -1
 */
int
wt_netfd_ssl_close(nfd)
	register wt_netfd_t *	nfd;
{
  if (nfd->ssl_ctx != NULL) {
    SSL_CTX_free(nfd->ssl_ctx);
    nfd->ssl_ctx = NULL;
  }
  return wt_netfd_close(nfd);
}

/*
 * Function Name:	wt_netfd_ssl_shutdown
 * Description:		shutdown netfd
 * Arguments:		netfd, how(WT_SHUT_RD,WT_SHUT_WR,WT_SHUT_RDWR)
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_netfd_ssl_shutdown(nfd, how)
	register wt_netfd_t *	nfd;
	register int		how;
{
  if (nfd->ssl != NULL) {
    if (SSL_shutdown(nfd->ssl) == 0) {
      SSL_shutdown(nfd->ssl);
    }
    SSL_clear(nfd->ssl);
    SSL_free(nfd->ssl);
    nfd->ssl = NULL;
  }
  return shutdown(nfd->osfd, how);
}

/*
 * FUNCTION NAME:	wt_netfd_ssl_read
 * DESCRIPTION:		read bytes from a SSL connection
 * ARGUMENTS:		netfd, buffer, buffer size, timeout
 * RETURN VALUE:	successful: read size, otherwise: -1
 */
ssize_t
wt_netfd_ssl_read(nfd, buf, nbyte, timeout)
	wt_netfd_t *	nfd;
	void *		buf;
	size_t		nbyte;
	wt_utime_t	timeout;
{
  register int		n;
  register char *	ptr = (char *)buf;

  if (wt_netfd_readbuf(nfd, timeout) < 0) {
    return -1;
  }
  if (nfd->rb.total == 0) {
    return 0;
  }
  for (n = 0; n < nbyte && nfd->rb.total > 0; n++) {
    *ptr++ = *nfd->rb.bufp++;
    nfd->rb.total--;
  }

  return n;
}

/*
 * FUNCTION NAME:	wt_netfd_ssl_write
 * DESCRIPTION:		write bytes to a SSL connection
 * ARGUMENTS:		netfd, buffer, buffer size, timeout
 * RETURN VALUE:	successful: read size, otherwise: -1
 */
ssize_t
wt_netfd_ssl_write(nfd, buf, nbyte, timeout)
	wt_netfd_t *	nfd;
	const void *	buf;
	size_t		nbyte;
	wt_utime_t	timeout;
{
  size_t	resid = nbyte;
  ssize_t	n;

  while (resid > 0) {
    n = SSL_write(nfd->ssl, buf, (int)resid);
    if (n <= 0) {
      if (errno == EINTR) {
	continue;
      }
      if (!WT_IO_NOT_READY) {
	return -1;
      }
      switch (SSL_get_error(nfd->ssl, n)) {
	case SSL_ERROR_NONE:
	  return n;
	case SSL_ERROR_WANT_WRITE:
#if	DEBUG
	  fprintf(stderr, "DEBUG[wt_netfd_ssl_write] more write errno=%d\n",
		  errno);
#endif	/* DEBUG */
	  break;
	default: 
	  return -1;
      }
      /* wait until the socket becoms writable */
      if (wt_netfd_poll(nfd, WT_POLL_OUT, timeout) == -1) {
	return -1;
      }
    }
    else {
      resid -= n;
      if (resid == 0) {
	break;
      }
      buf = (const void *)((const char *)buf + n);
    }
  }

  return (ssize_t)(nbyte - resid);
}

/*
 * FUNCTION NAME:	wt_netfd_ssl_accept
 * DESCRIPTION:		wait for a SSL client to initiate a SSL handshake
 * ARGUMENTS:		netfd
 * RETURN VALUE:	successful: new netfd, otherwise: NULL
 */
wt_netfd_t *
wt_netfd_ssl_accept(nsd, addr, addrlen, timeout)
	wt_netfd_t *		nsd;
	struct sockaddr *	addr;
	socklen_t *		addrlen;
	wt_utime_t		timeout;
{
  wt_netfd_t *	newfd;

  newfd = wt_netfd_accept(nsd, addr, addrlen, timeout);
  if (newfd != NULL) {
    int	status;

    newfd->ssl = SSL_new(nsd->ssl_ctx);
    if (newfd->ssl == NULL) {
      wt_netfd_close(newfd);
      errno = EBADF;
      return NULL;
    }
    newfd->ssl_ctx = NULL;
    SSL_clear(newfd->ssl);
    SSL_set_fd(newfd->ssl, newfd->osfd);
    SSL_set_read_ahead(newfd->ssl, 1);
    SSL_set_accept_state(newfd->ssl);

    while ((status = SSL_accept(newfd->ssl)) <= 0) {
      switch (SSL_get_error(newfd->ssl, status)) {
	case SSL_ERROR_NONE:
#if	DEBUG
	  fprintf(stderr, "DEBUG[SSL_accept] ok errno=%d\n", errno);
#endif	/* DEBUG */
	  break;
	case SSL_ERROR_WANT_READ:
#if	DEBUG
	  fprintf(stderr, "DEBUG[SSL_accept] more read data = %d\n",
		  SSL_want_read(newfd->ssl));
#endif	/* DEBUG */
	  /* wait until the socket becoms writable */
	  if (wt_netfd_poll(newfd, WT_POLL_IN, WT_UTIME_SEC(15)) == -1) {
	    return NULL;
	  }
	  break;
	case SSL_ERROR_SYSCALL:
#if	DEBUG
	  fprintf(stderr, "DEBUG[SSL_accept] some I/O error occurred\n");
#endif	/* DEBUG */
	case SSL_ERROR_WANT_WRITE:
	case SSL_ERROR_WANT_X509_LOOKUP:
	case SSL_ERROR_ZERO_RETURN:
	case SSL_ERROR_SSL:
	default: 
#if	DEBUG
	  fprintf(stderr, "DEBUG[SSL_accept] fatal=%d ret=%d errno=%d\n",
		  SSL_get_error(newfd->ssl, status), status, errno);
	  fprintf(stderr, "\tfailed to accept for SSL connection\n");
#endif	/* DEBUG */
	  wt_netfd_ssl_close(newfd);
	  errno = EPROTOTYPE;
	  return NULL;
      }
    }
  }

#if	DEBUG
  fprintf(stderr, "DEBUG[wt_netfd_ssl_accept] CC_CIPHER=%s:Version=%s;%s\n",
	  SSL_get_cipher(newfd->ssl),
	  SSL_get_cipher_version(newfd->ssl),
	  SSL_get_version(newfd->ssl));
#endif	/* DEBUG */

  return newfd;
}

/*
 * Function Name:	wt_netfd_ssl_sendfile
 * Description:		senf a file to a socket
 * Arguments:		nfd, nsd, offset, nbytes, hdtr, sbytes, flags, timeout
 * Return Value:	successful: 0, otherwise: -1
 */
int
wt_netfd_ssl_sendfile(nfd, nsd, offset, nbytes, hdtr, sbytes, flags, timeout)
	wt_netfd_t *		nfd;
	wt_netfd_t *		nsd;
	off_t			offset;
	size_t			nbytes;
#if	HAVE_FREEBSD_SENDFILE
	struct sf_hdtr *	hdtr __unused;
#elif	HAVE_LINUX_SENDFILE
	void *			hdtr __unused;
#else	/* No sendfile(2) */
	void *			hdtr __unused;
#endif
	off_t *			sbytes;
	int			flags __unused;
	wt_utime_t		timeout;
{
  void *	buf;
  size_t	len;
  ssize_t	n;
  int		status = 0;

  if (nbytes == 0) {
    /* get whole file size */
    len = (size_t)wt_netfd_seek(nfd, 0, WT_SEEK_END);
    if (len == -1) {
      return -1;
    }
  }
  else {
    len = nbytes;
  }
  buf = mmap(NULL, len, PROT_READ, MAP_PRIVATE, nfd->osfd, offset);
  if (buf == MAP_FAILED) {
    return -1;
  }

  n = wt_netfd_ssl_write(nsd, buf, nbytes, timeout);
  if (n == -1) {
    status = -1;
  }
  if (sbytes != NULL) {
    *sbytes = n;
  }
  munmap(buf, nbytes);

  return status;
}

/*
 * Function Name:	wt_netfd_ssl_readline
 * Description:		read a line
 * Arguments:		netfd, buffer, max buffer length, timeout
 * Return Value:	successful: read size, otherwise: -1
 */
ssize_t
wt_netfd_ssl_readline(nfd, buf, maxlen, timeout)
	wt_netfd_t *	nfd;
	void *		buf;
	size_t		maxlen;
	wt_utime_t	timeout;
{
  register int		n;
  register char *	ptr = (char *)buf;

  for (n = 1; n < maxlen; n++) {
    char *	c;
    int		status;

    c = (char *)alloca(sizeof(char));
    if (c == NULL) { return -1; }

    status = one_read(nfd, c, timeout);
    if (status == 1) {
      *ptr++ = *c;
      if (*c == LF)  { break; }
      if (*c == EOL) { return 0; }	/* EOF?, null data read */
    }
    else if (status == 0) {
      if (n == 1) { return 0; }		/* EOF, no data read */
      else	  { break; }		/* EOF, some data was read */
    }
    else {
      return -1;	/* error */
    }
  }

  *ptr = EOL;

  return n;
}

static ssize_t
one_read(nfd, ptr, timeout)
	wt_netfd_t *	nfd;
	char *		ptr;
	wt_utime_t	timeout;
{
  if (wt_netfd_readbuf(nfd, timeout) < 0) {
    return -1;
  }
  nfd->rb.total--;
  *ptr = *nfd->rb.bufp++;

  return 1;
}

static int
wt_netfd_readbuf(nfd, timeout)
	wt_netfd_t *	nfd;
	wt_utime_t	timeout;
{
  if (nfd->rb.total <= 0) {
    memset(nfd->rb.buf, 0, sizeof(nfd->rb.buf));
    while ((nfd->rb.total = SSL_read(nfd->ssl,
				     nfd->rb.buf, sizeof(nfd->rb.buf))) <= 0) {
      if (errno == EINTR) {
	continue;
      }
      if (!WT_IO_NOT_READY) {
	return -1;
      }
      switch (SSL_get_error(nfd->ssl, nfd->rb.total)) {
       	case SSL_ERROR_NONE:
  	  return nfd->rb.total;
   	case SSL_ERROR_WANT_READ:
#if	DEBUG
	  fprintf(stderr, "DEBUG[ssl_readbuf] more read errno=%d\n", errno);
#endif	/* DEBUG */
  	  break;
   	default: 
  	  return -1;
      }
      /* wait until the socket becomes readable */
      if (wt_netfd_poll(nfd, WT_POLL_IN, timeout) == -1) {
	return -1;
      }
    }
    nfd->rb.bufp = nfd->rb.buf;
  }

  return nfd->rb.total;
}

/*
 * Function Name:	wt_netfd_ssl_writev
 * Description:		write data
 * Arguments:		netfd, iovec buffers, iovec size, timeout
 * Return Value:	successful: printed size, otherwise: -1
 */
ssize_t
wt_netfd_ssl_writev(nfd, iov, iovcnt, timeout)
	wt_netfd_t *		nfd;
	const struct iovec *	iov;
	int			iovcnt;
	wt_utime_t		timeout;
{
  ssize_t	nbytes;
  int		i;

  for (nbytes = 0, i = 0; i < iovcnt; i++) {
    ssize_t	n;

    n = wt_netfd_ssl_write(nfd, iov[i].iov_base, iov[i].iov_len, timeout);
    if (n == -1) {
      break;
    }
    nbytes += n;
  }

  return nbytes;
}
