/*****************************************************************************
 *
 * FILE:	util.c
 * DESCRIPTION:	Wyvern: utility library
 * DATE:	Tue, May 27 2003
 * UPDATED:	Wed, Sep  5 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2003-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2003-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: util.c,v 1.32 2007/09/04 17:09:04 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <stdarg.h>
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#if	HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif	/* HAVE_SYS_TYPES_H */
#if	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif	/* HAVE_SYS_TIME_H */
#if	HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif	/* HAVE_SYS_RESOURCE_H */
#if	HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif	/* HAVE_SYS_PARAM_H */
#if	HAVE_SYSLOG_H
#include <syslog.h>
#endif	/* HAVE_SYSLOG_H */
#include <errno.h>
#include "wu.h"

/******************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	EOL	'\0'
#define	CR	'\015'
#define	LF	'\012'
#define	CRLF	"\015\012"
#define	SPC	'\040'
#define	HTAB	'\011'
#define	VTAB	'\013'

#if	HAVE_DEV_URANDOM
#define	DEV_URANDOM	"/dev/urandom"
#endif	/* HAVE_DEV_URANDOM */

#define	PATH_DEVNULL	"/dev/null"
#if	HAVE_SYSCONF
#define	MAXFD		sysconf(_SC_OPEN_MAX)
#else
#define	MAXFD		(64)
#endif	/* HAVE_SYSCONF */

/******************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/

/******************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wu_chomp
 * Description:		remove CRLF, CR or LF from string end, and
 *			remove white-space in front and behind
 * Arguments:		stirng
 * Return Value:	string
 * Note:		destroy original string
 */
char *
wu_chomp(s)
	char *	s;
{
  if (s != NULL) {
    register char *	p = s;
    size_t		len;
    size_t		offset = 0;

    while (*p == CR || *p == LF || *p == HTAB || *p == VTAB || *p == SPC) {
      offset++;
      p++;
    }
    if (offset > 0) {
      len = strlen(p);
      memmove(s, p, len);
      memset(s + len, 0, offset);
    }

    len = strlen(s);
    p	= &s[len - 1];
    while (*p == CR || *p == LF || *p == HTAB || *p == VTAB || *p == SPC) {
      *p-- = EOL;
    }
  }

  return s;
}

/*
 * Function Name:	wu_close
 * Description:		close a file/socket descriptor
 * Arguments:		descriptor
 * Return Value:	same as close(2)
 */
int
wu_close(d)
	int	d;
{
  int	status;

  while (((status = close(d)) < 0) && (errno == EINTR)) {
    ;
  }
  return status;
}

/*
 * Function Name:	wu_skip_line
 * Description:		skip white space, '#' and CR-LF
 * Arguments:		stirng
 * Return Value:	white space: 0, otherwise: -1
 */
int
wu_skip_line(s)
	const char *	s;
{
  switch (s[0]) {
    case '#':
    case SPC:
    case HTAB:
    case VTAB:
    case CR:
    case LF:
      return 0;
    default:
      break;
  }
  return -1;
}

/*
 * Function Name:	wu_strsep
 * Description:		separate strings
 * Arguments:		string pointer, delimitor, switch flag
 * Return Value:	same as strsep(3)
 * Notes:		based on FreeBSD:/usr/src/lib/libc/string/strsep.c
 *			The "text string" is separated as one token.
 *			Therefore, the delimitor must not include '"' character.
 */
char *
wu_strsep(stringp, delim, flag)
	register char **	stringp;
	register const char *	delim;
	register int		flag;	/* 0: same as strsep(3), 1: expansion */
{
  register char *	s;
  register const char *	spanp;
  register int		c;
  register int		sc;
  char *		tok;

  if ((s = *stringp) == NULL) {
    return NULL;
  }
  for (tok = s; ; ) {
    c	  = *s++;
    spanp = delim;
    if (flag == WUL_STRSEP_EXPANSION && c == '"') {
      register char *	p;

      for (p = s; *s != '"'; s++) {
	if (*s == '\0') {
	  s = p;	/* reset */
	  break;
	}
      }
      if (*s == '"') {
	c = *s++;
      }
    }
    do {
      if ((sc = *spanp++) == c) {
	if (c == '\0') {
	  s = NULL;
	}
	else {
	  s[-1] = '\0';
	}
	*stringp = s;
	return tok;
      }
    } while (sc != '\0');
  }
  /* NOTREACHED */
}

/*****************************************************************************/

/*
 * Function Name:	wu_timestamp
 * Description:		formatted time
 * Arguments:		return buffer, specified time_t, format type, GMT flag
 * Return Value:	successful: 0, otherwise: -1
 */
int
wu_timestamp(retbuf, t, form, gmt)
	char **	retbuf;
	time_t	t;
	int	form;
	int	gmt;	/* GMT: 1, JST: 0 */
{
  struct tm *	tmp;
  char		timebuf[40];
  size_t	len;
  int		status;

  if (t == 0) {	/* No specified time */
    time_t	now;
    time(&now);
    tmp = gmt ? gmtime(&now) : localtime(&now);
  }
  else {
    tmp = gmt ? gmtime(&t) : localtime(&t);
  }

  switch (form) {
    case WUL_TIME_FORMAT_RFC822:
      /* Sun, 06 Nov 1994 08:49:37 GMT ; RFC 822, updated by RFC 1123 */
      len = strftime(timebuf, sizeof(timebuf),
		     "%a, %d %b %Y %H:%M:%S %Z", tmp);
      break;
    case WUL_TIME_FORMAT_RFC850:
      /* Sunday, 06-Nov-94 08:49:37 GMT ; RFC 850, obsoleted by RFC 1036 */
      len = strftime(timebuf, sizeof(timebuf),
		     "%A, %d-%b-%y %H:%M:%S %Z", tmp);
      break;
    case WUL_TIME_FORMAT_APACHE: {	/* always use localtime zone */
      /* 15/Nov/2002:21:39:54 +0900 */
	if (t == 0) {	/* No specified time */
	  time_t	now;
	  time(&now);
	  tmp = localtime(&now);
	}
	else {
	  tmp = localtime(&t);
	}
	len = strftime(timebuf, sizeof(timebuf), "%d/%b/%Y:%H:%M:%S %z", tmp);
      }
      break;
    case WUL_TIME_FORMAT_SQL:	/* 2005-12-23 02:22:28 */
      len = strftime(timebuf, sizeof(timebuf), "%Y-%m-%d %H:%M:%S", tmp);
      break;
    case WUL_TIME_FORMAT_ANSI_C:
    case WUL_TIME_FORMAT_TYPICAL:
    default:
      /* Sun Nov  6 08:49:37 1994 ; ANSI C's asctime() format */
      len = strftime(timebuf, sizeof(timebuf), "%a %b %e %H:%M:%S %Y", tmp);
      break;
  }
  if (retbuf && len > 0) {
    *retbuf = strdup(timebuf);
    status  = *retbuf == NULL ? -1 : 0;
  }
  else {
    *retbuf = NULL;
    status  = -1;
  }

  return status;
}

/*
 * Function Name:	wu_uptime
 * Description:		calculate uptime
 * Arguments:		return buffer, start time, end time
 * Return Value:	successful: 0 , otherwise: -1
 */
int
wu_uptime(retbuf, s, e)
	char **			retbuf;
	struct timeval *	s;	/* start timestamp */
	struct timeval *	e;	/* end timestamp */
{
  long	uptime;
  int	days, hours, mins, secs, msecs;
  char	buf[48];
  int	n = 0;
  int	status;

  
  uptime  = e->tv_sec  - s->tv_sec;
  days	  = uptime / 86400;
  uptime %= 86400;
  hours	  = uptime / 3600;
  uptime %= 3600;
  mins	  = uptime / 60;
  secs	  = uptime % 60;
  msecs	  = e->tv_usec - s->tv_usec;
#if	DEBUG
  fprintf(stderr, "DEBUG[wu_uptime] usecs = %d\n", msecs);
#endif	/* DEBUG */
  msecs  /= 1000;	/* usecs to msecs */
  if (msecs < 0) {
    msecs += 1000;
  }

  if (msecs >= 1000) {
#if	DEBUG
    fprintf(stderr, "DEBUG[wu_uptime] msecs from %d", msecs);
#endif	/* DEBUG */
    secs += msecs / 1000;
    msecs = msecs % 1000;
#if	DEBUG
    fprintf(stderr, " to %d\n", msecs);
#endif	/* DEBUG */
  }
  if (secs >= 60) {
#if	DEBUG
    fprintf(stderr, "DEBUG[wu_uptime] secs from %d", secs);
#endif	/* DEBUG */
    mins += secs / 60;
    secs  = secs % 60;
#if	DEBUG
    fprintf(stderr, " to %d\n", secs);
#endif	/* DEBUG */
  }
  if (mins >= 60) {
#if	DEBUG
    fprintf(stderr, "DEBUG[wu_uptime] mins from %d", mins);
#endif	/* DEBUG */
    hours += mins / 60;
    mins   = mins % 60;
#if	DEBUG
    fprintf(stderr, " to %d\n", mins);
#endif	/* DEBUG */
  }
  if (hours >= 24) {
#if	DEBUG
    fprintf(stderr, "DEBUG[wu_uptime] hours from %d", hours);
#endif	/* DEBUG */
    days += hours / 24;
    hours = hours % 24;
#if	DEBUG
    fprintf(stderr, " to %d\n", hours);
#endif	/* DEBUG */
  }

#if	DEBUG
  fprintf(stderr, "DEBUG[wu_uptime] %d %d %d %d %d\n",
	  days, hours, mins, secs, msecs);
#endif	/* DEBUG */
  memset(buf, 0, sizeof(buf));
  if (days > 0) {
    n = snprintf(buf, sizeof(buf), "%d day%s,", days, days > 1 ? "s" : "");
  }
  if (hours > 0 && mins > 0 && secs > 0) {
    n += snprintf(buf + n, sizeof(buf) - n,
		  "%s%2d:%02d:%02d", n > 0 ? " " : "", hours, mins, secs);
  }
  else if (hours > 0 && mins > 0) {
    n += snprintf(buf + n, sizeof(buf) - n,
		  "%s%2d:%02d", n > 0 ? " " : "", hours, mins);
  }
  else {
    if (hours > 0) {
      n += snprintf(buf + n, sizeof(buf) - n,
		    "%s%d hour%s",
		    n > 0 ? " " : "", hours, hours > 1 ? "s" : "");
    }
    if (mins > 0) {
      n += snprintf(buf + n, sizeof(buf) - n,
		    "%s%d min%s", n > 0 ? " " : "", mins, mins > 1 ? "s" : "");
    }
    if (secs > 0) {
      n += snprintf(buf + n, sizeof(buf) - n,
		    "%s%d sec%s", n > 0 ? " " : "", secs, secs > 1 ? "s" : "");
    }
    if (msecs >= 0) {
      n += snprintf(buf + n, sizeof(buf) - n,
		    "%s%d msecs", n > 0 ? " " : "", msecs);
    }
  }
  if (retbuf && n > 0) {
    *retbuf = strdup(buf);
    status  = *retbuf == NULL ? -1 : 0;
#if	DEBUG
    fprintf(stderr, "DEBUG[wu_uptime] uptime: %s\n", buf);
#endif	/* DEBUG */
  }
  else {
    *retbuf = NULL;
    status = -1;
  }

  return status;
}

/*****************************************************************************/

char *
wu_uri_escaped(uri)
	const char *	uri;
{
  char *	o;	/* original pointer */
  char *	s;	/* working pointer */

  o = s = strdup(uri);
  if (s != NULL) {
    size_t	len;
    char *	enc;

    len = strlen(uri);
    enc = (char *)calloc(len * 3 + 1, sizeof(char));
    if (enc != NULL) {
      unsigned int	i;
      unsigned int	j;

      for (i = 0, j = 0; i < len; i++) {
	if ((s[i] & 0x80) == 0x80 ||	/* Japanese EUC code */
	    /* Reserved Characters */
	    s[i] == ';' || s[i] == '/' || s[i] == '?' || s[i] == ':' ||
	    s[i] == '@' || s[i] == '&' || s[i] == '=' || s[i] == '+' ||
	    s[i] == '$' || s[i] == ',' ||
	    /* Unreserved Characters */
	    s[i] == '-' || s[i] == '_' || s[i] == '.' || s[i] == '!' ||
	    s[i] == '~' || s[i] == '*' || s[i] == '\'' || s[i] == '(' ||
	    s[i] == ')' ||
	    /* control */
	    ((char)s[i] >= 0 && (char)s[i] <= 0x1f) || (char)s[i] == 0x7f ||
	    /* space */
	    s[i] == ' ' ||
	    /* delims */
	    s[i] == '<' || s[i] == '>' || s[i] == '#' || s[i] == '%' ||
	    s[i] == '"' ||
	    /* unwise */
	    s[i] == '{' || s[i] == '}' || s[i] == '|' || s[i] == '\\' ||
	    s[i] == '^' || s[i] == '[' || s[i] == ']' || s[i] == '`'
	    ) {
	  char	b1 = (s[i] >> 4) & 0x0f;
	  char	b2 = s[i] & 0x0f;
	  enc[j++] = '%';
	  enc[j++] = b1 < 10 ? b1 + '0' : b1 - 10 + 'A';
	  enc[j++] = b2 < 10 ? b2 + '0' : b2 - 10 + 'A';
	}
	else {
	  enc[j++] = s[i];
	}
      }
      free(o);

      return (char *)realloc(enc, (size_t)j);
    }
  }

  return NULL;
}

char *
wu_uri_unescaped(uri)
	const char *	uri;
{
  char *	tmp;	/* original string (to work) */
  char *	s;

  tmp = s = strdup(uri);
  if (s != NULL) {
    register char *	os  = tmp;
    register char *	ret = s;
    char		c, c1;

    while ((c = *os) != EOL) {
      /* escaped URI */
      if (c == '%') {
	os++;
	if ((c = *os++) && (c1 = *os++)) {
	  *s++ = ((c  >= 'A' ? ((c  & 0xdf) - 'A') + 10 : c  - '0') * 16) +
		 ((c1 >= 'A' ? ((c1 & 0xdf) - 'A') + 10 : c1 - '0'));
	  os--;	/* back one word */
	}
	else {
	  goto done;
	}
      }
      /* escaped white space */
      else if (c == '+') {
	*s++ = ' ';
      }
      /* normal characters */
      else {
	*s++ = c;
      }
      os++;
    }
    *s = EOL;

done:
    return ret;
  }

  return NULL;
}

/*****************************************************************************/

/*
 * Function Name:	wu_make_boundary
 * Description:		make a boundary string
 * Arguments:		seed for srandom
 * Return Value:	Successful: boundary string, otherwise: NULL
 * Reference:		RFC 1521
 *			MIME (Multipurpose Internet Mail Extensions) Part One:
 *			Mechanisms for Specifying and Describing
 *			the Format of Internet Message Bodies
 */
char *
wu_make_boundary(seed)
	unsigned long	seed;
{
#define	BCHARS_SIZE	(69)
  const char	bchars[] =
	"0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'()+_,-./:=?";
  char		boundary[BCHARS_SIZE + 1];
  register int	i;
#if	HAVE_DEV_URANDOM
  int		fd;

  fd = open(DEV_URANDOM, O_RDONLY, 0400);
  if (fd != -1) {
    for (i = 0; i < BCHARS_SIZE; i++) {
      unsigned int	c;

      read(fd, &c, sizeof(unsigned int));
      boundary[i] = bchars[c % BCHARS_SIZE];
    }
    close(fd);
    boundary[BCHARS_SIZE] = '\0';

    return strdup(boundary);
  }
#endif	/* HAVE_DEV_URANDOM */

  srandom(seed);
  for (i = 0; i < BCHARS_SIZE; i++) {
    unsigned int	c;

    c = (unsigned int)random();
    boundary[i] = bchars[c % BCHARS_SIZE];
  }
  boundary[BCHARS_SIZE] = '\0';

  return strdup(boundary);
}

/*****************************************************************************/

/*
 * Base64 alphabet
 */
static const char	base64[] =
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
/*
 * bit sequence for mask
 *
 * 0xfc = 11111100
 * 0x03 = 00000011
 * 0xf0 = 11110000
 * 0x0f = 00001111
 * 0xc0 = 11000000
 * 0x3f = 00111111
 * 0x30 = 00110000
 * 0x3c = 00111100
 */

/*
 * Function Name:	wu_base64_encode
 * Description:		encode a string to a base64 string
 * Arguments:		string
 * Return Value:	successful: base64 encoded string, otherwise: NULL
 */
char *
wu_base64_encode(s)
	register const char *	s;
{
  char *	enc;
  size_t	len;
  unsigned int	i;
  unsigned int	j;

  if (s == NULL) { return NULL; }	/* NULL checker */
  len = strlen(s);
  enc = (char *)calloc((size_t)((double)len * 1.5), sizeof(char));
  if (enc == NULL) {
    return NULL;
  }

  /*
   * encode 3-bytes (24-bits) at a time
   */
  for (i = j = 0; i < len - (len % 3); i += 3, j += 4) {
    enc[j]   = base64[(s[i]    & 0xfc) >> 2];
    enc[j+1] = base64[((s[i]   & 0x03) << 4) | ((s[i+1] & 0xf0) >> 4)];
    enc[j+2] = base64[((s[i+1] & 0x0f) << 2) | ((s[i+2] & 0xc0) >> 6)];
    enc[j+3] = base64[(s[i+2]  & 0x3f)];
  }

  i = len - (len % 3);	/* rest size */
  switch (len % 3) {
    case 2:	/* one character padding */
      enc[j]	= base64[(s[i]   & 0xfc) >> 2];
      enc[j+1]	= base64[((s[i]  & 0x03) << 4) | ((s[i+1] & 0xf0) >> 4)];
      enc[j+2]	= base64[(s[i+1] & 0x0f) << 2];
      enc[j+3]	= base64[64];	/* Pad	*/
      j += 4;
      break;
    case 1:	/* two character padding */
      enc[j]	= base64[(s[i] & 0xfc) >> 2];
      enc[j+1]	= base64[(s[i] & 0x03) << 4];
      enc[j+2]	= base64[64];	/* Pad	*/
      enc[j+3]	= base64[64];	/* Pad	*/
      j += 4;
      break;
  }
  enc[j] = '\0';

  return enc;
}

/*
 * Function Name:	wu_base64_decode
 * Description:		decode a base64 string
 * Arguments:		base64 encoded string
 * Return Value:	successful: plain string, otherwise: NULL
 */
char *
wu_base64_decode(s)
	register const char *	s;
{
#define	VAL(x)	(s[(x)] == '=' ? 0 : strchr(base64, s[(x)]) - base64)
  char *	dec;
  size_t	len;
  unsigned int	i;
  unsigned int	j;

  if (s == NULL) { return NULL; }	/* NULL checker */
  len = strlen(s);
  dec = (char *)calloc((size_t)len, sizeof(char));
  if (dec == NULL) {
    return NULL;
  }

  /*
   * work on 3-bytes (24-bits) at a time
   */
  for (i = j = 0; i < len; i += 4, j += 3) {
    dec[j]   = (VAL(i) << 2) | ((VAL(i+1) & 0x30) >> 4);
    dec[j+1] = ((VAL(i+1) & 0x0f) << 4) | ((VAL(i+2) & 0x3c) >> 2); 
    dec[j+2] = ((VAL(i+2) & 0x03) << 6) | (VAL(i+3) & 0x3f);
  }
  dec[j] = '\0';

  return dec;
}

/*
 * Function Name:	wu_mkstemp
 * Description:		make temporary file and return a file descriptor opened
 * Arguments:		temporary directory
 * Return Value:	successful: file descriptor, otherwise: -1
 */
int
wu_mkstemp(tmpdir)
	const char *	tmpdir;
{
  char *	tempfile;

  if (asprintf(&tempfile, "%s/.wuXXXXXX", tmpdir) != -1) {
    int		fd;
#if	HAVE_MKSTEMP
    fd = mkstemp(tempfile);
#else
    if (mktemp(tempfile) == NULL) {
      free((char *)tempfile);
      return -1;
    }
    fd = open(tempfile, O_CREAT|O_EXCL|O_RDWR, 0600);
#endif	/* HAVE_MKSTEMP */
    unlink(tempfile);
    free((char *)tempfile);

    return fd;
  }

  return -1;
}

/*****************************************************************************/

/*
 * Function Name:	wu_toupper
 * Description:		upper case letter conversion in string
 * Arguments:		string
 * Return Value:	upper letters string
 * Note:		destroy original string
 */
char *
wu_toupper(s)
	register char *	s;
{
  char *	retval = s;

  if (s == NULL) {
    return NULL;
  }

  for (; *s != EOL; s++) {
    if (*s >= 'a' && *s <= 'z') {
      *s -= 'a' - 'A';
    }
#if	0
    else if (*s == '-') {
      *s = '_';
    }
#endif
  }

  return retval;
}

/*
 * Function Name:	wu_tolower
 * Description:		lower case letter conversion in string
 * Arguments:		string
 * Return Value:	lower letters string
 */
char *
wu_tolower(s)
	register char *	s;
{
  char *	retval = s;

  if (s == NULL) {
    return NULL;
  }

  for (; *s != EOL; s++) {
    if (*s >= 'A' && *s <= 'Z') {
      *s += 'a' - 'A';
    }
#if	0
    else if (*s == '-') {
      *s = '_';
    }
#endif
  }

  return retval;
}

/*****************************************************************************/

/*
 * Function Name:	wu_dirname
 * Description:		extract the directory part of a pathname
 * Arguments:		path
 * Return Value:	successful: dirname, otherwise: NULL
 */
char *
wu_dirname(path)
	const char *	path;
{
  static char		dname[MAXPATHLEN];
  register char *	p;

  if (path == NULL || *path == EOL) {
    strlcpy(dname, ".", sizeof(dname));
    return dname;
  }

  if (strlcpy(dname, path, sizeof(dname)) >= sizeof(dname)) {
    return NULL;
  }

  p = path + strlen(path) - 1;
  while (p > path && *p != '/') {
    p--;
  }
  if (p == path) {
    strlcpy(dname, *p == '/' ? "/" : ".", sizeof(dname));
    return dname;
  }
  while (p > path && *p == '/') {
    p--;
  }
  dname[p - path + 1] = EOL;

  return dname;
}

/*****************************************************************************/

/*
 * Function Name:	wu_daemon
 * Description:		run in the background
 * Arguments:		no chdir, no close, syslog ident
 * Return Value:	successful: 0, otherwise: -1
 */
int
wu_daemon(nochdir, noclose, ident)
	int		nochdir;
	int		noclose;
	const char *	ident;
{
  struct rlimit	rl;

  /* detach the current process from TTY */
  signal(SIGHUP, SIG_IGN);
  switch (fork()) {
    case -1:	/* error */
      return -1;
    case 0:	/* child */
      break;
    default:	/* parent */
      _exit(0);
  }

  /* create new session */
  if (setsid() == -1) {
    return -1;
  }

  /* detach the process from the 1st child process */
  switch (fork()) {
    case -1:	/* error */
      exit(1);
    case 0:	/* child */
      break;
    default:	/* parent */
      _exit(0);
  }
  signal(SIGHUP, SIG_DFL);

  /* change directory */
  if (!nochdir) {
    (void)chdir("/");
  }

  /* clear the file mode mask */
  (void)umask(S_IWGRP|S_IWOTH);	/* 022 */

  /* close file descriptors */
  if (!noclose) {
    register int	fd;

    /* close all descriptors */
    for (fd = 0; fd < MAXFD; fd++) {
      close(fd);
    }

    fd = open(PATH_DEVNULL, O_RDWR, 0);
    if (fd != -1) {
      (void)dup2(fd, 0);
      (void)dup2(fd, 1);
      (void)dup2(fd, 2);
      if (fd > 2) {
	close(fd);
      }
    }
  }

  /* limit resource */
  if (getrlimit(RLIMIT_CORE, &rl) == 0) {
    rl.rlim_cur = rl.rlim_max = 0;	/* no core dump */
    (void)setrlimit(RLIMIT_CORE, &rl);
  }

  /* open syslog */
  if (ident != NULL) {
    openlog(ident, LOG_PID, LOG_DAEMON);
  }

  return 0;
}

/*
 * Function Name:	wu_store_pid
 * Description:		store the process id into pid file
 * Arguments:		pid file, uid, gid
 * Return Value:	successful: 0, error: -1
 */
int
wu_store_pid(pidfile, uid, gid)
	const char *	pidfile;
	uid_t		uid;
	gid_t		gid;
{
  FILE *	fout;
  int		fd;
  int		status;

  fd = open(pidfile, O_WRONLY|O_CREAT|O_EXCL, 0644);
  if (fd == -1) {
    return -1;
  }

  status = fchown(fd, uid, gid);
  fout	 = fdopen(fd, "w");
  if (fout == NULL) {
    unlink(pidfile);
    wu_close(fd);
    return -1;
  }

  fprintf(fout, "%u\n", (unsigned int)getpid());
  fclose(fout);

  return status;
}
