/*****************************************************************************
 *
 * FILE:	auth.c
 * DESCRIPTION:	Wyvern: common authentication library
 * DATE:	Mon, Sep 18 2000
 * UPDATED:	Wed, Sep 12 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2000-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2000-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: auth.c,v 1.25 2007/09/12 02:24:26 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <errno.h>
#if	ENABLE_SSL
#include <openssl/md5.h>
#define	MD5Init		MD5_Init
#define	MD5Update	MD5_Update
#define	MD5Final	MD5_Final
#else
#include <md5.h>
#endif	/* ENABLE_SSL */
#include "wyvern.h"
#include "auth.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	WM_MODULE_ID	"libauth"

#define	WA_SYSTEM_NAME	"<wyvern>"
typedef struct {
  char *	name;	/* username */
  char *	policy;	/* path of wauth.conf */
  char *	path;	/* path of password file */
  sxml_node_t *	root;
} Auth;

/* Authentication Scheme */
typedef enum {
  WA_SCHEME_BASIC,
  WA_SCHEME_DIGEST
} wa_scheme_t;

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static sxml_node_t *	policy_parse(const char *);
static Auth *		policy_insert(Cache *, const char *, const char *, const char *);
static Auth *		policy_lookup(Cache *, char *);
static int		compare_user(const void *, const void *);

static int	authenticate(Message *, sxml_node_t *, const char *,
			     const char *);
static int	user_checker(Message *, sxml_node_t *, const char *);
static int	basic_request(Message *, sxml_node_t *, char *, const char *);
static int	basic_response(Message *, const char *);
static int	digest_request(Message *, sxml_node_t *, char *, const char *);
static int	digest_response(Message *, const char *);
static int	digest_parse_args(const char **, char *);
static void	toHex(HASH, HASHHEX);
static void	digest_calc_nonce(Message *, HASHHEX);
static int	default_pw_checker(Message *, const char *, const char **,
				   wa_scheme_t);
static void	trashcan(const void *);
static void	wm_libauth_register_hooks(void);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static AuthProc	authProc = { NULL, NULL, default_pw_checker };

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/
static Auth *
policy_insert(top, name, policy, path)
	register Cache *	top;
	const char *		name;
	const char *		policy;
	const char *		path;
{
  sxml_node_t *	root;

  root = policy_parse(policy);
  if (root != NULL) {
    Cache *	new = wc_remove_cache(top->next);
    if (new != NULL) {
      Auth *	auth = (Auth *)ws_malloc(sizeof(Auth));

      auth->name   = ws_strdup(name);
      auth->policy = ws_strdup(policy);
      auth->path   = ws_strdup(path);
      auth->root   = root;
      if (new->data != NULL) {
#if	DEBUG
	fprintf(stderr, "DEBUG[free] %s: %s %s\n",
	     	((Auth *)new->data)->name,
	      	((Auth *)new->data)->policy,
	       	((Auth *)new->data)->path);
#endif	/* DEBUG */
	free((char *)((Auth *)new->data)->name);
	free((char *)((Auth *)new->data)->policy);
	free((char *)((Auth *)new->data)->path);
	sxml_delete_node(((Auth *)new->data)->root);
	free((Auth *)new->data);
      }
      new->data = auth;
      wc_insert_cache(top, new);
      DBG(fprintf(stderr, "DEBUG[insert] %s: %s %s\n", name, policy, path);)

      return auth;
    }
  }

  return NULL;
}

static sxml_node_t *
policy_parse(path)
	const char *	path;
{
  register FILE *	fin;

  fin = ws_safe_fopen(path, "r");
  if (fin != NULL) {
    sxml_node_t *	root;

    root = sxml_parse_file(fileno(fin));
    fclose(fin);

    return root;
  }

  return NULL;
}

static Auth *
policy_lookup(top, name)
	register Cache *	top;
	char *			name;
{
  Auth		auth;
  Auth *	p;

  auth.name = name;
  p = (Auth *)wc_search_cache(top, &auth, compare_user);
#if	DEBUG
  if (p != NULL) {
    fprintf(stderr, "DEBUG[search] %s: %s %s\n", p->name, p->policy, p->path);
  }
  else {
    fprintf(stderr, "DEBUG[search] %s not found\n", name);
  }
#endif	/* DEBUG */

  return p;
}

static int
compare_user(x, y)
	register const void *	x;
	register const void *	y;
{
  if (ws_strcmp(((Auth *)x)->name, ((Auth *)y)->name) == 0) {
    return 0;
  }
  return -1;
}


static int
authenticate(msg, tree, dir, path)
	register Message *	msg;
	register sxml_node_t *	tree;
	const char *		dir;
	const char *		path;	/* path of password file */
{
  if (user_checker(msg, tree->child, path) == -1) {
    sxml_node_t *	node;

    msg->response.status = STATUS_UNAUTHORIZED;
    node		 = sxml_find_element(tree->child, "auth", NULL, NULL);
    if (node != NULL) {
      const char *	type;

      type = sxml_get_attribute(node, "type");
      if (ws_strcasecmp(type, "digest") == 0) {
	return digest_response(msg, dir);
      }
    }
    return basic_response(msg, dir);
  }

  return 0;
}

/*
 * Function Name:	user_checker
 * Description:		check password of user to access
 * Arguments:		message, tree, password file
 * Return Value:	allow: 0, otherwise: -1
 */
static int
user_checker(msg, tree, path)
	register Message *	msg;
	register sxml_node_t *	tree;
	const char *		path;	/* path of password file */
{
  const char *	p;	/* Authorization */
  int		status = -1;

  p = wa_get(msg->request.header, "authorization");
  if (p != NULL) {
    char *	scheme;
    char *	param;

    /* divide 'Authorization' into 'scheme' and 'param' */
    scheme = ws_strdup(p);
    param  = strchr(scheme, SPC);
    if (param != NULL) {
      *param++ = EOL;

      wv_free(msg->request.auth_type);
      msg->request.auth_type = scheme;
      /* basic authentication scheme */
      if (ws_strcmp(scheme, "Basic") == 0) {
	status = basic_request(msg, tree, param, path);
      }
      /* digest authentication scheme */
      else if (ws_strcmp(scheme, "Digest") == 0) {
	status = digest_request(msg, tree, param, path);
      }
      /* unknown authentication scheme */
      else {
	msg->response.status = STATUS_UNAUTHORIZED;
      }
    }
  }

  return status;
}

static int
basic_request(msg, tree, param, path)
	register Message *	msg;
	register sxml_node_t *	tree;
	char *			param;
	const char *		path;	/* path of password file */
{
  const char **	ap;
  const char *	args[WA_BASIC_ARGS_SIZE];
  char *	p;
  int		status = -1;

  p = wu_decode_base64(param);
  if (p != NULL) {
    sxml_node_t *	node;

    /* set username and password */
    for (ap = args; (*ap = wu_strsep1(&p, ":")) != NULL; ) {
      if (**ap != EOL) {
       	if (++ap >= &args[WA_BASIC_ARGS_SIZE]) {
  	  break;
   	}
      }
    }
    free((char *)p);

    node = sxml_find_element(tree, "auth", "type", "Basic");
    if (node != NULL) {
      register sxml_node_t *	np;

      for (np = node->child; np != NULL; np = np->next) {
	const char *	name;

	if (np->type != SXML_ELEMENT)	    { continue; }
	if (np->value.element.name == NULL) { continue; }

	name = sxml_get_content(np->child);
	if (ws_strcmp(np->value.element.name, "user") == 0 &&
	    ws_strcmp(name, args[WA_BASIC_USERNAME]) == 0) {
	  DBG(fprintf(stderr, "[basic_request] %s(%s)\n",
		      args[WA_BASIC_USERNAME], args[WA_BASIC_PASSWORD]);)
	  msg->user.authName = ws_strdup(args[WA_BASIC_USERNAME]);
	  status = (*authProc.check_password)(msg, path, args, WA_SCHEME_BASIC);
	  break;
	}
      } /* end of for */
    }
  }

  return status;
}

static int
basic_response(msg, dir)
	register Message *	msg;
	const char *		dir;
{
  char *	wa;

  if (msg->user.name != NULL) {
    ws_asprintf(&wa, "Basic realm=\"%s@%s:%s\"",
		msg->user.name, msg->vhost->serverName, dir);
  }
  else {
    ws_asprintf(&wa, "Basic realm=\"%s:%s\"", msg->vhost->serverName, dir);
  }
  wa_put(msg->response.header, "WWW-Authenticate", wa);
  free((char *)wa);

  return -1;
}

static int
digest_request(msg, tree, param, path)
	register Message *	msg;
	register sxml_node_t *	tree;
	char *			param;
	const char *		path;	/* path of password file */
{
  unsigned int	n;
  char **	ap;
  char *	params[WA_DIGEST_ARGS_SIZE];
  char *	p = param;
  sxml_node_t *	node;
  int		status = -1;

  DBG(fprintf(stderr, "[digest_request] param=%s\n", p);)
  /* set username and password */
  for (n = 0, ap = params; (*ap = wu_strsep2(&p, ", \r\n")) != NULL; ) {
    if (**ap != EOL) {
      if (++ap >= &params[WA_DIGEST_ARGS_SIZE]) {
	break;
      }
      n++;
    }
  }

  node = sxml_find_element(tree, "auth", "type", "Digest");
  if (node != NULL) {
    register unsigned int	i;
    const char *		args[WA_DIGEST_ARGS_SIZE];
    register sxml_node_t *	np;

    /* parse arguments of digest response */
    for (i = 0; i < n; i++) {
      args[i] = NULL;
      digest_parse_args(args, params[i]);
    }

    for (np = node->child; np != NULL; np = np->next) {
      const char *	name;

      if (np->type != SXML_ELEMENT)	  { continue; }
      if (np->value.element.name == NULL) { continue; }

      name = sxml_get_content(np->child);
      if (ws_strcmp(np->value.element.name, "user") == 0 &&
	  ws_strcmp(name, args[WA_DIGEST_USERNAME]) == 0) {
	msg->user.authName = ws_strdup(args[WA_DIGEST_USERNAME]);
  	status = (*authProc.check_password)(msg, path, args, WA_SCHEME_DIGEST);
  	break;
      }
    } /* end of for */
    /* release memory */
    for (i = 0; i < n; i++) {
      wv_free(args[i]);
    }
  }

  return status;
}

static int
digest_response(msg, dir)
	register Message *	msg;
	const char *		dir;
{
  HASHHEX	nonce;
  char *	options;
  char *	opaque = NULL;
  char *	wa;	/* WWW-Authenticate */

  if (authProc.make_opaque != NULL) {
    opaque = (*authProc.make_opaque)(msg);
  }
  if (opaque != NULL) {
    ws_asprintf(&options,
		"\talgorithm=\"MD5\"," CRLF
		"\tqop=\"auth,auth-int\"," CRLF
		"\topaque=\"%s\"", opaque);
    free((char *)opaque);
  }
  else {
    ws_asprintf(&options,
		"\talgorithm=\"MD5\"," CRLF
		"\tqop=\"auth,auth-int\"");
  }

  digest_calc_nonce(msg, nonce);
  if (msg->user.name != NULL) {
    ws_asprintf(&wa,
		"Digest realm=\"%s@%s:%s\"," CRLF
		"\tdomain=\"%s\"," CRLF
		"\tnonce=\"%s\"," CRLF
		"%s",
		msg->user.name, msg->vhost->serverName, dir, msg->request.uri,
		nonce, options);
  }
  else {
    ws_asprintf(&wa,
	 	"Digest realm=\"%s:%s\"," CRLF
		"\tdomain=\"%s\"," CRLF
		"\tnonce=\"%s\"," CRLF
		"%s",
	 	msg->vhost->serverName, dir, msg->request.uri, nonce, options);
  }
  wa_put(msg->response.header, "WWW-Authenticate", wa);
  free((char *)wa);
  free((char *)options);

  return -1;
}

static int
digest_parse_args(args, arg)
	const char **	args;
	char *		arg;
{
#define	DIGEST_ARGS_USERNAME	"username"	// username
#define	DIGEST_ARGS_REALM	"realm"		// realm
#define	DIGEST_ARGS_NONCE	"nonce"		// nonce
#define	DIGEST_ARGS_URI		"uri"		// digest-uri
#define	DIGEST_ARGS_RESPONSE	"response"	// response
#define	DIGEST_ARGS_ALGORITHM	"algorithm"	// algorithm
#define	DIGEST_ARGS_CNONCE	"cnonce"	// cnonce
#define	DIGEST_ARGS_OPAQUE	"opaque"	// opaque
#define	DIGEST_ARGS_QOP		"qop"		// message-qop
#define	DIGEST_ARGS_NC		"nc"		// nonce-count
  register char *	sp;
  register char *	ep;

  ep = strchr(arg, '=');
  if (ep != NULL) {
    const char **	ap;

    *ep++ = EOL;
    sp	  = ep;
    if (*sp == '"') {
      sp++;
      ep = strchr(sp, '"');
      if (ep != NULL) {
	*ep = EOL;
      }
    }
    if (ws_strcmp(arg, DIGEST_ARGS_USERNAME) == 0) {
      ap = &args[WA_DIGEST_USERNAME];
    }
    else if (ws_strcmp(arg, DIGEST_ARGS_REALM) == 0) {
      ap = &args[WA_DIGEST_REALM];
    }
    else if (ws_strcmp(arg, DIGEST_ARGS_NONCE) == 0) {
      ap = &args[WA_DIGEST_NONCE];
    }
    else if (ws_strcmp(arg, DIGEST_ARGS_URI) == 0) {
      ap = &args[WA_DIGEST_URI];
    }
    else if (ws_strcmp(arg, DIGEST_ARGS_RESPONSE) == 0) {
      ap = &args[WA_DIGEST_RESPONSE];
    }
    else if (ws_strcmp(arg, DIGEST_ARGS_ALGORITHM) == 0) {
      ap = &args[WA_DIGEST_ALGORITHM];
    }
    else if (ws_strcmp(arg, DIGEST_ARGS_CNONCE) == 0) {
      ap = &args[WA_DIGEST_CNONCE];
    }
    else if (ws_strcmp(arg, DIGEST_ARGS_OPAQUE) == 0) {
      ap = &args[WA_DIGEST_OPAQUE];
    }
    else if (ws_strcmp(arg, DIGEST_ARGS_QOP) == 0) {
      ap = &args[WA_DIGEST_QOP];
    }
    else if (ws_strcmp(arg, DIGEST_ARGS_NC) == 0) {
      ap = &args[WA_DIGEST_NC];
    }
    else {
      return -1;
    }
    *ap = ws_strdup(sp);
  }

  return 0;
}

static void
toHex(bin, hex)
	HASH	bin;	/* IN */
	HASHHEX hex;	/* OUT */
{
  register unsigned short	i;
  register unsigned char	j;

  for (i = 0; i < WA_HASHLEN; i++) {
    j = (bin[i] >> 4) & 0xf;
    if (j <= 9) {
      hex[i*2] = (j + '0');
    }
    else {
      hex[i*2] = (j + 'a' - 10);
    }
    j = bin[i] & 0xf;
    if (j <= 9) {
      hex[i*2+1] = (j + '0');
    }
    else {
      hex[i*2+1] = (j + 'a' - 10);
    }
  }
  hex[WA_HASHHEXLEN] = '\0';
}

static void
digest_calc_nonce(msg, retval)
	Message *	msg;
	HASHHEX		retval;
{
  struct stat	sbuf;
  char *	etag;
  int		size;
  time_t	now;
  MD5_CTX	md5_ctx;
  HASH		md5_bin;

  if (stat(msg->file.path, &sbuf) == 0) {
    size = ws_asprintf(&etag, "\"%lx-%lx-%lx\"",
		       (long)sbuf.st_ino, (long)sbuf.st_size, sbuf.st_mtime);
  }
  else {
    size = ws_asprintf(&etag, "\"%s\"", msg->file.path);
  }
  time(&now);
  MD5Init(&md5_ctx);
  MD5Update(&md5_ctx, ctime(&now), 24);
  MD5Update(&md5_ctx, ":", 1);
  MD5Update(&md5_ctx, etag, (size_t)size);	free((char *)etag);
  MD5Update(&md5_ctx, ":", 1);
  MD5Update(&md5_ctx, msg->client.ip, strlen(msg->client.ip));
  MD5Final(md5_bin, &md5_ctx);
  toHex(md5_bin, retval);
}

static int
default_pw_checker(msg, path, args, scheme)
	register Message *	msg;
	const char *		path;	/* path of password file */
	const char *		args[];
	wa_scheme_t		scheme;	/* Basic | Digest */
{
#define	USERNAME	(0)
  char *	passwd;
  int		status = -1;

  passwd = (*authProc.get_password)(path, args[USERNAME]);
  if (passwd != NULL) {
    if (scheme == WA_SCHEME_DIGEST) {
      const char *	qop = "";
      HASHHEX		HA1;
      HASHHEX		HA2 = "";
      HASHHEX		response;

      /* qop checker */
      if (args[WA_DIGEST_QOP] != NULL) {
	if (ws_strcmp(args[WA_DIGEST_QOP], "auth") == 0 ||
	    ws_strcmp(args[WA_DIGEST_QOP], "auth-int") == 0) {
	  qop = args[WA_DIGEST_QOP];
	}
      }

      if (wauth_digest_calc_ha1(args[WA_DIGEST_ALGORITHM],
				args[WA_DIGEST_USERNAME], args[WA_DIGEST_REALM],
				passwd, args[WA_DIGEST_NONCE],
				args[WA_DIGEST_CNONCE], HA1) == 0) {
	if (wauth_digest_calc_response(HA1, args[WA_DIGEST_NONCE],
				       args[WA_DIGEST_NC],
				       args[WA_DIGEST_CNONCE], qop,
				       msg->request.method.name,
				       args[WA_DIGEST_URI],
				       HA2, response) == 0) {
	  status = ws_strcmp(response, args[WA_DIGEST_RESPONSE]);
	}
      }
    }
    else {
      status = ws_strcmp(passwd, args[WA_BASIC_PASSWORD]);
    }
    free((char *)passwd);
  }

  return status;
}

/*****************************************************************************/

/* calculate H(A1) as per HTTP Digest spec */
int
wauth_digest_calc_ha1(algorithm, username, realm, password, nonce, cnonce,
		      sessionKey)
	const char *	algorithm;	/* IN */
	const char *	username;	/* IN */
	const char *	realm;		/* IN */
	const char *	password;	/* IN */
	const char *	nonce;		/* IN */
	const char *	cnonce;		/* IN */
	HASHHEX		sessionKey;	/* OUT */
{
  MD5_CTX	md5_ctx;
  HASH		HA1;

  if (username == NULL || realm == NULL || password == NULL) {
    return -1;
  }
  MD5Init(&md5_ctx);
  MD5Update(&md5_ctx, username, strlen(username));
  MD5Update(&md5_ctx, ":", 1);
  MD5Update(&md5_ctx, realm, strlen(realm));
  MD5Update(&md5_ctx, ":", 1);
  MD5Update(&md5_ctx, password, strlen(password));
  MD5Final(HA1, &md5_ctx);

  if (algorithm != NULL && nonce != NULL && cnonce != NULL) {
    if (ws_strcmp(algorithm, "MD5-sess") == 0) {
      MD5Init(&md5_ctx);
      MD5Update(&md5_ctx, HA1, WA_HASHLEN);
      MD5Update(&md5_ctx, ":", 1);
      MD5Update(&md5_ctx, nonce, strlen(nonce));
      MD5Update(&md5_ctx, ":", 1);
      MD5Update(&md5_ctx, cnonce, strlen(cnonce));
      MD5Final(HA1, &md5_ctx);
    }
  }
  toHex(HA1, sessionKey);

  return 0;
}

/* calculate request-digest/response-digest as per HTTP Digest spec */
int
wauth_digest_calc_response(HA1, nonce, nonce_count, cnonce, qop, method,
			   digest_uri, HEntity, retval)
	HASHHEX		HA1;		/* IN: H(A1) */
	const char *	nonce;		/* IN: nonce from server */
	const char *	nonce_count;	/* IN: 8 hex digits */
	const char *	cnonce;		/* IN: client nonce */
	const char *	qop;	/* IN: qop-value: "", "auth", "auth-int" */
	const char *	method;		/* IN: method from the request */
	const char *	digest_uri;	/* IN: requested URL */
	HASHHEX		HEntity;/* IN: H(entity body) if qop="auth-int" */
	HASHHEX		retval;	/* OUT: request-digest or response-digest */
{
  MD5_CTX	md5_ctx;
  HASH		HA2;
  HASH		respHash;
  HASHHEX	HA2Hex;

  if (method == NULL || digest_uri == NULL || nonce == NULL) {
    return -1;
  }
  // calculate H(A2)
  MD5Init(&md5_ctx);
  MD5Update(&md5_ctx, method, strlen(method));
  MD5Update(&md5_ctx, ":", 1);
  MD5Update(&md5_ctx, digest_uri, strlen(digest_uri));
  if (ws_strcmp(qop, "auth-int") == 0) {
    MD5Update(&md5_ctx, ":", 1);
    MD5Update(&md5_ctx, HEntity, WA_HASHHEXLEN);
  }
  MD5Final(HA2, &md5_ctx);
  toHex(HA2, HA2Hex);

  // calculate response
  MD5Init(&md5_ctx);
  MD5Update(&md5_ctx, HA1, WA_HASHHEXLEN);
  MD5Update(&md5_ctx, ":", 1);
  MD5Update(&md5_ctx, nonce, strlen(nonce));
  MD5Update(&md5_ctx, ":", 1);
  if (*qop != '\0' && nonce_count != NULL && cnonce != NULL) {
    MD5Update(&md5_ctx, nonce_count, strlen(nonce_count));
    MD5Update(&md5_ctx, ":", 1);
    MD5Update(&md5_ctx, cnonce, strlen(cnonce));
    MD5Update(&md5_ctx, ":", 1);
    MD5Update(&md5_ctx, qop, strlen(qop));
    MD5Update(&md5_ctx, ":", 1);
  }
  MD5Update(&md5_ctx, HA2Hex, WA_HASHHEXLEN);
  MD5Final(respHash, &md5_ctx);
  toHex(respHash, retval);

  return 0;
}

int
wauth_checker(top, msg, db_suffix)
	register Cache *	top;
	register Message *	msg;
	const char *		db_suffix;
{
  Auth		auth;
  Auth *	ap;

  if (msg->user.name != NULL) {
    auth.name = msg->user.name;
    ws_asprintf((char **)&auth.policy, "%s/%s/%s",
		msg->user.homeDir, msg->vhost->userDir, WYVERN_USER_AUTH);
    ws_asprintf((char **)&auth.path, "%s/%s/%s%s",
		msg->user.homeDir, msg->vhost->userDir,
		WYVERN_USER_PASSWD, db_suffix);
  }
  else {
    auth.name = WA_SYSTEM_NAME;
    ws_asprintf((char **)&auth.policy, "%s/%s",
		httpd->configDir, WYVERN_SERVER_AUTH);
    ws_asprintf((char **)&auth.path, "%s/%s%s",
		httpd->serverRoot, WYVERN_SERVER_PASSWD, db_suffix);
  }

  ap = policy_lookup(top, auth.name);
  if (ap == NULL) {
    ap = policy_insert(top, auth.name, auth.policy, auth.path);
  }
  free((char *)auth.policy);
  free((char *)auth.path);

  if (ap != NULL) {
    register sxml_node_t *	np;

    for (np = sxml_find_element(ap->root, "limit", "directory", NULL);
	 np != NULL; np = np->next) {
      const char *	dir;

      dir = sxml_get_attribute(np, "directory");
      if (dir != NULL && strstr(msg->request.uri, dir) != NULL) {
	return authenticate(msg, np, dir, ap->path);
      }
    }
  }

  return 0;
}

void
wauth_cleaner(top)
	Cache *	top;
{
  if (top != NULL) {
    wc_flush_cache(top, trashcan);
    free((Cache *)top);
    top = NULL;
  }
}

static void
trashcan(x)
	const void *	x;
{
  Auth *	ap = (Auth *)x;
  wv_free(ap->name);
  wv_free(ap->policy);
  wv_free(ap->path);
  if (ap->root != NULL) {
    sxml_delete_node(ap->root);
  }
  free((Auth *)ap);
  ap = NULL;
}

void
wauth_set_subr(app)
	AuthProc *	app;
{
  if (app->get_password != NULL) {
    authProc.get_password = app->get_password;
  }
  if (app->make_opaque != NULL) {
    authProc.make_opaque = app->make_opaque;
  }
  if (app->check_password != NULL) {
    authProc.check_password = app->check_password;
  }
}

/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/
static void
wm_libauth_register_hooks(void)
{
  /* nothing to do */
}

static wmModule	libauth = {
  WM_MAGIC_NUMBER_MAJOR,		/* version */
  WM_FIXED,				/* secure */
  WM_MODULE_ID,				/* name */
  "Common Authentication Library 1.0",	/* info */
  "Kouichi ABE (WALL)",			/* author */
  NULL,					/* depend */
  NULL,					/* handle */
  (void *)NULL,				/* init */
  (void *)NULL,				/* finish */
  (void *)wm_libauth_register_hooks	/* register_handler */
};

wmModule *
wmModuleInfo(void)
{
  return &libauth;
}
