/*****************************************************************************
 *
 * FILE:	mod_admin.c
 * DESCRIPTION:	Wyvern: administrator session management module
 * DATE:	Sat, Nov  6 2004
 * UPDATED:	Wed, Sep 12 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2004-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2004-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_admin.c,v 1.14 2007/09/12 02:24:26 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#if	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif	/* HAVE_SYS_TIME_H */
#if	ENABLE_SSL
#include <openssl/sha.h>
#else
#if	HAVE_SHA_H
#include <sha.h>
#elif	HAVE_SHA1_H
#include <sha1.h>
#endif	/* HAVE_SHA_H */
#endif	/* ENABLE_SSL */
#if	ENABLE_SQLITE3
#include <sqlite3.h>
#elif	ENABLE_SQLITE2
#include <sqlite.h>
#endif	/* ENABLE_SQLITE3 */
#include <wyvern.h>

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	WM_MODULE_ID		"mod_admin"

#define	MAGIC_TYPE_ADMIN	"httpd/x-admin-*"
#define	MAGIC_TYPE_ADMIN_LOGIN	"httpd/x-admin-login"
#define	MAGIC_TYPE_ADMIN_LOGOUT	"httpd/x-admin-logout"

#define	ADMIN_AUTH_DB		"admin_auth.db"
#define	ADMIN_HISTORY_DB	"admin_history.db"
#define	ADMIN_CONF		"admin.conf"

#define	ADM_CACHE_CONTROL	"no-store, no-cache"
#define	ADM_PRAGMA		"no-cache"

typedef struct {
  char *	path;		/* path for administrator realm */
  size_t	plen;		/* path length */
  char *	page_login;	/* login dialog */
  char *	page_startup;	/* startup page after log-in */
  char *	page_expired;	/* optional */
  char *	page_invalid;	/* optional */
  char *	page_failure;	/* optional */
  char *	cmd_login;
  char *	cmd_logout;
  long		timeout;	/* timeout of session */
} wmConfig;

#define	MAX_SESSION	(7)
typedef struct {
  char *	id;
  long		expire;
  long		timeout;	/* interval until timeout */
} wmSession;

#define WM_SID_INVALID	(-1)
#define WM_SID_VALID	(0)
#define WM_SID_EXPIRED	(1)

#define WM_AUTH_NG	(-1)
#define WM_AUTH_OK	(0)

#define	SESSION_ID	"wmsid"
#define	PARAM_SID_SIZE	(2)
#define	PARAM_SID	SESSION_ID
#define	PARAM_AUTH_SIZE	(4)
#define	PARAM_USER_ID	"user_id"
#define	PARAM_PASSWORD	"password"

#define	DEF_HTTP_PORT		(80)
#define	DEF_SESSION_TIMEOUT	(5*60)	/* seconds */

#define	INFO_LOGOUT	"logout"
#define	INFO_EXPIRED	"expired"
#define	INFO_INVALID	"invalid"
#define	INFO_NO_COOKIE	"no cookie or session id expired"
#define	INSERT_HISTORY \
	"INSERT INTO admin_history VALUES(NULL, '%ld', '%s', '%s', '%q');"

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	wm_admin_init(void);
static void	wm_admin_finish(void);
static void	wm_admin_register_hooks(void);
static int	wm_admin_uri_translater(Message *);
static int	wm_admin_request_handler(Message *);
static int	wm_admin_validate_session(Message *);
static void	init_config(void);
static void	verify_config(void);
static void	parse_config(sxml_node_t *);
static void	parse_error(const char *);
static int	record_history(char *, char *, char *);
static int	get_values_from_query(char *, char **, size_t);
static int	authenticate(Message *, const char *, const char *);
static char *	generate_sid(Message *, const char *, const char *, long);
static int	set_cookie(Message *, const char *, const char *, long);
static int	set_redirect(Message *, const char *);
static int	set_cache_control(Message *);
#if	DEBUG
static void	debug_expires(long);
#endif	/* DEBUG */
static int	validate_sid(const char *, long *, long *);
static void	expire_sid(void);
static int	delete_sid(const char *);
static void	recycle_sid(int);
static int	lookup(const char *);
static unsigned int	rehash(unsigned int);
static unsigned int	hash(const char *);
#if	0
static char *	get_sid_from_query(char *);
#endif
static char *	get_sid_from_cookie(char *);
static int	parse_string(char *, char **, char **);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static wmConfig		config;
static wmSession	session[MAX_SESSION];
static unsigned int	active;	/* active session */
#if	ENABLE_SQLITE3
static sqlite3 *	sdb_history = NULL;	/* admin history */
#elif	ENABLE_SQLITE2
static sqlite *		sdb_history = NULL;	/* admin history */
#endif	/* ENABLE_SQLITE3 */

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wm_admin_init
 * Description:		initialize CGI object
 * Arguments:		none
 * Return Value:	none
 */
static void
wm_admin_init(void)
{
  char *	path;

  /* read configuration */
  init_config();
  ws_asprintf(&path, "%s/%s", httpd->configDir, ADMIN_CONF);
  if (path != NULL) {
    register FILE *	fin;
    register int	i;

    fin = fopen(path, "r");	free(path);
    if (fin != NULL) {
      sxml_node_t *	root;

      root = sxml_parse_file(fileno(fin));
      if (root != NULL) {
	parse_config(root);
	sxml_delete_node(root);
      }
      fclose(fin);
    }
    verify_config();

    /* initialize session variables */
    for (i = 0; i < MAX_SESSION; i++) {
      session[i].id = NULL;
    }
    active = 0;

    /* open history database */
    ws_asprintf(&path, "%s/%s", httpd->databaseDir, ADMIN_HISTORY_DB);
    if (path != NULL) {
#if	ENABLE_SQLITE3
      sqlite3_open(path, &sdb_history);
#elif	ENABLE_SQLITE2
      sdb_history = sqlite_open(path, 0, NULL);
#endif	/* ENABLE_SQLITE3 */
      free(path);
    }
  }
}

static void
init_config(void)
{
  config.path		= NULL;
  config.plen		= 0;
  config.page_login	= NULL;
  config.page_startup	= NULL;
  config.page_expired	= NULL;
  config.page_invalid	= NULL;
  config.page_failure	= NULL;
  config.cmd_login	= NULL;
  config.cmd_logout	= NULL;
  config.timeout	= DEF_SESSION_TIMEOUT;
}

static void
verify_config(void)
{
  if (config.path == NULL) {
    wv_die("[%s] no admin realm", WM_MODULE_ID);
  }
  if (config.page_login	== NULL) {
    wv_die("[%s] no login page", WM_MODULE_ID);
  }
  if (config.page_startup == NULL) {
    wv_die("[%s] no startup page", WM_MODULE_ID);
  }
  if (config.cmd_login == NULL) {
    wv_die("[%s] no login command", WM_MODULE_ID);
  }
  if (config.cmd_logout	== NULL) {
    wv_die("[%s] no logout command", WM_MODULE_ID);
  }
}

static void
parse_config(root)
	sxml_node_t *	root;
{
  sxml_node_t *	node;

  node = sxml_find_element(root, "wyvern-administrator", NULL, NULL);
  if (node != NULL) {
    const char *	attr;
    sxml_node_t *	cp;

    attr = sxml_get_attribute(node, "realm");
    if (attr == NULL) {
      attr = "/";
    }
    config.path = ws_strdup(attr);
    config.plen = strlen(config.path);

    for (cp = node->child; cp != NULL; cp = cp->next) {
      if (cp->type != SXML_ELEMENT) { continue; }
      if (cp->value.element.name == NULL) { continue; }

      if (ws_strcmp(cp->value.element.name, "page") == 0) {
#define	PG_IDX_LOGIN	(0)
#define	PG_IDX_STARTUP	(1)
#define	PG_IDX_EXPIRED	(2)
#define	PG_IDX_INVALID	(3)
#define	PG_IDX_FAILURE	(4)
	sxml_node_t *	np;
	const char *	args[5] = { NULL, NULL, NULL, NULL, NULL };

	for (np = cp->child; np != NULL; np = np->next) {
	  const char *	content;

	  if (np->type != SXML_ELEMENT) { continue; }
	  if (np->value.element.name == NULL) { continue; }

	  content = sxml_get_content(np->child);
	  if (ws_strcmp(np->value.element.name, "login") == 0) {
	    args[PG_IDX_LOGIN] = content;
	  }
	  else if (ws_strcmp(np->value.element.name, "startup") == 0) {
	    args[PG_IDX_STARTUP] = content;
	  }
	  else if (ws_strcmp(np->value.element.name, "expired-sid") == 0) {
	    args[PG_IDX_EXPIRED] = content;
	  }
	  else if (ws_strcmp(np->value.element.name, "invalid-sid") == 0) {
	    args[PG_IDX_INVALID] = content;
	  }
	  else if (ws_strcmp(np->value.element.name, "login-failure") == 0) {
	    args[PG_IDX_FAILURE] = content;
	  }
	  else {
	    parse_error(np->value.element.name);
	  }
	}
	if (args[PG_IDX_LOGIN] != NULL && args[PG_IDX_STARTUP] != NULL) {
	  config.page_login   = ws_strdup(args[PG_IDX_LOGIN]);
	  config.page_startup = ws_strdup(args[PG_IDX_STARTUP]);
	}
	if (args[PG_IDX_EXPIRED] != NULL) {
	  config.page_expired = ws_strdup(args[PG_IDX_EXPIRED]);
	}
	if (args[PG_IDX_INVALID] != NULL) {
	  config.page_invalid = ws_strdup(args[PG_IDX_INVALID]);
	}
	if (args[PG_IDX_FAILURE] != NULL) {
	  config.page_failure = ws_strdup(args[PG_IDX_FAILURE]);
	}
      }
      else if (ws_strcmp(cp->value.element.name, "command") == 0) {
#define	CMD_IDX_LOGIN	(0)
#define	CMD_IDX_LOGOUT	(1)
	sxml_node_t *	np;
	const char *	args[2] = { NULL, NULL };

	for (np = cp->child; np != NULL; np = np->next) {
	  const char *	content;

	  if (np->type != SXML_ELEMENT) { continue; }
	  if (np->value.element.name == NULL) { continue; }

	  content = sxml_get_content(np->child);
	  if (ws_strcmp(np->value.element.name, "login") == 0) {
	    args[CMD_IDX_LOGIN] = content;
	  }
	  else if (ws_strcmp(np->value.element.name, "logout") == 0) {
	    args[CMD_IDX_LOGOUT] = content;
	  }
	  else {
	    parse_error(np->value.element.name);
	  }
	}
	if (args[CMD_IDX_LOGIN] != NULL && args[CMD_IDX_LOGOUT] != NULL) {
	  config.cmd_login  = ws_strdup(args[CMD_IDX_LOGIN]);
	  config.cmd_logout = ws_strdup(args[CMD_IDX_LOGOUT]);
	}
      }
      else if (ws_strcmp(cp->value.element.name, "timeout") == 0) {
	const char *	content;

	content		= sxml_get_content(cp->child);
	config.timeout	= strtol(content, (char **)NULL, 10);
      }
      else {
	parse_error(cp->value.element.name);
      }
    }
  }
}

static void
parse_error(tag)
	const char *	tag;
{
  wv_die("[%s] error: unknown tag: %s", WM_MODULE_ID, tag);
}

/*****************************************************************************/

/*
 * Function Name:	wm_admin_finish
 * Description:		delete CGI object
 * Arguments:		none
 * Return Value:	none
 */
static void
wm_admin_finish(void)
{
  register int	i;

  for (i = 0; i < MAX_SESSION; i++) {
    wv_free(session[i].id);
  }
  wv_free(config.path);
  wv_free(config.page_login);
  wv_free(config.page_startup);
  wv_free(config.page_expired);
  wv_free(config.page_invalid);
  wv_free(config.page_failure);
  wv_free(config.cmd_login);
  wv_free(config.cmd_logout);

  if (sdb_history != NULL) {
#if	ENABLE_SQLITE3
    sqlite3_close(sdb_history);
#elif	ENABLE_SQLITE2
    sqlite_close(sdb_history);
#endif	/* ENABLE_SQLITE3 */
  }
}

/*****************************************************************************/

static int
record_history(ipaddr, sid, info)
	char *	ipaddr;
	char *	sid;
	char *	info;
{
  int	status = -1;

  if (sdb_history != NULL && ipaddr != NULL && sid != NULL && info != NULL) {
    time_t	now;
    char *	sql;

    time(&now);
#if	ENABLE_SQLITE3
    sql = sqlite3_mprintf(INSERT_HISTORY, now, ipaddr, sid, info);
#elif	ENABLE_SQLITE2
    sql = sqlite_mprintf(INSERT_HISTORY, now, ipaddr, sid, info);
#endif	/* ENABLE_SQLITE3 */
    if (sql != NULL) {
#define	INSERT_RETRY_MAX	(3)
      int	rc;
      int	c = 0;	/* retry counter */

retry:
#if	ENABLE_SQLITE3
      rc = sqlite3_exec(sdb_history, sql, NULL, NULL, NULL);
#elif	ENABLE_SQLITE2
      rc = sqlite_exec(sdb_history, sql, NULL, NULL, NULL);
#endif	/* ENABLE_SQLITE3 */
      switch (rc) {
	case SQLITE_OK:
	  status = 0;
	  break;
	case SQLITE_BUSY:
	  sleep(1);
	  if (c++ < INSERT_RETRY_MAX) {
	    goto retry;
	  }
	default:
	  break;
      }
#if	ENABLE_SQLITE3
      sqlite3_free(sql);
#elif	ENABLE_SQLITE2
      sqlite_freemem(sql);
#endif	/* ENABLE_SQLITE3 */
    }
  }

  return status;
}

/*****************************************************************************/

static int
wm_admin_uri_translater(msg)
	register Message *	msg;
{
  if (ws_strcmp(msg->request.uri, config.cmd_login) == 0) {
    msg->file.path = NULL;
    msg->file.mime = MAGIC_TYPE_ADMIN_LOGIN;
    return WM_RET_OK;
  }
  else if (ws_strcmp(msg->request.uri, config.cmd_logout) == 0) {
    msg->file.path = NULL;
    msg->file.mime = MAGIC_TYPE_ADMIN_LOGOUT;
    return WM_RET_OK;
  }

  return WM_RET_DECLINED;	/* continue to translate */
}

static int
wm_admin_request_handler(msg)
	register Message *	msg;
{
  if (ws_strcmp(msg->file.mime, MAGIC_TYPE_ADMIN_LOGIN) == 0) {
    char *	param[PARAM_AUTH_SIZE];
    int		n;

#define	PIDX_USER_KEY	(0)
#define	PIDX_USER_VALUE	(1)
#define	PIDX_PASS_KEY	(2)
#define	PIDX_PASS_VALUE	(3)
    n = get_values_from_query(msg->request.query_string, param,
			      PARAM_AUTH_SIZE);
    if (n == PARAM_AUTH_SIZE &&
	ws_strcmp(param[PIDX_USER_KEY], PARAM_USER_ID)  == 0 &&
	ws_strcmp(param[PIDX_PASS_KEY], PARAM_PASSWORD) == 0) {
      SHA_CTX		sha_ctx;
      unsigned char	password[SHA_DIGEST_LENGTH * 2 + 1];
      char *		sid;

      memset(password, 0, sizeof(password));
      SHA1_Init(&sha_ctx);
      SHA1_Update(&sha_ctx,
		  param[PIDX_PASS_VALUE], strlen(param[PIDX_PASS_VALUE]));
      SHA1_End(&sha_ctx, password);
      if (authenticate(msg, param[PIDX_USER_VALUE], password) == WM_AUTH_OK &&
	  (sid = generate_sid(msg, param[PIDX_USER_VALUE], password,
			      config.timeout)) != NULL) {
	char *	info;

	ws_asprintf(&info, "login: %s", param[PIDX_USER_VALUE]);
	record_history(msg->client.ip, sid, info);
	free(info);
	set_cookie(msg, sid, config.path, config.timeout);
	set_redirect(msg, config.page_startup);
      }
      else {
	char *	info;

	ws_asprintf(&info, "login failure: %s", param[PIDX_USER_VALUE]);
	record_history(msg->client.ip, "", info);
	free(info);
      }
    }
#if	DEBUG
    wv_syslog(WLL_LOG_DEBUG, "DEBUG[%s] login: %s=%s %s=%s",
	      WM_MODULE_ID, param[0], param[1], param[2], param[3]);
#endif	/* DEBUG */
    free(param[0]);
    free(param[1]);
    free(param[2]);
    free(param[3]);
    if (wa_get(msg->response.header, "Location") == NULL) {
      if (config.page_failure != NULL) {
	set_redirect(msg, config.page_failure);
      }
      else {
	set_redirect(msg, config.page_login);
      }
    }
    set_cache_control(msg);
    return WM_RET_OK;
  }
  else if (ws_strcmp(msg->file.mime, MAGIC_TYPE_ADMIN_LOGOUT) == 0) {
    register const char *	cookie;

    cookie = wa_get(msg->request.header, "cookie");
    if (cookie != NULL) {
      char *	s;
      
      s = ws_strdup(cookie);
      if (s != NULL) {
	char *	sid;

	sid = get_sid_from_cookie(s);
#if	DEBUG
	wv_syslog(WLL_LOG_DEBUG, "DEBUG[%s] logout: sid=%s", WM_MODULE_ID, sid);
#endif	/* DEBUG */
	if (sid != NULL && delete_sid(sid) == 0) {
	  set_cookie(msg, sid, config.path, 0);
   	}
	record_history(msg->client.ip, sid, INFO_LOGOUT);
      }
    }
    set_redirect(msg, config.page_login);
    set_cache_control(msg);
    return WM_RET_OK;
  }

  return WM_RET_DECLINED;	/* continue to handle request */
}

static int
get_values_from_query(query_string, retval, retsiz)
	char *	query_string;
	char **	retval;
	size_t	retsiz;
{
  char *	s = query_string;
  size_t	n = 0;

  while ((s = strtok(s, "&")) != NULL && n < retsiz) {
    char *	name;
    char *	value;

    name  = s;
    value = (char *)strchr(s, '=');
    if (value != NULL) {
      *value++ = EOL;
      if (n < retsiz) { retval[n++] = ws_strdup(name); }
      if (n < retsiz) { retval[n++] = ws_strdup(value); }
    }
    s = NULL;	/* DON'T REMOVE THIS LINE. */
  }
  if (n != retsiz) {
    register size_t	i;

    for (i = 0; i < n; i++) {
      free(retval[i]);
      retval[i] = NULL;
    }
  }

  return n;
}

#if	ENABLE_SQLITE3
static int
authenticate(msg, name, pass)
	register Message *	msg;
	const char *		name;
	const char *		pass;
{
  int		status = WM_AUTH_NG;
  char *	path;

  ws_asprintf(&path, "%s/%s", httpd->databaseDir, ADMIN_AUTH_DB);
  if (path != NULL) {
    sqlite3 *	sdb;
    int		rc;

    rc = sqlite3_open(path, &sdb);	free(path);
    if (rc == SQLITE_OK) {
      sqlite3_stmt *	stmt_admin;
      const char *	tail;

#define	SELECT_ADMIN \
	"SELECT password FROM administrators WHERE username = $NAME;"
      rc = sqlite3_prepare(sdb, SELECT_ADMIN, (int)strlen(SELECT_ADMIN),
			   &stmt_admin, &tail);
      if (rc == SQLITE_OK) {
	sqlite3_bind_text(stmt_admin, 1, name, (int)strlen(name),
			  SQLITE_TRANSIENT);
	rc = sqlite3_step(stmt_admin);
	if (rc == SQLITE_ROW && sqlite3_data_count(stmt_admin) > 0) {
	    const unsigned char *	value;

	    value = sqlite3_column_text(stmt_admin, 0);
#if	DEBUG
	    wv_syslog(WLL_LOG_DEBUG, "authenticate: %s=%s", pass, value);
#endif	/* DEBUG */
	    if (ws_strcmp(pass, value) == 0) {
	      status = WM_AUTH_OK;
	    }
	}
	sqlite3_finalize(stmt_admin);
      }
      sqlite3_close(sdb);
    }
  }

  return status;
}
#elif	ENABLE_SQLITE2
static int
authenticate(msg, name, pass)
	register Message *	msg;
	const char *		name;
	const char *		pass;
{
#define	QUERY_SELECT \
	"SELECT password FROM administrators WHERE username = '%s';"
  int		status = WM_AUTH_NG;
  char *	path;

  ws_asprintf(&path, "%s/%s", httpd->databaseDir, ADMIN_AUTH_DB);
  if (path != NULL) {
    sqlite *	sdb;

    sdb = sqlite_open(path, 0, NULL);	free(path);
    if (sdb != NULL) {
      char *	sql;

      ws_asprintf(&sql, QUERY_SELECT, name);
      if (sql != NULL) {
	sqlite_vm *	svm;
	int		rc;

	rc = sqlite_compile(sdb, sql, NULL, &svm, NULL);	free(sql);
	if (rc == SQLITE_OK) {
	  const char **	value;

	  rc = sqlite_step(svm, NULL, &value, NULL);
	  if (rc == SQLITE_ROW) {
#if	DEBUG
	    wv_syslog(WLL_LOG_DEBUG, "authenticate: %s=%s", pass, value[0]);
#endif	/* DEBUG */
	    if (ws_strcmp(pass, value[0]) == 0) {
	      status = WM_AUTH_OK;
	    }
	  }
	  sqlite_finalize(svm, NULL);
	}
      }
      sqlite_close(sdb);
    }
  }

  return status;
}
#else
#endif

static char *
generate_sid(msg, name, pass, expiry)
	register Message *	msg;
	const char *		name;
	const char *		pass;
	long			expiry;
{
  if (active < MAX_SESSION) {
    struct timeval	now;

    if (gettimeofday(&now, (struct timezone *)NULL) == 0) {
      SHA_CTX		sha_ctx;
      unsigned char	id[SHA_DIGEST_LENGTH * 2 + 1];
      unsigned int	h;

      memset(id, 0, sizeof(id));
      SHA1_Init(&sha_ctx);
      SHA1_Update(&sha_ctx, name, strlen(name));
      SHA1_Update(&sha_ctx, ":", 1);
      SHA1_Update(&sha_ctx, pass, strlen(pass));
      SHA1_Update(&sha_ctx, ":", 1);
      SHA1_Update(&sha_ctx, msg->client.ip, strlen(msg->client.ip));
      SHA1_Update(&sha_ctx, ":", 1);
      SHA1_Update(&sha_ctx, msg->client.port, strlen(msg->client.port));
      SHA1_End(&sha_ctx, id);

      h			 = hash(id);
      session[h].id	 = ws_strdup(id);
      session[h].expire  = now.tv_sec + expiry;
      session[h].timeout = expiry;
      active++;

      return (char *)session[h].id;
    }
  }

  return NULL;
}

/*****************************************************************************/

static int
set_cookie(msg, id, path, max_age)
	register Message *	msg;
	const char *		id;
	const char *		path;
	long			max_age;
{
  char *	cookie;

  ws_asprintf(&cookie, SESSION_ID "=%s; Path=%s; Max-Age=%ld; Version=1",
	      id, path, max_age);
  if (cookie != NULL) {
#if	DEBUG
    wv_syslog(WLL_LOG_DEBUG, "Set-Cookie: %s", cookie);
#endif	/* DEBUG */
    wa_put(msg->response.header, "Set-Cookie", cookie);
    return WM_RET_OK;
  }
  return WM_RET_ERROR;
}

static int
set_redirect(msg, page)
	register Message *	msg;
	const char *		page;
{
  char *	uri;

  if (msg->soci.port == DEF_HTTP_PORT) {
    ws_asprintf(&uri, "http://%s%s", msg->vhost->serverName, page);
  }
  else {
    ws_asprintf(&uri, "http://%s:%d%s",
		msg->vhost->serverName, msg->soci.port, page);
  }
  if (uri != NULL) {
#if	DEBUG
    wv_syslog(WLL_LOG_DEBUG, "Location: %s", uri);
#endif	/* DEBUG */
    wa_put(msg->response.header, "Location", uri);
    msg->response.status = STATUS_MOVED_PERMANENTLY;
    return WM_RET_OK;
  }

  return WM_RET_ERROR;
}

static int
set_cache_control(msg)
	register Message *	msg;
{
  wa_put(msg->response.header, "Cache-Control", ADM_CACHE_CONTROL);
  wa_put(msg->response.header, "Pragma", ADM_PRAGMA);

  return WM_RET_ERROR;
}

/*****************************************************************************/

/*
 * Function Name:	wm_admin_validate_session
 * Description:		validate a session
 * Arguments:		message
 * Return Value:	Successful: 0, otherwise: -1
 */
static int
wm_admin_validate_session(msg)
	register Message *	msg;
{
#if	DEBUG
  do {
    wv_syslog(WLL_LOG_DEBUG, "Request: %s:%d%s",
	      msg->vhost->serverName, msg->soci.port, msg->request.uri);
    fprintf(stderr, "DEBUG[%s] request: %s:%d%s\n",
	    WM_MODULE_ID, msg->vhost->serverName, msg->soci.port, msg->request.uri);
    fprintf(stderr, "DEBUG[%s] query_string: %s\n",
	    WM_MODULE_ID, msg->request.query_string);
    fprintf(stderr, "DEBUG[%s] cookie: %s\n", WM_MODULE_ID, wa_get(msg->request.header, "cookie"));
  } while (0);
#endif	/* DEBUG */
  if (ws_strcmp(msg->file.mime, MAGIC_TYPE_ADMIN_LOGIN)  != 0 &&
      ws_strcmp(msg->file.mime, MAGIC_TYPE_ADMIN_LOGOUT) != 0 &&
      ws_strcmp(msg->request.uri, config.page_login)   != 0 &&
      ws_strcmp(msg->request.uri, config.page_expired) != 0 &&
      ws_strcmp(msg->request.uri, config.page_invalid) != 0 &&
      ws_strcmp(msg->request.uri, config.page_failure) != 0 &&
      ws_strcmp(msg->request.uri, config.cmd_login)    != 0 &&
      ws_strcmp(msg->request.uri, config.cmd_logout)   != 0 &&
      ws_strncmp(msg->request.uri, config.path, config.plen) == 0) {
    register const char *	cookie;
    int				status = WM_RET_ERROR;

#if	0
    /* session id in query-string */
    if (msg->query_string != NULL) {
      char *	s;
      
      s = ws_strdup(msg->query_string);
      if (s != NULL) {
	char *	id;

	id = get_sid_from_query(s); free((char *)s);
	if (id != NULL) {
	  int	n;
	  long	expire;
	  long	timeout;

	  n = validate_sid(id, &expire, &timeout); free((char *)id);
	  if (n == WM_SID_VALID) {
	    return WM_RET_OK;
	  }
	}
      }
    }
#endif
    /* session id in cookie */
    cookie = wa_get(msg->request.header, "cookie");
    if (cookie != NULL) {
      char *	s;
      
      s = ws_strdup(cookie);
      if (s != NULL) {
	char *	sid;

	sid = get_sid_from_cookie(s);
	if (sid != NULL) {
	  int	n;
	  long	expire;
	  long	timeout;

	  free(s);
	  n = validate_sid(sid, &expire, &timeout);
#if	DEBUG
	  wv_syslog(WLL_LOG_DEBUG, "Cookie id: %s", sid);
	  debug_expires(expire);
#endif	/* DEBUG */
	  switch (n) {
	    case WM_SID_VALID:
	      status = set_cookie(msg, sid, config.path, timeout);
#if	DEBUG
	      wv_syslog(WLL_LOG_DEBUG, "Valid: %s", sid);
#endif	/* DEBUG */
	      break;
	    case WM_SID_EXPIRED:
	      if (config.page_expired != NULL) {
		set_redirect(msg, config.page_expired);
	      }
	      else {
		set_redirect(msg, config.page_login);
	      }
#if	DEBUG
	      wv_syslog(WLL_LOG_DEBUG, "Expired: %s", sid);
#endif	/* DEBUG */
	      record_history(msg->client.ip, sid, INFO_EXPIRED);
	      break;
	    case WM_SID_INVALID:
	      if (config.page_invalid != NULL) {
		set_redirect(msg, config.page_invalid);
	      }
	      else {
		set_redirect(msg, config.page_login);
	      }
#if	DEBUG
	      wv_syslog(WLL_LOG_DEBUG, "Invalid: %s", sid);
#endif	/* DEBUG */
	      record_history(msg->client.ip, sid, INFO_INVALID);
	      break;
	  }
	  free(sid);
	}
      }
      expire_sid();
    }
    else {
      set_redirect(msg, config.page_login);
#if	DEBUG
      wv_syslog(WLL_LOG_DEBUG, "No-Cookie: ");
#endif	/* DEBUG */
      record_history(msg->client.ip, "", INFO_NO_COOKIE);
    }
    set_cache_control(msg);

    return status;
  }

  return WM_RET_DECLINED;
}

#if	DEBUG
static void
debug_expires(expire)
	long	expire;
{
  static const char *	wdays[] = {"Sun","Mon","Tue","Wed","Thu","Fri","Sat"};
  static const char *	month[] = {"Jan","Feb","Mar","Apr","May","Jun",
				   "Jul","Aug","Sep","Oct","Nov","Dec"};
  struct tm *	tp;

  tp = localtime((const time_t *)&expire);
  if (tp) {
    wv_syslog(WLL_LOG_DEBUG,
	      "Expires: %s, %02d-%s-%4d %02d:%02d:%02d",
	      wdays[tp->tm_wday],
	      tp->tm_mday, month[tp->tm_mon], tp->tm_year + 1900,
	      tp->tm_hour, tp->tm_min, tp->tm_sec);
  }
}
#endif	/* DEBUG */

/*
 * Function Name:	validate_sid
 * Description:		validate a session id
 * Arguments:		session id
 * Return Value:	Successful: 0, otherwise: -1
 */
static int
validate_sid(id, expire, timeout)
	const char *	id;
	long *		expire;
	long *		timeout;
{
  struct timeval	now;

  if (gettimeofday(&now, (struct timezone *)NULL) != -1) {
    register int	c;
    unsigned int	h;

    for (c = 0, h = hash(id); c < MAX_SESSION; c++) {
      if (ws_strcmp(id, session[h].id) == 0) {
	*expire	 = session[h].expire;
	*timeout = session[h].timeout;
	if (now.tv_sec > *expire) {
	  return WM_SID_EXPIRED;
	}
	else {
	  *expire = session[h].expire = now.tv_sec + *timeout;
	  return WM_SID_VALID;
	}
      }
      h = rehash(h);
    }
#if	DEBUG
    wv_syslog(WLL_LOG_DEBUG, "No exist: id=%s", id);
#endif	/* DEBUG */
  }

  return WM_SID_INVALID;
}

static void
expire_sid(void)
{
  struct timeval	now;

  if (gettimeofday(&now, (struct timezone *)NULL) != -1) {
    register int	h;

    for (h = 0; h < MAX_SESSION; h++) {
      if (session[h].id != NULL && now.tv_sec > session[h].expire) {
	recycle_sid(h);
      }
    }
  }
}

static int
delete_sid(id)
	const char *	id;
{
  int	h;

  h = lookup(id);
  if (h != -1) {
    recycle_sid(h);
    return 0;
  }

  return -1;
}

static void
recycle_sid(h)
	int	h;
{
  free(session[h].id);
  session[h].id	     = NULL;
  session[h].expire  = 0L;
  session[h].timeout = 0L;
  active--;
}

static int
lookup(id)
	const char *	id;
{
  register int	c;
  unsigned int	h;

  for (c = 0, h = hash(id); c < MAX_SESSION; c++) {
    if (ws_strcmp(id, session[h].id) == 0) {
      return (int)h;
    }
    h = rehash(h);
  }

  return -1;
}

static unsigned int
rehash(h)
	unsigned int	h;
{
  return (h + 3) % MAX_SESSION;
}

static unsigned int
hash(s)
	register const char *	s;
{
  unsigned int	h;

  h = MAX_SESSION;
  if (s != NULL) {
    size_t	len;

    len = strlen(s);
    while (len > 0) {
      h += (h << 5);
      h ^= *s++;
      len--;
    }
  }

  return h % MAX_SESSION;
}

#if	0
/*
 * FUNCTION NAME:	get_sid_from_query
 * DESCRIPTION:		parse CGI query data
 * ARGUMENTS:		query string
 * RETURN VALUE:	Successful: id, otherwise: NULL
 */
static char *
get_sid_from_query(query_string)
	char *	query_string;
{
  char *	param[PARAM_SID_SIZE];
  int		n;

  n = get_values_from_query(query_string, param, PARAM_SID_SIZE);
  if (n == PARAM_SID_SIZE && ws_strcmp(param[0], SESSION_ID) == 0) {
    free((char *)param[0]);
    return param[1];
  }

  return NULL;
}
#endif

/*
 * FUNCTION NAME:	get_sid_from_cookie
 * DESCRIPTION:		parse Cookie string
 * ARGUMENTS:		cookie string
 * RETURN VALUE:	Successful: id, otherwise: NULL
 */
static char *
get_sid_from_cookie(http_cookie)
	char *	http_cookie;
{
  char *	string = http_cookie;

  while ((string = strtok(string, ";,")) != NULL) {
    char *	name;	/* parameter name */
    char *	value;	/* parameter value */

    /* skip white space */
    while (*string == SPC || *string == TAB) {
      string++;
    }

    if (parse_string(string, &name, &value) == 0) {
      if (ws_strcmp(name, SESSION_ID) == 0) {
	free(name);
	return value;
      }
    }
    string = NULL;	/* DON'T REMOVE! */
  }

  return NULL;
}

/*
 * FUNCTION NAME:	parse_string
 * DESCRIPTION:		parse NAME=VALUE string of cookie
 * ARGUMENTS:		string, name pointer, value pointer
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
parse_string(string, name, value)
	register char *	string;
	char **		name;
	char **		value;
{
  register char *	p;

#if	DEBUG
  wv_syslog(WLL_LOG_DEBUG, "Cookie: %s", string);
#endif	/* DEBUG */
  /* initialize */
  *name = *value = NULL;

  /* $Secure and $Discard is special parameter without a value */
  if (ws_strcasecmp(string, "$Secure")  == 0 ||
      ws_strcasecmp(string, "$Discard") == 0) {	/* see RFC2965 */
    *name = ws_strdup(string);
    return 0;
  }

  /* NAME */
  p = strchr(string, '=');
  if (p == NULL) {
    return -1;
  }
  *p++  = EOL;
  *name = ws_strdup(string);
  /* VALUE */
  if (*p == '"') {
    char *	q;

    q = ++p;
    p = strchr(q, '"');
    if (p) {
      *p = EOL;
    }
    *value = ws_strdup(q);
  }
  else {
    *value = ws_strdup(p);
  }

  if (*name == NULL || *value == NULL) {
    wv_free(*name);
    wv_free(*value);
    return -1;
  }

  return 0;
}


/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/
static void
wm_admin_register_hooks(void)
{
  wm_hook_uri_translater(wm_admin_uri_translater, HOOK_FIRST);
  wm_hook_request_handler(wm_admin_request_handler, HOOK_FIRST, MAGIC_TYPE_ADMIN);
  wm_hook_fixups(wm_admin_validate_session, HOOK_FIXUP_REQUEST);
}

static wmModule	mod_admin = {
  WM_MAGIC_NUMBER_MAJOR,	/* version */
  WM_UNSECURE,			/* unsecure */
  WM_MODULE_ID,			/* name */
  "Administrator Module version 1.1.4",	/* info */
  "Kouichi ABE (WALL)",		/* author */
  NULL,				/* depend */
  NULL,				/* handle */
  (void *)wm_admin_init,		/* init */
  (void *)wm_admin_finish,	/* finish */
  (void *)wm_admin_register_hooks	/* register_handler */
};

wmModule *
wmModuleInfo(void)
{
  return &mod_admin;
}
