/*****************************************************************************
 *
 * FILE:	mod_auth.c
 * DESCRIPTION:	Wyvern: authentication module
 * DATE:	Sun, Sep 17 2000
 * UPDATED:	Sat, Aug 25 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2000-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2000-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_auth.c,v 1.9 2007/08/24 18:01:55 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <errno.h>
#include "wyvern.h"
#include "auth.h"

/*****************************************************************************
 *
 *	definition of macros and structure
 *
 *****************************************************************************/
#define	WM_MODULE_ID	"mod_auth"

#define	DB_SUFFIX	".xml"
#define	USERNAME	(0)
#define	PASSWORD	(1)

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	wm_auth_init(void);
static void	wm_auth_finish(void);
static int	wm_auth_checker(Message *);
static char *	get_password(const char *, const char *);
static void	wm_auth_register_hooks(void);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static Cache *	authList = NULL;

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wm_auth_init
 * Description:		initialize cache for authentication table
 * Arguments:		none
 * Return Value:	none
 */
static void
wm_auth_init(void)
{
  AuthProc	ap = { get_password, NULL, NULL };

  authList = wc_new_cache(WA_AUTH_TABLE_SIZE);
  if (authList == NULL) {
    wv_die("[%s] failed to initialize cache", WM_MODULE_ID);
  }
  wauth_set_subr(&ap);
}

/*
 * Function Name:	wm_auth_finish
 * Description:		release cache memory area
 * Arguments:		none
 * Return Value:	none
 */
static void
wm_auth_finish(void)
{
  wauth_cleaner(authList);
}

/*****************************************************************************/

/*
 * Function Name:	wm_auth_checker
 * Description:		check user access
 * Arguments:		message
 * Return Value:	allow: 0, otherwise: -1
 */
static int
wm_auth_checker(msg)
	register Message *	msg;
{
  DBG(fputs("<auth_checker>\n", stderr);)
  return wauth_checker(authList, msg, DB_SUFFIX);
}

static char *
get_password(path, name)
	const char *	path;	/* path of password file */
	const char *	name;	/* username */
{
  register FILE *	fin;
  char *		status = NULL;

  if (path == NULL || name == NULL) {
    errno = EINVAL;
    return NULL;
  }

  fin = ws_safe_fopen(path, "r");
  if (fin != NULL) {
    sxml_node_t *	root;

    root = sxml_parse_file(fileno(fin));
    if (root != NULL) {
      sxml_node_t *	node;

      node = sxml_find_element(root, "authentication-info", NULL, NULL);
      if (node != NULL) {
	register sxml_node_t *	np;

	for (np = node->child; np != NULL; np = np->next) {
	  if (np->type != SXML_ELEMENT)		{ continue; }
	  if (np->value.element.name == NULL)	{ continue; }

	  if (ws_strcmp(np->value.element.name, "credential") == 0) {
	    register sxml_node_t *	nnp;
	    const char *		args[2] = { NULL, NULL };

	    for (nnp = np->child; nnp != NULL; nnp = nnp->next) {
	      const char *	content;

	      if (nnp->type != SXML_ELEMENT)	   { continue; }
	      if (nnp->value.element.name == NULL) { continue; }

	      content = sxml_get_content(nnp->child);
	      if (ws_strcmp(nnp->value.element.name, "username") == 0) {
		args[USERNAME] = content;
	      }
	      else if (ws_strcmp(nnp->value.element.name, "password") == 0) {
		args[PASSWORD] = content;
	      }
	    }
	    if (ws_strcmp(name, args[USERNAME]) == 0) {
	      status = ws_strdup(args[PASSWORD]);
	    }
#if	DEBUG
	    fprintf(stderr, "DEBUG[%s] %s: %s => %s\n",
		    WM_MODULE_ID, name, args[USERNAME], args[PASSWORD]);
#endif	/* DEBUG */
	  }	/* </credential> */
	}
      }	/* </authentication-info> */
      sxml_delete_node(root);
    }
    fclose(fin);
  }

#if	DEBUG
  if (status != NULL) {
    wv_syslog(WLL_LOG_DEBUG, "[%s] found: %s=%s", WM_MODULE_ID, name, status);
  }
  else {
    wv_syslog(WLL_LOG_DEBUG, "[%s] not found: %s", WM_MODULE_ID, name);
  }
#endif	/* DEBUG */

  return status;
}

/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/
static void
wm_auth_register_hooks(void)
{
  wm_hook_auth_checker(wm_auth_checker, HOOK_MIDDLE);
}

static wmModule	mod_auth = {
  WM_MAGIC_NUMBER_MAJOR,		/* version */
  WM_SECURE,				/* secure */
  WM_MODULE_ID,				/* name */
  "Authentication Module version 1.0.4",	/* info */
  "Kouichi ABE (WALL)",			/* author */
  "libauth",				/* depend */
  NULL,					/* handle */
  (void *)wm_auth_init,			/* init */
  (void *)wm_auth_finish,		/* finish */
  (void *)wm_auth_register_hooks	/* register_handler */
};

wmModule *
wmModuleInfo(void)
{
  return &mod_auth;
}
