/*****************************************************************************
 *
 * FILE:	mod_auth_pgsql.c
 * DESCRIPTION:	Wyvern: authentication module with PostgreSQL
 * DATE:	Tue, Dec  6 2005
 * UPDATED:	Sat, Aug 25 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2005-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2005-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_auth_pgsql.c,v 1.7 2007/08/24 18:01:55 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#include <errno.h>
#include <libpq-fe.h>
#include "wyvern.h"
#include "auth.h"

/*****************************************************************************
 *
 *	definition of macros and structure
 *
 *****************************************************************************/
#define	WM_MODULE_ID	"mod_auth_pgsql"

#define	DB_SUFFIX	".pgsql"
#define	PGSQL_DB_NAME	"wyvern"
#define	PGSQL_USERNAME	"wyvern"
#define	PGSQL_PASSWORD	"admin@wyvern"
/*
 * CREATE TABLE auth (
 *   username	text UNIQUE PRIMARY KEY,
 *   password	text
 * );
 */
#define	SELECT_PASSWORD	"SELECT password FROM auth WHERE username='%s';"

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	wm_auth_pgsql_init(void);
static void	wm_auth_pgsql_finish(void);
static int	wm_auth_pgsql_checker(Message *);
static char *	get_password(const char *, const char *);
static void	wm_auth_pgsql_register_hooks(void);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static Cache *	authList = NULL;

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wm_auth_pgsql_init
 * Description:		initialize cache for authentication table
 * Arguments:		none
 * Return Value:	none
 */
static void
wm_auth_pgsql_init(void)
{
  AuthProc	ap = { get_password, NULL, NULL };

  authList = wc_new_cache(WA_AUTH_TABLE_SIZE);
  if (authList == NULL) {
    wv_die("[%s] failed to initialize cache", WM_MODULE_ID);
  }
  wauth_set_subr(&ap);
}

/*
 * Function Name:	wm_auth_pgsql_finish
 * Description:		release cache memory area
 * Arguments:		none
 * Return Value:	none
 */
static void
wm_auth_pgsql_finish(void)
{
  wauth_cleaner(authList);
}

/*****************************************************************************/

/*
 * Function Name:	wm_auth_pgsql_checker
 * Description:		check user access
 * Arguments:		message
 * Return Value:	allow: 0, otherwise: -1
 */
static int
wm_auth_pgsql_checker(msg)
	register Message *	msg;
{
  DBG(fputs("<auth_pgsql_checker>\n", stderr);)
  return wauth_checker(authList, msg, DB_SUFFIX);
}

static char *
get_password(path, name)
	const char *	path;	/* dummy */
	const char *	name;	/* username */
{
  PGconn *	conn	= NULL;
  const char *	pghost	= NULL;
  const char *	pgport	= NULL;
  const char *	pgopts	= NULL;
  const char *	pgtty	= NULL;
  const char *	dbName	= PGSQL_DB_NAME;
  const char *	user	= PGSQL_USERNAME;
  const char *	passwd	= PGSQL_PASSWORD;
  char *	status	= NULL;

  if (name == NULL) {
    errno = EINVAL;
    return NULL;
  }

  /* make a connection to the database */
  conn = PQsetdbLogin(pghost, pgport, pgopts, pgtty, dbName, user, passwd);

  /* check to see that the backend connection was successfully made */
  if (PQstatus(conn) == CONNECTION_BAD) {
    wv_syslog(WLL_LOG_ERR,
	      "[%s] failed to connection to database '%s': %s",
	      WM_MODULE_ID, dbName, PQerrorMessage(conn));
  }
  else {
    char *	sql;
    PGresult *	res;

    ws_asprintf(&sql, SELECT_PASSWORD, name);
    res = PQexec(conn, sql);
    if (res == NULL || PQresultStatus(res) != PGRES_TUPLES_OK) {
      wv_syslog(WLL_LOG_ERR, "[%s] failed to send query '%s': %s",
    		WM_MODULE_ID, sql, PQerrorMessage(conn));
    }
    else {
      status = ws_strdup(PQgetvalue(res, 0, 0));
      PQclear(res);
    }
    free(sql);
  }

  if (conn != NULL) {
    PQfinish(conn);	/* close the connection to the database and cleanup */
  }

#if	DEBUG
  if (status != NULL) {
    wv_syslog(WLL_LOG_DEBUG, "[%s] found: %s=%s", WM_MODULE_ID, name, status);
  }
  else {
    wv_syslog(WLL_LOG_DEBUG, "[%s] not found: %s", WM_MODULE_ID, name);
  }
#endif	/* DEBUG */

  return status;
}

/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/
static void
wm_auth_pgsql_register_hooks(void)
{
  wm_hook_auth_checker(wm_auth_pgsql_checker, HOOK_MIDDLE);
}

static wmModule	mod_auth_pgsql = {
  WM_MAGIC_NUMBER_MAJOR,		/* version */
  WM_SECURE,				/* secure */
  WM_MODULE_ID,				/* name */
  "Authentication Module with PostgreSQL version 1.0.3",	/* info */
  "Kouichi ABE (WALL)",			/* author */
  "libauth",				/* depend */
  NULL,					/* handle */
  (void *)wm_auth_pgsql_init,		/* init */
  (void *)wm_auth_pgsql_finish,		/* finish */
  (void *)wm_auth_pgsql_register_hooks	/* register_handler */
};

wmModule *
wmModuleInfo(void)
{
  return &mod_auth_pgsql;
}
