/*****************************************************************************
 *
 * FILE:	mod_autoindex.c
 * DESCRIPTION:	Wyvern: directory indexing module
 * DATE:	Mon, Aug 26 2002
 * UPDATED:	Sat, Aug 25 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2002-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2002-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_autoindex.c,v 1.27 2007/08/24 18:01:55 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_DIRENT_H
#include <dirent.h>
#else
#define	dirent	direct
#if	HAVE_SYS_NDIR_H
#include <sys/ndir.h>
#endif	/* HAVE_SYS_NDIR_H */
#if	HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif	/* HAVE_SYS_DIR_H */
#if	HAVE_NDIR_H
#include <ndir.h>
#endif	/* HAVE_NDIR_H */
#endif	/* HAVE_DIRENT_H */
#ifdef HAVE_D_NAMLEN
#define NAMLEN(d)	(d->d_namlen)
#else	/* !HAVE_D_NAMLEN */
#define NAMLEN(d)	(strlen(d->d_name))
#endif	/* !HAVE_D_NAMLEN */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include "wyvern.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	WM_MODULE_ID	"mod_autoindex"

#ifndef	dirfd
#define	dirfd(dirp)	((dirp)->dd_fd)
#endif	/* dirfd */

#define	DIR_TABLE_SIZE	(11)	/* directory indexing cache */
#define	MAX_DIR_ENTRY	(1024)

#define	U_KILO		(1024L)
#define	U_MEGA		(1024L*1024L)
#define	U_GIGA		(1024L*1024L*1024L)

#define	HEADER_FILE	".HEADER"
#define	FOOTER_FILE	".FOOTER"

typedef struct _DirIndex {	/* directory cache structure */
  char *	name;	/* path of request directory */
  char *	path;	/* path of cache file */
} DirIndex;

#define TOTAL_KEYWORDS	(47)
#define MIN_WORD_LENGTH (2)
#define MAX_WORD_LENGTH (6)
#define MIN_HASH_VALUE	(0)
#define MAX_HASH_VALUE	(31)
/* maximum key range = 32, duplicates = 25 */

typedef enum {		/* file form */
  FF_COMPRESSED,	/* compressed */
  FF_TAR,		/* tar archived */
  FF_ARCHIVE,		/* archived */
  FF_BINARY,		/* binary */
  FF_HTML,		/* HTML */
  FF_TEXT,		/* text */
  FF_PDF,		/* PDF */
  FF_DVI,		/* DVI */
  FF_POSTSCRIPT,	/* PostScript */
  FF_C,			/* C language */
  FF_F77,		/* Fortran language */
  FF_PERL,		/* Perl language */
  FF_TEX,		/* TeX/LaTeX */
  FF_PATCH,		/* patch file */
  FF_SHELL,		/* shell script */
  FF_IMAGE,		/* image */
  FF_MOVIE,		/* movie */
  FF_SOUND		/* sound/audio */
} fform_t;

typedef struct fileForm {
  char *	name;
  fform_t	form;
} FileForm;

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void		wm_autoindex_init(void);
static void		wm_autoindex_finish(void);
static void		trashcan(const void *);
static int		wm_directory_indexor(Message *);
static DirIndex *	didx_insert(const char *, const char *);
static DirIndex *	didx_lookup(char *);
static int		compare(const void *, const void *);
static int		didx_update(Message *, DirIndex *);
static int		didx_open(Message *, DirIndex *);
static void		make_content(Message *, FILE *, char **, int);
static void		include_file(const char *, const char *, FILE *);
static void		print_file_form(const char *, mode_t, FILE *);
#ifdef __GNUC__
__inline
#endif
static unsigned int	hash(const char *, unsigned int);
#ifdef __GNUC__
__inline
#endif
static const FileForm *	get_fform(const char *, unsigned int);
#ifdef __GNUC__
__inline
#endif
static void		quick_sort(char * *, int, int);
#ifdef __GNUC__
__inline
#endif
static void		swap(char * *, int, int);
static void		wm_autoindex_register_hooks(void);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static Cache *	dirList = NULL;

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

static void
wm_autoindex_init(void)
{
  dirList = wc_new_cache(DIR_TABLE_SIZE);
  if (dirList == NULL) {
    wv_die("[%s] failed to initialize cache", WM_MODULE_ID);
  }
}

static void
wm_autoindex_finish(void)
{
  if (dirList != NULL) {
    wc_flush_cache(dirList, trashcan);
    free((Cache *)dirList);
  }
}

static void
trashcan(x)
	const void *	x;
{
  DirIndex *	dp = (DirIndex *)x;
  wv_free(dp->name);
  if (dp->path != NULL) {
    unlink(dp->path);
    free((char *)dp->path);
  }
  free((DirIndex *)dp);
}

/*
 * Function Name:	wm_directory_indexor
 * Description:		file listing in the directory
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
static int
wm_directory_indexor(msg)
	register Message *	msg;
{
  register DirIndex *	dp;

  DBG(fputs("<wm_directory_indexor>\n", stderr);)
  if (msg->vhost->autoIndexing == false) {
    msg->response.status = STATUS_FORBIDDEN;
    return WM_RET_ERROR;
  }

  /* obtain directory information */
  stat(msg->file.path, &msg->file.status);

  /* search cache */
  dp = didx_lookup(msg->file.path);
  if (dp == NULL) {
    DirIndex	didx;

    didx.name = msg->file.path;
    ws_asprintf((char **)&didx.path,  "%s/dc%lx%lx%lx",
  		httpd->cacheDir,
  		(long)msg->file.status.st_ino,
  		(long)msg->file.status.st_size,
  		msg->file.status.st_mtime);
    dp = didx_insert(didx.name, didx.path);
    free((char *)didx.path);
  }
  if (dp == NULL) {
    wv_syslog(WLL_LOG_ERR, "[%s] failed to get cache entry", WM_MODULE_ID);
    msg->response.status = STATUS_INTERNAL_SERVER_ERROR;
    return WM_RET_ERROR;
  }

  return didx_update(msg, dp);
}

/*****************************************************************************/

static DirIndex *
didx_insert(name, path)
	const char *	name;
	const char *	path;
{
  Cache *	new;
  
  new = wc_remove_cache(dirList->next);
  if (new != NULL) {
    DirIndex *	didx = (DirIndex *)ws_malloc(sizeof(DirIndex));

    didx->name = ws_strdup(name);
    didx->path = ws_strdup(path);
    if (new->data != NULL) {
#if	DEBUG
      fprintf(stderr, "DEBUG[free] %s: %s\n",
	      ((DirIndex *)new->data)->name,
	      ((DirIndex *)new->data)->path);
#endif	/* DEBUG */
      unlink(((DirIndex *)new->data)->path);
      free((char *)((DirIndex *)new->data)->name);
      free((char *)((DirIndex *)new->data)->path);
      free((DirIndex *)new->data);
    }
    new->data = didx;
    wc_insert_cache(dirList, new);
    DBG(fprintf(stderr, "DEBUG[insert] %s: %s\n", name, path);)

    return didx;
  }

  return NULL;
}

static DirIndex *
didx_lookup(name)
	char *	name;
{
  DirIndex	didx;
  DirIndex *	p;

  didx.name = name;
  p = (DirIndex *)wc_search_cache(dirList, &didx, compare);
#if	DEBUG
  if (p != NULL) {
    fprintf(stderr, "DEBUG[search] %s: %s\n", p->name, p->path);
  }
  else {
    fprintf(stderr, "DEBUG[search] %s not found\n", name);
  }
#endif	/* DEBUG */

  return p;
}

static int
compare(x, y)
	register const void *	x;
	register const void *	y;
{
  if (ws_strcmp(((DirIndex *)x)->name, ((DirIndex *)y)->name) == 0) {
    return 0;
  }
  return -1;
}

/*
 * Function Name:	didx_update
 * Description:		update cache file for directory indexing
 * Arguments:		message, cache
 * Return Value:	successful: 0, otherwise: -1
 */
static int
didx_update(msg, dip)
	register Message *	msg;
	register DirIndex *	dip;
{
  register DIR *	dfd;
  struct dirent *	dp;
  struct stat		sbuf;
  size_t		items;	/* total entries in directory */
  char **		dir;
  int			dnum = 0;
  register FILE *	fout;
  int			i;
  int			status = -1;

  /* open directory */
  dfd = opendir(msg->file.path);
  if (dfd == NULL) {
    msg->response.status = STATUS_FORBIDDEN;
    return WM_RET_ERROR;
  }

  /* set current response status */
  msg->response.status = STATUS_INTERNAL_SERVER_ERROR;

  /* estimate total entries in directory */
  if (fstat(dirfd(dfd), &sbuf) == -1) {
    wv_syslog(WLL_LOG_ERR, "[%s] failed to get file status", WM_MODULE_ID);
    goto error0;
  }
  items = (sbuf.st_size / 24);
  dir	= (char **)ws_calloc(items, sizeof(char *));

  /* directory walking */
  while ((dp = readdir(dfd)) != NULL) {
    register char *	p;

    p = strrchr(dp->d_name, '.');
    /* skip dot files */
    if (dp->d_name[0] == '.' || dnum > MAX_DIR_ENTRY - 1 ||
    /* skip CVS/RCS directory */
	ws_strcmp(dp->d_name, "CVS") == 0 ||
	ws_strcmp(dp->d_name, "RCS") == 0 ||
    /* skip *,? and *~ files (for CVS and emacs series) */
	dp->d_name[NAMLEN(dp) - 2] == ',' ||
	dp->d_name[NAMLEN(dp) - 1] == '~' ||
    /* skip #* and ,* files (for MH) */
	dp->d_name[0] == '#' || dp->d_name[0] == ',' ||
    /* skip *.{bak,org} files (for vim series) */
	ws_strcasecmp(p, ".bak") == 0 ||
	ws_strcasecmp(p, ".org") == 0 ||
    /* skip *.{orig,rej} files (for patch series) */
	ws_strcasecmp(p, ".orig") == 0 ||
	ws_strcasecmp(p, ".rej")  == 0) {
      continue;
    }
    dir[dnum] = ws_strdup(dp->d_name);
    dnum++;
  }

  /* sort directory indexing */
  quick_sort(dir, 0, dnum - 1);

  /* update cache file */
  fout = ws_safe_fopen(dip->path, "w");
  if (fout == NULL) {
    wv_syslog(WLL_LOG_ERR, "[%s] failed to open file: %s",
	      WM_MODULE_ID, dip->path);
    goto error;
  }
  make_content(msg, fout, dir, dnum);
  fclose(fout);

  /* bind new file descriptor to current message */
  status = didx_open(msg, dip);

error:
  for (i = 0; i < dnum; i++) {
    wv_free(dir[i]);
  }
  free((char **)dir);
error0:
  closedir(dfd);

  return status;
}

/*
 * Function Name:	didx_open
 * Description:		open cache file
 * Arguments:		message, cache
 * Return Value:	successful: 0, otherwise: -1
 */
static int
didx_open(msg, dip)
	register Message *	msg;
	register DirIndex *	dip;
{
  int		fd;
  struct stat	sbuf;

  /* open directory listed file */
  fd = ws_safe_open(dip->path, O_RDONLY, &sbuf, false);
  if (fd == -1) {
    wv_syslog(WLL_LOG_ERR, "[%s] failed to open file: %s",
	      WM_MODULE_ID, dip->path);
    msg->response.status = STATUS_INTERNAL_SERVER_ERROR;
    return WM_RET_ERROR;
  }

  /* close previous opened file */
  if (msg->file.nfd != NULL) {
    wt_netfd_close(msg->file.nfd);
  }
  msg->file.nfd			= wt_netfd_fdopen(fd);
  msg->file.status.st_size	= sbuf.st_size;
  msg->response.status		= STATUS_REQUEST_OK;
  msg->file.mime		= MAGIC_TYPE_HTML;

  return WM_RET_OK;
}

/*****************************************************************************/

/*
 * Function Name:	make_content
 * Description:		make a directory listing HTML file
 * Arguments:		message, output file pointer, directory name,
 *			# of entry in directory
 * Return Value:	none
 */
static void
make_content(msg, fout, dir, dnum)
	register Message *	msg;
	register FILE *		fout;
	register char **	dir;
	int			dnum;
{
  int	i;

  /* make HTML text */
  fputs(
    "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\">\n"
    "<html lang=\"ja\">\n"
    "<head>\n"
    "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=EUC-JP\">\n"
    "<meta http-equiv=\"Content-Style-Type\" content=\"text/css\">\n"
    "<style type=\"text/css\">\n"
    "<!--\n"
    ".size_b { text-align: right; padding-right: 20px; color: orange }\n"
    ".size_k { text-align: right; padding-right: 20px; color: green }\n"
    ".size_m { text-align: right; padding-right: 20px; color: red }\n"
    ".size_g { text-align: right; padding-right: 20px; color: purple }\n"
    ".text { color: brown }\n"
    "span { color: #ff0033 }\n"
    "table { width: 100%; background-color: #bec3d2 }\n"
    "tr.none { background-color: transparent }\n"
    "td.timestamp { text-align: center }\n"
    "td.size { text-align: right }\n"
    "tr { background-color: #f0f1f5 }\n"
    "td { color: #000033 }\n"
    "-->\n"
    "</style>\n",
    fout);

  fprintf(fout,
	"<title>Directory list of %s</title>\n"
	"</head>\n"
	"<body lang=\"ja\">\n"
	"<h1>Directory list of <span>%s</span></h1>\n",
	msg->request.uri, msg->request.uri);

  /* include HeaderFile if it exists */
  include_file(msg->file.path, HEADER_FILE, fout);

  fprintf(fout,
	"<div>\n"
	"<table border=\"0\" cellspacing=\"0\" summary=\"%s\">\n"
	"<thead>"
	"<tr class=\"none\">"
	"<th colspan=\"2\">Name</th>"
	"<th>Last modified</th>"
	"<th>Size</th></tr>\n"
	"</thead>\n<tbody>\n",
	msg->request.uri);

  for (i = 0; i < dnum; i++) {
    char *	path;
    struct stat	sbuf;
    int		n;
    char	tbuf[32];

    ws_asprintf(&path, "%s/%s", msg->file.path, dir[i]);
    /* failed to get file status */
    n = stat(path, &sbuf);	free((char *)path);
    if (n == -1) {
      fprintf(fout,
	"<tr>\n<td>&nbsp;</td>"
	"<td><a href=\"%s\">%s</a></td>\n"
	"<td colspan=\"2\"></td>\n</tr>\n",
	dir[i], dir[i]);
      continue;
    }

    /* print the icon of the file type */
    fputs("<tr>\n<td>", fout);
    print_file_form(dir[i], (mode_t)sbuf.st_mode, fout);
    fputs("</td>\n", fout);

    /* print file name (directory link is a special) */
    if (S_ISDIR(sbuf.st_mode)) {
      fprintf(fout, "<td><a href=\"%s/\">%s</a>/</td>\n", dir[i], dir[i]);
    }
    else {
      fprintf(fout, "<td><a href=\"%s\">%s</a></td>\n", dir[i], dir[i]);
    }

    /* print last modified */
    memset(tbuf, 0, sizeof(tbuf));
    strftime(tbuf, sizeof(tbuf),
	     "%d-%b-%Y&nbsp;%H:%M", localtime(&sbuf.st_mtime));
    fprintf(fout, "<td class=\"timestamp\">&nbsp;%s&nbsp;</td>\n", tbuf);

    /* print file size and file description */
    if (S_ISDIR(sbuf.st_mode)) {
      fputs("<td class=\"size\">&nbsp;-&nbsp;</td>\n</tr>\n", fout);
    }
    else {
      fputs("<td class=\"", fout);
      if (sbuf.st_size < 1024) {
	fprintf(fout, "size_b\">%ldB", (long)sbuf.st_size);
      }
      else if (sbuf.st_size < 1024 * 1024) {
	fprintf(fout, "size_k\">%ld.%ldK",
		(long)sbuf.st_size / U_KILO,
		((long)sbuf.st_size % U_KILO) * 10 / U_KILO);
      }
      else if (sbuf.st_size < 1024 * 1024 * 1024) {
	fprintf(fout, "size_m\">%ld.%ldM",
		(long)sbuf.st_size / U_MEGA,
		((long)sbuf.st_size % U_MEGA) * 10 / U_MEGA);
      }
      else {
	fprintf(fout, "size_g\">%ld.%ldG",
		(long)sbuf.st_size / U_GIGA,
		((long)sbuf.st_size % U_GIGA) * 10 / U_GIGA);
      }
      fputs("</td>\n</tr>\n", fout);
    }
  }
  fputs("</tbody>\n"
	"</table>\n"
	"</div>\n",
	fout);

  /* include FooterFile if it exists */
  include_file(msg->file.path, FOOTER_FILE, fout);

  fprintf(fout,
	"<hr>\n"
	"<address>\n"
	"<a href=\"%s\">%s</a> Server at %s Port %d\n"
	"</address>\n"
	"</body>\n"
	"</html>\n",
	SERVER_ADDRESS, SERVER_VERSION,
	msg->vhost->serverName, msg->soci.port);
}

/*
 * Function Name:	include_file
 * Description:		include HeaderFile/FooterFile file
 * Arguments:		file path, file name, output file pointer
 * Return Value:	none
 */
static void
include_file(directory, filename, fout)
	const char *	directory;
	const char *	filename;
	register FILE *	fout;
{
  register FILE *	fin;
  char *		path;

  ws_asprintf(&path, "%s%s", directory, filename);
  fin = ws_safe_fopen(path, "r");	free((char *)path);
  if (fin != NULL) {
    char	buf[BUFSIZ];

    memset(buf, 0, sizeof(buf));
    while (fgets(buf, sizeof(buf), fin) != NULL) {
      fputs(buf, fout);
      memset(buf, 0, sizeof(buf));
    }
    fputs("\n<hr>\n", fout);
    fclose(fin);
  }
}

/*****************************************************************************/

/*
 * Function Name:	print_file_form
 * Description:		print the file form
 * Arguments:		file name, file mode, file pointer
 * Return Value:	none
 */
static void
print_file_form(fname, fmode, fout)
	const char *	fname;
	mode_t		fmode;
	register FILE *	fout;
{
  fputs("<img src=\"/icons/", fout);

  if (S_ISDIR(fmode)) {
    fputs("dir.gif\" alt=\"[DIR]\" ", fout);
  }
  else if (S_ISREG(fmode)) {
#if	ENABLE_SIMPLE_INDEXING
    fputs("generic.gif\" alt=\"[   ]\" ", fout);
#else
    char *	p = NULL;

    p = strrchr(fname, '.');
    if (p == NULL) {
      fputs("unknown.gif\" alt=\"[   ]\" ", fout);
    }
    else {
      const FileForm *	ff = get_fform(p, strlen(p));

      if (ff == NULL) {
	fputs("generic.gif\" alt=\"[   ]\" ", fout);
      }
      else {
	switch (ff->form) {
	  case FF_COMPRESSED:	/* compressed */
	    fputs("compressed.gif\" alt=\"[CMP]\" ", fout);
	    break;
	  case FF_TAR:		/* tar archived */
	    fputs("tar.gif\" alt=\"[TAR]\" ", fout);
	    break;
	  case FF_ARCHIVE:	/* archived */
	    fputs("compressed.gif\" alt=\"[ARC]\" ", fout);
	    break;
	  case FF_BINARY:		/* binary */
	    fputs("binary.gif\" alt=\"[BIN]\" ", fout);
	    break;
	  case FF_HTML:		/* HTML */
	    fputs("layout.gif\" alt=\"[HTM]\" ", fout);
	    break;
	  case FF_TEXT:		/* text */
	    fputs("text.gif\" alt=\"[TXT]\" ", fout);
	    break;
	  case FF_PDF:		/* PDF */
	    fputs("pdf.gif\" alt=\"[PDF]\" ", fout);
	    break;
	  case FF_DVI:		/* DVI */
	    fputs("dvi.gif\" alt=\"[DVI]\" ", fout);
	    break;
	  case FF_POSTSCRIPT:	/* PostScript */
	    fputs("ps.gif\" alt=\"[PS ]\" ", fout);
	    break;
	  case FF_C:		/* C language */
	    fputs("c.gif\" alt=\"[C  ]\" ", fout);
	    break;
	  case FF_F77:		/* Fortran language */
	    fputs("f.gif\" alt=\"[F77]\" ", fout);
	    break;
	  case FF_PERL:		/* Perl language */
	    fputs("p.gif\" alt=\"[P  ]\" ", fout);
	    break;
	  case FF_TEX:		/* TeX/LaTeX */
	    fputs("tex.gif\" alt=\"[TEX]\" ", fout);
	    break;
	  case FF_PATCH:		/* patch file */
	    fputs("patch.gif\" alt=\"[PAT]\" ", fout);
	    break;
	  case FF_SHELL:		/* shell script */
	    fputs("script.gif\" alt=\"[SH ]\" ", fout);
	    break;
	  case FF_IMAGE:		/* image */
	    fputs("image1.gif\" alt=\"[IMG]\" ", fout);
	    break;
	  case FF_MOVIE:		/* movie */
	    fputs("movie.gif\" alt=\"[MOV]\" ", fout);
	    break;
	  case FF_SOUND:		/* sound/audio */
	    fputs("sound2.gif\" alt=\"[SND]\" ", fout);
	    break;
	  default:
	    break;
	}
      }
    }
#endif	/* ENABLE_SIMPLE_INDEXING */
  }
  else if (S_ISLNK(fmode)) {
    fputs("link.gif\" alt=\"[LNK]\" ", fout);
  }
  else {
    fputs("unknown.gif\" alt=\"[UNK]\" ", fout);
  }

  fputs("width=\"20\" height=\"22\">", fout);
}

/*
 * following is automatically generated by gperf
 */
#ifdef __GNUC__
__inline
#else
#ifdef __cplusplus
inline
#endif
#endif
static unsigned int
hash(str, len)
	register const char *	str;
	register unsigned int	len;
{
  static const unsigned char asso_values[] =
    {
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32,  0, 32, 32, 32,
       6, 11, 32, 32, 32,  9, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
       1, 32, 32, 32, 32, 32, 32, 32, 32,  4,
      30, 18,  5, 10,  0, 25, 32, 32, 15,  8,
      23, 32,  3, 32, 28, 31, 26, 32, 21, 32,
      13, 16, 20, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
      32, 32, 32, 32, 32, 32
    };
  return asso_values[(unsigned char)str[len - 1]] + asso_values[(unsigned char)str[0]];
}

#ifdef __GNUC__
__inline
#endif
static const FileForm *
get_fform(str, len)
	register const char *	str;
	register unsigned int	len;
{
  static const unsigned char lengthtable[] =
    {
       4,  4,  3,  6,  3,  2,  2,  4,  4,  2,  4,  5,  4,  4,
       5,  2,  4,  4,  3,  4,  4,  5,  4,  5,  4,  4,  4,  4,
       5,  5,  3,  6,  4,  4,  3,  4,  3,  4,  4,  4,  5,  4,
       4,  4,  4,  4,  3
    };
  static const FileForm wordlist[] =
    {
      {".lzh",	FF_ARCHIVE},
      {".csh",	FF_SHELL},
      {".sh",	FF_SHELL},
      {".patch",	FF_PATCH},
      {".ph",	FF_PERL},
      {".h",	FF_C},
      {".Z",	FF_COMPRESSED},
      {".zip",	FF_ARCHIVE},
      {".bmp",	FF_IMAGE},
      {".c",	FF_C},
      {".pdf",	FF_PDF},
      {".tiff",	FF_IMAGE},
      {".tif",	FF_IMAGE},
      {".gif",	FF_IMAGE},
      {".diff",	FF_PATCH},
      {".f",	FF_F77},
      {".bz2",	FF_COMPRESSED},
      {".htm",	FF_HTML},
      {".pm",	FF_PERL},
      {".f77",	FF_F77},
      {".png",	FF_IMAGE},
      {".mpeg",	FF_MOVIE},
      {".mpg",	FF_MOVIE},
      {".jpeg",	FF_IMAGE},
      {".jpg",	FF_IMAGE},
      {".mp3",	FF_SOUND},
      {".hqx",	FF_BINARY},
      {".tex",	FF_TEX},
      {".html",	FF_HTML},
      {".perl",	FF_PERL},
      {".pl",	FF_PERL},
      {".shtml",	FF_HTML},
      {".sty",	FF_TEX},
      {".exe",	FF_BINARY},
      {".gz",	FF_COMPRESSED},
      {".tgz",	FF_COMPRESSED},
      {".bz",	FF_COMPRESSED},
      {".mov",	FF_MOVIE},
      {".bin",	FF_BINARY},
      {".dvi",	FF_DVI},
      {".midi",	FF_SOUND},
      {".avi",	FF_MOVIE},
      {".txt",	FF_TEXT},
      {".tar",	FF_TAR},
      {".snd",	FF_SOUND},
      {".mid",	FF_SOUND},
      {".ps",	FF_POSTSCRIPT}
    };

  static const short lookup[] =
    {
      -98,   6,  -1, -96,   9, -94,  16,  -1, -92,  19,
      -90,  25,  -1, -88,  -1, -86,  32,  -1,  33,  -1,
      -84,  37,  -1,  38,  -1, -82,  42,  -1,  43,  -1,
      -80,  46,  -3,  -2,  -8,  -3, -13,  -3, -19,  -4,
      -21,  -2, -27,  -5, -30,  -2, -37,  -6, -40,  -2,
      -47,  -6
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH) {
    register int	key = hash(str, len);

    if (key <= MAX_HASH_VALUE && key >= 0) {
      register int	idx = lookup[key];

      if (idx >= 0) {
	if (len == lengthtable[idx]) {
	  register const char *	s = wordlist[idx].name;

	  if (*str == *s && !memcmp(str + 1, s + 1, len - 1)) {
	    return &wordlist[idx];
	  }
	}
      }
      else if (idx < -TOTAL_KEYWORDS) {
	register int	offset = - 1 - TOTAL_KEYWORDS - idx;
	register const unsigned char *	lengthptr = &lengthtable[TOTAL_KEYWORDS + lookup[offset]];
	register const FileForm *	wordptr = &wordlist[TOTAL_KEYWORDS + lookup[offset]];
	register const FileForm *	wordendptr = wordptr + -lookup[offset + 1];

	while (wordptr < wordendptr) {
	  if (len == *lengthptr) {
	    register const char *	s = wordptr->name;

	    if (*str == *s && !memcmp(str + 1, s + 1, len - 1)) {
	      return wordptr;
	    }
	  }
	  lengthptr++;
	  wordptr++;
	}
      }
    }
  }

  return 0;
}

/*****************************************************************************/

/*
 * Function Name:	quick_sort
 * Description:		sort string by upper order
 * Arguments:		string array, left value, right value
 * Return Value:	none
 */
#ifdef __GNUC__
__inline
#endif
static void
quick_sort(s, left, right)
	char *	s[];
	int	left;
	int	right;
{
  register int	last;
  register int	i;

  if (left >= right) {
    return;
  }

  swap(s, left, (left + right) / 2);
  last = left;

  for (i = left + 1; i <= right; i++) {
    if (ws_strcmp(s[i], s[left]) < 0) {
      swap(s, ++last, i);
    }
  }

  swap(s, left, last);
  quick_sort(s, left, last - 1);
  quick_sort(s, last + 1, right);
}

/*
 * Function Name:	swap
 * Description:		swap the string
 * Arguments:		string array, swapped position (1), swapped position (2)
 * Return Value:	none
 */
#ifdef __GNUC__
__inline
#endif
static void
swap(s, i, j)
	char *	s[];
	int	i;
	int	j;
{
  char *	t;

  t	= s[i];
  s[i]	= s[j];
  s[j]	= t;
}

/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/
static void
wm_autoindex_register_hooks(void)
{
  wm_hook_request_handler(wm_directory_indexor, HOOK_MIDDLE, MAGIC_TYPE_DIR);
}

static wmModule	mod_autoindex = {
  WM_MAGIC_NUMBER_MAJOR,			/* version */
  WM_UNSECURE,					/* unsecure */
  WM_MODULE_ID,					/* name */
  "Directory Listing Module version 1.2.4",	/* info */
  "Kouichi ABE (WALL)",				/* author */
  NULL,						/* depend */
  NULL,						/* handle */
  (void *)wm_autoindex_init,			/* init */
  (void *)wm_autoindex_finish,			/* finish */
  (void *)wm_autoindex_register_hooks		/* register_handler */
};

wmModule *
wmModuleInfo(void)
{
  return &mod_autoindex;
}

