/*****************************************************************************
 *
 * FILE:	mod_cgi.c
 * DESCRIPTION:	Wyvern: CGI module
 * DATE:	Sat, Sep 16 2000
 * UPDATED:	Sun, Aug 19 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2000-2007 졿Kouichi ABE (WALL), All rights reserved.
 * REFERENCE:	draft-coar-cgi-v11-03.txt
 * LICENSE:
 *
 *  Copyright (c) 2000-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_cgi.c,v 1.26 2007/08/19 06:40:28 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <signal.h>
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#include <errno.h>
#include "wyvern.h"

/*****************************************************************************
 *
 *	Macros and structure definition
 *
 *****************************************************************************/
#define	WM_MODULE_ID	"mod_cgi"

/* KR-C code produced by gperf version 2.7.2 */
/* Command-line: gperf -k'1-43,$' -L KR-C -K code -N lookup -I -t -o -C -n cgi.gperf  */
struct StatusCodeMap {
  const char *	code;
  status_t	status;
};
#define TOTAL_KEYWORDS	(40)
#define MIN_WORD_LENGTH (3)
#define MAX_WORD_LENGTH (3)
#define MIN_HASH_VALUE	(1)
#define MAX_HASH_VALUE	(125)
/* maximum key range = 125, duplicates = 0 */

#define	CGI_TIMEOUT	(60)	/* seconds */

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static int		wm_cgi_run_script(Message *);
#ifdef __GNUC__
__inline
#endif
static unsigned int	hash(char *, unsigned int);
#ifdef __GNUC__
__inline
#endif
static struct StatusCodeMap *	lookup(char *, unsigned int);
static void		wm_cgi_register_hooks(void);

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * SERVER_SOFTWARE:	the name and version of the server software
 * SERVER_NAME:		the name of the server
 * GATEWAY_INTERFACE:	the revision of the CGI specification
 *
 * SERVER_PROTOCOL:	the protocol version of the HTTP
 * SERVER_PORT:		the port which the request is received on the server
 *
 * REQUEST_METHOD:	the method of the HTTP
 * SCRIPT_NAME:		the identity of the CGI script
 * PATH_INFO:		the path to be interpreted by the CGI script
 * PATH_TRANSLATED:	the location in the server's document
 * QUERY_STRING:	a URL-encoded string
 * REMOTE_HOST:		the FQDN of the client
 * REMOTE_ADDR:		the IP address of the client
 * REMOTE_PORT:		the remote port of the client
 * AUTH_TYPE:		the requests "Authorization"
 * REMOTE_USER:		the remote user-ID
 * REMOTE_IDENT:	the identity information reported about the connection
 *			by a RFC 1413
 * CONTENT_TYPE:	the Internet Media Type
 * CONTENT_LENGTH:	the size of the message-body entity
 *
 *
 * HTTP_ACCEPT:
 * HTTP_ACCEPT_ENCODING:
 * HTTP_ACCEPT_LANGUAGE:
 * HTTP_HOST:
 * HTTP_USER_AGENT:
 * HTTP_REFERER:
 * HTTP_COOKIE:
 * HTTP_COOKIE2:
 *
 * DOCUMENT_ROOT:	NCSA and APACHE added -- not in CGI spec
 * SERVER_ROOT:		NCSA added
 * SERVER_ADMIN:	APACHE added
 */

/*****************************************************************************/

/*
 * Function Name:	wm_cgi_run_script
 * Description:		execute CGI script
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
#define	READ_FD		(0)	/* descriptor to read */
#define	WRITE_FD	(1)	/* descriptor to write */
static int
wm_cgi_run_script(msg)
	register Message *	msg;
{
  char *		cwd;	/* CGI working directory */
#if	ENABLE_SUEXEC
  char *		args[3] = { SUEXEC_PATH, NULL, NULL };
  char *		cmd	= SUEXEC_PATH;
#else
  char *		args[2] = { NULL, NULL };
#endif	/* ENABLE_SUEXEC */
  int			fd;
  register FILE *	fin;
  char			buf[BUF_SIZE];
  size_t		len;
  unsigned int		timeout;
  int			rc;

  DBG(fputs("<wm_cgi_run_script>\n", stderr);)
  /* default response status */
  msg->response.status = STATUS_INTERNAL_SERVER_ERROR;

  /* Set CGI working directory */
  cwd = wu_dirname(msg->file.path);
  if (cwd == NULL) {
    wv_syslog(WLL_LOG_ERR, "[%s] failed to change directory", WM_MODULE_ID);
    return WM_RET_ERROR;
  }

  /* Make temporary file to read/write between parent and child */
  fd = wu_mkstemp(httpd->tmpDir);
  if (fd == -1) {
    wv_syslog(WLL_LOG_ERR, "[%s] failed to create temporary file",
	      WM_MODULE_ID);
    return WM_RET_ERROR;
  }

  /* timeout */
  timeout = msg->vhost->timeout > 0 ? msg->vhost->timeout : CGI_TIMEOUT;

  /* Execute the CGI script */
#if	ENABLE_SUEXEC
  if (msg->user.homeDir != NULL) {
    args[1] = msg->file.path;
  }
  else {
    args[0] = cmd = msg->file.path;
  }
  DBG(fprintf(stderr, "DEBUG[%s] cmd=%s args[0]=%s args[1]=%s\n",
	      WM_MODULE_ID, cmd, args[0], args[1]);)
  rc = wt_spawn(cwd, cmd, args, msg->cgi.env, NULL,
	        msg->request.post_fd, fd, fileno(msg->vhost->errorLog.fp),
	        WT_DEFAULT_CPU_LIMIT, WT_DEFAULT_FSIZE_LIMIT,
	        WT_UTIME_SEC(timeout));
#else
  args[0] = msg->file.path;
  rc = wt_spawn(cwd, msg->file.path, args, msg->cgi.env, NULL,
	        msg->request.post_fd, fd, fileno(msg->vhost->errorLog.fp),
	        WT_DEFAULT_CPU_LIMIT, WT_DEFAULT_FSIZE_LIMIT,
	        WT_UTIME_SEC(timeout));
#endif	/* ENABLE_SUEXEC */
  if (rc == -1) {
    wv_syslog(WLL_LOG_ERR, "[%s] failed to execute %s (errno=%d)",
	      WM_MODULE_ID, args[0], errno);
    switch (errno) {
      case EACCES:	msg->response.status = STATUS_FORBIDDEN;	break;
      case ENOENT:	msg->response.status = STATUS_NOT_FOUND;	break;
      case ETIMEDOUT:	msg->response.status = STATUS_GATEWAY_TIMEOUT;	break;
      default:		msg->response.status = STATUS_INTERNAL_SERVER_ERROR;
	break;
    }
    return WM_RET_ERROR;
  }

  /* Get file status */
  lseek(fd, (off_t)0L, SEEK_SET);
  if (fstat(fd, &msg->file.status) == -1) {
    wv_syslog(WLL_LOG_ERR, "[%s] failed to fstat(3)", WM_MODULE_ID);
    return WM_RET_ERROR;
  }

  fin = fdopen(fd, "r");
  if (fin == NULL) {
    wv_syslog(WLL_LOG_ERR, "[%s] failed to fdopen(3)", WM_MODULE_ID);
    return WM_RET_ERROR;
  }

  /* Parse CGI headers */
  len = 0;
  while (fgets(buf, sizeof(buf), fin) != NULL) {
    char *	name;
    char *	value;
    char	c;

    DBG(fprintf(stderr, "DEBUG[%s] %s", WM_MODULE_ID, buf);)
    len += strlen(buf);
    /* non-parsed header */
    if (ws_strncasecmp(buf, "HTTP/1.1", 8) == 0 ||
	ws_strncasecmp(buf, "HTTP/1.0", 8) == 0 ||
	ws_strncasecmp(buf, "HTTP/0.9", 8) == 0) {
      register char *	sp;
      register char *	ep;

      sp = strchr(buf, ' ');
      if (sp != NULL) {
	struct StatusCodeMap *	mp;

	sp++;
	ep = strchr(sp, ' ');
	if (ep != NULL) {
	  *ep = EOL;
	}

	mp = lookup(sp, strlen(sp));
	if (mp != NULL) {
	  msg->response.status = mp->status;
	}
      }
      continue;
    }

    name  = wu_chomp(buf);
    if (buf[0] == EOL) {
      break;
    }
    value = (char *)strchr(buf, ':');
    if (value == NULL) {
      continue;	/* invalid field */
    }
    *value++ = EOL;
    while ((c = *value) && (c == SPC || c == TAB)) {
      value++;
    }
#if	ENABLE_STRICT_PARSING
    wu_tolower(name);

    switch (name[0]) {
      case 'a':
	/* See RFC 2616 Section 14.5 */
	if (ws_strcmp(name, "accept-range") == 0) {
	  /* skip */
	}
	/* See RFC 2616 Section 14.6 */
	else if (ws_strcmp(name, "age") == 0) {
	  wa_put(msg->response.header, "Age", value);
	}
	/* See RFC 2616 Section 14.7 */
	else if (ws_strcmp(name, "allow") == 0) {
	  wa_put(msg->response.header, "Allow", value);
	}
	else {
	  goto unknown;
	}
	break;
      case 'c':
	/* See RFC 2616 Section 14.17 */
	if (ws_strcmp(name, "content-type") == 0) {
	  msg->file.mime = wa_put(msg->response.header, "Content-Type", value);
	}
	/* See RFC 2109 */
	else if (ws_strcmp(name, "cookie") == 0) {
	  wa_put(msg->response.header, "Cookie", value);
	}
	/* See RFC 2965 */
	else if (ws_strcmp(name, "cookie2") == 0) {
	  wa_put(msg->response.header, "Cookie2", value);
	}
	/* See RFC 2616 Section 14.10 */
	else if (ws_strcmp(name, "connection") == 0) {
	  /* skip */
	}
	/* See RFC 2616 Section 14.19 */
	else if (ws_strcmp(name, "cache-control") == 0) {
	  wa_put(msg->response.header, "Cache-Control", value);
	}
	/* See RFC 2616 Section 14.11 */
	else if (ws_strcmp(name, "content-encoding") == 0) {
	  wa_put(msg->response.header, "Content-Encoding", value);
	}
	/* See RFC 2616 Section 14.12 */
	else if (ws_strcmp(name, "content-language") == 0) {
	  wa_put(msg->response.header, "Content-Language", value);
	}
	/* See RFC 2616 Section 14.13 */
	else if (ws_strcmp(name, "content-length") == 0) {
	  msg->response.content_length = (unsigned int)strtol(value, NULL, 10);
	}
	/* See RFC 2616 Section 14.15 */
	else if (ws_strcmp(name, "content-md5") == 0) {
	  wa_put(msg->response.header, "Content-MD5", value);
	}
	/* See RFC 2616 Section 14.14 */
	else if (ws_strcmp(name, "content-location") == 0) {
	  wa_put(msg->response.header, "Content-Location", value);
	}
	/* See RFC 2616 Section 14.16 */
	else if (ws_strcmp(name, "content-range") == 0) {
	  /* skip */
	}
	/* See RFC 2616 Section 19.5.1 */
	else if (ws_strcmp(name, "content-disposition") == 0) {
	  wa_put(msg->response.header, "Content-Disposition", value);
	}
	else {
	  goto unknown;
	}
	break;
      case 'd':
	/* See RFC 2616 Section 14.18 */
	if (ws_strcmp(name, "date") == 0) {
	  /* skip */
	}
	else {
	  goto unknown;
	}
	break;
      case 'e':
	/* See RFC 2616 Section 14.19 */
	if (ws_strcmp(name, "etag") == 0) {
	  wa_put(msg->response.header, "ETag", value);
	}
	/* See RFC 2616 Section 14.21 */
	else if (ws_strcmp(name, "expires") == 0) {
	  wa_put(msg->response.header, "Expires", value);
	}
	else {
	  goto unknown;
	}
	break;
      case 'k':
	/* See RFC 2068 Section 19.7.1.1 */
	if (ws_strcmp(name, "keep-alive") == 0) {
	  /* skip */
	}
	else {
	  goto unknown;
	}
	break;
      case 'l':
	/* See RFC 2616 Section 14.30 */
	if (ws_strcmp(name, "location") == 0) {
	  wa_put(msg->response.header, "Location", value);
	}
	/* See RFC 2616 Section 14.29 */
	else if (ws_strcmp(name, "last-modified") == 0) {
	  /* skip */
	}
	else {
	  goto unknown;
	}
	break;
      case 'p':
	/* See RFC 2616 Section 14.32 */
	if (ws_strcmp(name, "pragma") == 0) {
	  wa_put(msg->response.header, "Pragma", value);
	}
	else {
	  goto unknown;
	}
	break;
      case 'r':
	/* See RFC 2616 Section 14.37 */
	if (ws_strcmp(name, "retry-after") == 0) {
	  /* skip */
	}
	else {
	  goto unknown;
	}
	break;
      case 's':
	if (ws_strcmp(name, "status") == 0) {
	  register char *			p;
	  register struct StatusCodeMap *	mp;

	  p = strchr(value, ' ');
	  if (p != NULL) {
	    *p = EOL;
	  }

	  mp = lookup(value, strlen(value));
	  if (mp != NULL) {
	    msg->response.status = mp->status;
	  }
	}
	/* See RFC 2109 */
	else if (ws_strcmp(name, "set-cookie") == 0) {
	  wa_put_plus(msg->response.header, "Set-Cookie", value);
	}
	/* See RFC 2965 */
	else if (ws_strcmp(name, "set-cookie2") == 0) {
	  wa_put_plus(msg->response.header, "Set-Cookie2", value);
	}
	else if (ws_strcmp(name, "server") == 0) {
	  /* skip */
	}
	else {
	  goto unknown;
	}
	break;
      case 't':
	/* See RFC 2616 Section 14.41 */
	if (ws_strcmp(name, "transfer-encoding") == 0) {
	  wa_put(msg->response.header, "Transfer-Encoding", value);
	}
	/* See RFC 2616 Section 14.40 */
	else if (ws_strcmp(name, "trailer") == 0) {
	  wa_put(msg->response.header, "Trailer", value);
	}
	else {
	  goto unknown;
	}
	break;
      case 'u':
	/* See RFC 2616 Section 14.42 */
	if (ws_strcmp(name, "upgrade") == 0) {
	  wa_put(msg->response.header, "Upgrade", value);
	}
	else {
	  goto unknown;
	}
	break;
      case 'v':
	/* See RFC 2616 Section 14.45 */
	if (ws_strcmp(name, "via") == 0) {
	  wa_put(msg->response.header, "Via", value);
	}
	/* See RFC 2616 Section 14.44 */
	else if (ws_strcmp(name, "vary") == 0) {
	  wa_put(msg->response.header, "Vary", value);
	}
	else {
	  goto unknown;
	}
	break;
      case 'w':
	/* See RFC 2616 Section 14.46 */
	if (ws_strcmp(name, "warning") == 0) {
	  wa_put(msg->response.header, "Warning", value);
	}
	/* See RFC 2616 Section 14.47 */
	else if (ws_strcmp(name, "www-authenticate") == 0) {
	  wa_put(msg->response.header, "WWW-Authenticate", value);
	}
	else {
	  goto unknown;
	}
	break;
      default:
      unknown:
	break;
    }
#else
    if (ws_strcmp(name, "Content-Type") == 0) {
      msg->file.mime = (char *)wa_put(msg->response.header, name, value);
    }
    else if (ws_strcmp(name, "Content-Length") == 0) {
      msg->response.content_length = (unsigned int)strtol(value, NULL, 10);
    }
    else if (ws_strcmp(name, "Status") == 0) {
      register char *			p;
      register struct StatusCodeMap *	mp;

      p = strchr(value, ' ');
      if (p != NULL) {
	*p = EOL;
      }

      mp = lookup(value, strlen(value));
      if (mp != NULL) {
	msg->response.status = mp->status;
      }
    }
    else if (ws_strcmp(name, "Set-Cookie") == 0) {	/* See RFC 2109 */
      wa_put_plus(msg->response.header, name, value);
    }
    else if (ws_strcmp(name, "Set-Cookie2") == 0) {	/* See RFC 2965 */
      wa_put_plus(msg->response.header, name, value);
    }
    else {
      wa_put(msg->response.header, name, value);
    }
#endif	/* ENABLE_STRICT_PARSING */
  }

  msg->file.nfd		    = wt_netfd_fdopen(fd);
  msg->file.offset	    = len;
  msg->file.status.st_size -= len;
  msg->file.status.st_mtime = 0;
  msg->response.status	    = STATUS_REQUEST_OK;
  wt_netfd_seek(msg->file.nfd, (off_t)len, WT_SEEK_SET);

  return WM_RET_OK;
}

#ifdef __GNUC__
__inline
#endif
static unsigned int
hash(str, len)
	register char *		str;
	register unsigned int	len;
{
  static unsigned char	asso_values[] =
    {
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126,   0,  20,
       40,   1,   5,   6,  25,  45,  60,  21, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126, 126, 126, 126, 126,
      126, 126, 126, 126, 126, 126
  };
  register int hval = 0;

  switch (len) {
    default:
    case 3:
      hval += asso_values[(unsigned char)str[2]];
    case 2:
      hval += asso_values[(unsigned char)str[1]];
    case 1:
      hval += asso_values[(unsigned char)str[0]];
      break;
  }
  return hval + asso_values[(unsigned char)str[len - 1]];
}

#ifdef __GNUC__
__inline
#endif
static struct StatusCodeMap *
lookup(str, len)
	register char *		str;
	register unsigned int	len;
{
  static struct StatusCodeMap wordlist[] =
    {
      {""},
      {"300",	STATUS_MULTIPLE_CHOICES},
      {""},
      {"303",	STATUS_SEE_OTHER},
      {""},
      {"400",	STATUS_BAD_REQUEST},
      {"500",	STATUS_INTERNAL_SERVER_ERROR},
      {"403",	STATUS_FORBIDDEN},
      {"503",	STATUS_SERVICE_UNAVAILABLE},
      {""}, {""},
      {"304",	STATUS_NOT_MODIFIED},
      {""},
      {"305",	STATUS_USE_PROXY},
      {""},
      {"404",	STATUS_NOT_FOUND},
      {"504",	STATUS_GATEWAY_TIMEOUT},
      {"405",	STATUS_METHOD_NOT_ALLOWED},
      {"505",	STATUS_HTTP_VERSION_NOT_SUPPORTED},
      {""},
      {"100",	STATUS_CONTINUE},
      {""}, {""}, {""}, {""},
      {"410",	STATUS_GONE},
      {""},
      {"413",	STATUS_REQUEST_ENTITY_TOO_LARGE},
      {""}, {""}, {""}, {""}, {""}, {""}, {""},
      {"414",	STATUS_REQUEST_URI_TOO_LONG},
      {""},
      {"415",	STATUS_UNSUPPORTED_MEDIA_TYPE},
      {""}, {""},
      {"200",	STATUS_REQUEST_OK},
      {"301",	STATUS_MOVED_PERMANENTLY},
      {"203",	STATUS_NON_AUTH_INFO},
      {""}, {""},
      {"401",	STATUS_UNAUTHORIZED},
      {"501",	STATUS_NOT_IMPLEMENTED},
      {"409",	STATUS_CONFLICT},
      {""}, {""},
      {"204",	STATUS_NO_CONTENT},
      {""},
      {"205",	STATUS_RESET_CONTENT},
      {""}, {""},
      {"406",	STATUS_NOT_ACCEPTABLE},
      {""}, {""}, {""}, {""},
      {"101",	STATUS_SWITCHING_PROTOCOLS},
      {""}, {""}, {""}, {""},
      {"411",	STATUS_LENGTH_REQUIRED},
      {""}, {""}, {""}, {""}, {""}, {""}, {""}, {""}, {""},
      {"416",	STATUS_REQUESTED_RANGE_NOT_SATISFIABLE},
      {""}, {""}, {""}, {""},
      {"201",	STATUS_CREATED},
      {"302",	STATUS_FOUND},
      {""}, {""}, {""},
      {"402",	STATUS_PAYMENT_REQUIRED},
      {"502",	STATUS_BAD_GATEWAY},
      {""}, {""}, {""},
      {"206",	STATUS_PARTIAL_CONTENT},
      {"307",	STATUS_TEMPORARY_REDIRECT},
      {""}, {""}, {""},
      {"407",	STATUS_PROXY_AUTH_REQUIRED},
      {""}, {""}, {""}, {""}, {""}, {""}, {""}, {""}, {""},
      {"412",	STATUS_PRECONDITION_FAILED},
      {""}, {""}, {""}, {""}, {""}, {""}, {""}, {""}, {""},
      {"417",	STATUS_EXPECTATION_FAILED},
      {""}, {""}, {""}, {""},
      {"202",	STATUS_ACCEPTED},
      {""}, {""}, {""}, {""},
      {"408",	STATUS_REQUEST_TIMEOUT}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH) {
    register int	key = hash (str, len);

    if (key <= MAX_HASH_VALUE && key >= 0) {
      register const char *	s = wordlist[key].code;

      if (*str == *s && ws_strcmp(str + 1, s + 1) == 0) {
        return &wordlist[key];
      }
    }
  }

  return 0;
}

/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/
static void
wm_cgi_register_hooks(void)
{
  wm_hook_request_handler(wm_cgi_run_script, HOOK_MIDDLE, MAGIC_TYPE_CGI);
}

static wmModule	mod_cgi = {
  WM_MAGIC_NUMBER_MAJOR,	/* version */
  WM_UNSECURE,			/* unsecure */
  WM_MODULE_ID,			/* name */
  "CGI Module version 1.1.11",	/* info */
  "Kouichi ABE (WALL)",		/* author */
  "mod_env",			/* depend */
  NULL,				/* handle */
  NULL,				/* init */
  NULL,				/* finish */
  (void *)wm_cgi_register_hooks	/* register_handler */
};

wmModule *
wmModuleInfo(void)
{
  return &mod_cgi;
}
