/*****************************************************************************
 *
 * FILE:	mod_ident.c
 * DESCRIPTION:	Identification Protocol (RFC1413) module for wyvern
 * DATE:	Sat, Sep 16 2000
 * UPDATED:	Wed, Mar  1 2006
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2000-2006 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2000-2006 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_ident.c,v 1.7 2006/03/01 13:21:37 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_NETDB_H
#include <netdb.h>
#endif	/* HAVE_NETDB_H */
#include <sys/types.h>
#if	HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif	/* HAVE_SYS_SOCKET_H */
#if	HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif	/* HAVE_NETINET_IN_H */
#include <setjmp.h>
#include <signal.h>
#include <errno.h>
#include "wyvern.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	WM_MODULE_ID	"mod_ident"

#define	IDENT_TIMEOUT	(10)	/* wait at least 30 seconds (or longer) */
#define	IDENT_PORT	"113"	/* authentication service in /etc/service */

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static int	wm_query_remote_ident(Message *);
static void	timeout(int);
static void	wm_ident_register_hooks(void);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static jmp_buf	id_env;	/* to store program environments */

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * FUNCTION NAME:	wm_query_remote_ident
 * DESCRIPTION:		identify remote user
 * ARGUMENTS:		thread
 * RETURN VALUE:	successful: 0, otherwise: -1
 */
static int
wm_query_remote_ident(msg)
	register Message *	msg;
{
  /* set up a timer to avoid blocking the server */
  if (setjmp(id_env) == 0) {
    struct addrinfo *		our;

    signal(SIGALRM, timeout);
    alarm(IDENT_TIMEOUT);

    our = wt_host_serv("127.0.0.1", NULL, AF_INET, SOCK_STREAM);
    if (our != NULL) {
      int			sd;
      struct addrinfo *		rmt;

      /* create the socket */
      sd = socket(our->ai_family, our->ai_socktype, our->ai_protocol);
      if (sd == -1) {
	wv_syslog(WLL_LOG_ERR, "[%s] failed to create a socket", WM_MODULE_ID);
	goto our_free;
      }
      /* bind the socket */
      if (bind(sd, our->ai_addr, our->ai_addrlen) == -1) {
        wv_syslog(WLL_LOG_ERR, "[%s] failed to bind the name", WM_MODULE_ID);
        goto sd_free;
      }

      rmt = wt_host_serv(msg->client.ip, IDENT_PORT, 0, SOCK_STREAM);
      if (rmt != NULL) {
	unsigned int	our_port;
	unsigned int	rmt_port;
	char		buf[BUFSIZ];
	int		n;

	/* connect the ident server */
	if (connect(sd, rmt->ai_addr, rmt->ai_addrlen) == -1) {
	  if ((errno != ECONNREFUSED) && (errno != EHOSTUNREACH) &&
	      (errno != ECONNRESET)) {
	    wv_syslog(WLL_LOG_ERR,
		      "[%s] failed to connect a remote identd", WM_MODULE_ID);
	  }
	  goto rmt_free;
	}

	our_port = (unsigned int)httpd->sockets->info.port;
	rmt_port = (unsigned int)strtol(msg->client.port, NULL, 10);

	memset(buf, 0, sizeof(buf));
	n = snprintf(buf, sizeof(buf), "%u,%u" CRLF, rmt_port, our_port);
	if (n >= sizeof(buf)) {
	  wv_syslog(WLL_LOG_ALERT, "[%s] caused buffer overflow", WM_MODULE_ID);
	  goto rmt_free;
	}

	/* send the query to the ident server */
	if (write(sd, buf, (size_t)n) != n) {
	  wv_syslog(WLL_LOG_ERR, "[%s] failed to write a data", WM_MODULE_ID);
	  goto rmt_free;
	}

	/* nothing more to send, so shutdown the write socket */
	if (shutdown(sd, SHUT_WR) < 0) {
	  wv_syslog(WLL_LOG_ERR, "[%s] failed to shutdown", WM_MODULE_ID);
	  goto rmt_free;
	}

	/* read response from the ident server */
	memset(buf, 0, sizeof(buf));
	n = read(sd, buf, (size_t)sizeof(buf));
	if (n > 0) {
	  char		user[MAX_NAME_LEN];
	  unsigned int	s_port;
	  unsigned int	c_port;

	  /* parse the response */
	  memset(user, 0, sizeof(user));
	  n = sscanf(buf, "%5u , %5u : USERID :%*[^:]:%8s",
			&s_port, &c_port, user);
	  if ((n == 3) && (rmt_port == s_port) && (our_port == c_port)) {
	    msg->user.identName = ws_strdup(user);
	  }
	}
	else {
	  wv_syslog(WLL_LOG_ERR, "[%s] failed to read a data", WM_MODULE_ID);
	}
#if	DEBUG
	wv_syslog(WLL_LOG_DEBUG, "[%s] %s", WM_MODULE_ID, buf);
#endif	/* DEBUG */

rmt_free:
	freeaddrinfo(rmt);
      }
sd_free:
      wu_close(sd);
our_free:
      freeaddrinfo(our);
    }
    alarm(0);	/* reset timer */
  }

  return WM_RET_OK;
}

/*
 * FUNCTION NAME:	timeout
 * DESCRIPTION:		timeout to request
 * ARGUMENTS:		signal id
 * RETURN VALUE:	none
 */
static void
timeout(sig)
	int	sig;
{
  longjmp(id_env, sig);
}

/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/
static void
wm_ident_register_hooks(void)
{
  wm_hook_user_id_checker(wm_query_remote_ident, HOOK_MIDDLE);
}

static wmModule	mod_ident = {
  WM_MAGIC_NUMBER_MAJOR,	/* version */
  WM_SECURE,			/* secure */
  WM_MODULE_ID,			/* name */
  "Identification Protocol Module version 1.1.1",	/* info */
  "Kouichi ABE (WALL)",		/* author */
  NULL,				/* depend */
  NULL,				/* handle */
  NULL,				/* init */
  NULL,				/* finish */
  (void *)wm_ident_register_hooks	/* register_handler */
};

wmModule *
wmModuleInfo(void)
{
  return &mod_ident;
}
