/*****************************************************************************
 *
 * FILE:	mod_log_pgsql.c
 * DESCRIPTION:	Wyvern: logging module with PostgreSQL
 * DATE:	Tue, Dec  6 2005
 * UPDATED:	Wed, Sep 12 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2005-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2005-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_log_pgsql.c,v 1.13 2007/09/12 02:24:26 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <libpq-fe.h>
#include <wyvern.h>

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	WM_MODULE_ID	"mod_log_pgsql"

#define	PGSQL_DB_NAME	"wyvern"
#define	PGSQL_USERNAME	"wyvern"
#define	PGSQL_PASSWORD	"$wyvern"
#define	INSERT_ACCESS \
	"INSERT INTO access_log VALUES('%s','%s.%lu','%s','%s','%s','%s','%s','%s','%s',%d,%ld,'%s','%s','%s',%.3f);"

#define	UNIX_EPOC	"1970-01-01 00:00:00"

#ifndef	XPUT
#define	XPUT(a)	(a != NULL ? a : "")
#endif	/* XPUT */

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	wm_log_pgsql_init(void);
static void	wm_log_pgsql_finish(void);
static int	wm_log_pgsql_writer(Message *);
static void	wm_log_pgsql_register_hooks(void);
static int	escape(const char * s, char ** r);
static double	proctime(struct timeval *, struct timeval *);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static PGconn *	conn = NULL;

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/
static void
wm_log_pgsql_init(void)
{
  const char *	pghost	= NULL;
  const char *	pgport	= NULL;
  const char *	pgopts	= NULL;
  const char *	pgtty	= NULL;
  const char *	dbName	= PGSQL_DB_NAME;
  const char *	user	= PGSQL_USERNAME;
  const char *	passwd	= PGSQL_PASSWORD;

  /* make a connection to the database */
  conn = PQsetdbLogin(pghost, pgport, pgopts, pgtty, dbName, user, passwd);

  /* check to see that the backend connection was successfully made */
  if (PQstatus(conn) == CONNECTION_BAD) {
    PQfinish(conn);
    wv_die("[%s] failed to connection to database '%s': %s",
	   WM_MODULE_ID, dbName, PQerrorMessage(conn));
  }
}

static void
wm_log_pgsql_finish(void)
{
  if (conn != NULL) {
    PQfinish(conn);	/* close the connection to the database and cleanup */
    conn = NULL;
  }
}

static int
wm_log_pgsql_writer(msg)
	register Message *	msg;
{
  int		status	= WM_RET_DECLINED;
  char *	tbuf;
  char *	sql	= NULL;
  long		length	= 0;	/* transfer length */
  const char *	phrase	= NULL;	/* text phrase of error status */
  double	uptime;
  const char *	xff;
  const char *	ref;
  char *	ref_esc	= NULL;	/* escaped Referer */
  const char *	ua;
  char *	ua_esc	= NULL;	/* escaped User-Agent */

  wu_timestamp(&tbuf, msg->time.request.tv_sec, WUL_TIME_FORMAT_SQL,
	       WUL_ZONE_LOCAL);

  /* Informational: 1xx */
  /* Successful:    2xx */
  /* Redirection:   3xx */
  if (msg->response.status <= STATUS_CODE_3XX) {
    switch (msg->request.method.type) {
      case M_HEAD: length = 0;				break;
      case M_GET:  length = msg->file.status.st_size;	break;
      case M_POST: length = msg->request.content_length;break;
      default:     length = -1;				break;
    }
  }
  /* Client Error: 4xx */
  /* Server Error: 5xx */
  if (msg->response.status > STATUS_CODE_4XX) {
    phrase = wvResponseStatusTable[msg->response.status].phrase;
  }

  /* calculate uptime */
  uptime = proctime(&msg->time.request, &msg->time.response);

  xff = wa_get(msg->request.header, "x-forwarded-for");
  ref = wa_get(msg->request.header, "referer");		escape(ref, &ref_esc);
  ua  = wa_get(msg->request.header, "user-agent");	escape(ua, &ua_esc);

  ws_asprintf(&sql, INSERT_ACCESS,
	msg->vhost->serverName,
	tbuf ? tbuf : UNIX_EPOC, tbuf ? msg->time.request.tv_usec : 0,
	msg->client.ip,
	XPUT(xff),
	XPUT(msg->user.identName),
	XPUT(msg->user.authName),
	XPUT(msg->request.method.name),
	XPUT(msg->request.uri),
	XPUT(msg->request.http.protocol),
	wvResponseStatusTable[msg->response.status].code,
	length,
	XPUT(phrase),
	XPUT(ref_esc),
	XPUT(ua_esc),
	uptime);
  wv_free(tbuf);
  wv_free(ref_esc);
  wv_free(ua_esc);

  if (sql != NULL) {
    static bool	in_progress = false;
    int		rc;

    /* check connection */
    if (PQstatus(conn) != CONNECTION_OK) {
      PQreset(conn);
      wv_syslog(WLL_LOG_INFO, "[%s] re-established the connection",
		WM_MODULE_ID);
    }

    /* wait to achieve a previous query */
    if (in_progress) {
      PGresult *	res;

      while ((res = PQgetResult(conn)) != NULL) { continue; }
      PQclear(res);	/* to avoid memory leaks */
      in_progress = false;
    }

    rc = PQsendQuery(conn, sql);
    if (rc == 0) {
      status = WM_RET_ERROR;
      wv_syslog(WLL_LOG_ERR, "[%s] INSERT command failed: %s [%s]",
  		WM_MODULE_ID, PQerrorMessage(conn), sql);
    }
    else {
      status = WM_RET_OK;
      in_progress = true;
    }
    free(sql);
  }

  return status;
}

static int
escape(s, r)
	const char *	s;
	char **		r;	/* return value */
{
  int	error = -1;

  if (s != NULL) {
    size_t	l;

    l  = strlen(s);
    *r = (char *)malloc(l * 2 + 1);
    if (*r != NULL) {
      PQescapeStringConn(conn, *r, s, l, &error);
      if (error != 0) {	/* failed */
	free(*r);
	*r = NULL;
      }
    }
  }

  return error;
}

static double
proctime(s, e)
	struct timeval *	s;	/* start timestamp */
	struct timeval *	e;	/* end timestamp */
{
  long		secs;
  long		msecs;
  char *	buf;
  double	status = 0;
  
  secs   = e->tv_sec  - s->tv_sec;
  msecs  = e->tv_usec - s->tv_usec;
  msecs /= 1000;
  if (msecs < 0) {
    msecs += 1000;
  }
  if (msecs >= 1000) {
    secs += msecs / 1000;
    msecs = msecs % 1000;
  }

  ws_asprintf(&buf, "%ld.%ld", secs, msecs);
  if (buf != NULL) {
    status = strtod(buf, (char **)NULL);
    free((char *)buf);
  }

  return status;
}

/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/
static void
wm_log_pgsql_register_hooks(void)
{
  wm_hook_logger(wm_log_pgsql_writer, HOOK_FIRST);
}

static wmModule	mod_log_pgsql = {
  WM_MAGIC_NUMBER_MAJOR,	/* version */
  WM_SECURE,			/* secure */
  WM_MODULE_ID,			/* name */
  "PostgreSQL Logging Module version 1.0.6",	/* info */
  "Kouichi ABE (WALL)",		/* author */
  NULL,				/* depend */
  NULL,				/* handle */
  (void *)wm_log_pgsql_init,	/* init */
  (void *)wm_log_pgsql_finish,	/* finish */
  (void *)wm_log_pgsql_register_hooks	/* register_handler */
};

wmModule *
wmModuleInfo(void)
{
  return &mod_log_pgsql;
}
