/*****************************************************************************
 *
 * FILE:	mod_log_sqlite.c
 * DESCRIPTION:	Wyvern: logging module with SQLite
 * DATE:	Wed, Apr 13 2005
 * UPDATED:	Wed, Sep 12 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2005-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2005-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_log_sqlite.c,v 1.14 2007/09/12 02:24:26 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	ENABLE_SQLITE3
#include <sqlite3.h>
#elif	ENABLE_SQLITE2
#include <sqlite.h>
#endif	/* ENABLE_SQLITE3 */
#include <wyvern.h>

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	WM_MODULE_ID	"mod_log_sqlite"

#define	LOG_DB_NAME	"sqlite_log.db"
#define	INSERT_ACCESS \
	"INSERT INTO access_log VALUES('%q','%q.%lu','%q','%q','%q','%q','%q','%q','%q','%d','%ld','%q','%q','%q','%.3f');"

#define	UNIX_EPOC	"1970-01-01 00:00:00"

#ifndef	XPUT
#define	XPUT(a)	(a != NULL ? a : "")
#endif	/* XPUT */

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	wm_log_sqlite_init(void);
static void	wm_log_sqlite_finish(void);
static int	wm_log_sqlite_writer(Message *);
static void	wm_log_sqlite_register_hooks(void);
static double	proctime(struct timeval *, struct timeval *);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
#if	ENABLE_SQLITE3
static sqlite3 *	sdb = NULL;
#elif	ENABLE_SQLITE2
static sqlite *		sdb = NULL;
#endif	/* ENABLE_SQLITE3 */

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/
static void
wm_log_sqlite_init(void)
{
  char *	path;

  ws_asprintf(&path, "%s/%s", httpd->databaseDir, LOG_DB_NAME);
  if (path != NULL) {
    /* open history database */
#if	ENABLE_SQLITE3
    if (sqlite3_open(path, &sdb) != SQLITE_OK) {
      wv_die("[%s] failed to open database: %s", WM_MODULE_ID, path);
    }
#elif	ENABLE_SQLITE2
    sdb = sqlite_open(path, 0, NULL);
    if (sdb != NULL) {
      wv_die("[%s] failed to open database: %s", WM_MODULE_ID, path);
    }
#endif	/* ENABLE_SQLITE3 */
    free(path);
  }
  else {
    wv_die("[%s] no enough memory", WM_MODULE_ID);
  }
}

static void
wm_log_sqlite_finish(void)
{
  if (sdb != NULL) {
#if	ENABLE_SQLITE3
    sqlite3_close(sdb);
#elif	ENABLE_SQLITE2
    sqlite_close(sdb);
#endif	/* ENABLE_SQLITE3 */
  }
}

static int
wm_log_sqlite_writer(msg)
	register Message *	msg;
{
#define	INSERT_RETRY_MAX	(3)
  int		status	= WM_RET_DECLINED;
  char *	tbuf;
  char *	sql	= NULL;
  long		length	= 0;	/* transfer length */
  const char *	phrase	= NULL;	/* text phrase of error status */
  double	uptime;
  register const char *	xff;
  register const char *	ref;
  register const char *	ua;

  wu_timestamp(&tbuf, msg->time.request.tv_sec, WUL_TIME_FORMAT_SQL,
	       WUL_ZONE_LOCAL);

  /* Informational: 1xx */
  /* Successful:    2xx */
  /* Redirection:   3xx */
  if (msg->response.status <= STATUS_CODE_3XX) {
    switch (msg->request.method.type) {
      case M_HEAD: length = 0;				break;
      case M_GET:  length = msg->file.status.st_size;	break;
      case M_POST: length = msg->request.content_length;break;
      default:     length = -1;				break;
    }
  }
  /* Client Error: 4xx */
  /* Server Error: 5xx */
  if (msg->response.status > STATUS_CODE_4XX) {
    phrase = wvResponseStatusTable[msg->response.status].phrase;
  }

  /* calculate uptime */
  uptime = proctime(&msg->time.request, &msg->time.response);

  xff = wa_get(msg->request.header, "x-forwarded-for");
  ref = wa_get(msg->request.header, "referer");
  ua  = wa_get(msg->request.header, "user-agent");
#if	ENABLE_SQLITE3
  sql = sqlite3_mprintf
#elif	ENABLE_SQLITE2
  sql = sqlite_mprintf
#endif	/* ENABLE_SQLITE3 */
       (INSERT_ACCESS,
	msg->vhost->serverName,
	tbuf ? tbuf : UNIX_EPOC, tbuf ? msg->time.request.tv_usec : 0,
	msg->client.ip,
	XPUT(xff),
	XPUT(msg->user.identName),
	XPUT(msg->user.authName),
	XPUT(msg->request.method.name),
	XPUT(msg->request.uri),
	XPUT(msg->request.http.protocol),
	wvResponseStatusTable[msg->response.status].code,
	length,
	XPUT(phrase),
	XPUT(ref),
	XPUT(ua),
	uptime);
  wv_free(tbuf);

  if (sql != NULL) {
    int	rc;
    int	c = 0;

retry:
#if	ENABLE_SQLITE3
    rc = sqlite3_exec(sdb, sql, NULL, NULL, NULL);
#elif	ENABLE_SQLITE2
    rc = sqlite_exec(sdb, sql, NULL, NULL, NULL);
#endif	/* ENABLE_SQLITE3 */
    switch (rc) {
      case SQLITE_OK:
	status = WM_RET_OK;
	break;
      case SQLITE_BUSY:
	sleep(1);
	if (c++ < INSERT_RETRY_MAX) {
	  goto retry;
	}
      default:
	status = WM_RET_ERROR;
	break;
    }
#if	ENABLE_SQLITE3
    sqlite3_free(sql);
#elif	ENABLE_SQLITE2
    sqlite_freemem(sql);
#endif	/* ENABLE_SQLITE3 */
  }

  return status;
}

static double
proctime(s, e)
	struct timeval *	s;	/* start timestamp */
	struct timeval *	e;	/* end timestamp */
{
  long		secs;
  long		msecs;
  char *	buf;
  double	status = 0;
  
  secs   = e->tv_sec  - s->tv_sec;
  msecs  = e->tv_usec - s->tv_usec;
  msecs /= 1000;
  if (msecs < 0) {
    msecs += 1000;
  }
  if (msecs >= 1000) {
    secs += msecs / 1000;
    msecs = msecs % 1000;
  }

  ws_asprintf(&buf, "%ld.%ld", secs, msecs);
  if (buf != NULL) {
    status = strtod(buf, (char **)NULL);
    free((char *)buf);
  }

  return status;
}

/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/
static void
wm_log_sqlite_register_hooks(void)
{
  wm_hook_logger(wm_log_sqlite_writer, HOOK_FIRST);
}

static wmModule	mod_log_sqlite = {
  WM_MAGIC_NUMBER_MAJOR,	/* version */
  WM_SECURE,			/* secure */
  WM_MODULE_ID,			/* name */
  "SQLite Logging Module version 1.0.7",	/* info */
  "Kouichi ABE (WALL)",		/* author */
  NULL,				/* depend */
  NULL,				/* handle */
  (void *)wm_log_sqlite_init,	/* init */
  (void *)wm_log_sqlite_finish,	/* finish */
  (void *)wm_log_sqlite_register_hooks	/* register_handler */
};

wmModule *
wmModuleInfo(void)
{
  return &mod_log_sqlite;
}
