/*****************************************************************************
 *
 * FILE:	mod_ssi.c
 * DESCRIPTION:	Wyvern: SSI module
 * DATE:	Sat, Sep 16 2000
 * UPDATED:	Sat, Aug 25 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2000-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2000-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_ssi.c,v 1.30 2007/08/24 18:01:55 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_VFORK_H
#include <vfork.h>
#endif	/* HAVE_VFORK_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <signal.h>
#include <time.h>
#include "wyvern.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	WM_MODULE_ID	"mod_ssi"

#define	DEF_SSI_SUFFIX	".shtml"
#ifndef	MAP_FAILED
#define	MAP_FAILED	((caddr_t)-1)
#endif	/* MAP_FAILED */
#define	U_KILO		(1024L)
#define	U_MEGA		(1024L*1024L)
#define	U_GIGA		(1024L*1024L*1024L)
#define	TIMEFMT_LEN	(64)
#define	SSI_BUFSIZE	(256)

typedef enum {
  SSI_CONFIG,
  SSI_INCLUDE,
  SSI_ECHO,
  SSI_FSIZE,
  SSI_FLASTMOD,
  SSI_EXEC
} ssi_directives_t;

typedef enum {	/* ssi token parsing state */
  ST_GND,
  ST_LBR,
  ST_EXC,
  ST_HYP1,
  ST_HYP2,
  ST_CMD
} state_t;

typedef enum {
  SZ_BYTES,
  SZ_ABBREV
} sizefmt_t;

typedef enum {
  ET_TAG,
  ET_VALUE,
  ET_FORMAT
} error_ssi_t;

#define	SSI_TIMEOUT	(10)

#define	SSI_CACHE_CONTROL	"no-store, no-cache"
#define	SSI_PRAGMA		"no-cache"

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static int	wm_ssi_type_checker(Message *);
static int	wm_ssi_filter(Message *);
static void	ssi_exec_directive(Message *, FILE *, char *);

static void	do_config(Message *, const char *, const char *);
static void	do_include(Message *, FILE *, const char *, const char *);
static void	do_echo(Message *, FILE *, const char *, const char *);
static void	do_fsize(Message *, FILE *, const char *, const char *);
static void	do_flastmod(Message *, FILE *, const char *, const char *);
static void	do_exec(Message *, FILE *, const char *, const char *);
static void	ssi_run_program(Message *, FILE *, const char *);

static void	print_formatted_time(Message *, FILE *, time_t, bool);
static int	get_file_path(Message *, const char *, const char *, char * *);
static int	ssi_error(error_ssi_t, const char *);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static char		timefmt[TIMEFMT_LEN];
static sizefmt_t	sizefmt;

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wm_ssi_type_checker
 * Description:		check mimetype of supplied file
 * Arguments:		message
 * Return Value:	successful: 0, continued: 1, otherwise: -1
 */
static int
wm_ssi_type_checker(msg)
	register Message *	msg;
{
  DBG(fprintf(stderr, "[wm_ssi_type_checker] mime = %s\n", msg->file.mime);)
  if (msg->file.mime == NULL && msg->file.path != NULL) {
    char *	suffix;	/* suffix of file */

    DBG(fprintf(stderr, "[wm_ssi_type_checker] %s\n", msg->file.path);)
    suffix = strrchr(msg->file.path, '.');
    if (suffix != NULL) {	/* no suffix */
      if (ws_strncmp(suffix, DEF_SSI_SUFFIX, 6) == 0) {
	msg->file.mime = MAGIC_TYPE_SSI;
	return WM_RET_OK;
      }
      return WM_RET_ERROR;
    }
  }

  return WM_RET_DECLINED;
}

/*****************************************************************************/

/*
 * FUNCTION NAME:	wm_ssi_filter
 * DESCRIPTION:		address an SSI content
 * ARGUMENTS:		message
 * RETURN VALUE:	successful: 0, otherwise: -1
 */
static int
wm_ssi_filter(msg)
	register Message *	msg;
{
  int			fd;
  char *		file;
  register char *	ch;
  register FILE *	fout;
  state_t		st;
  char *		directive;
  char *		dirp;
  char *		text;
  size_t		txtc;	/* length counter */
  struct stat		sbuf;

  DBG(fputs("<wm_ssi_filter>\n", stderr);)
  /* Default formats */
  				/* ex. Sat Nov 27 01:26:37 JST 1999	*/
  strlcpy(timefmt, "%a %b %e %T %Z %Y", sizeof(timefmt));
  sizefmt = SZ_BYTES;

  /* Map a file */
  file = (char *)mmap(NULL, (size_t)msg->file.status.st_size,
		      PROT_READ|PROT_WRITE, MAP_PRIVATE,
		      wt_netfd_fileno(msg->file.nfd), (off_t)0);
  if (file == MAP_FAILED) {
    wv_syslog(WLL_LOG_ERR, "[%s] mmap(2) error", WM_MODULE_ID);
    goto noparse;
  }

  /* Open temporary file */
  fd = wu_mkstemp(httpd->tmpDir);
  if (fd == -1) {
    wv_syslog(WLL_LOG_ERR, "[%s] failed to create temporary file",
	      WM_MODULE_ID);
    goto error;
  }

  /* Open output file */
  fout = fdopen(fd, "r+");
  if (fout == NULL) {
    wv_syslog(WLL_LOG_ERR, "[%s] failed to open file descriptors",
	      WM_MODULE_ID);
    close(fd);
    goto error;
  }

  /* parsing */
  text = directive = file;
  dirp = NULL;
  txtc = 0;
  st   = ST_GND;
  for (ch = file; *ch != EOL; ch++) {
    switch (st) {
      case ST_LBR:
	txtc++;
	if (*ch == '!')	{ st = ST_EXC; }
	else		{ st = ST_GND; }
	break;
      case ST_EXC:
	txtc++;
	if (*ch == '-')	{ st = ST_HYP1; }
	else		{ st = ST_GND; }
	break;
      case ST_HYP1:
	txtc++;
	if (*ch == '-')	{ st = ST_HYP2; }
	else		{ st = ST_GND; }
	break;
      case ST_HYP2:
	txtc++;
	if (*ch == '#')	{ st = ST_CMD; txtc -= 5; directive = &ch[1];}
	else		{ st = ST_GND;
			  fwrite(text, 1, txtc, fout);
			  text = &ch[1];
			  txtc = 0;
			}
	break;
      case ST_CMD:
	dirp = ch;
	if (*ch == '-') {
	  if (ch[1] == '-' && ch[2] == '>') {
	    ch += 3;
	    fwrite(text, 1, txtc, fout);
	    text  = ch;
	    txtc  = 1;
	    *dirp = EOL;
	    ssi_exec_directive(msg, fout, directive);
	    directive	= NULL;
	    dirp	= NULL;
	    st		= ST_GND;
	  }
	}
	break;
      case ST_GND:
      default:
	if (*ch == '<') {
	  st = ST_LBR;
	}
	txtc++;
	break;
    }
  }
  fputs(text, fout);
  fflush(fout);

  if (fstat(fd, &sbuf) == -1) {
    wv_syslog(WLL_LOG_ERR, "[%s] failed to get file status", WM_MODULE_ID);
    close(fd);
    goto error;
  }
  DBG(fprintf(stderr, "[ssi] Content-Length: %d\n", (int)sbuf.st_size);)

  wt_netfd_close(msg->file.nfd);
  msg->file.nfd			= wt_netfd_fdopen(fd);
  msg->file.status.st_size	= sbuf.st_size - msg->file.offset;
  wa_put(msg->response.header, "Cache-Control", SSI_CACHE_CONTROL);
  wa_put(msg->response.header, "Pragma", SSI_PRAGMA);
error:
  munmap((char *)file, (size_t)msg->file.status.st_size);
noparse:
  msg->file.mime	= MAGIC_TYPE_HTML;
  msg->response.status	= STATUS_REQUEST_OK;
  wt_netfd_seek(msg->file.nfd, (off_t)msg->file.offset, WT_SEEK_SET);

  return WM_RET_OK;
}

/*
 * FUNCTION NAME:	ssi_exec_directive
 * DESCRIPTION:		execute directive
 * ARGUMENTS:		message, mapped buffer
 * RETURN VALUE:	none
 */
static void
ssi_exec_directive(msg, fout, cmd)
	register Message *	msg;
	register FILE *		fout;
	char *			cmd;
{
  char *	directive;
  char *	tag;
  char *	val;
  char *	p;

  DBG(fprintf(stderr, "[ssi_exec_directive] %s\n", cmd);)
  /* SSI directive */
  directive  = cmd;
  directive += strspn(cmd, " \t\r\n");

  /* SSI tag */
  tag = (char *)strpbrk(directive, " \t\r\n");
  if (tag == NULL) {
    wv_syslog(WLL_LOG_ERR, "[%s] illegal format: %s", WM_MODULE_ID, directive);
    return;
  }
  *tag++ = EOL;
  tag	+= strspn(tag, " \t\r\n=");

  /* SSI value */
  val = (char *)strpbrk(tag, " \t\r\n=");
  if (val == NULL) {
    wv_syslog(WLL_LOG_ERR, "[%s] %s needs a tag", WM_MODULE_ID, directive);
    return;
  }
  *val++ = EOL;
  val	 = (char *)strchr(val, '"');
  if (val++ == NULL) {
    wv_syslog(WLL_LOG_ERR, "[%s] %s's %s needs a value",
	      WM_MODULE_ID, directive, tag);
    return;
  }
  p = (char *)strchr(val, '"');
  if (p == NULL) {
    wv_syslog(WLL_LOG_ERR, "[%s] %s's value of %s isn't closed by \".",
	      WM_MODULE_ID, directive, tag);
    return;
  }
  *p = EOL;

  /* act by the directive */
  if (ws_strcmp(directive, "config") == 0) {
    do_config(msg, tag, val);
  }
  else if (ws_strcmp(directive, "include") == 0) {
    do_include(msg, fout, tag, val);
  }
  else if (ws_strcmp(directive, "echo") == 0) {
    do_echo(msg, fout, tag, val);
  }
  else if (ws_strcmp(directive, "fsize") == 0) {
    do_fsize(msg, fout, tag, val);
  }
  else if (ws_strcmp(directive, "flastmod") == 0) {
    do_flastmod(msg, fout, tag, val);
  }
  else if (ws_strcmp(directive, "exec") == 0) {
    do_exec(msg, fout, tag, val);
  }
  else {
    ssi_error(ET_FORMAT, directive);
  }
}

/*****************************************************************************/

/*
 * FUNCTION NAME:	do_config
 * DESCRIPTION:		address SSI config format
 * ARGUMENTS:		message, tag, val
 * RETURN VALUE:	none
 */
static void
do_config(msg, tag, val)
	register Message *	msg;
	const char *		tag;
	const char *		val;
{
  if (ws_strcmp(tag, "timefmt") == 0) {
    strlcpy(timefmt, val, TIMEFMT_LEN);
  }
  else if (ws_strcmp(tag, "sizefmt") == 0) {
    if (ws_strcmp(val, "bytes") == 0) {
      sizefmt = SZ_BYTES;
    }
    else if (ws_strcmp(val, "abbrev") == 0) {
      sizefmt = SZ_ABBREV;
    }
    else {
      ssi_error(ET_VALUE, val);
    }
  }
  else {
    ssi_error(ET_TAG, tag);
  }
}

/*
 * FUNCTION NAME:	do_include
 * DESCRIPTION:		address SSI include format
 * ARGUMENTS:		message, output file pointer, tag, val
 * RETURN VALUE:	none
 */
static void
do_include(msg, fout, tag, val)
	register Message *	msg;
	register FILE *		fout;
	const char *		tag;
	const char *		val;
{
  char *	fpath;
  
  if (get_file_path(msg, tag, val, &fpath) == 0) {
    register FILE *	fin;

    fin = ws_safe_fopen(fpath, "r");	free((char *)fpath);
    if (fin != NULL) {
      char	buf[BUF_SIZE];

      while (!feof(fin)) {
	size_t	size = fread(buf, sizeof(char), (size_t)sizeof(buf), fin);
	fwrite(buf, sizeof(char), size, fout);
      }
      fclose(fin);
    }
  }
}

/*
 * FUNCTION NAME:	do_echo
 * DESCRIPTION:		address SSI echo format
 * ARGUMENTS:		message, output file pointer, tag, val
 * RETURN VALUE:	none
 */
static void
do_echo(msg, fout, tag, val)
	register Message *	msg;
	register FILE *		fout;
	const char *		tag;
	const char *		val;
{
  time_t	now;

  if (ws_strcmp(tag, "var") != 0) {
    ssi_error(ET_TAG, tag);
    return;
  }

  time(&now);	/* set current time in cur_time */
  if (ws_strcmp(val, "DOCUMENT_NAME") == 0) {
    /* the current filename */
    char *	p = (char *)strrchr(msg->file.path, '/');
    if (p != NULL) {
      fputs(p + 1, fout);
    }
  }
  else if (ws_strcmp(val, "DOCUMENT_URI") == 0) {
    /* the virtual path to this document (such as /docs/tutorials/foo.shtml) */
    fputs(msg->request.uri, fout);
  }
  else if (ws_strcmp(val, "QUERY_STRING_UNESCAPED") == 0) {
    if (msg->request.query_string != NULL) {
      fputs(msg->request.query_string, fout);
    }
  }
  else if (ws_strcmp(val, "DATE_LOCAL") == 0) {
    /* the current date, local time zone */
    print_formatted_time(msg, fout, now, false);	/* localtime */
  }
  else if (ws_strcmp(val, "DATE_GMT") == 0) {
    /* the current date in Greenwich mean  time */
    print_formatted_time(msg, fout, now, true);	/* Greenwich */
  }
  else if (ws_strcmp(val, "LAST_MODIFIED") == 0) {
    struct stat	sbuf;

    if (lstat(msg->file.path, &sbuf) == 0) {
      /* the last modification date of the current document */
      print_formatted_time(msg, fout, sbuf.st_mtime, false);	/* localtime */
    }
  }
}

/*
 * FUNCTION NAME:	do_fsize
 * DESCRIPTION:		address SSI fsize format
 * ARGUMENTS:		message, output file pointer, tag, val
 * RETURN VALUE:	none
 */
static void
do_fsize(msg, fout, tag, val)
	register Message *	msg;
	register FILE *		fout;
	const char *		tag;
	const char *		val;
{
  char *	fpath;

  if (get_file_path(msg, tag, val, &fpath) == 0) {
    struct stat	sbuf;
    int		rc;

    rc = lstat(fpath, &sbuf);	free((char *)fpath);
    if (rc == -1) {
      wv_syslog(WLL_LOG_ERR, "[%s] failed to get file status: %s",
		WM_MODULE_ID, fpath);
      return;
    }

    switch (sizefmt) {
      case SZ_BYTES:
	fprintf(fout, "%ld", (long)sbuf.st_size);
	break;
      case SZ_ABBREV:
	if (sbuf.st_size < 1024) {
	  fprintf(fout, "%ld", (long)sbuf.st_size);
	}
	else if (sbuf.st_size < 1024 * 1024) {
	  fprintf(fout, "%ld.%ldK", (long)sbuf.st_size / U_KILO,
		  ((long)sbuf.st_size % U_KILO) * 10 / U_KILO);
	}
	else if (sbuf.st_size < 1024 * 1024 * 1024) {
	  fprintf(fout, "%ld.%ldM", (long)sbuf.st_size / U_MEGA,
		  ((long)sbuf.st_size % U_MEGA) * 10 / U_MEGA);
	}
	else {
	  fprintf(fout, "%ld.%ldG", (long)sbuf.st_size / U_GIGA,
		  ((long)sbuf.st_size % U_GIGA) * 10 / U_GIGA);
	}
      default:
	break;
    }
  }
}

/*
 * FUNCTION NAME:	do_flastmod
 * DESCRIPTION:		address SSI flastmod format
 * ARGUMENTS:		message, output file pointer, tag, val
 * RETURN VALUE:	none
 */
static void
do_flastmod(msg, fout, tag, val)
	register Message *	msg;
	register FILE *		fout;
	const char *		tag;
	const char *		val;
{
  char *	fpath;

  if (get_file_path(msg, tag, val, &fpath) == 0) {
    struct stat	sbuf;
    int		rc;

    rc = lstat(fpath, &sbuf);	free((char *)fpath);
    if (rc == -1) {
      wv_syslog(WLL_LOG_ERR, "[%s] failed to get file status: %s",
		WM_MODULE_ID, fpath);
    }
    else {
      print_formatted_time(msg, fout, sbuf.st_mtime, false);	/* localtime */
    }
  }
}

/*
 * FUNCTION NAME:	do_exec
 * DESCRIPTION:		address SSI exec format
 * ARGUMENTS:		message, output file pointer, tag, val
 * RETURN VALUE:	none
 */
static void
do_exec(msg, fout, tag, val)
	register Message *	msg;
	register FILE *		fout;
	const char *		tag;
	const char *		val;
{
  if (ws_strcmp(tag, "cmd") == 0) {
    ssi_run_program(msg, fout, val);
  }
#if	0
  else if (ws_strcmp(tag, "cgi") == 0) {
    /* not implemented... */
  }
#endif
  else {
    ssi_error(ET_TAG, tag);
  }
}

/*
 * FUNCTION NAME:	ssi_run_program
 * DESCRIPTION:		execute a program for SSI's cmd
 * ARGUMENTS:		message, output file pointer, command
 * RETURN VALUE:	none
 */
#define	READ_FD		(0)	/* descriptor to read */
#define	WRITE_FD	(1)	/* descriptor to write */
#define	MAX_ARGS	(5)	/* maximum number of arguments for program */
static void
ssi_run_program(msg, fout, cmd)
	register Message *	msg;
	register FILE *		fout;
	const char *		cmd;
{
  char *	str = (char *)cmd;
  char *	args[MAX_ARGS];
  char **	ap;
  int		c;
  char *	cwd;	/* command working directory */
  unsigned int	timeout;
  /*
   * shell meta characters
   * \134 = backslash
   * \042 = double quotation
   * \012 = new line
   * \015 = carriage return
   */
  const char	sh_meta[] = "&;`'\134\042|*?~<>^()[]{}$\012\015";

  /* Does cmd include shell meta characters? */
  if (strpbrk(cmd, sh_meta) != NULL) {
    wv_syslog(WLL_LOG_SECURITY,
	     "[%s] unsecure/untrusted command string: #cmd='%s'",
	     WM_MODULE_ID, cmd);
    return;
  }

  /* parsing arguments */
  for (c = 0, ap = args; (*ap = wu_strsep1(&str, " \t\r\n")) != NULL; c++) {
    args[c] = *ap;
    if (**ap != EOL) {
      if (++ap >= &args[MAX_ARGS]) {
	wv_syslog(WLL_LOG_ERR,
		 "[%s] unexpected too many arguments: #cmd='%s'",
		 WM_MODULE_ID, cmd);
	return;	/* error: arguments overflow */
      }
    }
  }
  args[c] = EOL;

  /* set working directory */
  cwd = wu_dirname(msg->file.path);
  if (cwd == NULL) {
    wv_syslog(WLL_LOG_ERR, "[%s] failed to change directory", WM_MODULE_ID);
    return;
  }

  /* timeout */
  timeout = msg->vhost->timeout > 0 ? msg->vhost->timeout : SSI_TIMEOUT;

  /* execute a program */
  fflush(fout);
  if (wt_spawn(cwd, args[0], args, msg->cgi.env, NULL,
	       -1, fileno(fout), fileno(fout),
	       WT_DEFAULT_CPU_LIMIT, WT_DEFAULT_FSIZE_LIMIT,
	       WT_UTIME_SEC(timeout)) == -1) {
    wv_syslog(WLL_LOG_ERR, "[%s] failed to exec: #cmd='%s'",
	      WM_MODULE_ID, args[0]);
  }
}

/*****************************************************************************/

/*
 * FUNCTION NAME:	print_formatted_time
 * DESCRIPTION:		print the formatted time
 * ARGUMENTS:		message, output file pointer,
 *			time, gmt flag (true: GMT, false: Localtime)
 * RETURN VALUE:	none
 */
static void
print_formatted_time(msg, fout, t, gmt)
	register Message *	msg;
	register FILE *		fout;
	time_t			t;
	bool			gmt;
{
  struct tm *	tp;
  char		tbuf[128];

  tp = gmt ? gmtime(&t) : localtime(&t);
  if (strftime(tbuf, sizeof(tbuf), timefmt, tp) > 0) {
    fputs(tbuf, fout);
  }
}

/*
 * FUNCTION NAME:	get_file_path
 * DESCRIPTION:		get the filepath
 * ARGUMENTS:		message, tag, val, fpath
 * RETURN VALUE:	successful: 0, otherwise: -1
 */
static int
get_file_path(msg, tag, val, fpath)
	register Message *	msg;
	const char *		tag;
	const char *		val;
	char **			fpath;	/* return value */
{
  if (ws_strcmp(tag, "file") == 0) {
    char *	tmp;

    if (val[0] == '/' || val[0] == '.') {
      goto invalid;
    }

    tmp = wu_dirname(msg->file.path);
    if (tmp != NULL) {
      ws_asprintf(fpath, "%s/%s", tmp, val);
    }
  }
  else if (ws_strcmp(tag, "virtual") == 0) {
    if (ws_strcmp(val, WYVERN_USER_AUTH) == 0 ||
     	ws_strcmp(val, WYVERN_USER_PASSWD) == 0) {
      goto invalid;
    }
    if (msg->user.homeDir != NULL) {
      ws_asprintf(fpath, "%s/%s", msg->user.homeDir, val);
    }
    else {
      ws_asprintf(fpath, "%s/%s", msg->vhost->documentRoot, val);
    }
  }
  else {
    return ssi_error(ET_TAG, tag);
  }

  return 0;

invalid:
  wv_syslog(WLL_LOG_SECURITY, "[%s] invalid file access: '%s'",
	    WM_MODULE_ID, val);
  wv_syslog(WLL_LOG_ERR, "[%s] failed to get file path: %s=%s",
	    WM_MODULE_ID, tag, val);
  return -1;
}

/*****************************************************************************/

/*
 * FUNCTION NAME:	ssi_error
 * DESCRIPTION:		unsupported SSI tag/value/format
 * ARGUMENTS:		type, name
 * RETURN VALUE:	always: -1
 */
static int
ssi_error(type, name)
	error_ssi_t	type;
	const char *	name;
{
  switch (type) {
    case ET_TAG:
      wv_syslog(WLL_LOG_WARNING, "[%s] unsupported tag: %s",
		WM_MODULE_ID, name);
      break;
    case ET_VALUE:
      wv_syslog(WLL_LOG_WARNING, "[%s] unsupported value: %s",
		WM_MODULE_ID, name);
      break;
    case ET_FORMAT:
      wv_syslog(WLL_LOG_WARNING, "[%s] unsupported format: #%s",
		WM_MODULE_ID, name);
    default:
      break;
  }

  return -1;
}

/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/
static void
wm_ssi_register_hooks(void)
{
  wm_hook_type_checker(wm_ssi_type_checker, HOOK_MIDDLE);
  wm_hook_contents_filter(wm_ssi_filter, HOOK_FIRST, MAGIC_TYPE_SSI);
}

static wmModule	mod_ssi = {
  WM_MAGIC_NUMBER_MAJOR,	/* version */
  WM_UNSECURE,			/* unsecure */
  WM_MODULE_ID,			/* name */
  "SSI Module version 1.1.10",	/* info */
  "Kouichi ABE (WALL)",		/* author */
  "mod_env",			/* depend */
  NULL,				/* handle */
  NULL,				/* init */
  NULL,				/* finish */
  (void *)wm_ssi_register_hooks	/* register_handler */
};

wmModule *
wmModuleInfo(void)
{
  return &mod_ssi;
}
