/*****************************************************************************
 *
 * FILE:	cond.c
 * DESCRIPTION:	Wyvern: addressing the conditional request
 * DATE:	Wed, Dec 15 1999
 * UPDATED:	Wed, Sep 12 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 1999-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 1999-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: cond.c,v 1.16 2007/09/12 02:24:27 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <ctype.h>
#include "wyvern.h"

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static int	if_modified_since(Message *);
static int	if_unmodified_since(Message *);
static int	if_match(Message *);
static int	if_none_match(Message *);
static int	if_range(Message *);
static int	set_content_location(Message *);
static int	parse_range(Message *);

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wv_conditional_request_checker
 * Description:		perform the conditional request
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
wv_conditional_request_checker(msg)
	register Message *	msg;
{
  if (parse_range(msg) == -1 ||
      if_range(msg) == -1 ||
      if_modified_since(msg)   == -1 || if_none_match(msg) == -1 ||
      if_unmodified_since(msg) == -1 || if_match(msg)      == -1) {
    return -1;
  }

  return 0;
}

/*****************************************************************************/

/*
 * Function Name:	if_modified_since
 * Description:		perform the conditional request with If-Modified-Since
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
static int
if_modified_since(msg)
	register Message *	msg;
{
  register const char *	p;

  if ((p = wa_get(msg->request.header, "if-modified-since")) != NULL) {
    time_t	ims;

    ims = wv_convert_date(p);
    if (ims == (time_t)0) {
      msg->response.status = STATUS_FORBIDDEN;
      return -1;
    }
    if (msg->file.status.st_mtime <= ims) {
      msg->response.status = STATUS_NOT_MODIFIED;
      set_content_location(msg);
    }
  }

  return 0;
}

/*
 * Function Name:	if_unmodified_since
 * Description:		perform the conditional request with If-Unmodified-Since
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
static int
if_unmodified_since(msg)
	register Message *	msg;
{
  register const char *	p;

  if ((p = wa_get(msg->request.header, "if-unmodified-since")) != NULL) {
    time_t	ius;

    ius = wv_convert_date(p);
    if (ius == (time_t)0) {
      msg->response.status = STATUS_FORBIDDEN;
      return -1;
    }
    if (msg->file.status.st_mtime > ius) {
      msg->response.status = STATUS_NOT_MODIFIED;
      set_content_location(msg);
    }
  }

  return 0;
}

/*
 * Function Name:	if_match
 * Description:		perform the conditional request with If-Match
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
static int
if_match(msg)
	register Message *	msg;
{
  register const char *	p;

  if ((p = wa_get(msg->request.header, "if-match")) != NULL) {
    if (ws_strcmp(p, wa_get(msg->request.header, "etag")) != 0 &&
	ws_strcmp(p, "*") != 0) {
      msg->response.status = STATUS_PRECONDITION_FAILED;
      return -1;
    }
  }
  return 0;
}

/*
 * Function Name:	if_none_match
 * Description:		check the conditional request with If-None-Match
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
static int
if_none_match(msg)
	register Message *	msg;
{
  register const char *	p;

  if ((p = wa_get(msg->request.header, "if-none-match")) != NULL) {
    if (ws_strcmp(p, wa_get(msg->request.header, "etag")) == 0 ||
	ws_strcmp(p, "*") == 0) {
      return -1;
    }
    msg->response.status = STATUS_NOT_MODIFIED;
  }
  return 0;
}

/*
 * Function Name:	if_range
 * Description:		perform the conditional request with If-Range
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
static int
if_range(msg)
	register Message *	msg;
{
  register const char *	p;

  if ((p = wa_get(msg->request.header, "if-range")) != NULL) {
    /* First, check the entity-tag */
    if (ws_strcmp(p, wa_get(msg->request.header, "etag")) == 0) {
      return 0;
    }
    /* Second, check the HTTP-date */
    else {
      time_t	t;

      t = wv_convert_date(p);
      if (msg->file.status.st_mtime == t) {
	return 0;
      }
    }
    msg->response.status = STATUS_REQUESTED_RANGE_NOT_SATISFIABLE;
    return -1;
  }

  return 0;
}

/*
 * Function Name:	set_content_location
 * Description:		set Content-Location field
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
static int
set_content_location(msg)
	register Message *	msg;
{
  if (msg->request.http.version == HTTP_11) {
    register const char *	host;
    char *			location;

    host = wa_get(msg->request.header, "host");
#if	ENABLE_SSL
    if (msg->soci.secure == SOCKET_SECURE) {
      ws_asprintf(&location, "https://%s%s", host, msg->request.uri);
    }
    else {
      ws_asprintf(&location, "http://%s%s", host, msg->request.uri);
    }
#else
    ws_asprintf(&location, "http://%s%s", host, msg->request.uri);
#endif	/* ENABLE_SSL */
    wa_put(msg->response.header, "Content-Location", location);
    free(location);
  }

  return 0;
}

/*
 * Function Name:	parse_range
 * Description:		parse Range header
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
static int
parse_range(msg)
	register Message *	msg;
{
  register const char *	range;
  register int		idx;
  unsigned int		num;
  register char *	p;
  char			buf[10];

  if ((range = wa_get(msg->request.header, "range")) == NULL) {
    return 0;	/* No Range header */
  }

  num = 1;
  p   = (char *)range;

  /* count range items */
  while (*p != EOL) {
    if (*p == ',') {
      num++;
    }
    p++;
  }

  p = strchr((char *)range, '=');
  if (p == NULL) {	/* Invalid Range format. No 'bytes=' string */
    goto error0;
  }
  p++;

  /* make range array */
  msg->file.ranges = ws_malloc(num * sizeof(Range));
  msg->file.ritems = num;

  for (idx = 0; *p != EOL; ) {
    if (isdigit(*p)) {		/* Found start range */
      register int	i;

      for (i = 0; isdigit(*p); i++) {
	buf[i] = *p++;
      }
      buf[i] = EOL;
      msg->file.ranges[idx].start = (size_t)strtol(buf, NULL, 10);

      if (*p++ != '-') {
	goto error;
      }
      if (isdigit(*p)) {
	for (i = 0; isdigit(*p); i++) {
	  buf[i] = *p++;
	}
	buf[i] = EOL;
	msg->file.ranges[idx].end = (size_t)strtol(buf, NULL, 10);
      }
      else {
	msg->file.ranges[idx].end = msg->file.status.st_size - 1;
      }
      idx++;
    }
    else if (*p == '-') {	/* No start range */
      register int	i;

      p++;
      if (!isdigit(*p)) {
	goto error;
      }

      for (i = 0; isdigit(*p); i++) {
	buf[i] = *p++;
      }
      buf[i] = EOL;
      i = (size_t)strtol(buf, NULL, 10);
      msg->file.ranges[idx].start = msg->file.status.st_size - i - 1;
      msg->file.ranges[idx].end	  = msg->file.status.st_size - 1;
      idx++;
    }
    else {
      p++;
    }
  }
  msg->response.status = STATUS_PARTIAL_CONTENT;

  return 0;

error:
  wv_free(msg->file.ranges);
  msg->file.ritems = 0;
error0:
  msg->response.status = STATUS_REQUESTED_RANGE_NOT_SATISFIABLE;

  return -1;
}
