/*****************************************************************************
 *
 * FILE:	config.c
 * DESCRIPTION:	Wyvern: read configuration file
 * DATE:	Wed, Sep  8 1999
 * UPDATED:	Sun, Aug 19 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 1999-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 1999-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: config.c,v 1.29 2007/08/19 06:45:31 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <pwd.h>
#include <grp.h>
#include <errno.h>
#include "wyvern.h"

/*****************************************************************************
 *
 *	Local functions definition
 *
 *****************************************************************************/
static void	active_parser(const char *, int (*)(FILE *));
static int	parse_mime(FILE *);
static int	parse(FILE *);
static void	parse_config(sxml_node_t *);
static void	parse_listen_socket(sxml_node_t *);
static void	parse_security(sxml_node_t *);
#if	ENABLE_SSL
static void	parse_ssl(sxml_node_t *);
#endif	/* ENABLE_SSL */
static void	parse_limits(sxml_node_t *);
static void	parse_paths(sxml_node_t *);
static void	parse_module(sxml_node_t *);
static void	parse_host(sxml_node_t *, const char *);
static void	parse_add_type(sxml_node_t *);
static void	parse_error_document(sxml_node_t *);
static void	parse_map(sxml_node_t *, const char *);
static void	parse_error(const char *);
static int	c_get_int(const char *);
static void	c_set_string(char **, const char *);

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wv_init_httpd
 * Description:		initialize server entity
 * Arguments:		none
 * Return Value:	httpd entity
 */
void
wv_init_httpd(void)
{
  httpd = (Httpd *)ws_malloc(sizeof(Httpd));	/* leak */
  /* initialized by default values */
  httpd->listen		  = NULL;			/* Listen */
  httpd->maxThread	  = DEF_MAX_THREAD;		/* MaxThread */
  httpd->maxRequest	  = DEF_MAX_REQUEST;		/* MaxRequest */
  httpd->maxPostSize	  = DEF_MAX_POST_SIZE;		/* MaxPostSize */
  httpd->user		  = DEF_USER;			/* User */
  httpd->group		  = DEF_GROUP;			/* Group */
  httpd->secureLevel	  = DEF_SECURE_LEVEL;		/* SecureLevel */
  httpd->securityLog.path = ws_strdup(DEF_SECURITY_LOG);/* SecurityLog */
  httpd->securityLog.fp	  = NULL;
  httpd->pidFile	  = ws_strdup(DEF_PID_FILE);	/* PidFile */
  httpd->pid		  = -1;				/* Pid */
  httpd->serverRoot	  = ws_strdup(DEF_SERVER_ROOT);	/* ServerRoot */
  httpd->tmpDir		  = ws_strdup(DEF_TMP_DIR);	/* TmpDir */
  httpd->cacheDir	  = ws_strdup(DEF_CACHE_DIR);	/* CacheDir */
  httpd->configDir	  = ws_strdup(WYVERN_CONF_DIR);	/* ConfigDir */
  httpd->databaseDir	  = ws_strdup(WYVERN_DB_DIR);	/* DatabaseDir */
  httpd->typesConfig	  = ws_strdup(DEF_TYPES_CONFIG);/* TypesConfig */
#if	ENABLE_SSL
  httpd->sslCertificateFile	= ws_strdup(SSL_SERVER_CRT);
  httpd->sslCertificateKeyFile	= ws_strdup(SSL_SERVER_KEY);
  httpd->sslCipherList		= NULL;
#endif	/* ENABLE_SSL */
  /**/
  httpd->vhostTable	  = NULL;		/* virtual hosts manager */
  httpd->host		  = wv_virtual_host_new();	/* main server host */
  /* initialize main server host entity */
  httpd->host->errorLog.path	= ws_strdup(DEF_ERROR_LOG);
  httpd->host->errorLog.fp	= stderr;
  httpd->host->accessLog.path	= ws_strdup(DEF_ACCESS_LOG);
  httpd->host->accessLog.fp	= NULL;
  httpd->host->documentRoot	= ws_strdup(DEF_DOC_ROOT);
  httpd->host->next		= NULL;
  httpd->vhostTable		= NULL;	/* reset virtual host table */
  /**/
  httpd->vhost			= httpd->host;	/* set current virtual host */
  /**/
  httpd->sockets		= NULL;
  /**/
  httpd->sighup			= false;
  /**/
  httpd->bootTime.tv_sec	= (long)0;	/* The 0 shows no boot */
  httpd->bootTime.tv_usec	= (long)0;	/* The 0 shows no boot */
}

/*
 * Function Name:	wv_read_config_file
 * Description:		read the configuration file via yyparse.
 * Arguments:		configuration file
 * Return Value:	none
 */
void
wv_read_config_file(configFile)
	const char *	configFile;
{
  /* confirm config file whether it is valid or not */
  switch (configFile[0]) {
    case EOL: configFile = DEF_WYVERN_CONFIG;	/* FALLTHROUGH */
    case '/': break;
     default: wv_die("No absolute path: %s", configFile);
  }

  /* open config file */
  active_parser(configFile, parse);

  /* open mime.types file */
  active_parser(httpd->typesConfig, parse_mime);

  /* verify configuration */
  if (httpd->host->serverName == NULL) {
    wv_die("No ServerName in %s", configFile);
  }
  if (httpd->host->serverAdmin == NULL) {
    wv_die("No ServerAdmin in %s", configFile);
  }

  /* useful trick */
  httpd->host->next = httpd->vhostTable;
}

static void
active_parser(fname, parser)
	const char *	fname;
	int		(*parser)(FILE *);
{
  register FILE *	fin;

  fin = ws_safe_fopen(fname, "r");
  if (fin == NULL) {
    switch (errno) {
      case ENOENT: wv_die("Not exist: %s", fname);
      case EACCES: wv_die("Failed to access %s", fname);
      case EMLINK: wv_die("Untrusted symbolic link: %s", fname);
	  default: wv_die("Failed to open %s", fname);
    }
  }
  if ((*parser)(fin) == -1) {
    wv_die("Failed to parse %s", fname);
  }
  fclose(fin);
}

/* media-type file parser */
#define	PARSE_TOKEN	" \t\r\n"
static int
parse_mime(fin)
	register FILE *	fin;
{
  char	buf[BUFSIZ];

  memset(buf, 0, sizeof(buf));
  while (fgets(buf, sizeof(buf), fin) != NULL) {
    const char *	m = buf;	/* media-type */
    if (*m != ' ' && *m != '\t' && *m != '#' && *m != '\n') {
      register char *	s;		/* suffix */

      s = strpbrk(buf, PARSE_TOKEN);
      if (s != NULL) {
	*s = '\0';
	for (++s; *s == ' ' || *s == '\t'; s++) {
	  ;
	}
	while ((s = strtok(s, PARSE_TOKEN)) != NULL) {
	  wv_add_media_types(s, m);
	  s = NULL;
	}
      }
    }
    memset(buf, 0, sizeof(buf));
  }

  return 0;
}

/*****************************************************************************/

static int
parse(fin)
	register FILE *	fin;
{
  sxml_node_t *	root;

  root = sxml_parse_file(fileno(fin));
  if (root != NULL) {
    parse_config(root);
    sxml_delete_node(root);
    return 0;
  }

  return -1;
}

static void
parse_config(root)
	sxml_node_t *	root;
{
  sxml_node_t *	node;

  node = sxml_find_element(root, "wyvern", NULL, NULL);
  if (node != NULL) {
    register sxml_node_t *	np;

    for (np = node->child; np != NULL; np = np->next) {
      if (np->type != SXML_ELEMENT)	  { continue; }
      if (np->value.element.name == NULL) { continue; }

      if (ws_strcmp(np->value.element.name, "listen-socket") == 0) {
	parse_listen_socket(np->child);
      }
      else if (ws_strcmp(np->value.element.name, "security") == 0) {
	parse_security(np->child);
      }
#if	ENABLE_SSL
      else if (ws_strcmp(np->value.element.name, "ssl") == 0) {
	parse_ssl(np->child);
      }
#endif	/* ENABLE_SSL */
      else if (ws_strcmp(np->value.element.name, "limits") == 0) {
	parse_limits(np->child);
      }
      else if (ws_strcmp(np->value.element.name, "paths") == 0) {
	parse_paths(np->child);
      }
      else if (ws_strcmp(np->value.element.name, "module") == 0) {
	parse_module(np->child);
      }
      else if (ws_strcmp(np->value.element.name, "host") == 0) {
	parse_host(np->child, sxml_get_attribute(np, "type"));
      }
      else {
	parse_error(np->value.element.name);
      }
    }
  }
}

static void
parse_listen_socket(node)
	sxml_node_t *	node;
{
  if (node != NULL) {
    register sxml_node_t *	np;
    const char *		args[3] = { NULL, NULL, NULL };

    for (np = node; np != NULL; np = np->next) {
      const char *	content;

      if (np->type != SXML_ELEMENT)	  { continue; }
      if (np->value.element.name == NULL) { continue; }

      content = sxml_get_content(np->child);
      if (ws_strcmp(np->value.element.name, "port") == 0) {
	args[0] = content;
      }
#if	ENABLE_SSL
      else if (ws_strcmp(np->value.element.name, "ssl-port") == 0) {
	args[1] = content;
      }
#endif	/* ENABLE_SSL */
      else if (ws_strcmp(np->value.element.name, "bind-address") == 0) {
	args[2] = content;
      }
      else {
	parse_error(np->value.element.name);
      }
    }
    if (args[0] != NULL) {
      wv_add_sockets(args[2], args[0], SOCKET_UNSECURE);
    }
    if (args[1] != NULL) {
      wv_add_sockets(args[2], args[1], SOCKET_SECURE);
    }
  }
}

static void
parse_security(node)
	sxml_node_t *	node;
{
  if (node != NULL) {
    register sxml_node_t *	np;

    for (np = node; np != NULL; np = np->next) {
      const char *	content;

      if (np->type != SXML_ELEMENT)	  { continue; }
      if (np->value.element.name == NULL) { continue; }

      content = sxml_get_content(np->child);
      if (ws_strcmp(np->value.element.name, "change-owner") == 0) {
	register sxml_node_t *	cp;

	for (cp = np->child; cp != NULL; cp = cp->next) {
	  const char *	id = sxml_get_content(cp->child);

	  if (ws_strcmp(cp->value.element.name, "user") == 0) {
	    if (id != NULL) {
	      struct passwd *	pbuf;
	      char *		endp;
	      long		val;
	      
	      val = strtol(id, &endp, 10);
	      if (*id != EOL && *endp == EOL) {
		pbuf = getpwuid((uid_t)val);
	      }
	      else {
		pbuf = getpwnam(id);
	      }
	      if (pbuf == NULL) {
		wv_die("No such user/uid: %s", id);
	      }
	      httpd->user = pbuf->pw_uid;
	    }
	  }
	  else if (ws_strcmp(cp->value.element.name, "group") == 0) {
	    if (id != NULL) {
	      struct group *	gbuf;
	      char *		endp;
	      long		val;
	      
	      val = strtol(id, &endp, 10);
	      if (*id != EOL && *endp == EOL) {
		gbuf = getgrgid((gid_t)val);
	      }
	      else {
		gbuf = getgrnam(id);
	      }
	      if (gbuf == NULL) {
		wv_die("No such group/gid: %s", id);
	      }
	      httpd->group = gbuf->gr_gid;
	    }
	  }
	  else {
	    parse_error(np->value.element.name);
	  }
	}
      }
      else if (ws_strcmp(np->value.element.name, "secure-level") == 0) {
	if (content != NULL) {
	  httpd->secureLevel = c_get_int(content);
	  if (httpd->secureLevel < DEF_SECURE_LEVEL ||
	      httpd->secureLevel > MAX_SECURE_LEVEL) {
	    httpd->secureLevel = MAX_SECURE_LEVEL;
	  }
	}
      }
      else if (ws_strcmp(np->value.element.name, "security-log") == 0) {
	c_set_string(&httpd->securityLog.path, content);
      }
      else {
	parse_error(np->value.element.name);
      }
    }
  }
}

#if	ENABLE_SSL
static void
parse_ssl(node)
	sxml_node_t *	node;
{
  if (node != NULL) {
    register sxml_node_t *	np;

    for (np = node; np != NULL; np = np->next) {
      const char *	content;

      if (np->type != SXML_ELEMENT)	  { continue; }
      if (np->value.element.name == NULL) { continue; }

      content = sxml_get_content(np->child);
      if (ws_strcmp(np->value.element.name, "certificate-file") == 0) {
	c_set_string(&httpd->sslCertificateFile, content);
      }
      else if (ws_strcmp(np->value.element.name, "certificate-key-file") == 0) {
	c_set_string(&httpd->sslCertificateKeyFile, content);
      }
      else if (ws_strcmp(np->value.element.name, "cipher-list") == 0) {
	c_set_string(&httpd->sslCipherList, content);
      }
      else {
	parse_error(np->value.element.name);
      }
    }
  }
}
#endif	/* ENABLE_SSL */

static void
parse_limits(node)
	sxml_node_t *	node;
{
  if (node != NULL) {
    register sxml_node_t *	np;

    for (np = node; np != NULL; np = np->next) {
      const char *	content;

      if (np->type != SXML_ELEMENT)	  { continue; }
      if (np->value.element.name == NULL) { continue; }

      content = sxml_get_content(np->child);
      if (ws_strcmp(np->value.element.name, "max-thread") == 0) {
	if (content != NULL) {
	  int	n;

	  n = c_get_int(content);
	  httpd->maxThread = n <= 0 ? DEF_MIN_THREAD : n;
	}
      }
      else if (ws_strcmp(np->value.element.name, "max-request") == 0) {
	if (content != NULL) {
	  int	n;

	  n = c_get_int(content);
	  httpd->maxRequest = n <= 0 ? DEF_MIN_REQUEST : n;
	}
      }
      else if (ws_strcmp(np->value.element.name, "max-post-size") == 0) {
	if (content != NULL) {
	  int	n;

	  n = c_get_int(content);
	  httpd->maxPostSize = n <= 0 ? DEF_MAX_POST_SIZE : n * 1024 * 1024;
	}
      }
      else if (ws_strcmp(np->value.element.name, "timeout") == 0) {
	if (content != NULL) {
	  int	n;

	  n = c_get_int(content);
	  httpd->vhost->timeout = n <= 0 ? DEF_TIMEOUT : n;
	}
      }
      else {
	parse_error(np->value.element.name);
      }
    }
  }
}

static void
parse_paths(node)
	sxml_node_t *	node;
{
  if (node != NULL) {
    register sxml_node_t *	np;

    for (np = node; np != NULL; np = np->next) {
      const char *	content;

      if (np->type != SXML_ELEMENT)	  { continue; }
      if (np->value.element.name == NULL) { continue; }

      content = sxml_get_content(np->child);
      if (ws_strcmp(np->value.element.name, "server-root") == 0) {
	c_set_string(&httpd->serverRoot, content);
      }
      else if (ws_strcmp(np->value.element.name, "pid-file") == 0) {
	c_set_string(&httpd->pidFile, content);
      }
      else if (ws_strcmp(np->value.element.name, "types-config") == 0) {
	c_set_string(&httpd->typesConfig, content);
      }
      else if (ws_strcmp(np->value.element.name, "tmp-dir") == 0) {
      	c_set_string(&httpd->tmpDir, content);
      }
      else if (ws_strcmp(np->value.element.name, "cache-dir") == 0) {
	c_set_string(&httpd->cacheDir, content);
      }
      else if (ws_strcmp(np->value.element.name, "config-dir") == 0) {
	c_set_string(&httpd->configDir, content);
      }
      else if (ws_strcmp(np->value.element.name, "database-dir") == 0) {
	c_set_string(&httpd->databaseDir, content);
      }
      else {
	parse_error(np->value.element.name);
      }
    }
  }
}

static void
parse_module(node)
	sxml_node_t *	node;
{
  if (node != NULL) {
    register sxml_node_t *	np;

    for (np = node; np != NULL; np = np->next) {
      if (np->type != SXML_ELEMENT)	  { continue; }
      if (np->value.element.name == NULL) { continue; }

      if (ws_strcmp(np->value.element.name, "path") == 0) {
	const char *	path;

	path = sxml_get_content(np->child);
	if (path != NULL) {
	  wv_load_module(path);
	}
      }
      else {
	parse_error(np->value.element.name);
      }
    }
  }
}

static void
parse_host(node, type)
	sxml_node_t *	node;
	const char *	type;	/* host type */
{
  if (node != NULL) {
    register sxml_node_t *	np;

    if (ws_strcmp(type, "virtual") == 0) {
      httpd->vhost = wv_virtual_host_new();
    }

    for (np = node; np != NULL; np = np->next) {
      const char *	content;

      if (np->type != SXML_ELEMENT)	  { continue; }
      if (np->value.element.name == NULL) { continue; }

      content = sxml_get_content(np->child);
      if (ws_strcmp(np->value.element.name, "server-name") == 0) {
	c_set_string(&httpd->vhost->serverName, content);
      }
      else if (ws_strcmp(np->value.element.name, "server-admin") == 0) {
	c_set_string(&httpd->vhost->serverAdmin, content);
      }
      else if (ws_strcmp(np->value.element.name, "error-log") == 0) {
	c_set_string(&httpd->vhost->errorLog.path, content);
      }
      else if (ws_strcmp(np->value.element.name, "access-log") == 0) {
	c_set_string(&httpd->vhost->accessLog.path, content);
      }
      else if (ws_strcmp(np->value.element.name, "document-root") == 0) {
	c_set_string(&httpd->vhost->documentRoot, content);
      }
      else if (ws_strcmp(np->value.element.name, "user-dir") == 0) {
	c_set_string(&httpd->vhost->userDir, content);
      }
      else if (ws_strcmp(np->value.element.name, "directory-indexes") == 0) {
	httpd->vhost->directoryIndexes = wv_add_indexes((char *)content);
      }
      else if (ws_strcmp(np->value.element.name, "auto-indexing") == 0) {
	if (ws_strcasecmp(content, "on") == 0) {
	  httpd->vhost->autoIndexing = true;
	}
	else {
	  httpd->vhost->autoIndexing = false;
	}
      }
      else if (ws_strcmp(np->value.element.name, "follow-symlinks") == 0) {
	if (ws_strcasecmp(content, "on") == 0) {
	  httpd->vhost->followSymlinks = true;
	}
	else {
	  httpd->vhost->followSymlinks = false;
	}
      }
      else if (ws_strcmp(np->value.element.name, "default-type") == 0) {
	c_set_string(&httpd->vhost->defaultType, content);
      }
      else if (ws_strcmp(np->value.element.name, "add-type") == 0) {
	parse_add_type(np->child);
      }
      else if (ws_strcmp(np->value.element.name, "error-document") == 0) {
	parse_error_document(np->child);
      }
      else if (ws_strcmp(np->value.element.name, "map") == 0) {
	parse_map(np->child, sxml_get_attribute(np, "type"));
      }
      else {
	parse_error(np->value.element.name);
      }
    }

    if (ws_strcmp(type, "virtual") == 0) {
      httpd->vhost = httpd->host;
    }
  }
}

static void
parse_add_type(node)
	sxml_node_t *	node;
{
  if (node != NULL) {
    register sxml_node_t *	np;
    const char *		args[2] = { NULL, NULL };

    for (np = node; np != NULL; np = np->next) {
      const char *	content;

      if (np->type != SXML_ELEMENT)	  { continue; }
      if (np->value.element.name == NULL) { continue; }

      content = sxml_get_content(np->child);
      if (ws_strcmp(np->value.element.name, "media-type") == 0) {
	args[1] = content;
      }
      else if (ws_strcmp(np->value.element.name, "suffix") == 0) {
	args[0] = content;
      }
      else {
	parse_error(np->value.element.name);
      }
    }
    wv_add_types(&httpd->vhost->mimeTable, args[0], args[1]);
  }
}

static void
parse_error_document(node)
	sxml_node_t *	node;
{
  if (node != NULL) {
    register sxml_node_t *	np;
    const char *		args[2] = { NULL, NULL };
    errdoc_t			type;

    for (np = node; np != NULL; np = np->next) {
      const char *	content;

      if (np->type != SXML_ELEMENT)	  { continue; }
      if (np->value.element.name == NULL) { continue; }

      content = sxml_get_content(np->child);
      if (ws_strcmp(np->value.element.name, "code") == 0) {
	args[0] = content;
      }
      else if (ws_strcmp(np->value.element.name, "string") == 0) {
	args[1] = content;
	type	= ED_STRING;
      }
      else if (ws_strcmp(np->value.element.name, "file") == 0) {
	args[1] = content;
	type	= ED_FILE;
      }
      else if (ws_strcmp(np->value.element.name, "uri") == 0) {
	args[1] = content;
	type	= ED_URL;
      }
      else {
	parse_error(np->value.element.name);
      }
    }
    if (args[0] != NULL && args[1] != NULL) {
      int	code;

      code = c_get_int(args[0]);
      wv_add_error_documents(&httpd->vhost->errorDocTable, code, args[1], type);
    }
  }
}

static void
parse_map(node, type)
	sxml_node_t *	node;
	const char *	type;	/* map type */
{
  if (node != NULL) {
    register sxml_node_t *	np;
    const char *		args[2] = { NULL, NULL };

    for (np = node; np != NULL; np = np->next) {
      const char *	content;

      if (np->type != SXML_ELEMENT)	  { continue; }
      if (np->value.element.name == NULL) { continue; }

      content = sxml_get_content(np->child);
      if (ws_strcmp(np->value.element.name, "path") == 0) {
	args[0] = content;
      }
      else if (ws_strcmp(np->value.element.name, "uri") == 0) {
	args[1] = content;
      }
      else {
	parse_error(np->value.element.name);
      }
    }

    if (ws_strcmp(type, "redirect") == 0) {
      wv_add_aliases(&httpd->vhost->aliasTable, args[0], args[1], A_REDIRECT);
    }
    else if (ws_strcmp(type, "alias") == 0) {
      wv_add_aliases(&httpd->vhost->aliasTable, args[0], args[1], A_ALIAS);
    }
    else if (ws_strcmp(type, "script") == 0) {
      wv_add_aliases(&httpd->vhost->aliasTable, args[0], args[1], A_SCRIPT);
    }
    else {
      wv_die("Error: Unknown map type: %s", type);
    }
  }
}

static void
parse_error(tag)
	const char *	tag;
{
  wv_die("Error: Unknown tag: %s", tag);
}

/*****************************************************************************/

static int
c_get_int(np)
	const char *	np;
{
  char *	endp;
  long		val;

  val = strtol(np, &endp, 10);
  if (!(*np != EOL && *endp == EOL)) {
    wv_die("Invalid number: %s", np);
  }

  return (int)val;
}

static void
c_set_string(d, s)
	char **		d;
	const char *	s;
{
  wv_free(*d);
  *d = ws_strdup(s);
}
