/*****************************************************************************
 *
 * FILE:	file.c
 * DESCRIPTION:	Wyvern: HEAD/GET method operation
 * DATE:	Fri, Aug 16 2002
 * UPDATED:	Sat, Aug 18 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2002-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2002-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: file.c,v 1.23 2007/08/19 06:45:31 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#include <errno.h>
#include "wyvern.h"

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static StrList *	new_index(StrList **, const char *);

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wv_add_indexes
 * Description:		register index file of DirectoryIndexes directive
 * Arguments:		indexes string (CSV)
 * Return Value:	StrList pointer
 */
StrList *
wv_add_indexes(indexes)
	char *	indexes;
{
#define	MAX_INDEXES	(10)
  StrList *	new = NULL;
  char *	s   = indexes;
  char **	ap;
  char *	args[MAX_INDEXES];
  register int	n;

  for (n = 0, ap = args; (*ap = wu_strsep1(&s, ",")) != NULL; ) {
    if (**ap != EOL) {
      n++;
      if (++ap >= &args[MAX_INDEXES]) {
	break;
      }
    }
  }

  if (n > 0) {
    register int	i;

    for (i = n; i > 0; i--) {
      new_index(&new, args[i - 1]);
    }
  }

  return new;
}

/*
 * Function Name:	new_index
 * Description:		register index file into string list
 * Arguments:		StrList, index filename
 * Return Value:	StrList pointer
 */
static StrList *
new_index(list, name)
	StrList **	list;
	const char *	name;
{
  StrList *	new;

  new	    = (StrList *)ws_malloc(sizeof(StrList));
  new->text = ws_strdup(name);
  new->next = *list;
  *list	    = new;

  return new;
}

/*****************************************************************************/

/*
 * Function Name:	wv_document_root_translater
 * Description:		translate a requests URI
 * Arguments:		message
 * Return Value:	successful: 0, no translated: 1, otherwise: -1
 */
int
wv_document_root_translater(msg)
	register Message *	msg;
{
  ws_asprintf(&msg->file.path, "%s%s",
	      msg->vhost->documentRoot, msg->request.uri);
  DBG(fprintf(stderr, "[translateDocRoot] %s\n", msg->file.path);)
  return wv_complement_path(msg);
}

/*
 * Function Name:	wv_complement_path
 * Description:		complement the request path
 * Arguments:		message
 * Return Value:	always 0
 */
int
wv_complement_path(msg)
	register Message *	msg;
{
  /* The GET,HEAD method only can access static contents.  */
  if (msg->request.method.type != M_GET && msg->request.method.type != M_HEAD) {
    msg->response.status = STATUS_METHOD_NOT_ALLOWED;
    return WM_RET_ERROR;
  }
  if (msg->request.uri[strlen(msg->request.uri) - 1] == '/') {
    register StrList *	slp;
    char *		path;	/* base path */

    path = ws_strdup(msg->file.path);	wv_free(msg->file.path);
    for (slp = msg->vhost->directoryIndexes; slp != NULL; slp = slp->next) {
      struct stat	sbuf;

      ws_asprintf(&msg->file.path, "%s%s", path, slp->text);
      if (lstat(msg->file.path, &sbuf) == 0) {
	free(path);
	DBG(fprintf(stderr, "[complementPath] %s\n", msg->file.path);)
	return WM_RET_OK;
      }
      wv_free(msg->file.path);
    }
    msg->file.path	 = (char *)path;	/* Restore original path */
    msg->file.mime	 = MAGIC_TYPE_DIR;
    msg->response.status = STATUS_FORBIDDEN;	/* Initialize */
  }
  DBG(fprintf(stderr, "<complementPath> %s\n", msg->file.path);)
  return WM_RET_OK;
}

/*
 * Function Name:	wv_open_request_file
 * Description:		open request file
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
wv_open_request_file(msg)
	register Message *	msg;
{
  int		fd;
  char *	etag;

  DBG(fprintf(stderr, "[openRequestFile] status(%d)\n",
	      wvResponseStatusTable[msg->response.status].code);)
again:
  fd = ws_safe_open(msg->file.path, O_RDONLY, &msg->file.status,
		    httpd->secureLevel >= SECURE_LEVEL3 ? false : true);
  if (fd == -1) {
    switch (errno) {
      case ENOENT:
	msg->response.status = STATUS_NOT_FOUND;
	break;
      case EMLINK:
      case EACCES:
	msg->response.status = STATUS_FORBIDDEN;
	break;
      case EISDIR:
	/* redirect an illegal request URI */
	if (msg->request.uri[strlen(msg->request.uri)-1] != '/') {
	  char *	location;
	  ws_asprintf(&location, "%s/", msg->request.uri);
	  wa_put(msg->response.header, "Location", location);
	  free(location);
	  msg->response.status = STATUS_MOVED_PERMANENTLY;
	}
	break;
      case EINTR:
	goto again;
      default:
	wv_syslog(WLL_LOG_ERR, "failed to open file: %s", msg->file.path);
	msg->response.status = STATUS_INTERNAL_SERVER_ERROR;
	break;
    }
    return -1;
  }

  /* Following symbolic link is restricted to secure region */
  if (msg->vhost->followSymlinks == false &&
      wv_safe_path_checker(msg, msg->file.path) == -1) {
    wu_close(fd);
    return -1;
  }
  msg->file.nfd = wt_netfd_fdopen(fd);
  if (msg->file.nfd == NULL) {
    wu_close(fd);
    return -1;
  }

  /* Set ETag ( For conditional request */
  ws_asprintf(&etag, "\"%lx-%lx-%lx\"",
		(long)msg->file.status.st_ino,
		(long)msg->file.status.st_size,
		msg->file.status.st_mtime);
  wa_put(msg->response.header, "ETag", etag);
  free(etag);

  return 0;
}
