/*****************************************************************************
 *
 * FILE:	headers.c
 * DESCRIPTION:	Wyvern: manipulate request/response headers
 * DATE:	Sun, Aug  4 2002
 * UPDATED:	Wed, Sep 12 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2002-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2002-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: headers.c,v 1.31 2007/09/12 02:24:27 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <errno.h>
#include "wyvern.h"

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static int	receive_entity(Message *);

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wv_header_parser
 * Description:		parse request headers
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
wv_header_parser(msg)
	register Message *	msg;
{
#if	ENABLE_STRICT_PARSING
  char	buf[BUF_SIZE];
  int	status = WM_RET_OK;

  do {
    int		n;
    char *	name;
    char *	value;
    char	c;

    /* Read each request header field */
    memset(buf, 0, sizeof(buf));
    n = msg->readline(msg->nsd, buf, sizeof(buf),
		      WT_UTIME_SEC(msg->ka.timeout));
    if (n <= 0) {
      return WM_RET_ERROR;
    }
    name  = wu_chomp(buf);
    value = (char *)strchr(buf, ':');
    if (value == NULL) {
      continue;	/* invalid field */
    }
    *value++ = EOL;
    while ((c = *value) && (c == SPC || c == TAB)) {
      value++;
    }
    wu_tolower(name);

    /* Header Parsing */
    switch (name[0]) {
      case 'a':
	    /* See RFC 2616 Section 14.1 */
	if (ws_strcmp(name, "accept") == 0 ||
	    /* See RFC 2616 Section 14.2 */
	    ws_strcmp(name, "accept-charset") == 0 ||
	    /* See RFC 2616 Section 14.3 */
	    ws_strcmp(name, "accept-encoding") == 0 ||
	    /* See RFC 2616 Section 14.4 */
	    ws_strcmp(name, "accept-language") == 0) {
	  wa_put_plus(msg->request.header, name, value);
	}
	/* See RFC 2616 Section 14.8 */
	else if (ws_strcmp(name, "authorization") == 0) {
	  wa_put(msg->request.header, name, value);
	}
	else {
	  goto unknown;
	}
	break;
      case 'c':
	/* RFC 2616 Section 14.10 */
	if (ws_strcmp(name, "connection") == 0) {
	  if (ws_strcasecmp(value, "close") == 0) {
	    msg->ka.status = KA_CLOSE;
	  }
	  else if (ws_strcasecmp(value, "keep-alive") == 0) {
	    if (msg->request.http.version == HTTP_11 ||
		(msg->request.http.version == HTTP_10 && msg->ka.max-- > 0)) {
	      msg->ka.status = KA_CONTINUE;
	    }
	    else {
	      msg->ka.status = KA_CLOSE;
	    }
	  }
	  else {
	    msg->ka.status = KA_DISABLED;
	  }
	}
	/* See RFC 2616 Section 14.13 */
	else if (ws_strcmp(name, "content-length") == 0) {
	  msg->request.content_length = (unsigned int)strtol(value, NULL, 10);
	}
		 /* See RFC 2616 Section 14.9 */
	else if (ws_strcmp(name, "cache-control") == 0 ||
		 /* See RFC 2616 Section 14.17 */
		 ws_strcmp(name, "content-type") == 0 ||
		 /* See RFC 2616 Section 14.11 */
		 ws_strcmp(name, "content-encoding") == 0 ||
		 /* See RFC 2616 Section 14.12 */
		 ws_strcmp(name, "content-language") == 0 ||
		 /* See RFC 2616 Section 14.15 */
		 ws_strcmp(name, "content-md5") == 0 ||
		 /* See RFC 2109	*/
		 ws_strcmp(name, "cookie") == 0 ||
		 /* See RFC 2965	*/
		 ws_strcmp(name, "cookie2") == 0) {
	  wa_put(msg->request.header, name, value);
	}
	else {
	  goto unknown;
	}
	break;
      case 'd':
	/* See RFC 2616 Section 14.18 */
	if (ws_strcmp(name, "date") == 0) {
	  wa_put(msg->request.header, name, value);
	}
	else {
	  goto unknown;
	}
	break;
      case 'e':
	/* See RFC 2616 Section 14.20 */
	if (ws_strcmp(name, "expect") == 0) {
	  wa_put(msg->request.header, name, value);
	}
	else {
	  goto unknown;
	}
	break;
      case 'f':
	/* See RFC 2616 Section 14.22 */
	if (ws_strcmp(name, "from") == 0) {
	  wa_put(msg->request.header, name, value);
	}
	else {
	  goto unknown;
	}
	break;
      case 'h':
	/* See RFC 2616 Section 14.23 */
	if (ws_strcmp(name, "host") == 0) {
	  wa_put(msg->request.header, name, value);
	  status = wv_switch_virtual_host(msg);	/* Switching virtual host */
	}
	else {
	  goto unknown;
	}
	break;
      case 'i':
	    /* See RFC 2616 Section 14.25	*/
	if (ws_strcmp(name, "if-modified-since") == 0 ||
	    /* See RFC 2616 Section 14.24	*/
	    ws_strcmp(name, "if-match") == 0 ||
	    /* See RFC 2616 Section 14.26	*/
	    ws_strcmp(name, "if-none-match") == 0 ||
	    /* See RFC 2616 Section 14.27	*/
	    ws_strcmp(name, "if-range") == 0 ||
	    /* See RFC 2616 Section 14.28 */
	    ws_strcmp(name, "if-unmodified-since") == 0) {
	  wa_put(msg->request.header, name, value);
	}
	else {
	  goto unknown;
	}
	break;
      case 'k':
	/* RFC 2068 Section 19.7.1.1 */
	if (ws_strcmp(name, "keep-alive") == 0) {
	  ;	/* ignore this header */
	}
	else {
	  goto unknown;
	}
	break;
      case 'm':
	/* See RFC 2616 Section 14.31	*/
	if (ws_strcmp(name, "max-forwards") == 0) {
	  wa_put(msg->request.header, name, value);
	}
	else {
	  goto unknown;
	}
	break;
      case 'p':
	    /* See RFC 2616 Section 14.32	*/
	if (ws_strcmp(name, "pragma") == 0 ||
	    /* See RFC 2616 Section 14.34	*/
	    ws_strcmp(name, "proxy-authorization") == 0) {
	  wa_put(msg->request.header, name, value);
	}
	else {
	  goto unknown;
	}
	break;
      case 'r':
	    /* See RFC 2616 Section 14.36	*/
	if (ws_strcmp(name, "referer") == 0 ||
	    /* See RFC 2616 Section 14.35	*/
	    ws_strcmp(name, "range") == 0) {
	  wa_put(msg->request.header, name, value);
	}
	else {
	  goto unknown;
	}
	break;
      case 't':
	    /* See RFC 2616 Section 14.41 */
	if (ws_strcmp(name, "transfer-encoding") == 0 ||
	    /* See RFC 2616 Section 14.40 */
	    ws_strcmp(name, "trailer") == 0 ||
	    /* See RFC 2616 Section 14.39	*/
	    ws_strcmp(name, "te") == 0) {
	  wa_put(msg->request.header, name, value);
	}
	else {
	  goto unknown;
	}
	break;
      case 'u':
	    /* See RFC 2616 Section 14.43	*/
	if (ws_strcmp(name, "user-agent") == 0 ||
	    /* See RFC 2616 Section 14.42 */
	    ws_strcmp(name, "upgrade") == 0 ||
	    /* XXX: added 2007/08/22(Wed) */
	    ws_strcmp(name, "ua-cpu") == 0 ||
	    /* XXX: added 2007/08/24(Fri) */
	    ws_strcmp(name, "ua-os") == 0) {
	  wa_put(msg->request.header, name, value);
	}
	else {
	  goto unknown;
	}
	break;
      case 'v':
	/* See RFC 2616 Section 14.45 */
	if (ws_strcmp(name, "via") == 0) {
	  wa_put(msg->request.header, name,  value);
	}
	else {
	  goto unknown;
	}
	break;
      case 'x':
	    /* Squid uses this request header */
	if (ws_strcmp(name, "x-forwarded-for") == 0 ||
	    /* XXX: added 2007/08/24(Fri) */
	    ws_strncmp(name, "x-", 2) == 0) {
	  wa_put(msg->request.header, name, value);
	}
	else {
	  goto unknown;
	}
	break;
      /* XXX: added 2007/08/24(Fri) */
      case '-': case '~':
	/* ignore */
	break;
      default:
      unknown: {
	  const char *	ua;
	  
	  ua = wa_get(msg->request.header, "user-agent");
	  wv_syslog(WLL_LOG_NOTICE,
		    "unknown request header: %s=%s from %s (%s)",
		    name, value, msg->client.ip, ua ? ua : "???");
	}
	break;
    }
  } while (buf[0] != EOL);

  /* The HTTP/1.1 request needs Host request header field. */
  if (msg->request.http.version == HTTP_11 &&
      wa_get(msg->request.header, "host") == NULL) {
    msg->response.status = STATUS_BAD_REQUEST;
    return WM_RET_ERROR;
  }

  /* Request has entity */
  if (msg->request.method.type == M_POST) {
    if (msg->request.http.version == HTTP_09) {
      msg->response.status = STATUS_METHOD_NOT_ALLOWED;
      return WM_RET_ERROR;
    }
    return receive_entity(msg);
  }

  return status;
#else
  char		buf[BUF_SIZE];
  const char *	p;
  int		status = WM_RET_OK;

  /* Header Parsing */
  do {
    int		n;
    char *	name;
    char *	value;
    char	c;

    /* Read each request header field */
    memset(buf, 0, sizeof(buf));
    n = msg->readline(msg->nsd, buf, sizeof(buf),
		      WT_UTIME_SEC(msg->ka.timeout));
    if (n <= 0) {
      return WM_RET_ERROR;
    }
    name  = wu_chomp(buf);
    value = (char *)strchr(buf, ':');
    if (value == NULL) {
      continue;	/* invalid field */
    }
    *value++ = EOL;
    while ((c = *value) && (c == SPC || c == TAB)) {
      value++;
    }
    wu_tolower(name);
    wa_put_plus(msg->request.header, name, value);
  } while (buf[0] != EOL);

  /* Connection request header */
  if ((p = wa_get(msg->request.header, "connection")) != NULL) {
    if (ws_strcasecmp(p, "keep-alive") == 0) {
      if (msg->request.http.version == HTTP_10 && msg->ka.max-- > 0) {
	msg->ka.status = KA_CONTINUE;
      }
    }
    else if (ws_strcmp(p, "close") == 0) {
      msg->ka.status = KA_CLOSE;
    }
    else {
      msg->ka.status = KA_DISABLED;
    }
  }
  /* Content-Length request header for POST method */
  if ((p = wa_get(msg->request.header, "content-length")) != NULL) {
    msg->request.content_length = (unsigned int)strtol(p, NULL, 10);
  }
  /* The HTTP/1.1 request needs Host request header field. */
  if ((p = wa_get(msg->request.header, "host")) == NULL &&
      msg->request.http.version == HTTP_11) {
    msg->response.status = STATUS_BAD_REQUEST;
    return WM_RET_ERROR;
  }
  if (p != NULL) {
    status = wv_switch_virtual_host(msg);	/* Switching virtual host */
  }

  /* Request has entity */
  if (msg->request.method.type == M_POST) {
    if (msg->request.http.version == HTTP_09) {
      msg->response.status = STATUS_METHOD_NOT_ALLOWED;
      return WM_RET_ERROR;
    }
    return receive_entity(msg);
  }

  return status;
#endif	/* ENABLE_STRICT_PARSING */
}

/*
 * Function Name:	receive_entity
 * Description:		receive the request entity from a client
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
static int
receive_entity(msg)
	register Message *	msg;
{
  size_t	len;
  ssize_t	n;

  /* No Content-Length request header */
  if (msg->request.content_length == 0) {
    msg->response.status = STATUS_LENGTH_REQUIRED;
    return WM_RET_ERROR;
  }
  /* Too large request entity */
  if (msg->request.content_length > httpd->maxPostSize) {
    msg->response.status = STATUS_REQUEST_ENTITY_TOO_LARGE;
    return WM_RET_ERROR;
  }

  /* Create temporary filename */
  ws_asprintf(&msg->request.post_tmp, "%s/wvXXXXXX", httpd->tmpDir);

  /* Open temporary file */
  msg->request.post_fd = mkstemp(msg->request.post_tmp);
  if (msg->request.post_fd == -1) {
    goto error;
  }
  /* The file will be removed when it's closed. */
  unlink(msg->request.post_tmp);

  /* Read data from client */
  for (n = 0, len = 0; len < msg->request.content_length; len += n) {
    char	buf[BUF_SIZE];

    memset(buf, 0, sizeof(buf));
    n = msg->read(msg->nsd, buf, sizeof(buf), WT_UTIME_SEC(msg->ka.timeout));
    if (n > 0) {
      if (write(msg->request.post_fd, buf, (size_t)n) != n) {
	goto error;
      }
    }
    else if (n < 0) {
      goto error;
    }
    else {
      break;
    }
  }
#if	HAVE_FTRUNCATE
  /*
   * The posted contents have CRLF at end of streams.
   * Wyvern removes it to ftruncate(2) system call.
   * I think that this way is dirty.
   */
  if (len > msg->request.content_length) {
    ftruncate(msg->request.post_fd, (off_t)msg->request.content_length);
  }
#endif	/* HAVE_FTRUNCATE */

  /* Rewind file pointer */
  lseek(msg->request.post_fd, (off_t)0, SEEK_SET);

  return WM_RET_OK;

error:
  wv_syslog(WLL_LOG_ERR, "failed to receive the request entity");
  msg->response.status = STATUS_INTERNAL_SERVER_ERROR;

  return WM_RET_ERROR;
}
