/*****************************************************************************
 *
 * FILE:	hook.c
 * DESCRIPTION:	Wyvern: hook procedure manager
 * DATE:	Tue, Aug 13 2002
 * UPDATED:	Sat, Aug 25 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2002-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2002-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: hook.c,v 1.19 2007/08/26 16:37:36 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include "wyvern.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
typedef struct _HookList {
  hook_callback_t	routine;
  const char *		mime;		/* mime type for RequestHandler */
  size_t		mlen;		/* mime type valid length */
  hook_ordering_t	position;	/* hooked position */
  struct _HookList *	next;
} HookList;

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static int	run_contents_filter(HookList *, Message *, hook_ordering_t);
static int	run_hook_list_until_fails(HookList *, Message *);
static int	run_hook_list_until_succeeds(HookList *, Message *);
static int	run_hook_list_while_not_succeed(HookList *, Message *);
static void	make_hook_list(HookList **, hook_callback_t,
			       hook_ordering_t, const char *);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static HookList *	requestReader	= NULL;
static HookList *	headerParser	= NULL;
static HookList *	accessChecker	= NULL;
static HookList *	uriTranslater	= NULL;
static HookList *	userIDChecker	= NULL;
static HookList *	authChecker	= NULL;
static HookList *	typeChecker	= NULL;
static HookList *	fixups		= NULL;
static HookList *	requestHandler	= NULL;
static HookList *	contentsFilter	= NULL;
static HookList *	responseHandler	= NULL;
static HookList *	logger		= NULL;

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wm_run_request_reader
 * Description:		run hooked request reader
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
wm_run_request_reader(msg)
	register Message *	msg;
{
  return run_hook_list_until_succeeds(requestReader, msg);
}

/*
 * Function Name:	wm_run_header_parser
 * Description:		run hooked header parsers
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
wm_run_header_parser(msg)
	register Message *	msg;
{
  return run_hook_list_until_fails(headerParser, msg);
}

/*
 * Function Name:	wm_run_access_checker
 * Description:		run hooked access checkers
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
wm_run_access_checker(msg)
	register Message *	msg;
{
  return run_hook_list_while_not_succeed(accessChecker, msg);
}

/*
 * Function Name:	wm_run_uri_translater
 * Description:		run hooked uri translators
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
wm_run_uri_translater(msg)
	register Message *	msg;
{
  return run_hook_list_until_succeeds(uriTranslater, msg);
}

/*
 * Function Name:	wm_run_user_id_checker
 * Description:		run hooked uid checkers
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
wm_run_user_id_checker(msg)
	register Message *	msg;
{
  return run_hook_list_while_not_succeed(userIDChecker, msg);
}

/*
 * Function Name:	wm_run_auth_checker
 * Description:		run hooked authentication checkers
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
wm_run_auth_checker(msg)
	register Message *	msg;
{
  return run_hook_list_while_not_succeed(authChecker, msg);
}

/*
 * Function Name:	wm_run_type_checker
 * Description:		run hooked mime-type checkers
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
wm_run_type_checker(msg)
	register Message *	msg;
{
  return run_hook_list_until_succeeds(typeChecker, msg);
}

/*
 * Function Name:	wm_run_fixups
 * Description:		run hooked fixups
 * Arguments:		message, position
 * Return Value:	successful: 0, otherwise: -1
 */
int
wm_run_fixups(msg, position)
	register Message *	msg;
	hook_ordering_t		position;
{
  register HookList *	hlp;

  for (hlp = fixups; hlp != NULL; hlp = hlp->next) {
    if (hlp->position == position && hlp->routine(msg) == -1) {
      return -1;
    }
  }

  return 0;
}

/*
 * Function Name:	wm_run_request_handler
 * Description:		run hooked request handlers
 * Arguments:		message
 * Return Value:	successful: 0, cgi: RET_WAIT, otherwise: -1
 */
int
wm_run_request_handler(msg)
	register Message *	msg;
{
  register HookList *	hlp;

  for (hlp = requestHandler; hlp != NULL; hlp = hlp->next) {
    if (ws_strncmp(hlp->mime, msg->file.mime, hlp->mlen) == 0) {
      int	status;

      status = hlp->routine(msg);
      if (status != WM_RET_DECLINED) {
       	return status;
      }
    }
  }

  return 0;
}

/*
 * Function Name:	wm_run_contents_filter
 * Description:		run hooked contents filters
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
wm_run_contents_filter(msg)
	register Message *	msg;
{
  if (run_contents_filter(contentsFilter, msg, HOOK_FIRST)  == -1 ||
      run_contents_filter(contentsFilter, msg, HOOK_MIDDLE) == -1 ||
      run_contents_filter(contentsFilter, msg, HOOK_LAST)   == -1 ||
      run_contents_filter(contentsFilter, msg, HOOK_FINAL)  == -1) {
    return -1;
  }
  return 0;
}

/*
 * Function Name:	run_contents_filter
 * Description:		run hooked contents filters
 * Arguments:		hookList, message, position
 * Return Value:	successful: 0, otherwise: -1
 */
static int
run_contents_filter(hookList, msg, position)
	register HookList *	hookList;
	register Message *	msg;
	hook_ordering_t		position;
{
  register HookList *	hlp;
  
  if (msg->file.mime == NULL) {
    return -1;
  }

  for (hlp = hookList; hlp != NULL; hlp = hlp->next) {
    if (hlp->position == position &&
	ws_strncmp(hlp->mime, msg->file.mime, hlp->mlen) == 0) {
      return hlp->routine(msg);
    }
  }

  return 0;
}

/*
 * Function Name:	wm_run_response_handler
 * Description:		run hooked response handler
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
wm_run_response_handler(msg)
	register Message *	msg;
{
  return run_hook_list_until_succeeds(responseHandler, msg);
}

/*
 * Function Name:	wm_run_logger
 * Description:		run hooked loggers
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
wm_run_logger(msg)
	register Message *	msg;
{
  return run_hook_list_until_succeeds(logger, msg);
}


/*
 * Function Name:	run_hook_list_until_fails
 * Description:		run hooked functions
 * Arguments:		hookList, message
 * Return Value:	successful: 0, otherwise: -1
 */
static int
run_hook_list_until_fails(hookList, msg)
	register HookList *	hookList;
	register Message *	msg;
{
  register HookList *	hlp;

  /* call functions until it fails. */
  for (hlp = hookList; hlp != NULL; hlp = hlp->next) {
    if (hlp->routine(msg) == -1) {
      return -1;
    }
  }
  return 0;
}

/*
 * Function Name:	run_hook_list_until_succeeds
 * Description:		run hooked functions
 *			allow first matching, empty list is bad
 * Arguments:		hookList, message
 * Return Value:	successful: 0, otherwise: -1
 */
static int
run_hook_list_until_succeeds(hookList, msg)
	register HookList *	hookList;
	register Message *	msg;
{
  register HookList *	hlp;

  /* call functions until it succeeds. */
  for (hlp = hookList; hlp != NULL; hlp = hlp->next) {
    if (hlp->routine(msg) == 0) {
      return 0;
    }
  }
  return -1;
}

/*
 * Function Name:	run_hook_list_while_not_succeed
 * Description:		run hooked functions
 *			allow first matching, empty list is Ok
 * Arguments:		hookList, message
 * Return Value:	successful: 0, otherwise: -1
 */
static int
run_hook_list_while_not_succeed(hookList, msg)
	register HookList *	hookList;
	register Message *	msg;
{
  register HookList *	hlp;
  volatile int		status = 0;

  /* call functions until it succeeds. */
  for (hlp = hookList; hlp != NULL; hlp = hlp->next) {
    status = hlp->routine(msg);
    if (status == 0) {
      return 0;
    }
  }

  return status;	/* 0 returns if the list is empty. */
}

/*****************************************************************************/

/*
 * Function Name:	wm_hook_request_reader
 * Description:		hook request readers
 * Arguments:		routine, position
 * Return Value:	successful: 0, otherwise: -1
 */
void
wm_hook_request_reader(routine, position)
	hook_callback_t	routine;
	hook_ordering_t	position;
{
  make_hook_list(&requestReader, routine, position, NULL);
}

/*
 * Function Name:	wm_hook_header_parser
 * Description:		hook header parsers
 * Arguments:		routine, position
 * Return Value:	successful: 0, otherwise: -1
 */
void
wm_hook_header_parser(routine, position)
	hook_callback_t	routine;
	hook_ordering_t	position;
{
  make_hook_list(&headerParser, routine, position, NULL);
}

/*
 * Function Name:	wm_hook_access_checker
 * Description:		hook access checkers
 * Arguments:		routine, position
 * Return Value:	successful: 0, otherwise: -1
 */
void
wm_hook_access_checker(routine, position)
	hook_callback_t	routine;
	hook_ordering_t	position;
{
  make_hook_list(&accessChecker, routine, position, NULL);
}

/*
 * Function Name:	wm_hook_uri_translater
 * Description:		hook uri translators
 * Arguments:		routine, position
 * Return Value:	successful: 0, otherwise: -1
 */
void
wm_hook_uri_translater(routine, position)
	hook_callback_t	routine;
	hook_ordering_t	position;
{
  make_hook_list(&uriTranslater, routine, position, NULL);
}

/*
 * Function Name:	wm_hook_user_id_checker
 * Description:		hook uid checkers
 * Arguments:		routine, position
 * Return Value:	successful: 0, otherwise: -1
 */
void
wm_hook_user_id_checker(routine, position)
	hook_callback_t	routine;
	hook_ordering_t	position;
{
  make_hook_list(&userIDChecker, routine, position, NULL);
}

/*
 * Function Name:	wm_hook_auth_checker
 * Description:		hook authentication checkers
 * Arguments:		routine, position
 * Return Value:	successful: 0, otherwise: -1
 */
void
wm_hook_auth_checker(routine, position)
	hook_callback_t	routine;
	hook_ordering_t	position;
{
  make_hook_list(&authChecker, routine, position, NULL);
}

/*
 * Function Name:	wm_hook_type_checker
 * Description:		hook mime-type checkers
 * Arguments:		routine, position
 * Return Value:	successful: 0, otherwise: -1
 */
void
wm_hook_type_checker(routine, position)
	hook_callback_t	routine;
	hook_ordering_t	position;
{
  make_hook_list(&typeChecker, routine, position, NULL);
}

/*
 * Function Name:	wm_hook_fixups
 * Description:		hook fixups
 * Arguments:		routine, position
 * Return Value:	successful: 0, otherwise: -1
 */
void
wm_hook_fixups(routine, position)	
	hook_callback_t	routine;
	hook_ordering_t	position;
{
  make_hook_list(&fixups, routine, position, NULL);
}

/*
 * Function Name:	wm_hook_request_handler
 * Description:		hook request handlers
 * Arguments:		routine, position, mime-type
 * Return Value:	successful: 0, otherwise: -1
 */
void
wm_hook_request_handler(routine, position, mime)
	hook_callback_t	routine;
	hook_ordering_t	position;
	const char *	mime;
{
  make_hook_list(&requestHandler, routine, position, mime);
}

/*
 * Function Name:	wm_hook_contents_filter
 * Description:		hook contents filters
 * Arguments:		routine, position, mime-type
 * Return Value:	successful: 0, otherwise: -1
 */
void
wm_hook_contents_filter(routine, position, mime)	
	hook_callback_t	routine;
	hook_ordering_t	position;
	const char *	mime;
{
  make_hook_list(&contentsFilter, routine, position, mime);
}

/*
 * Function Name:	wm_hook_response_handler
 * Description:		hook response handlers
 * Arguments:		routine, position
 * Return Value:	successful: 0, otherwise: -1
 */
void
wm_hook_response_handler(routine, position)	
	hook_callback_t	routine;
	hook_ordering_t	position;
{
  make_hook_list(&responseHandler, routine, position, NULL);
}

/*
 * Function Name:	wm_hook_logger
 * Description:		hook loggers
 * Arguments:		routine, position
 * Return Value:	successful: 0, otherwise: -1
 */
void
wm_hook_logger(routine, position)	
	hook_callback_t	routine;
	hook_ordering_t	position;
{
  make_hook_list(&logger, routine, position, NULL);
}

/*
 * Function Name:	make_hook_list
 * Description:		make hook list
 * Arguments:		hookList, routine, mime-type, position
 * Return Value:	none
 */
static void
make_hook_list(hookList, routine, position, mime)
	HookList **	hookList;
	hook_callback_t	routine;
	hook_ordering_t	position;
	const char *	mime;
{
  HookList *		new;
  register HookList *	hlp;
  register HookList *	prev_hlp;

  new		= (HookList *)ws_malloc(sizeof(HookList));	/* leak */
  new->routine	= routine;
  new->next	= NULL;
  if (mime != NULL) {
    size_t	len = strlen(mime);

    new->mime = ws_strdup(mime);
    new->mlen = mime[len - 1] == '*' ? len - 1 : len;
  }
  else {
    new->mime = NULL;
    new->mlen = 0;
  }
  new->position	= position;

  /* find position */
  for (prev_hlp = NULL, hlp = *hookList;
       hlp != NULL && hlp->position < position; hlp = hlp->next) {
    prev_hlp = hlp;
  }
  /* adjust position */
  for ( ; hlp != NULL && hlp->position == position; hlp = hlp->next) {
    prev_hlp = hlp;
  }

  if (hlp == NULL) {
    if (*hookList == NULL) {	/* at first */
      new->next = *hookList;
      *hookList	= new;
    }
    else if (prev_hlp != NULL) {
      prev_hlp->next = new;
    }
  }
  else {
    new->next = hlp;
    if (prev_hlp != NULL) {
      prev_hlp->next = new;
    }
    else {
      *hookList = new;
    }
  }
}
