/*****************************************************************************
 *
 * FILE:	log.c
 * DESCRIPTION:	Wyvern: record server logging
 * DATE:	Sun, Sep 26 1999
 * UPDATED:	Wed, Sep 12 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 1999-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 1999-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: log.c,v 1.29 2007/09/12 02:24:27 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#if	__STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif	/* __STDC__ */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include "wyvern.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#ifndef	HAVE_SETLINEBUF
#define	setlinebuf(x)	setvbuf((x), (char *)NULL, _IOLBF, 0)
#endif	/* HAVE_SETLINEBUF */

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	open_log_files(register VirtualHost *);
static void	subsyslog(FILE *, int, const char *, va_list);

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wv_open_log_files
 * Description:		open log files
 * Arguments:		none
 * Return Value:	none
 */
void
wv_open_log_files(void)
{
  register VirtualHost *	vp;

  /* virtual hosts */
  for (vp = httpd->host; vp != NULL; vp = vp->next) {
    open_log_files(vp);
    /* Open error-log file */
    if (vp->errorLog.path == NULL) {
      vp->errorLog.fp = httpd->host->errorLog.fp;
    }
    /* Open access-log file */
    if (vp->accessLog.path == NULL) {
      vp->accessLog.fp = httpd->host->accessLog.fp;
    }
  }

  /* security log */
  if (httpd->securityLog.path != NULL) {
    httpd->securityLog.fp = wl_openlog(httpd->securityLog.path);
    if (httpd->securityLog.fp == NULL) {
      wv_die("Failed to open securityLog: %s", httpd->securityLog.path);
    }
  }
}

/*
 * Function Name:	open_log_files
 * Description:		open log files
 * Arguments:		virtual host list
 * Return Value:	none
 */
static void
open_log_files(vp)
	register VirtualHost *	vp;
{
  /* errorLog */
  if (vp->errorLog.path != NULL) {
    vp->errorLog.fp = wl_openlog(vp->errorLog.path);
    if (vp->errorLog.fp == NULL) {
      wv_die("Failed to open errorLog: %s", vp->errorLog.path);
    }
  }

  /* accessLog */
  if (vp->accessLog.path != NULL) {
    vp->accessLog.fp = wl_openlog(vp->accessLog.path);
    if (vp->accessLog.fp == NULL) {
      wv_die("Failed to open accessLog: %s", vp->accessLog.path);
    }
  }
}

/*
 * Function Name:	wv_close_log_files
 * Description:		close log files
 * Arguments:		none
 * Return Value:	none
 */
void
wv_close_log_files(void)
{
  register VirtualHost *	vp;

  /* access/error log */
  for (vp = httpd->host; vp != NULL; vp = vp->next) {
    if (vp->errorLog.fp == NULL || vp->accessLog.fp == NULL) {
      continue;
    }
    wv_close_log(vp);
  }

  /* security log */
  if (httpd->securityLog.path != NULL && httpd->securityLog.fp != NULL) {
    wl_closelog(httpd->securityLog.fp);
    httpd->securityLog.fp = NULL;
  }
}

/*
 * Function Name:	wv_close_log
 * Description:		close log files
 * Arguments:		virtual host list
 * Return Value:	none
 */
void
wv_close_log(vp)
	register VirtualHost *	vp;
{
  /* Close error-log file */
  if (vp->errorLog.path != NULL) {
    wl_closelog(vp->errorLog.fp);
    vp->errorLog.fp = NULL;
  }

  /* Close access-log file */
  if (vp->accessLog.path != NULL) {
    wl_closelog(vp->accessLog.fp);
    vp->accessLog.fp = NULL;
  }
}

/*****************************************************************************/

/*
 * Function Name:	wv_log_writer
 * Description:		write the log message in log file
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
wv_log_writer(msg)
	register Message *	msg;
{
  register const char *	xff;	// X-Forwarded-For
  register const char *	ref;	// Referer
  register const char *	ua;	// User-Agent
  char *		clientName = NULL;
  char *		tbuf	   = NULL;

  /* Set client name */
  if ((xff = wa_get(msg->request.header, "x-forwarded-for")) != NULL) {
    ws_asprintf(&clientName, "%s(%s)", msg->client.ip, xff);
  }
  else {
    clientName = msg->client.ip;
  }

  /* calculate timestamp */
  wu_timestamp(&tbuf, msg->time.request.tv_sec, WUL_TIME_FORMAT_RFC822,
	       WUL_ZONE_LOCAL);

  /* Client Error: 4xx */
  /* Server Error: 5xx */
  if (msg->response.status > STATUS_CODE_4XX) {
    fprintf(msg->vhost->accessLog.fp, "%s %s %s [%s] \"%s %s %s\" %d \"%s\" ",
	    clientName,
	    msg->user.identName == NULL ? "-" : msg->user.identName,
	    msg->user.authName  == NULL ? "-" : msg->user.authName,
	    tbuf == NULL ? "-" : tbuf,
	    msg->request.method.name == NULL ? "-" : msg->request.method.name,
	    msg->request.uri[0]      == EOL  ? "-" : msg->request.uri,
	    msg->request.http.protocol == NULL ? "-" : msg->request.http.protocol,
	    wvResponseStatusTable[msg->response.status].code,
	    wvResponseStatusTable[msg->response.status].phrase);
  }
  /* Informational: 1xx */
  /* Successful:    2xx */
  /* Redirection:   3xx */
  else {
    fprintf(msg->vhost->accessLog.fp, "%s %s %s [%s] \"%s %s %s\" %d ",
	    clientName,
	    msg->user.identName == NULL ? "-" : msg->user.identName,
	    msg->user.authName  == NULL ? "-" : msg->user.authName,
	    tbuf == NULL ? "-" : tbuf,
	    msg->request.method.name,
	    msg->request.uri,
	    msg->request.http.protocol,
	    wvResponseStatusTable[msg->response.status].code);
    if (msg->response.status > STATUS_CODE_3XX) {
      fputs("-", msg->vhost->accessLog.fp);
    }
    else {
      switch (msg->request.method.type) {
	case M_HEAD:
	  fputs("0", msg->vhost->accessLog.fp);
	  break;
	case M_GET:
	  fprintf(msg->vhost->accessLog.fp, "%ld",
		  (long)msg->file.status.st_size);
	  break;
	case M_POST:
	  fprintf(msg->vhost->accessLog.fp, "%zu", msg->request.content_length);
	  break;
	default:
	  fputs("-", msg->vhost->accessLog.fp);
	  break;
      }
    }
  }
  wv_free(tbuf);

  fprintf(msg->vhost->accessLog.fp, " \"%s\" \"%s\"",
	  (ref = wa_get(msg->request.header, "referer")) == NULL    ? "-" : ref,
	  (ua  = wa_get(msg->request.header, "user-agent")) == NULL ? "-" : ua);
#if	ENABLE_SSL
  if (msg->soci.secure == SOCKET_SECURE) {
    fprintf(msg->vhost->accessLog.fp, " \"%s\" \"%s\"",
	    SSL_get_version(wt_netfd_ssl(msg->nsd)),
	    SSL_get_cipher(wt_netfd_ssl(msg->nsd)));
  }
#endif	/* ENABLE_SSL */
  if (wu_uptime(&tbuf, &msg->time.request, &msg->time.response) == 0) {
    fprintf(msg->vhost->accessLog.fp, " (%s)", tbuf);
    free(tbuf);
  }
  fputs("\n", msg->vhost->accessLog.fp);

  if (xff != NULL) { free(clientName); }

  return WM_RET_OK;
}

/*****************************************************************************/

/*
 * Function Name:	wv_die
 * Description:		print an error message and terminate process
 * Arguments:		type of error
 * Return Value:	none
 */
void
#if	__STDC__
wv_die(const char *fmt, ...)
#else
wv_die(fmt, va_alist)
	char *	fmt;
	va_dcl
#endif	/* __STDC__ */
{
  va_list	ap;

#if	__STDC__
  va_start(ap, fmt);
#else
  va_start(ap);
#endif	/* __STDC__ */
  subsyslog(httpd->host->errorLog.fp, WLL_LOG_FATAL, fmt, ap);
  va_end(ap);

  if (httpd->pid > 0) {
    (void)unlink(httpd->pidFile);	/* ignore error */
  }
  exit(1);
}

/*
 * Function Name:	wv_syslog
 * Description:		write log messages
 * Arguments:		level, log messages
 * Return Value:	none
 */
void
#if	__STDC__
wv_syslog(int level, const char * message, ...)
#else
wv_syslog(level, message, va_alist)
	int	level;
	char *	message;
	va_dcl
#endif	/* __STDC__ */
{
  va_list	ap;
#if	__STDC__
  va_start(ap, message);
#else
  va_start(ap);
#endif	/* __STDC__ */
  subsyslog(httpd->vhost->errorLog.fp, level, message, ap);
  va_end(ap);
}

/*
 * Function Name:	subsyslog
 * Description:		write log messages
 * Arguments:		output FILE pointer, log level, log messages
 * Return Value:	none
 */
static void
subsyslog(fout, level, fmt, ap)
	register FILE *	fout;
	int		level;
	const char *	fmt;
	va_list		ap;
{
  char *		tbuf;
  static const char *	priority[] = {
    "unk",	/* dummy */
    "fatal",
    "emerg", "alert", "crit", "error", "warning", "notice", "info", "debug",
    "security"
  };

  /* Get current time and date */
  wu_timestamp(&tbuf, 0, WUL_TIME_FORMAT_RFC822, WUL_ZONE_LOCAL);

  /* Write security log */
  if (level & WLL_LOG_SECURITY) {
    if (httpd->securityLog.fp != NULL) {
      fprintf(httpd->securityLog.fp, "[%s] ", tbuf == NULL ? "-" : tbuf);
      vfprintf(httpd->securityLog.fp, fmt, ap);
      fputs("\n", httpd->securityLog.fp);
    }
    level &= WLL_LOG_MASK;
  }

  /* countermeasure for first boot time error */
  if (fout == NULL) { fout = stderr; }

  /* Write error log */
  if (level) {
    fprintf(fout, "[%s] [%s] ", tbuf == NULL ? "-" : tbuf, priority[level]);
    vfprintf(fout, fmt, ap);
  }
  wv_free(tbuf);
  fputs("\n", fout);
}
