/*****************************************************************************
 *
 * FILE:	mime.c
 * DESCRIPTION:	Wyvern: register/get MIME types
 * DATE:	Sat, Sep 18 1999
 * UPDATED:	Sun, Jan 29 2006
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 1999-2006 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 1999-2006 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mime.c,v 1.13 2006/08/27 07:30:30 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include "wyvern.h"

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static Hash *	mediaTable = NULL;

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static char *	lookup(Hash *, const char *);

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wv_add_media_types
 * Description:		register the suffix and the media-type into hash table
 * Arguments:		suffix, media-type
 * Return Value:	none
 */
void
wv_add_media_types(suffix, type)
	const char *	suffix;
	const char *	type;
{
  wv_add_types(&mediaTable, suffix, type);
}

/*
 * Function Name:	wv_add_types
 * Description:		register the suffix and the media-type into hash table
 * Arguments:		mime table, suffix, media-type
 * Return Value:	none
 */
void
wv_add_types(mimeTable, suffix, type)
	Hash **		mimeTable;
	const char *	suffix;
	const char *	type;
{
  if (suffix != NULL && lookup(*mimeTable, suffix) == NULL) {
    Hash *	new;

    new		= (Hash *)ws_malloc(sizeof(Hash));	/* leak */
    new->name	= ws_strdup(suffix);
    new->value	= ws_strdup(type);
    new->next	= *mimeTable;
    *mimeTable	= new;
  }
}

/*
 * Function Name:	lookup
 * Description:		lookup mime-type with suffix
 * Arguments:		mime table, suffix
 * Return Value:	mime-type
 */
static char *
lookup(mimeTable, suffix)
	Hash *		mimeTable;
	const char *	suffix;
{
  register Hash *	hp;

  for (hp = mimeTable; hp != NULL; hp = hp->next) {
    if (ws_strcmp(hp->name, suffix) == 0) {
      return hp->value;
    }
  }

  return NULL;
}

/*
 * Function Name:	wv_fixup_mime_table
 * Description:		fixup mime-type table each virtual host
 * Arguments:		mime table
 * Return Value:	none
 */
void
wv_fixup_mime_table(mimeTable)
	Hash **	mimeTable;
{
  if (*mimeTable != NULL) {
    register Hash *	hp;

    /* move to end */
    for (hp = *mimeTable; hp->next != NULL; hp = hp->next) {
      ;	/* nothing to do */
    }
    hp->next = mediaTable;
  }
  else {
    *mimeTable = mediaTable;
  }
}

/*****************************************************************************/


/*
 * Function Name:	wv_media_type_checker
 * Description:		check mime-type of supplied file
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
wv_media_type_checker(msg)
	register Message *	msg;
{
  if (msg->file.mime == NULL && msg->file.path != NULL) {
    char *	suffix;	/* suffix of file */

    msg->file.mime	= msg->vhost->defaultType;
    suffix		= strrchr(msg->file.path, '.');
    if (suffix++ != NULL) {
      char *	mtype;
      
      mtype = lookup(msg->vhost->mimeTable, suffix);
      if (mtype != NULL) {
	msg->file.mime = mtype;
      }
    }
  }
  DBG(fprintf(stderr, "[checkMediaType] %s\n", msg->file.mime);)

  return WM_RET_OK;
}

/*****************************************************************************/

#if	DEBUG
void
debugMimeTable(mimeTable)
	Hash *	mimeTable;
{
  register Hash *	hp;

  for (hp = mimeTable; hp != NULL; hp = hp->next) {
    fprintf(stderr, "%s => %s\n", hp->name, hp->value);
  }
}

#endif	/* DEBUG */

