/*****************************************************************************
 *
 * FILE:	module.c
 * DESCRIPTION:	Wyvern: module manager
 * DATE:	Wed, Sep 13 2000
 * UPDATED:	Sat, Aug 25 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2000-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2000-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: module.c,v 1.28 2007/08/26 16:37:36 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#if	HAVE_DLOPEN
#include <dlfcn.h>
#else
#if	HAVE_SHL_LOAD
#include <dl.h>
#else
#if	HAVE_LIBDLD
#include <dld.h>
#endif	/* HAVE_LIBDLD */
#endif	/* HAVE_SHL_LOAD */
#endif	/* HAVE_DLOPEN */
#include <errno.h>
#include "wyvern.h"

/******************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
typedef struct _ModuleList {
  struct _ModuleList *	prev;
  struct _ModuleList *	next;
  const wmModule *	module;
} ModuleList;

/******************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void *		load(const char *);
static void *		fetch(void *);
static void		install(ModuleList *);
static int		verify(const wmModule *);
static ModuleList *	lookup(const char *);
static ModuleList *	new_module_list(const wmModule *);
static int		unload(void *);
static void		uninstall(ModuleList *);

/******************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static ModuleList 	moduleList = { &moduleList, &moduleList, NULL };

/******************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wv_load_module
 * Description:		load a module and add it into the module manager
 * Arguments:		module path
 * Return Value:	successful: 0, otherwise: -1
 */
int
wv_load_module(path)
	const char *	path;
{
  void *	handle;
  wmModule *	module;
  int		rc;

  /* Phase 1: load the module */
  handle = load(path);
  if (handle == NULL) {
    goto error1;
  }

  /* Phase 2: fetch the symbol function */
  module = (wmModule *)fetch(handle);
  if (module == NULL) {
    goto error;
  }

  /* Phase 3: verify the module */
  rc = verify(module);
  if (rc == -1) {
    goto error;
  }

  /* Phase 4: install the module */
  module->handle = handle;
  install(new_module_list(module));

  return 0;

error:
  unload(handle);
error1:
  return -1;
}

/*
 * Function Name:	wv_unload_module
 * Description:		unload the module and remove it from the module manager
 * Arguments:		module name
 * Return Value:	successful: 0, otherwise: -1
 */
int
wv_unload_module(name)
	const char *	name;
{
  ModuleList *	mlp;
  int		status;

  if (name == NULL) {
    errno = EINVAL;
    return -1;
  }

  mlp = lookup(name);
  if (mlp == NULL) {
    wv_syslog(WLL_LOG_ERR, "%s wasn't loaded", name);
    return -1;
  }

  if (mlp->module->finish != NULL) {
    mlp->module->finish();
  }

  status = unload(mlp->module->handle);
  uninstall(mlp);
  free(mlp);

  return status;
}

/*
 * Function Name:	wv_initialize_modules
 * Description:		call initialize routine of modules
 * Arguments:		none
 * Return Value:	none
 */
void
wv_initialize_modules(void)
{
  DBG(fputs("[wv_initialize_modules]\n", stderr);)
  if (ws_revoke_privilege(httpd->user, httpd->group) != -1) {
    register ModuleList *	mlp;

    for (mlp = moduleList.next; mlp != &moduleList; mlp = mlp->next) {
      if (mlp->module->init != NULL) {
       	mlp->module->init();
      }
      if (mlp->module->register_hooks != NULL) {
	mlp->module->register_hooks();
      }

      /* Show module information */
      wv_syslog(WLL_LOG_INFO, "%s: %s, '%s'",
		mlp->module->name, mlp->module->info, mlp->module->author);
    }
    (void)ws_grant_privilege();
  }
}

/*
 * Function Name:	wv_remove_modules
 * Description:		remove modules
 *			remove all modules if unsecure flag is 0,
 *			remove unsecured modules if the flag is 1.
 * Arguments:		unsecure flag
 * Return Value:	none
 */
void
wv_remove_modules(unsecure)
	int	unsecure;
{
  register ModuleList *	mlp;
  register ModuleList *	mlp_next;

  for (mlp = moduleList.next; mlp != &moduleList; mlp = mlp_next) {
    mlp_next = mlp->next;
    switch (unsecure & mlp->module->flags) {
      case WM_FIXED:
	break;
      case WM_SECURE:
      case WM_UNSECURE:
      default:
	if (mlp->module->finish != NULL) {
	  mlp->module->finish();
	}
	wv_syslog(WLL_LOG_INFO, "unload '%s'", mlp->module->name);
	uninstall(mlp);
	unload(mlp->module->handle);
	free(mlp);
	break;
    }
  }
}

/*****************************************************************************/

/*
 * Function Name:	load
 * Description:		load a module
 * Arguments:		module path
 * Return Value:	successful: module handle, otherwise: NULL
 */
static void *
load(path)
	const char *	path;
{
#if	HAVE_DLOPEN
  void *	handle;

  if (path[0] != '/') {
    wv_syslog(WLL_LOG_ERR, "module: no absolute path: %s", path);
    return NULL;
  }
  handle = dlopen(path, RTLD_NOW|RTLD_GLOBAL);
#elif	HAVE_SHL_LOAD
  shl_t		handle;

  if (path[0] != '/') {
    wv_syslog(WLL_LOG_ERR, "module: no absolute path: %s", path);
    return NULL;
  }
  handle = shl_load(path, BIND_IMMEDIATE|BIND_VERBOSE|BIND_NOSTART, 0L);
#elif	HAVE_LIBDLD
#endif	/* HAVE_DLOPEN */
  if (handle == NULL) {
    wv_syslog(WLL_LOG_ERR, "failed to load a module: %s: %s", path,
#if	HAVE_DLOPEN
	     (char *)dlerror()
#elif	HAVE_SHL_LOAD
	     (char *)strerror(errno)
#elif	HAVE_LIBDLD
	     "link failed"
#endif	/* HAVE_DLOPEN */
	);
  }

  return handle;
}

/*
 * Function Name:	fetch
 * Description:		fetch a symbol from loaded module
 * Arguments:		module handle
 * Return Value:	successful: symbol pointer, otherwise: NULL
 */
static void *
fetch(handle)
	void *	handle;
{
  void *	(*symbol)(void);

#if	HAVE_DLOPEN
#if defined(USCORE) && !defined(DLSYM_ADDS_USCORE)
  char *	usymb;

  ws_asprintf(&usymb, "_%s", WM_DLSYM_NAME);
  symbol = dlsym(handle, usymb);
  free(usymb);
#else
  symbol = dlsym(handle, WM_DLSYM_NAME);
#endif	/* USCORE */
  if (symbol == NULL) {
    wv_syslog(WLL_LOG_ERR, "failed to get a symbol: %s", WM_DLSYM_NAME);
    return NULL;
  }
#elif	HAVE_SHL_LOAD
  int	status;

  errno	 = 0;
  status = shl_findsym((shl_t *)&handle, WM_DLSYM_NAME, TYPE_PROCEDURE,
		       (int *)&ret);
  if (status == -1 && errno == 0) {
    status = shl_findsym((shl_t *)&handle, WM_DLSYM_NAME, TYPE_DATA,
			 (int *)&ret);
  }
  if (status == -1) {
    wv_syslog(WLL_LOG_ERR, "failed to get a symbol: %s", WM_DLSYM_NAME);
    return NULL;
  }
#elif	HAVE_LIBDLD
#endif	/* HAVE_DLOPEN */

  return symbol();
}

/*
 * Function Name:	verify
 * Description:		verify a loaded module
 * Arguments:		module
 * Return Value:	successful: 0, otherwise: -1
 */
static int
verify(module)
	const wmModule *	module;
{
  if (module == NULL) {
    return -1;
  }

  /* Check the module name */
  if (module->name == NULL) {
    wv_syslog(WLL_LOG_ERR, "module: unidentified name");
    return -1;
  }

  /* Check the API version */
  if (module->version != WM_MAGIC_NUMBER_MAJOR) {
    wv_syslog(WLL_LOG_ERR, "module: %s has invalid version %d",
	      module->name, module->version);
    return -1;
  }

  /* Check the duplicate loading */
  if (lookup(module->name) != NULL) {
    wv_syslog(WLL_LOG_ERR, "module: %s is already loaded", module->name);
    return -1;
  }

  /* Check the module description */
  if (module->info == NULL) {
    wv_syslog(WLL_LOG_ERR, "module: %s has no description", module->name);
  }

  /* Check the module creator */
  if (module->author == NULL) {
    wv_syslog(WLL_LOG_ERR, "module: %s has no creator name", module->name);
    return -1;
  }

  /* Check the depended module */
  if (module->depend != NULL && lookup(module->depend) == NULL) {
    wv_syslog(WLL_LOG_ERR, "module: %s depends on %s; but not loaded",
	     module->name, module->depend);
    return -1;
  }

  return 0;
}

/*
 * Function Name:	lookup
 * Description:		lookup module
 * Arguments:		module name
 * Return Value:	successful: pointer, otherwise: NULL
 */
static ModuleList *
lookup(name)
	const char *	name;
{
  register ModuleList *	mlp;

  for (mlp = moduleList.next; mlp != &moduleList; mlp = mlp->next) {
    if (ws_strcmp(mlp->module->name, name) == 0) {
      return mlp;
    }
  }

  return NULL;
}

/*
 * Function Name:	new_module_list
 * Description:		make a new module list entity
 * Arguments:		module
 * Return Value:	new pointer
 */
static ModuleList *
new_module_list(module)
	const wmModule *	module;
{
  ModuleList *	new;

  new		= (ModuleList *)ws_malloc(sizeof(ModuleList));
  new->next	= new;
  new->prev	= new;
  new->module	= module;

  return new;
}


/*
 * Function Name:	install
 * Description:		install new module to module list
 * Arguments:		module list
 * Return Value:	none
 */
static void
install(mlist)
	register ModuleList *	mlist;
{
  register ModuleList *	tail = moduleList.prev;

  mlist->next	    = tail->next;
  mlist->prev	    = tail;
  mlist->prev->next = mlist;
  mlist->next->prev = mlist;
}

/*****************************************************************************/

/*
 * Function Name:	unload
 * Description:		unload a module
 * Arguments:		module handle
 * Return Value:	successful: 0, otherwise: -1
 */
static int
unload(handle)
	void *	handle;
{
  if (handle != NULL) {
#if	HAVE_DLOPEN
    return dlclose(handle);
#elif	HAVE_SHL_LOAD
    return shl_unload((shl_t)handle);
#elif	HAVE_LIBDLD
#endif  /* HAVE_DLOPEN */
  }

  return 0;
}

/*
 * Function Name:	uninstall
 * Description:		uninstall the module from module list
 * Arguments:		module list
 * Return Value:	none
 */
static void
uninstall(mlist)
	register ModuleList *	mlist;
{
  mlist->prev->next = mlist->next;
  mlist->next->prev = mlist->prev;
}
