/*****************************************************************************
 *
 * FILE:	network.c
 * DESCRIPTION:	Wyvern: lookup the client host
 * DATE:	Sun, Sep 26 1999
 * UPDATED:	Sat, Aug 25 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 1999-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 1999-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: network.c,v 1.46 2007/08/26 16:37:36 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <errno.h>
#include "wyvern.h"

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static ListenAddr *	lookup(const char *, const char *);
static void		request_loop(void *);

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wv_add_sockets
 * Description:		register bind-address, port into ListenAddr list
 * Arguments:		bind address, port, secure flag
 * Return Value:	none
 */
void
wv_add_sockets(ipaddr, port, secure)
	const char *	ipaddr;
	const char *	port;
	const int	secure;
{
  DBG(fprintf(stderr, "addSockets '%s' '%s' %s\n",
      ipaddr, port, secure ? "secure" : "unsecure");)
  if (lookup(ipaddr, port) == NULL) {	/* no conflicted */
    ListenAddr *	new;

    new		  = (ListenAddr *)ws_malloc(sizeof(ListenAddr));/* leak */
    new->ipaddr	  = ws_strdup(ipaddr);
    new->port	  = ws_strdup(port);
    new->secure	  = secure;
    new->next	  = httpd->listen;

    httpd->listen = new;
  }
}

static ListenAddr *
lookup(ipaddr, port)
	const char *	ipaddr;
	const char *	port;
{
  register ListenAddr *	lap;

  for (lap = httpd->listen; lap != NULL; lap = lap->next) {
    if (ws_strcmp(port, lap->port) == 0) {
      if (ipaddr == NULL || lap->ipaddr == NULL || 
	  ws_strcmp(ipaddr, lap->ipaddr) == 0) {
  	return lap;
      }
    }
  }

  return NULL;
}

/*****************************************************************************/

/*
 * Function Name:	wv_open_channels
 * Description:		open available all communication channels
 * Arguments:		none
 * Return Value:	none
 */
void
wv_open_channels(void)
{
#if	SO_ACCEPTFILTER
  struct accept_filter_arg	afa;
#endif	/* SO_ACCEPTFILTER */
  register ListenAddr *		lap;

#if	SO_ACCEPTFILTER
  /* preparation of ACCEPT_FILTER_HTTP */
  memset((char *)&afa, 0, sizeof(afa));
  strlcpy(afa.af_name, "httpready", sizeof(afa));
#endif	/* SO_ACCEPTFILTER */

  /* open all available channels */
  for (lap = httpd->listen; lap != NULL; lap = lap->next) {
    wt_netfd_t *	nsd;

    /* open port */
#if	ENABLE_SSL
    if (lap->secure == SOCKET_SECURE) {
      nsd = wt_netfd_ssl_tcp_servers(lap->ipaddr, lap->port,
				     (int)httpd->maxRequest,
				     httpd->sslCertificateFile,
				     httpd->sslCertificateKeyFile,
				     httpd->sslCipherList);
    }
    else {
      nsd = wt_netfd_tcp_servers(lap->ipaddr, lap->port,
				     (int)httpd->maxRequest);
    }
#else
    nsd = wt_netfd_tcp_servers(lap->ipaddr, lap->port, (int)httpd->maxRequest);
#endif	/* ENABLE_SSL */
    if (nsd == NULL) {
      wv_die("Failed to open port %s at %s", lap->port, lap->ipaddr);
    }
    for (; nsd != NULL; nsd = nsd->next) {
      /*
       * linger:	off (default)	on	on
       * timeout:	---		0	>0
       * packet:	FIN		RST	FIN (EAGAIN by timeout)
       */
      struct linger	lval = { 1 /* on */, 30 /* sec. */ };
      Socket *		soc;
      unsigned int	i;
      int		bsize;
      socklen_t		bsize_len;

      /* make new socket */
      soc = (Socket *)ws_malloc(sizeof(Socket));	/* leak */

      /* set socket information (secure/unsecure and port) */
      soc->info.secure = lap->secure;
      soc->info.port   = (unsigned int)strtol(lap->port, NULL, 10);

      /*
       * Programmers must describe the following code
       * in case of using wt_netfd_tcp_servers().
       */
      wt_netfd_assign(soc->nsd, nsd);	/* It's a magic! */

      /* set send buffer size to improve performance */
      if (getsockopt(wt_netfd_fileno(soc->nsd), SOL_SOCKET, SO_SNDBUF,
		     &bsize, &bsize_len) == 0) {
	const int	size = 229376;	/* 224KB */

	if (setsockopt(wt_netfd_fileno(soc->nsd), SOL_SOCKET, SO_SNDBUF,
 		      &size, sizeof(size)) == 0) {
  	  wv_syslog(WLL_LOG_INFO, "SNDBUF changed: %d -> %d", bsize, size);
   	}
      }
      /* set linger filter to improve performance */
      if (setsockopt(wt_netfd_fileno(soc->nsd), SOL_SOCKET, SO_LINGER,
		     &lval, sizeof(lval)) == 0) {
	wv_syslog(WLL_LOG_INFO, "linger: %d %d", lval.l_onoff, lval.l_linger);
      }
#if	SO_ACCEPTFILTER
      /* set accept filter to improve performance */
      if (setsockopt(wt_netfd_fileno(soc->nsd), SOL_SOCKET, SO_ACCEPTFILTER,
		     &afa, sizeof(afa)) == -1) {
	wv_syslog(WLL_LOG_NOTICE, "unsupported ACCEPT_FILTER_HTTP");
      }
#endif	/* SO_ACCEPTFILTER */

      /* make mutex lock */
      if (wt_thread_mutex_init(&soc->mlock) == -1) {
	wv_die("Failed to get mutex lock");
      }

      /* make threads */
      soc->thread = (wt_thread_t *)ws_calloc((size_t)httpd->maxThread,
					    sizeof(wt_thread_t));/* leak */
      for (i = 0; i < httpd->maxThread; i++) {
#if	DEBUG
	wt_thread_attr_t	attr = wt_thread_attr_defaults;

	ws_asprintf((char **)&attr.name, "thread%02d", i); 
	if (wt_thread_create(&soc->thread[i], &attr, (void *)&request_loop,
			    (void *)soc) == -1) {
	  wv_die("Failed to create threads for %s", lap->ipaddr);
	}
#else
	if (wt_thread_create(&soc->thread[i], NULL, (void *)&request_loop,
			    (void *)soc) == -1) {
	  wv_die("Failed to create threads for %s", lap->ipaddr);
	}
#endif	/* DEBUG */
      }
      soc->next		= httpd->sockets;
      httpd->sockets	= soc;
    }
  }
}

/*
 * Function Name:	request_loop
 * Description:		thread main routine
 * Arguments:		socket pointer
 * Return Value:	none
 */
static void
request_loop(arg)
	void *	arg;
{
  Socket *	sp = (Socket *)arg;
  Message * 	msg;

  msg		   = wv_request_new(&sp->info);
  msg->client.addr = ws_malloc(wv_socklen(sp));	/* leak */
#if	ENABLE_SSL
  if (msg->soci.secure == SOCKET_SECURE) {
    msg->accept	  = (void *)(wt_netfd_ssl_accept);
    msg->close	  = (void *)(wt_netfd_ssl_close);
    msg->shutdown = (void *)(wt_netfd_ssl_shutdown);
    msg->read	  = (void *)(wt_netfd_ssl_read);
    msg->write	  = (void *)(wt_netfd_ssl_write);
    msg->writev	  = (void *)(wt_netfd_ssl_writev);
    msg->readline = (void *)(wt_netfd_ssl_readline);
    msg->sendfile = (void *)(wt_netfd_ssl_sendfile);
  }
  else {
    msg->accept	  = (void *)(wt_netfd_accept);
    msg->close	  = (void *)(wt_netfd_close);
    msg->shutdown = (void *)(wt_netfd_shutdown);
    msg->read	  = (void *)(wt_netfd_read);
    msg->write	  = (void *)(wt_netfd_write);
    msg->writev	  = (void *)(wt_netfd_writev);
    msg->readline = (void *)(wt_netfd_readline);
    msg->sendfile = (void *)(wt_netfd_sendfile);
  }
#else
  msg->accept	= (void *)(wt_netfd_accept);
  msg->close	= (void *)(wt_netfd_close);
  msg->shutdown = (void *)(wt_netfd_shutdown);
  msg->read	= (void *)(wt_netfd_read);
  msg->write	= (void *)(wt_netfd_write);
  msg->writev	= (void *)(wt_netfd_writev);
  msg->readline	= (void *)(wt_netfd_readline);
  msg->sendfile	= (void *)(wt_netfd_sendfile);
#endif	/* ENABLE_SSL */

  /* Wyvern Thread Request Engine */
  for (;;) {
    socklen_t	clilen = wv_socklen(sp);

    /* Accept new connection */
    DBG(fprintf(stderr, "DEBUG[request] waiting for sd=%d port=%d\n",
		sp->nsd->osfd, sp->info.port);)
    wt_thread_mutex_lock(sp->mlock);
    msg->nsd = msg->accept(sp->nsd, msg->client.addr, &clilen,
			   WT_UTIME_NO_TIMEOUT);
    wt_thread_mutex_unlock(sp->mlock);
    DBG(fprintf(stderr, "DEBUG[request] errno=%d\n", errno);)
    if (msg->nsd != NULL) {
      register char *	p;

      /* initialize request object */
      wv_request_clear(msg, true);
      /* Set peer IP address and port */
      msg->client.ip = wt_sock_ntop(msg->client.addr, true);
      p = strrchr(msg->client.ip, '.');
      if (p != NULL) {
	*p++ = EOL;
	msg->client.port = ws_strdup(p);
      }

      /* Convert IPv4 mapped IPv6 address to IPv4 address */
      if (ws_strncmp(msg->client.ip, "::ffff:", 7) == 0) {
	register char *	tmp = msg->client.ip;
	msg->client.ip = ws_strdup(tmp + 7);
	wv_free(tmp);
      }

      DBG(fprintf(stderr, "DEBUG: New connection from %s:%s\n",
		  msg->client.ip, msg->client.port);)
      do {
	int	rc;

	rc = wv_request_engine(msg);		/* Process a request */
	DBG(wv_syslog(WLL_LOG_DEBUG, "[request] rc=%d errno=%d", rc, errno);)
	if (errno == EAGAIN) {
	  wv_request_clear(msg, true);
	  continue;
	}
#if	0
	if (msg->response.status == STATUS_REQUEST_TIMEOUT) {
	  break;
	}
	if (msg->response.status == STATUS_NONE) {	/* XXX */
	  wv_syslog(WLL_LOG_DEBUG, "status=%d errno=%d",
		    msg->response.status, errno);
	  break;
	}
#else
	if (msg->response.status == STATUS_REQUEST_TIMEOUT ||
	    msg->response.status == STATUS_NONE) {	/* XXX */
	  break;
	}
#endif
	if (wv_response_engine(msg) == 0) {	/* Process a response */
	  wm_run_logger(msg);			/* Logging */
	}
	wv_request_clear(msg, rc == -1 ? true : false);
	DBG(fprintf(stderr, "DEBUG: keepalive = %d\n", msg->ka.max);)
      } while (msg->ka.status == KA_CONTINUE);

      msg->shutdown(msg->nsd, WT_SHUT_RDWR);	errno = 0;
      if (msg->close(msg->nsd) == -1 && errno == EAGAIN) {
	wv_syslog(WLL_LOG_NOTICE, "incomplete closing");
      }
      DBG(fprintf(stderr, "DEBUG: connection closed\n");)
    }
    /* SIGHUP rotates log-files (after several delays) */
    if (httpd->sighup == true) {
      wv_syslog(WLL_LOG_INFO, "log files rotated");
      wv_close_log_files();
      wv_open_log_files();
      httpd->sighup = false;
    }
  }
  /* NOTREACHED */
}


/*
 * Function Name:	wv_close_channels
 * Description:		close all opened communication channels
 * Arguments:		none
 * Return Value:	none
 */
void
wv_close_channels(void)
{
  register Socket *	sp;

  for (sp = httpd->sockets; sp != NULL; sp = sp->next) {
    if (sp->nsd == NULL) {
      continue;
    }
#if	ENABLE_SSL
    if (sp->info.secure == SOCKET_SECURE) {
      /* shutdown full-duplex connection */
      wt_netfd_ssl_shutdown(sp->nsd, WT_SHUT_RDWR);
      /* close the socket */
      wt_netfd_ssl_close(sp->nsd);
    }
    else {
      /* shutdown full-duplex connection */
      wt_netfd_shutdown(sp->nsd, WT_SHUT_RDWR);
      /* close the socket */
      wt_netfd_close(sp->nsd);
    }
#else
    /* shutdown full-duplex connection */
    wt_netfd_shutdown(sp->nsd, WT_SHUT_RDWR);
    /* close the socket */
    wt_netfd_close(sp->nsd);
#endif	/* ENABLE_SSL */
  }
}
