/*****************************************************************************
 *
 * FILE:	response.c
 * DESCRIPTION:	Wyvern: response the data to the client
 * DATE:	Sun, Oct  3 1999
 * UPDATED:	Wed, Sep  5 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 1999-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 1999-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: response.c,v 1.36 2007/09/12 02:23:38 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <errno.h>
#include "wyvern.h"

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	mk_response_headers(const char * key, const char * val,
				    void * args, unsigned int n);
static int	send_response_entity(Message *);
static int	send_multipart_byte_ranges(Message *);

/*****************************************************************************
 *
 *	Local variables definition
 *
 *****************************************************************************/

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

int
wv_response_engine(msg)
	register Message *	msg;
{
  /* Responding */
  wm_run_fixups(msg, HOOK_FIXUP_RESPONSE);
  wm_run_fixups(msg, HOOK_FINAL);
  if (wm_run_response_handler(msg) == -1) {
    wv_syslog(WLL_LOG_ERR, "failed to response data: %s from %s (errno = %d)",
	      msg->request.uri[0] == EOL ? "-" : msg->request.uri,
	      msg->client.ip, errno);
    return -1;
  }

  return 0;
}

/*
 * Function Name:	wv_response_handler
 * Description:		default response handler
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
wv_response_handler(msg)
	register Message *	msg;
{
  int	status;

  /* Timestamp of Response */
  gettimeofday(&msg->time.response, (struct timezone *)0);

  /* Send response header field and entity */
  status = wv_send_response_header(msg);
  if (status == WM_RET_OK) {
    switch (msg->request.method.type) {
      case M_GET:
      case M_POST:
       	status = send_response_entity(msg);
      default:
	break;
    }
  }

  return status;
}

/*
 * Function Name:	wv_make_response_header
 * Description:		make response header fields
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
wv_make_response_header(msg)
	register Message *	msg;
{
  char *	tbuf;
  char *	length = NULL;

  /* Connection: */
  wa_delete(msg->response.header, "Connection");
  switch (msg->request.http.version) {
    case HTTP_10:
      if (msg->ka.status == KA_CONTINUE && msg->ka.max > 0) {
	const char *	ka;
	char *		ka_val;

  	ka = wa_put(msg->response.header, "Connection", "Keep-Alive");
  	ws_asprintf(&ka_val, "timeout=%d, max=%d", DEF_KA_TIMEOUT, msg->ka.max);
  	wa_put(msg->response.header, ka, ka_val);
	free(ka_val);
	break;
      }
    case HTTP_09:
      wa_put(msg->response.header, "Connection", "close");
      msg->ka.status = KA_CLOSE;
    default:
      break;
  }
  /* Date: */
  if (wu_timestamp(&tbuf, 0, WUL_TIME_FORMAT_RFC822, WUL_ZONE_GMT) == 0) {
    wa_put(msg->response.header, "Date", tbuf);
    free(tbuf);
  }
  /* Server: */
  wa_put(msg->response.header, "Server", SERVER_VERSION);
  /* Accept-Ranges: */
  wa_put(msg->response.header, "Accept-Ranges", RANGE_UNIT);

  DBG(fprintf(stderr, "DEBUG: Status = %d\n",
	      wvResponseStatusTable[msg->response.status].code);)
  if (msg->response.status > STATUS_CODE_3XX) {		/* 3xx/4xx/5xx */
    if (msg->response.status > STATUS_CODE_4XX) {	/* 4xx/5xx */
      /* Content-Type: */
      wa_put(msg->response.header, "Content-Type", "text/html");
    }
    return WM_RET_OK;
  }

  /* 2xx */
  DBG(fprintf(stderr, "DEBUG: Content-Type = %s\n", wa_get(msg->request.header, "content-type"));)
  if (msg->response.status == STATUS_PARTIAL_CONTENT) {
    if (msg->file.ritems == 1) {	/* single part */
      /* Content-Length: */
      ws_asprintf(&length, "%zu",
		  msg->file.ranges[0].end - msg->file.ranges[0].start + 1);
    }
    else {				/* multipart */
      char *	ctype;
      /* Content-Type: */
      msg->file.boundary =
	wu_make_boundary((unsigned long)msg->file.status.st_mtime);
      ws_asprintf(&ctype, "multipart/byteranges; boundary=\"%s\"",
		  msg->file.boundary);
      msg->file.mime = wa_put(msg->response.header, "Content-Type", ctype);
      free(ctype);
    }
  }
  else {
    DBG(fprintf(stderr, "DEBUG: Media = %s\n", msg->file.mime);)
    /* Content-Type: */
    if (msg->file.mime != NULL) {
      wa_put(msg->response.header, "Content-Type", msg->file.mime);
    }
    /* Content-Length: */
    if (msg->file.status.st_size > 0) {
      ws_asprintf(&length, "%zu", msg->file.status.st_size);
    }
  }
  /* Content-Length */
  if (length != NULL) {
    wa_put(msg->response.header, "Content-Length", length);
    free(length);
  }
  /* Last-Modified: */
  if (msg->file.status.st_mtime > 0) {
    if (wu_timestamp(&tbuf, msg->file.status.st_mtime,
		     WUL_TIME_FORMAT_RFC822, WUL_ZONE_GMT) == 0) {
      wa_put(msg->response.header, "Last-Modified", tbuf);
      free(tbuf);
    }
  }

  return WM_RET_OK;
}

/*
 * Function Name:	wv_send_response_header
 * Description:		send response header fields
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
wv_send_response_header(msg)
	register Message *	msg;
{
  size_t		nums;	/* the number of response headers */
  struct iovec *	iov = NULL;
  unsigned int		n   = 0;

  nums	= wa_get_items(msg->response.header);
  nums++;	/* for response status line */
  iov	= (struct iovec *)ws_calloc(nums + 1, sizeof(struct iovec));

  /* Make the response status line */
  iov[n].iov_len = ws_asprintf((char **)&iov[n].iov_base, "%s %d %s" CRLF,
			msg->request.http.protocol,
			wvResponseStatusTable[msg->response.status].code,
			wvResponseStatusTable[msg->response.status].phrase);
  n++;

  /* Make response header fields */
  n += wa_foreach(msg->response.header, mk_response_headers, iov);
  iov[n].iov_base = CRLF;
  iov[n].iov_len  = 2;
  n++;

  /* Send response headers */
  if (msg->writev(msg->nsd, iov, n, WT_UTIME_SEC(msg->ka.timeout)) == -1) {
    return WM_RET_ERROR;
  }

  /* clear */
  for (n = 0; n < nums; n++) { free(iov[n].iov_base); }
  free(iov);

  return WM_RET_OK;
}

static void
mk_response_headers(key, val, args, n)
	const char *	key;
	const char *	val;
	void *		args;
	unsigned int	n;
{
  struct iovec *	iov = (struct iovec *)args;

  iov[n].iov_len = ws_asprintf((char **)&iov[n].iov_base,
			       "%s: %s" CRLF, key, val ? val : "");
#if	DEBUG
  fprintf(stderr, "DEBUG[callback] %u: %s: %s\n", n, key, val);
#endif	/* DEBUG */
}

/*
 * Function Name:	send_response_entity
 * Description:		send response entity
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
static int
send_response_entity(msg)
	register Message *	msg;
{
  if (msg->response.status > STATUS_CODE_4XX && msg->file.nfd == NULL) {
    return wv_error_message_sender(msg);
  }
  else if (msg->response.status > STATUS_CODE_3XX) {
    ;	/* Send no entity */
  }
  else if (msg->response.status == STATUS_PARTIAL_CONTENT) {
    return send_multipart_byte_ranges(msg);
  }
  else {
    off_t	sbytes;	/* for error check */

    if (msg->sendfile(msg->file.nfd, msg->nsd, msg->file.offset,
		      (size_t)msg->file.status.st_size, NULL, &sbytes, 0,
		      WT_UTIME_SEC(msg->ka.timeout)) == -1) {
      msg->ka.status = KA_CLOSE;
      return -1;
    }
  }

  return 0;
}

static int
send_multipart_byte_ranges(msg)
	register Message *	msg;
{
  /* single part */
  if (msg->file.ritems == 1) {
    if (msg->sendfile(msg->file.nfd, msg->nsd, (off_t)msg->file.ranges[0].start,
		      msg->file.ranges[0].end - msg->file.ranges[0].start + 1,
		      NULL, NULL, 0, WT_UTIME_SEC(msg->ka.timeout)) == -1) {
      msg->ka.status = KA_CLOSE;
      return -1;
    }
  }
  /* multipart */
  else {
    unsigned int	i;
    char		buf[BUF_SIZE];
    register int	n;

    for (i = 0; i < msg->file.ritems; i++) {
      /* Send boundary */
      memset(buf, 0, sizeof(buf));
      n = snprintf(buf, sizeof(buf), "--%s" CRLF, msg->file.boundary);
      msg->write(msg->nsd, buf, (size_t)n, WT_UTIME_SEC(msg->ka.timeout));

      /* Content-Type */
      memset(buf, 0, sizeof(buf));
      n = snprintf(buf, sizeof(buf), "Content-Type: %s" CRLF,
		   wa_get(msg->response.header, "Content-Type"));
      msg->write(msg->nsd, buf, (size_t)n, WT_UTIME_SEC(msg->ka.timeout));

      /* Content-Range */
      memset(buf, 0, sizeof(buf));
      n = snprintf(buf, sizeof(buf), "Content-Range: bytes %d-%d/%ld" CRLF CRLF,
		   msg->file.ranges[i].start,
		   msg->file.ranges[i].end,
		   (long)msg->file.status.st_size);
      msg->write(msg->nsd, buf, (size_t)n, WT_UTIME_SEC(msg->ka.timeout));

      /* Send entity */
      if (msg->sendfile(msg->file.nfd, msg->nsd,
			(off_t)msg->file.ranges[i].start,
			msg->file.ranges[i].end - msg->file.ranges[i].start + 1,
			NULL, NULL, 0, WT_UTIME_SEC(msg->ka.timeout)) == -1) {
	msg->ka.status = KA_CLOSE;
      }
    }

    /* Send end boundary */
    memset(buf, 0, sizeof(buf));
    n = snprintf(buf, sizeof(buf), "--%s--" CRLF, msg->file.boundary);
    msg->write(msg->nsd, buf, (size_t)n, WT_UTIME_SEC(msg->ka.timeout));
  }

  return 0;
}
