/*****************************************************************************
 *
 * FILE:	security.c
 * DESCRIPTION:	Wyvern: security functions
 * DATE:	Tue, Aug  6 2002
 * UPDATED:	Sat, Aug 18 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2002-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2002-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: security.c,v 1.20 2007/08/19 06:45:31 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <pwd.h>
#include <grp.h>
#include <errno.h>
#include "wyvern.h"

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	check_directory_condition(void);
static void	alienate_property(void);
static int	check_safe_path(Message *, const char *);

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wv_enforce_secure_level
 * Description:		enforce secure level
 * Arguments:		none
 * Return Value:	none
 */
void
wv_enforce_secure_level(void)
{
  register VirtualHost *	vp;

  DBG(fprintf(stderr, "[enforceSecureLevel] SecureLevel=%d\n",
	      httpd->secureLevel);)
  switch (httpd->secureLevel) {
    case SECURE_LEVEL7:	/* Disable virtual hosting */
      wv_delete_virtual_host();
    case SECURE_LEVEL6:	/* Disable all modules */
      wv_remove_modules(WM_ALL_MODULES);
    case SECURE_LEVEL5:	/* Forbid accessing UserDir */
      for (vp = httpd->host; vp != NULL; vp = vp->next) {
	wv_free(vp->userDir);
      }
    case SECURE_LEVEL4:	/* Disable modules to make dynamic pages */
      wv_remove_modules(WM_UNSECURE_MODULES);
    case SECURE_LEVEL3:	/* Forbid accessing symbolic linked files */
      for (vp = httpd->host; vp != NULL; vp = vp->next) {
	vp->followSymlinks = false;
      }
    case SECURE_LEVEL2:	/* Change owner and permission */
      alienate_property();
    case SECURE_LEVEL1:	/* Forbid running as root:wheel */
      /* dropPrivilege() will be called by main() in wyvern.c */
    case SECURE_LEVEL0:
      check_directory_condition();
    default:
      break;
  }

  /* fixups mime table for all hosts */
  for (vp = httpd->host; vp != NULL; vp = vp->next) {
    wv_fixup_mime_table(&vp->mimeTable);
  }
}

/*
 * Function Name:	check_directory_condition
 * Description:		confirm existence of CacheDir/TmpDir directories
 * Arguments:		none
 * Return Value:	none
 */
static void
check_directory_condition(void)
{
  struct stat	sbuf;

  /* check a temporary directory */
  if (lstat(httpd->tmpDir, &sbuf) == -1 && errno == ENOENT) {
    wv_die("No directory: %s", httpd->tmpDir);
  }
  if (S_ISLNK(sbuf.st_mode)) {
    wv_die("Untrusted symbolic link: %s", httpd->tmpDir);
  }

  /* check a cache directory */
  if (lstat(httpd->cacheDir, &sbuf) == -1 && errno == ENOENT) {
    wv_die("No directory: %s", httpd->cacheDir);
  }
  if (S_ISLNK(sbuf.st_mode)) {
    wv_die("Untrusted symbolic link: %s", httpd->cacheDir);
  }
}

/*
 * Function Name:	alienate_property
 * Description:		change owner/group and permission
 * Arguments:		none
 * Return Value:	none
 * Security Level:	level2
 */
static void
alienate_property(void)
{
  if (getuid() == UID_ROOT) {
    register VirtualHost *	vp;

    /* temporary dir */
    if (ws_chown(httpd->tmpDir, httpd->user, httpd->group) == -1) {
      wv_die("Failed to change owner of %s", httpd->tmpDir);
    }
    if (ws_chmod(httpd->tmpDir, S_IRWXU|S_IRGRP|S_IXGRP) == -1) {
      wv_die("Failed to change mode of %s", httpd->tmpDir);
    }

    /* cache dir */
    if (ws_chown(httpd->cacheDir, httpd->user, httpd->group) == -1) {
      wv_die("Failed to change owner of %s", httpd->cacheDir);
    }
    if (ws_chmod(httpd->cacheDir, S_IRWXU|S_IRGRP|S_IXGRP) == -1) {
      wv_die("Failed to change mode of %s", httpd->cacheDir);
    }

    /* log files */
    if (fchown(fileno(httpd->securityLog.fp),
		httpd->user, httpd->group) == -1) {
      wv_die("Failed to change owner of %s", httpd->securityLog.path);
    }
    for (vp = httpd->host; vp != NULL; vp = vp->next) {
      if (fchown(fileno(vp->errorLog.fp), httpd->user, httpd->group) == -1) {
	wv_die("Failed to change owner of %s", vp->errorLog.path);
      }
      if (fchown(fileno(vp->accessLog.fp), httpd->user, httpd->group) == -1) {
	wv_die("Failed to change owner of %s", vp->accessLog.path);
      }
    }
  }
}

/*
 * Function Name:	wv_drop_privilege
 * Description:		drop the privilege of wyvern process
 * Arguments:		none
 * Return Value:	none
 * Security Level:	level1
 */
void
wv_drop_privilege(void)
{
  if (httpd->secureLevel > SECURE_LEVEL0) {
    /* drop privilege */
    switch (ws_drop_privilege(httpd->user, httpd->group)) {
      case -1:
	wv_die("Failed to drop privilege: uid = %d, gid = %d",
		httpd->user, httpd->group);
      case 1:
	wv_syslog(WLL_LOG_WARNING,
		  "No running as root: failed to change uid %d and gid %d",
		  httpd->user, httpd->group);
	httpd->user  = getuid();
	httpd->group = getgid();
      default:
	break;
    }
  }
}


/*****************************************************************************/

/*
 * Function Name:	wv_safe_path_checker
 * Description:		track a symbolic link file
 * Arguments:		message, file path
 * Return Value:	Safe path: 0, otherwise: -1
 */
#define	MAX_SYMLINKS	(5)
int
wv_safe_path_checker(msg, path)
	register Message *	msg;
	const char *		path;
{
  int		safety	 = -1;
  static int	symlinks = 0;
  struct stat	sbuf;

  if (lstat(path, &sbuf) == 0) {
    if (S_ISLNK(sbuf.st_mode)) {	/* symbolic link */
      char	realPath[MAX_PATH_LEN + 1];

      memset(realPath, 0, sizeof(realPath));
      if (++symlinks <= MAX_SYMLINKS &&
	  readlink(path, realPath, sizeof(realPath)) != -1) {
	return wv_safe_path_checker(msg, realPath);
      }
    }
    safety = check_safe_path(msg, path);
  }
  else if (path[0] != '/' && path[0] != '.' &&
	   path[1] != '/' && path[1] != '.') {
    DBG(fprintf(stderr, "[safe] path = %s\n", path);)
    safety = 0;
  }
  else if (symlinks == 0 && errno == ENOENT) {
    msg->response.status = STATUS_NOT_FOUND;
  }
  else {
    msg->response.status = STATUS_FORBIDDEN;
  }

  if (symlinks > 0) {
    symlinks = 0;
    wv_syslog(WLL_LOG_SECURITY,
	      "request URI '%s' is %s symbolic file '%s'",
	      msg->request.uri, safety == 0 ? "a" : "an unsecure", path);
  }

  return safety;
}

/*
 * Function Name:	check_safe_path
 * Description:		check the path which is safe
 * Arguments:		message, file path
 * Return Value:	Successful: 0, otherwise: -1
 */
static int
check_safe_path(msg, path)
	register Message *	msg;
	const char *		path;
{
  /* in DocumentRoot? */
  if (ws_strncmp(path, msg->vhost->documentRoot,
		 strlen(msg->vhost->documentRoot)) == 0) {
    return 0;
  }

  /* in alias? */
  if (msg->vhost->aliasTable != NULL) {
    register Alias *	ap;

    for (ap = msg->vhost->aliasTable; ap != NULL; ap = ap->next) {
      if (ap->type == A_REDIRECT) {
       	continue;
      }
      if (ws_strncmp(path, ap->realName, ap->realLen) == 0) {
	return 0;
      }
    }
  }

  /* in ServerRoot? */
  if (ws_strncmp(path, httpd->serverRoot, strlen(httpd->serverRoot)) == 0) {
    return 0;
  }

  /* in UserDir? */
  if (msg->user.indexDir != NULL &&
      ws_strncmp(path, msg->user.indexDir, strlen(msg->user.indexDir)) == 0) {
    return 0;
  }

  msg->response.status = STATUS_FORBIDDEN;

  return -1;
}
