/*****************************************************************************
 *
 * FILE:	user.c
 * DESCRIPTION:	Wyvern: management user information
 * DATE:	Sat, Sep 18 1999
 * UPDATED:	Sat, Aug 18 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 1999-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 1999-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: user.c,v 1.18 2007/08/19 06:45:31 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <pwd.h>
#include "wyvern.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
typedef struct {
  char *	name;	/* owner of UserDir */
  char *	home;	/* home directory */
} UserInfo;

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static char *	get_home_directory(const char *);
static int	compare(const void *, const void *);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static Cache *	userList;	/* header of cache */

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wv_user_cache_init
 * Description:		initialize userTable called at only once
 * Arguments:		none
 * Return Value:	none
 */
void
wv_user_cache_init(void)
{
  userList = wc_new_cache(USER_TABLE_SIZE);
  if (userList == NULL) {
    wv_die("Failed to initialize user cache");
  }
}

/*****************************************************************************/


/*
 * Function Name:	wv_user_dir_translater
 * Description:		translate an user directory of requests URI
 * Arguments:		message
 * Return Value:	successful: 0, no translated: 1, otherwise: -1
 */
int
wv_user_dir_translater(msg)
	register Message *	msg;
{
  size_t	len;

  if (msg->vhost->userDir == NULL || msg->request.uri[1] != '~') {
    return WM_RET_DECLINED;	/* Continue to translate */
  }

  len		    = strcspn(msg->request.uri    + TILDE_OFFSET, "/");
  msg->user.name    = ws_strndup(msg->request.uri + TILDE_OFFSET, len);
  msg->user.homeDir = get_home_directory(msg->user.name);
  if (msg->user.homeDir == NULL) {
    msg->response.status = STATUS_NOT_FOUND;
    return WM_RET_ERROR;
  }

  /* Actual path of request file */
  ws_asprintf(&msg->user.indexDir, "%s/%s",
	      msg->user.homeDir, msg->vhost->userDir);
  ws_asprintf(&msg->file.path, "%s%s",
	      msg->user.indexDir, msg->request.uri + TILDE_OFFSET + len);
  DBG(fprintf(stderr, "[translateUserDir] %s\n", msg->file.path);)
  return wv_complement_path(msg);
}

/*
 * Function Name:	get_home_directory
 * Description:		return the user's home directory
 *			(user record is registered into cache)
 * Arguments:		user name
 * Return Value:	user's home directory
 */
static char *
get_home_directory(username)
	const char *	username;
{
  struct passwd	*	pwd;	/* pointer to password entry 	*/

  /* Get user information */
  pwd = getpwnam(username);
  if (pwd != NULL) {
    register UserInfo *	uip;
    UserInfo		ui;
    register Cache *	cp;

    /* Check cache */
    ui.name = (char *)username;
    uip	    = (UserInfo *)wc_search_cache(userList, &ui, compare);
    if (uip != NULL) {
      DBG(fprintf(stderr, "[userDir] found %s(%s)\n", uip->name, uip->home);)
      return uip->home;
    }

    /* Get top entry from list, and move one to end of list */
    cp	= wc_remove_cache(userList->next);
    uip = (UserInfo *)cp->data;
    if (uip != NULL) {	/* release old entry */
      wv_free(uip->name);
      wv_free(uip->home);
    }
    else {
      uip	= ws_malloc(sizeof(UserInfo));	/* leak */
      cp->data	= uip;
    }
    /* Register new entry */
    uip->name = ws_strdup(pwd->pw_name);
    uip->home = ws_strdup(pwd->pw_dir);
    wc_insert_cache(userList, cp);
    DBG(fprintf(stderr, "[userDir] new %s(%s)\n", uip->name, uip->home);)

    return uip->home;
  }

  return NULL;
}

static int
compare(x, y)
	register const void *	x;
	register const void *	y;
{
  if (ws_strcmp(((UserInfo *)x)->name, ((UserInfo *)y)->name) == 0) {
    return 0;
  }
  return -1;
}
