/*****************************************************************************
 *
 * FILE:	vhost.c
 * DESCRIPTION:	Wyvern: virtual host
 * DATE:	Sun, Apr 22 2001
 * UPDATED:	Wed, Sep 12 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2001-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2001-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: vhost.c,v 1.20 2007/09/12 02:24:27 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include "wyvern.h"

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static VirtualHost *	lookup(const char *);

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wv_virtual_host_new
 * Description:		make a new virtual host in vhostList
 * Arguments:		none
 * Return Value:	new virtual host pointer
 */
VirtualHost *
wv_virtual_host_new(void)
{
  VirtualHost *	new;

  new			= (VirtualHost *)ws_calloc(1, sizeof(VirtualHost));
  new->timeout		= DEF_TIMEOUT;	/* Timeout */
  new->defaultType	= ws_strdup(DEF_DEFTYPE);
  new->next		= httpd->vhostTable;
  httpd->vhostTable	= new;

  return new;
}

/*****************************************************************************/

/*
 * Function Name:	wv_delete_virtual_host
 * Description:		delete virtual hosts in vhostList
 * Arguments:		none
 * Return Value:	none
 */
void
wv_delete_virtual_host(void)
{
  register VirtualHost *	vp;
  register VirtualHost *	vp_next;

  for (vp = httpd->vhostTable; vp != NULL; vp = vp_next) {
    register StrList *	slp;
    register StrList *	slp_next;
    register ErrorDoc *	edp;
    register ErrorDoc *	edp_next;
    register Alias *	ap;
    register Alias *	ap_next;
    register Hash *	hp;
    register Hash *	hp_next;

    vp_next = vp->next;
    for (slp = vp->directoryIndexes; slp != NULL; slp = slp_next) {
      slp_next = slp->next;
      wv_free(slp->text);
      free(slp);
    }
    for (edp = vp->errorDocTable; edp != NULL; edp = edp_next) {
      edp_next = edp->next;
      wv_free(edp->text);
      free(edp);
    }
    for (ap = vp->aliasTable; ap != NULL; ap = ap_next) {
      ap_next = ap->next;
      wv_free(ap->fakeName);
      wv_free(ap->realName);
      free(ap);
    }
    for (hp = vp->mimeTable; hp != NULL; hp = hp_next) {
      hp_next = hp->next;
      wv_free(hp->name);
      wv_free(hp->value);
      free(hp);
    }

    wv_close_log(vp);
    wv_free(vp->serverName);
    wv_free(vp->serverAdmin);
    wv_free(vp->errorLog.path);
    wv_free(vp->accessLog.path);
    wv_free(vp->documentRoot);
    wv_free(vp->userDir);
    wv_free(vp->defaultType);
    free(vp);
  }
  httpd->host->next = NULL;
  httpd->vhostTable = NULL;
}

/*
 * Function Name:	wv_switch_virtual_host
 * Description:		search virtual host and set current virtual host
 * Arguments:		message
 * Return Value:	Successful: 0, otherwise: -1
 */
int
wv_switch_virtual_host(msg)
	register Message *	msg;
{
  register const char *	host;
  char *		hostname;
  register ListenAddr *	lap;
  int			status = -1;

  if ((host = wa_get(msg->request.header, "host")) == NULL) {
    msg->vhost = httpd->host;
    return 0;
  }

  /* Get hostname from Host request header (Host: hostname[:port]) */
  hostname = ws_strndup(host, strcspn(host, ":"));

  /* Get virtual ServerName */
  msg->vhost = lookup((const char *)hostname);
  if (msg->vhost != NULL) {
    httpd->vhost = msg->vhost;
    status	 = 0;
    goto done;
  }

  /* Set default server */
  msg->vhost = httpd->host;

  /* localhost/127.0.0.1 are special */
  if (ws_strcmp(hostname, "localhost") == 0 ||
      ws_strcmp(hostname, "127.0.0.1") == 0 ||
      ws_strcmp(hostname, "::") == 0) {
    status = 0;
    goto done;
  }

  /* No ServerName, try confirming IP address */
  for (lap = httpd->listen; lap != NULL; lap = lap->next) {
    if (lap->ipaddr == NULL ||			/* wildcard address */
  	ws_strcmp(lap->ipaddr, hostname) == 0) {/* Ok, match IP address */
      status = 0;
      goto done;
    }
  }
  /* Why Bad Request? See RFC2616 Section 5.2 */
  msg->response.status = STATUS_BAD_REQUEST;

done:
  free(hostname);

  return status;
}

/*
 * Function Name:	lookup
 * Description:		lookup the virtual hosts in the list
 * Arguments:		hostname
 * Return Value:	successful: VirtualHost pointer, otherwise: NULL
 */
static VirtualHost *
lookup(name)
	const char *	name;
{
  register VirtualHost *	vp;

  for (vp = httpd->host; vp != NULL; vp = vp->next) {
    if (ws_strcasecmp(vp->serverName, name) == 0) {
      return vp;
    }
  }

  return NULL;
}
