/*****************************************************************************
 *
 * FILE:	wyvern.c
 * DESCRIPTION:	Wyvern: main source file
 * DATE:	Sun, Aug 29 1999
 * UPDATED:	Sat, Aug 18 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 1999-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 1999-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: wyvern.c,v 1.22 2007/08/19 06:45:31 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <signal.h>
#include <errno.h>

#define	MAIN_MODULE
#include "wyvern.h"
#undef	MAIN_MODULE

/*****************************************************************************
 *
 *	Program & Copyright
 *
 *****************************************************************************/
const char	Program[]   = "wyvern";
const char	Copyright[] =
	"Copyright (c) 1999-2007 Kouichi ABE (WALL), All rights reserved.";

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	detach(void);
static void	store_pid(void);
static void	register_default_hooks(void);
static void	version(void);
static void	usage(void);

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	detach
 * Description:		detach the controlling terminal and create a new session
 * Arguments:		none
 * Return Value:	none
 */
static void
detach(void)
{
#ifndef	DEBUG
  if (wu_daemon(0, 0, NULL) == -1) {
    wv_die("Failed to detach the terminal");
  }
#endif	/* DEBUG */
  /* Initialize the log files */
  wv_open_log_files();		
  /* Initialize the Wyvern Thread Library */
  if (wt_init(WT_VP_SS_MEDIUM) == -1) {
    wv_die("Failed to initialize the WTL");
  }
  /* Change top directory */
  if (chdir(httpd->serverRoot) == -1) {
    wv_die("Failed to change directory: %s", httpd->serverRoot);
  }
}

/*
 * Function Name:	store_pid
 * Description:		store the process id into the PidFile
 * Arguments:		none
 * Return Value:	none
 */
static void
store_pid(void)
{
  if (wu_store_pid(httpd->pidFile, UID_ROOT, GID_WHEEL) == -1) {
    if (errno == EPERM) {
      wv_syslog(WLL_LOG_WARNING, "No changed owner/group: %s", httpd->pidFile);
    }
    else {
      wv_die("Failed to open %s", httpd->pidFile);
    }
  }
  httpd->pid = getpid();
}

/*
 * Function Name:	wv_quit
 * Description:		stop the service and exit
 * Arguments:		none
 * Return Value:	none
 */
void
wv_quit(void)
{
  sigset_t	set;

  sigfillset(&set);	/* block all signals */
  if (sigprocmask(SIG_BLOCK, &set, NULL) == 0) {
    struct timeval	now;
    char *		uptime;

    /* Close service channel */
    wv_close_channels();

    /* Remove all modules */
    wv_remove_modules(WM_ALL_MODULES);

    /* Record uptime */
    gettimeofday(&now, (struct timezone *)0);
    if (wu_uptime(&uptime, &httpd->bootTime, &now) == 0) {
      wv_syslog(WLL_LOG_INFO, "uptime: %s", uptime);
      free(uptime);
    }

    /* Close all log files */
    wv_close_log_files();
  }

  /* Remove PID file to unlock */
  if (httpd->pid > 0) {	
    (void)unlink(httpd->pidFile);
  }

  exit(0);	/* release allocated memories automatically */
}

/*****************************************************************************/

/*
 * Function Name:	registerDefaultHook
 * Description:		register default hooks
 * Arguments:		none
 * Return Value:	none
 */
static void
register_default_hooks(void)
{
  wm_hook_request_reader(wv_request_reader, HOOK_FINAL);
  wm_hook_header_parser(wv_header_parser, HOOK_INITIAL);
  wm_hook_fixups(wv_parse_uri, HOOK_FIXUP_URI);
  wm_hook_uri_translater(wv_alias_translater, HOOK_FINAL);
  wm_hook_uri_translater(wv_user_dir_translater, HOOK_FINAL);
  wm_hook_uri_translater(wv_document_root_translater, HOOK_FINAL);
  wm_hook_type_checker(wv_media_type_checker, HOOK_FINAL);
  wm_hook_request_handler(wv_request_handler, HOOK_FINAL, MAGIC_TYPE_ANY);
  wm_hook_fixups(wv_make_response_header, HOOK_FIXUP_RESPONSE);
  wm_hook_response_handler(wv_response_handler, HOOK_FINAL);
  wm_hook_logger(wv_log_writer, HOOK_FINAL);
}

/*****************************************************************************/

/*
 * Function Name:	version
 * Description:		show the version of Wyvern
 * Arguments:		none
 * Return Value:	none
 */
static void
version(void)
{
  fprintf(stderr, "%s\n%s\n", SERVER_SOFTWARE, Copyright);
#if	DEBUG
  fputs("+DEBUG ", stderr);
#endif	/* DEBUG */
#if	ENABLE_JP_MESG
  fputs("+JMESG ", stderr);
#endif	/* ENABLE_JP_MESG */
#if	ENABLE_SYSCTL_PS
  fputs("+SYSCTL ", stderr);
#endif	/* ENABLE_SYSCTL_PS */
#if	ENABLE_SIMPLE_INDEXING
  fputs("+SIMPLE ", stderr);
#endif	/* ENABLE_SIMPLE_INDEXING */
#if	ENABLE_STRICT_PARSING
  fputs("+STRICT ", stderr);
#endif	/* ENABLE_STRICT_PARSING */
#if	ENABLE_SUEXEC
  fputs("+SUEXEC ", stderr);
#endif	/* ENABLE_SUEXEC */
#if	ENABLE_SSL
  fputs("+SSL ", stderr);
#endif	/* ENABLE_SSL */
  fputs("\n\n", stderr);
  exit(64);
}

/*
 * Function Name:	usage
 * Description:		show the simple usage for Wyvern
 * Arguments:		none
 * Return Value:	none
 */
static void
usage(void)
{
  fprintf(stderr,
	  "%s\n%s\n%s\n"
	  "usage: %s [-r config_file]\n\n",
	  SERVER_SOFTWARE, Copyright, SERVER_ADDRESS,
	  Program);
  exit(64);
}

/*****************************************************************************/

/*
 *	Wyvern Main
 */
int
main(argc, argv)
	int	argc;
	char *	argv[];
{
  char *	configFile = "";

  switch (argc) {
    case 3:
      if (ws_strncmp(argv[1], "-r", 2) == 0) {
	configFile = argv[2];
      }
    case 1:	/* use default config */
      break;
    default:
      if (ws_strncmp(argv[1], "-v", 2) == 0) {
	version();
      }
      usage();
  }

  /*
   * bootstrap
   */
  wv_init_httpd();			/* Make server entity */
  wv_read_config_file(configFile);	/* Read configuration file */
  register_default_hooks();		/* Register default hooks */
  detach();				/* Detach the control terminal */
  wv_enforce_secure_level();		/* Enforce security level */
  wv_initialize_modules();		/* Initialize all modules */
  wv_open_channels();			/* Open service channel (socket) */
  store_pid();				/* Preserve server process ID	*/
  wv_drop_privilege();			/* Drop privilege eternally */
  wv_user_cache_init();			/* Initialize the user cache list */
  wv_signal_trap();			/* Set signal handlers */
  /* Main request loop */
  gettimeofday(&httpd->bootTime, (struct timezone *)0);
  wt_thread_run();

  wv_quit();

  return 0;
}
